SystemOrganization addCategory: #'Morphic-Core'!
SystemOrganization addCategory: 'Morphic-Core-Announcements'!
SystemOrganization addCategory: 'Morphic-Core-Borders'!
SystemOrganization addCategory: 'Morphic-Core-Canvas'!
SystemOrganization addCategory: 'Morphic-Core-Events'!
SystemOrganization addCategory: 'Morphic-Core-Kernel'!
SystemOrganization addCategory: 'Morphic-Core-Support'!
SystemOrganization addCategory: 'Morphic-Core-Worlds'!

!Form methodsFor: '*Morphic-Core' stamp: 'PavelKrivanek 11/18/2012 20:37'!
defaultCanvasClass

	"Return the default canvas used for drawing onto the receiver"
	^ FormCanvas! !

!Form methodsFor: '*Morphic-Core' stamp: 'ar 5/28/2000 12:03'!
getCanvas
	"Return a Canvas that can be used to draw onto the receiver"
	^self defaultCanvasClass on: self! !

BitBlt subclass: #GrafPort
	instanceVariableNames: 'alpha fillPattern lastFont lastFontForegroundColor lastFontBackgroundColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Support'!

!GrafPort commentStamp: 'TorstenBergmann 2/20/2014 18:35' prior: 0!
Provides extended Graphic support like translucency!

!GrafPort methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:07'!
alphaBits: a
	alpha := a! !

!GrafPort methodsFor: 'copying' stamp: 'ar 12/30/2001 20:32'!
clippedBy: aRectangle
	^ self copy clipBy: aRectangle! !

!GrafPort methodsFor: 'accessing' stamp: 'ar 5/28/2000 14:41'!
contentsOfArea: aRectangle into: aForm
	destForm 
		displayOn: aForm 
		at:  aRectangle origin
		clippingBox: (0@0 extent: aRectangle extent).
	^aForm! !

!GrafPort methodsFor: 'copying' stamp: 'MarcusDenker 10/21/2013 14:22'!
copyBits
	"Override copybits to do translucency if desired"

	(combinationRule between: 30 and: 31)
		ifTrue: [ self copyBitsTranslucent: (alpha ifNil: [ 255 ]) ]
		ifFalse: [ super copyBits ]! !

!GrafPort methodsFor: 'accessing' stamp: 'nice 11/2/2013 17:29'!
displayScannerFor: para foreground: foreColor background: backColor ignoreColorChanges: shadowMode
	^ (BitBltDisplayScanner new text: para text textStyle: para textStyle
			foreground: foreColor background: backColor fillBlt: self
			ignoreColorChanges: shadowMode)
		setPort: self shallowCopy! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/16/2000 22:32'!
fillOval: rect
	| centerX centerY nextY yBias xBias outer nextOuterX |
	rect area <= 0 ifTrue: [^ self].
	height := 1.
	yBias := rect height odd ifTrue: [0] ifFalse: [-1].
	xBias := rect width odd ifTrue: [1] ifFalse: [0].
	centerX := rect center x.
	centerY := rect center y.
	outer := EllipseMidpointTracer new on: rect.
	nextY := rect height // 2.
	[nextY > 0] whileTrue:[
		nextOuterX := outer stepInY.
		width := (nextOuterX bitShift: 1) + xBias.
		destX := centerX - nextOuterX.
		destY := centerY - nextY.
		self copyBits.
		destY := centerY + nextY + yBias.
		self copyBits.
		nextY := nextY - 1.
	].
	destY := centerY.
	height := 1 + yBias.
	width := rect width.
	destX := rect left.
	self copyBits.
! !

!GrafPort methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:09'!
fillPattern: anObject
	fillPattern := anObject.
	self fillColor: anObject.! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 5/17/2000 21:20'!
fillRect: rect offset: aPoint
	"The offset is really just for stupid InfiniteForms."
	| fc |
	fillPattern class == InfiniteForm ifTrue:[
		fc := halftoneForm.
		self fillColor: nil.
		fillPattern displayOnPort: ((self clippedBy: rect) colorMap: nil) at: aPoint.
		halftoneForm := fc.
		^self].

	destX := rect left.
	destY := rect top.
	sourceX := 0.
	sourceY := 0.
	width := rect width.
	height := rect height.
	self copyBits.! !

!GrafPort methodsFor: 'drawing support' stamp: 'MarcusDenker 10/28/2010 14:02'!
frameOval: rect borderWidth: borderWidth
	| centerX centerY nextY yBias xBias wp outer inner nextOuterX nextInnerX |
	rect area <= 0 ifTrue: [^ self].
	height := 1.
	wp := borderWidth asPoint.
	yBias := rect height odd ifTrue: [0] ifFalse: [-1].
	xBias := rect width odd ifTrue: [1] ifFalse: [0].
	centerX := rect center x.
	centerY := rect center y.
	outer := EllipseMidpointTracer new on: rect.
	inner := EllipseMidpointTracer new on: (rect insetBy: wp).
	nextY := rect height // 2.
	1 to: (wp y min: nextY) do:[:i|
		nextOuterX := outer stepInY.
		width := (nextOuterX bitShift: 1) + xBias.
		destX := centerX - nextOuterX.
		destY := centerY - nextY.
		self copyBits.
		destY := centerY + nextY + yBias.
		self copyBits.
		nextY := nextY - 1.
	].
	[nextY > 0] whileTrue:[
		nextOuterX := outer stepInY.
		nextInnerX := inner stepInY.
		destX := centerX - nextOuterX.
		destY := centerY - nextY.
		width := nextOuterX - nextInnerX.
		self copyBits.
		destX := centerX + nextInnerX + xBias.
		self copyBits.
		destX := centerX - nextOuterX.
		destY := centerY + nextY + yBias.
		self copyBits.
		destX := centerX + nextInnerX + xBias.
		self copyBits.
		nextY := nextY - 1.
	].
	destY := centerY.
	height := 1 + yBias.
	width := wp x.
	destX := rect left.
	self copyBits.
	destX := rect right - wp x.
	self copyBits.
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 14:44'!
frameRect: rect borderWidth: borderWidth
	sourceX := 0.
	sourceY := 0.
	(rect areasOutside: (rect insetBy: borderWidth)) do:
		[:edgeStrip | self destRect: edgeStrip; copyBits].
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 01:08'!
frameRectBottom: rect height: h

	destX := rect left + 1.
	destY := rect bottom - 1.
	width := rect width - 2.
	height := 1.
	1 to: h do: [:i |
		self copyBits.
		destX := destX + 1.
		destY := destY - 1.
		width := width - 2].
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 01:08'!
frameRectRight: rect width: w

	width := 1.
	height := rect height - 1.
	destX := rect right - 1.
	destY := rect top + 1.
	1 to: w do: [:i |
		self copyBits.
		destX := destX - 1.
		destY := destY + 1.
		height := height - 2].
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 01:08'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."

	sourceForm := aForm.
	combinationRule := rule.
	self sourceRect: sourceRect.
	self destOrigin: aPoint.
	self copyBits! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 8/8/2001 14:26'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule alpha: sourceAlpha
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."

	sourceForm := aForm.
	combinationRule := rule.
	self sourceRect: sourceRect.
	self destOrigin: aPoint.
	self copyBitsTranslucent: (alpha := (sourceAlpha * 255) truncated min: 255 max: 0).! !

!GrafPort methodsFor: 'private' stamp: 'yo 1/8/2005 09:12'!
installStrikeFont: aStrikeFont

	^ self installStrikeFont: aStrikeFont foregroundColor: (lastFontForegroundColor ifNil: [Color black]) backgroundColor: (lastFontBackgroundColor ifNil: [Color transparent]).
! !

!GrafPort methodsFor: 'private' stamp: 'tween 6/8/2008 12:29'!
installStrikeFont: aStrikeFont foregroundColor: foregroundColor backgroundColor: backgroundColor
	super installStrikeFont: aStrikeFont foregroundColor: foregroundColor backgroundColor: backgroundColor.
	aStrikeFont glyphs depth = 1 ifTrue: [
		alpha := foregroundColor privateAlpha.
		"dynamically switch between blend modes to support translucent text"
		"To handle the transition from TTCFont to StrikeFont, rule 34 must be taken into account."
		alpha = 255 ifTrue:[
			combinationRule = 30 ifTrue: [combinationRule := Form over].
			combinationRule = 31 ifTrue: [combinationRule := Form paint].
			combinationRule = 34 ifTrue: [combinationRule := Form paint].
			combinationRule = 41 ifTrue: [combinationRule := Form paint]. "41 is  SPRmode"
		] ifFalse:[
			combinationRule = Form over ifTrue: [combinationRule := 30].
			combinationRule = Form paint ifTrue: [combinationRule := 31].
			combinationRule = 34 ifTrue: [combinationRule := 31].
			combinationRule = 41 ifTrue: [combinationRule := 31]. "41 is SPR mode"
		]
	].
	lastFont := aStrikeFont.
	lastFontForegroundColor := foregroundColor.
	lastFontBackgroundColor := backgroundColor.
! !

!GrafPort methodsFor: 'private' stamp: 'yo 1/8/2005 09:13'!
lastFont

	^ lastFont.
! !

!GrafPort methodsFor: 'accessing' stamp: 'tween 4/5/2007 08:03'!
lastFontForegroundColor
	^lastFontForegroundColor! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 00:31'!
stencil: stencilForm at: aPoint sourceRect: aRect
	"Paint using aColor wherever stencilForm has non-zero pixels"
	self sourceForm: stencilForm;
		destOrigin: aPoint;
		sourceRect: aRect.
	self copyBits! !

Announcement subclass: #MorphAnnouncement
	instanceVariableNames: 'morph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Announcements'!

!MorphAnnouncement commentStamp: '<historical>' prior: 0!
I am an abstract announcement for morphic events.  I carry the morph where the event has ocurred.

My subclasses should add information for more precise events.!

!MorphAnnouncement class methodsFor: 'instance creation' stamp: 'GuillermoPolito 5/1/2012 19:48'!
morph: aMorph
	^self new
			morph: aMorph;
			yourself! !

!MorphAnnouncement methodsFor: 'accessing' stamp: 'GuillermoPolito 5/1/2012 19:47'!
morph
	^ morph! !

!MorphAnnouncement methodsFor: 'accessing' stamp: 'GuillermoPolito 5/1/2012 19:47'!
morph: anObject
	morph := anObject! !

MorphAnnouncement subclass: #MorphChanged
	instanceVariableNames: 'selector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Announcements'!

!MorphChanged commentStamp: '<historical>' prior: 0!
I am raised when a morph property changed, and should be propagated to dependents!

!MorphChanged methodsFor: 'delivering' stamp: 'IgorStasenko 8/27/2013 16:27'!
deliverTo: aHandler

	^ aHandler update: selector

	! !

!MorphChanged methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 7/22/2013 15:05'!
selector
	^ selector! !

!MorphChanged methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 7/22/2013 15:05'!
selector: anObject
	selector := anObject! !

MorphChanged subclass: #MorphChangedWithArguments
	instanceVariableNames: 'arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Announcements'!

!MorphChangedWithArguments commentStamp: '<historical>' prior: 0!
I am an announcement raised when a Morph changed and need to propagate the information to dependents. In addition, I also propagate a value!

!MorphChangedWithArguments methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 7/23/2013 12:32'!
arguments
	^ arguments! !

!MorphChangedWithArguments methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 7/23/2013 12:32'!
arguments: anObject
	arguments := anObject! !

!MorphChangedWithArguments methodsFor: 'delivering' stamp: 'IgorStasenko 8/27/2013 16:28'!
deliverTo: aHandler

	^ aHandler update: selector with: arguments

	! !

MorphAnnouncement subclass: #MorphDeleted
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Announcements'!

!MorphDeleted commentStamp: '<historical>' prior: 0!
I'm an announcement raised when a morph is deleted. Subscribe to me when you want to do things after the morph is opened.

Take into account that when a morph is deleted, all it's children also get deleted and they raise a similar announcement.!

MorphAnnouncement subclass: #MorphGotFocus
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Announcements'!

!MorphGotFocus commentStamp: '<historical>' prior: 0!
I'm an announcement raised when a morph gets keyboard focus. !

MorphAnnouncement subclass: #MorphLostFocus
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Announcements'!

!MorphLostFocus commentStamp: '<historical>' prior: 0!
I'm an announcement raised when a morph losts keyboard focus.!

MorphAnnouncement subclass: #MorphOpened
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Announcements'!

!MorphOpened commentStamp: '<historical>' prior: 0!
I'm an announcement raised when a morph is opened.  Subscribe to me when you want to do things after the morph is opened.

Take into account that when a morph is opened, all it's children also get opened and they raise a similar announcement.!

Object subclass: #BorderStyle
	instanceVariableNames: ''
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Morphic-Core-Borders'!

!BorderStyle commentStamp: 'kfr 10/27/2003 10:19' prior: 0!
See BorderedMorph

BorderedMorh new borderStyle: (BorderStyle inset width: 2); openInWorld.!

!BorderStyle class methodsFor: 'instance creation' stamp: 'sw 11/26/2001 16:05'!
borderStyleChoices
	"Answer the superset of all supported borderStyle symbols"

	^ #(simple inset raised complexAltFramed complexAltInset complexAltRaised complexFramed complexInset complexRaised)! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'MarcusDenker 7/18/2010 15:43'!
borderStyleForSymbol: sym
	"Answer a border style corresponding to the given symbol"

	| aSymbol |
	aSymbol := sym == #none ifTrue: [#simple] ifFalse: [sym].
	^ self perform: aSymbol! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 23:52'!
color: aColor width: aNumber
	^self width: aNumber color: aColor! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:01'!
complexAltFramed
	^ComplexBorder style: #complexAltFramed! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:01'!
complexAltInset
	^ComplexBorder style: #complexAltInset! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexAltRaised
	^ComplexBorder style: #complexAltRaised! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexFramed
	^ComplexBorder style: #complexFramed! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexInset
	^ComplexBorder style: #complexInset! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexRaised
	^ComplexBorder style: #complexRaised! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 17:26'!
default
	^Default ifNil:[Default := self new]! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 14:59'!
inset
	^InsetBorder new! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 14:59'!
raised
	^RaisedBorder new! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'sw 11/27/2001 15:22'!
simple
	"Answer a simple border style"

	^ SimpleBorder new! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'rr 6/21/2005 13:50'!
thinGray
	^ self width: 1 color: Color gray! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 16:19'!
width: aNumber
	^self width: aNumber color: Color black! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 16:19'!
width: aNumber color: aColor
	^SimpleBorder new color: aColor; width: aNumber; yourself! !

!BorderStyle methodsFor: 'comparing' stamp: 'ar 8/25/2001 18:38'!
= aBorderStyle
	^self species = aBorderStyle species
		and:[self style == aBorderStyle style
		and:[self width = aBorderStyle width
		and:[self color = aBorderStyle color]]].! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:15'!
baseColor
	^Color transparent! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:15'!
baseColor: aColor
	"Ignored"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
color
	^Color transparent! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
color: aColor
	"Ignored"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 11/26/2001 15:22'!
colorsAtCorners
	^Array new: 4 withAll: self color! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
dotOfSize: diameter forDirection: aDirection
	| form |
	form := Form extent: diameter@diameter depth: Display depth.
	form getCanvas fillOval: form boundingBox color: self color.
	^form! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 17:01'!
drawLineFrom: startPoint to: stopPoint on: aCanvas
	^aCanvas line: startPoint to: stopPoint width: self width color: self color! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 16:24'!
frameOval: aRectangle on: aCanvas
	"Frame the given rectangle on aCanvas"
	aCanvas frameOval: aRectangle width: self width color: self color! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 16:57'!
framePolygon: vertices on: aCanvas
	"Frame the given rectangle on aCanvas"
	self framePolyline: vertices on: aCanvas.
	self drawLineFrom: vertices last to: vertices first on: aCanvas.! !

!BorderStyle methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:59'!
framePolyline: vertices on: aCanvas 
	"Frame the given rectangle on aCanvas"

	| prev next |
	prev := vertices first.
	2 to: vertices size
		do: 
			[:i | 
			next := vertices at: i.
			self 
				drawLineFrom: prev
				to: next
				on: aCanvas.
			prev := next]! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 16:24'!
frameRectangle: aRectangle on: aCanvas
	"Frame the given rectangle on aCanvas"
	aCanvas frameRectangle: aRectangle width: self width color: self color! !

!BorderStyle methodsFor: 'testing' stamp: 'gvc 6/25/2008 12:09'!
hasFillStyle
	"Answer false."

	^false! !

!BorderStyle methodsFor: 'comparing' stamp: 'ar 8/25/2001 16:08'!
hash
	"hash is implemented because #= is implemented"
	^self species hash bitXor: (self width hash bitXor: self color hash)! !

!BorderStyle methodsFor: 'testing' stamp: 'ar 8/25/2001 16:08'!
isBorderStyle
	^true! !

!BorderStyle methodsFor: 'testing' stamp: 'ar 8/26/2001 19:30'!
isComplex
	^false! !

!BorderStyle methodsFor: 'testing' stamp: 'gvc 3/14/2007 10:31'!
isComposite
	"Answer false."

	^false! !

!BorderStyle methodsFor: 'printing' stamp: 'gvc 2/25/2009 15:35'!
printOn: aStream
	"Print a description of the
	receiver on the given stream."

	self storeOn: aStream! !

!BorderStyle methodsFor: 'initialize' stamp: 'ar 8/25/2001 16:06'!
releaseCachedState
	"Release any associated cached state"! !

!BorderStyle methodsFor: 'printing' stamp: 'MarcusDenker 10/21/2013 14:23'!
storeOn: aStream
	"Store a reconstructable representation of the
	receiver on the given stream."

	aStream
		nextPutAll: '(';
		nextPutAll: self class name;
		nextPutAll: ' width: ';
		print: self width;
		nextPutAll: ' color: ';
		print: self color;
		nextPutAll: ')'! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 17:51'!
style
	^#none! !

!BorderStyle methodsFor: 'color tracking' stamp: 'ar 8/25/2001 17:29'!
trackColorFrom: aMorph
	"If necessary, update our color to reflect a change in aMorphs color"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
width
	^0! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
width: aNumber
	"Ignored"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:08'!
widthForRounding
	^self width! !

Object subclass: #Canvas
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Canvas'!

!Canvas commentStamp: '<historical>' prior: 0!
A canvas is a two-dimensional medium on which morphs are drawn in a device-independent manner. Canvases keep track of the origin and clipping rectangle, as well as the underlying drawing medium (such as a window, pixmap, or postscript script).

Subclasses must implement (at least) the following methods:
	* Drawing:
		#fillOval:color:borderWidth:borderColor:
		#frameAndFillRectangle:fillColor:borderWidth:borderColor:
		#drawPolygon:color:borderWidth:borderColor:
		#image:at:sourceRect:rule:
		#stencil:at:sourceRect:rule:
		#line:to:width:color:
		#paragraph:bounds:color:
		#text:bounds:font:color:
	* Support
		#clipBy:during:
		#translateBy:during:
		#translateBy:clippingTo:during:
		#transformBy:clippingTo:during:
!

!Canvas methodsFor: 'converting' stamp: 'ar 8/8/2001 14:22'!
asAlphaBlendingCanvas: alpha
	^(AlphaBlendingCanvas on: self) alpha: alpha! !

!Canvas methodsFor: 'converting' stamp: 'ar 6/24/1999 17:46'!
asShadowDrawingCanvas
	^self asShadowDrawingCanvas: (Color black alpha: 0.5).! !

!Canvas methodsFor: 'converting' stamp: 'ar 8/8/2001 14:14'!
asShadowDrawingCanvas: aColor
	^(ShadowDrawingCanvas on: self) shadowColor: aColor! !

!Canvas methodsFor: 'drawing-support' stamp: 'gm 2/22/2003 14:53'!
cache: aRectangle using: aCache during: aBlock 
	"Cache the execution of aBlock by the given cache.
	Note: At some point we may want to actually *create* the cache here;
		for now we're only using it."

	(aCache notNil 
		and: [(aCache isForm) and: [aCache extent = aRectangle extent]]) 
			ifTrue: [^self paintImage: aCache at: aRectangle origin].
	aBlock value: self! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 02:53'!
clipBy: aRectangle during: aBlock
	"Set a clipping rectangle active only during the execution of aBlock.
	Note: In the future we may want to have more general clip shapes - not just rectangles"
	^self subclassResponsibility! !

!Canvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:07'!
clipRect
	"Return the currently active clipping rectangle"
	^self subclassResponsibility! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/12/2000 18:17'!
contentsOfArea: aRectangle
	"Return the contents of the given area"
	^self contentsOfArea: aRectangle into: (Form extent: aRectangle extent depth: self depth)! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/12/2000 18:17'!
contentsOfArea: aRectangle into: aForm
	"Return the contents of the given area"
	^self subclassResponsibility! !

!Canvas methodsFor: 'copying' stamp: 'ls 3/20/2000 21:24'!
copyClipRect: newClipRect

	^ ClippingCanvas canvas: self clipRect: newClipRect
! !

!Canvas methodsFor: 'accessing' stamp: ''!
depth

	^ Display depth
! !

!Canvas methodsFor: 'testing' stamp: 'di 8/12/2000 15:04'!
doesRoundedCorners 

	^ true! !

!Canvas methodsFor: 'drawing-general' stamp: 'ar 5/29/1999 05:14'!
draw: anObject
	^anObject drawOn: self! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/16/2000 23:45'!
drawImage: aForm at: aPoint
	"Draw the given Form, which is assumed to be a Form or ColorForm"

	self drawImage: aForm
		at: aPoint
		sourceRect: aForm boundingBox! !

!Canvas methodsFor: 'drawing-images' stamp: 'IgorStasenko 7/18/2011 18:08'!
drawImage: aForm at: aPoint sourceRect: sourceRect
	"Draw the given form."

	^self image: aForm
		at: aPoint
		sourceRect: sourceRect
		rule: Form over! !

!Canvas methodsFor: 'drawing-general' stamp: 'StephaneDucasse 2/8/2011 09:10'!
drawMorph: aMorph
	"Draw the receiver morph on the receiver"
	
	"Changed to improve performance. Have seen a 30% improvement."	
	(aMorph fullBounds intersects: self clipRect)
		ifFalse: [^self].
	self draw: aMorph! !

!Canvas methodsFor: 'drawing-polygons' stamp: 'ar 6/18/1999 08:56'!
drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc
	"Draw the given polygon."
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-polygons' stamp: 'ar 6/25/1999 12:18'!
drawPolygon: vertices fillStyle: aFillStyle
	"Fill the given polygon."
	self drawPolygon: vertices fillStyle: aFillStyle borderWidth: 0 borderColor: Color transparent! !

!Canvas methodsFor: 'drawing-polygons' stamp: 'ar 6/18/1999 08:58'!
drawPolygon: vertices fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given polygon.
	Note: The default implementation does not recognize any enhanced fill styles"
	self drawPolygon: vertices color: aFillStyle asColor borderWidth: bw borderColor: bc! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/30/2001 20:35'!
drawString: s at: pt

	^ self drawString: s from: 1 to: s size at: pt font: nil color: Color black! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:25'!
drawString: s at: pt font: aFont color: aColor

	^ self drawString: s from: 1 to: s size at: pt font: aFont color: aColor! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/30/2001 20:36'!
drawString: s from: firstIndex to: lastIndex at: pt font: font color: aColor
	self drawString: s from: firstIndex to: lastIndex in: (pt extent: 10000@10000) font: font color: aColor! !

!Canvas methodsFor: 'drawing-text' stamp: 'tween 3/10/2009 08:07'!
drawString: s from: firstIndex to: lastIndex at: pt font: font color: aColor underline: underline underlineColor: uc strikethrough: strikethrough strikethroughColor: sc
	self drawString: s from: firstIndex to: lastIndex in: (pt extent: 10000@10000) font: font color: aColor underline: underline underlineColor: uc strikethrough: strikethrough strikethroughColor: sc! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/30/2001 20:37'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-text' stamp: 'tween 3/10/2009 07:42'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c underline: underline underlineColor: uc strikethrough: strikethrough strikethroughColor: sc
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:39'!
drawString: s in: boundsRect
	^self drawString: s from: 1 to: s size in: boundsRect font: nil color: Color black! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:13'!
drawString: s in: boundsRect font: fontOrNil color: c
	^self drawString: s from: 1 to: s size in: boundsRect font: fontOrNil color: c! !

!Canvas methodsFor: 'drawing-text' stamp: 'tween 3/10/2009 08:12'!
drawString: s in: boundsRect font: fontOrNil color: c underline: underline underlineColor: uc strikethrough: strikethrough strikethroughColor: sc
	^self drawString: s from: 1 to: s size in: boundsRect font: fontOrNil color: c underline: underline underlineColor: uc strikethrough: strikethrough strikethroughColor: sc! !

!Canvas methodsFor: 'drawing' stamp: 'BenjaminVanRyseghem 5/4/2012 16:47'!
encryptedParagraph: para bounds: bounds color: c
	"Draw the given paragraph"
	^self subclassResponsibility! !

!Canvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:15'!
extent
	"Return the physical extent of the output device"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:18'!
fillColor: aColor
	"Fill the receiver with the given color.
	Note: This method should be named differently since it is intended to fill the background and thus fills even if the color is transparent"
	^self fillRectangle: self clipRect color: (aColor alpha: 1.0).! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
fillOval: r color: c

	self fillOval: r color: c borderWidth: 0 borderColor: Color transparent.
! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
fillOval: r color: c borderWidth: borderWidth borderColor: borderColor
	"Fill the given oval."
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:51'!
fillOval: aRectangle fillStyle: aFillStyle
	"Fill the given oval."
	^self fillOval: aRectangle fillStyle: aFillStyle borderWidth: 0 borderColor: Color transparent! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:50'!
fillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given oval.
	Note: The default implementation does not recognize any enhanced fill styles"
	self fillOval: aRectangle color: aFillStyle asColor borderWidth: bw borderColor: bc! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'gvc 3/21/2008 16:34'!
fillRectangle: aRectangle basicFillStyle: aFillStyle
	"Fill the given rectangle with the given, non-composite, fill style
	Note: The default implementation does not recognize any enhanced fill styles."
	
	self fillRectangle: aRectangle color: aFillStyle asColor.! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:32'!
fillRectangle: r color: c
	"Fill the rectangle using the given color"
	^self 
		frameAndFillRectangle: r
		fillColor: c
		borderWidth: 0
		borderColor: Color transparent! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'StephaneDucasse 2/9/2011 14:36'!
fillRectangle: aRectangle fillStyle: aFillStyle
	"Fill the given rectangle. Double-dispatched via the fill style."
	
	aFillStyle fillRectangle: aRectangle on: self! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'AlexandreBergel 1/29/2013 14:13'!
fillRectangle: aRectangle fillStyle: aFillStyle borderStyle: aBorderStyle
	"Fill the given rectangle."
       aFillStyle isTransparent ifFalse: [
               self fillRectangle: (aRectangle insetBy: aBorderStyle width) fillStyle: aFillStyle ].
       
       (aBorderStyle notNil and: [aBorderStyle width]) >= 0 ifTrue: [ aBorderStyle frameRectangle: aRectangle on: self ].
       ! !

!Canvas methodsFor: 'initialization' stamp: 'ar 5/27/2000 21:50'!
finish
	"If there are any pending operations on the receiver complete them. Do not return before all modifications have taken effect."
	^self flush! !

!Canvas methodsFor: 'initialization' stamp: 'ar 2/9/1999 06:29'!
flush! !

!Canvas methodsFor: 'other' stamp: 'StephaneDucasse 2/9/2011 14:51'!
flushDisplay
	"Empty hook method."! !

!Canvas methodsFor: 'other' stamp: 'StephaneDucasse 2/9/2011 14:51'!
forceToScreen: rect
	"Empty hook method"
! !

!Canvas methodsFor: 'accessing' stamp: 'jm 6/2/1998 06:39'!
form

	^ Display
! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:32'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor
	"Draw the rectangle using the given attributes"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'RAA 8/14/2000 14:22'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor
	"Draw the rectangle using the given attributes.
	Note: This is a *very* simple implementation"
	| bw pt |
	self frameAndFillRectangle: r
		fillColor: fillColor
		borderWidth: borderWidth
		borderColor: bottomRightColor.
	bottomRightColor = topLeftColor ifFalse: [
		bw := borderWidth asPoint.
		pt := r topLeft + (bw // 2).
		self line: pt to: pt + ((r extent x - bw x)@0) width: borderWidth color: topLeftColor.
		self line: pt to: pt + (0@(r extent y - bw y)) width: borderWidth color: topLeftColor.
	].! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
frameOval: r color: c

	self fillOval: r color: Color transparent borderWidth: 1 borderColor: c.
! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
frameOval: r width: w color: c

	self fillOval: r color: Color transparent borderWidth: w borderColor: c.
! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:33'!
frameRectangle: r color: c

	self frameRectangle: r width: 1 color: c.
! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'marcus.denker 8/15/2008 17:43'!
frameRectangle: r width: w color: c
	"Draw a frame around the given rectangle"
	^self frameAndFillRectangle: r
			fillColor: Color transparent
			borderWidth: w
			borderColor: c! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'gvc 9/14/2006 14:33'!
frameRectangle: aRectangle width: width colors: colors dashes: dashes
	"Draw a rectangle with the given width, colors and dash lengths."

	self
		frameRectangle: aRectangle
		width: width
		colors: colors
		dashes: dashes
		offset: self origin! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'gvc 9/18/2006 16:52'!
frameRectangle: aRectangle width: width colors: colors dashes: dashes offset: offset
	"Draw a rectangle with the given width, colors and dash lengths.
	The offset specifies how the coordinate system is translated from the screen origin
	(infinite forms are 0@0 screen based)."

	|o s hf vf c r ds di d os l|
	width < 1 ifTrue: [^self].
	dashes size < 2
		ifTrue: [^self  frameRectangle: aRectangle width: width color: colors first].
	r := aRectangle.
	s := dashes sum * width.
	ds := dashes size.
	di := 1.
	d := (dashes at: di) * width.
	c := colors at: di.
	hf := Form extent: s @ 1 depth: 32.
	r height >= width ifTrue: [
		o := r left + offset x \\ s.
		0 to: s - 1 do: [:x |
			hf colorAt: x + o \\ s  @ 0 put: c.
			d := d - 1.
			d = 0 ifTrue: [
				di := di \\ ds + 1.
				d := (dashes at: di) * width.
				c := colors at: di]].
		os := 0.
		l := r width truncateTo: width.
		self
			fillRectangle: (r topLeft + (os@0) extent: l@width)
			color: (InfiniteForm with: hf)].
	vf := Form extent: 1 @ s depth: 32.
	r width >= width ifTrue: [
		o := r top + offset y + width + (s - (r width \\ s)) \\ s.
		0 to: s - 1 do: [:y |
			vf colorAt: 0 @ (y + o \\ s) put: c.
			d := d - 1.
			d = 0 ifTrue: [
				di := di \\ ds + 1.
				d := (dashes at: di) * width.
				c := colors at: di]].
		os := width - (r width \\ width).
		l := r height - os truncateTo: width.
		self
			fillRectangle: (r topRight + (width negated @ os) extent: width@l)
			color: (InfiniteForm with: vf)].
	r height > width ifTrue: [
		o := r right + offset x - (width * 2) + (r height \\ s) + (r width \\ s) \\ s.
		0 to: s - 1 do: [:x |
			hf colorAt: o + s -1 - x \\ s  @ 0 put: c.
			d := d - 1.
			d = 0 ifTrue: [
				di := di \\ ds + 1.
				d := (dashes at: di) * width.
				c := colors at: di]].
		os := width - (r width \\ width + (r height \\ width) \\ width).
		l := r width - os truncateTo: width.
		os := (r width - os) \\ width.
		self
			fillRectangle: (r bottomLeft + (os @ width negated) extent: l@width)
			color: (InfiniteForm with: hf)].
	r width > width ifTrue: [
		o :=  r top + offset y + (r height * 2 \\ s) + (r width * 2 \\ s) - (width * 3) \\ s.
		0 to: s - 1 do: [:y |
			vf colorAt: 0 @ (o + s -1 - y \\ s) put: c.
			d := d - 1.
			d = 0 ifTrue: [
				di := di \\ ds + 1.
				d := (dashes at: di) * width.
				c := colors at: di]].
		l := r height - (2 * width) + os.
		os := width.
		self
			fillRectangle: (r topLeft + (0@os) extent: width@l)
			color: (InfiniteForm with: vf)]! !

!Canvas methodsFor: 'drawing-general' stamp: ''!
fullDraw: anObject
	^anObject fullDrawOn: self! !

!Canvas methodsFor: 'drawing-general' stamp: 'StephaneDucasse 2/8/2011 09:30'!
fullDrawMorph: aMorph
	"Hook method for potential other canvases. In the core, this method looks supefluous but PostscriptCanvases and other canvases can specialized it for Morph rendering. Therefore it should not be merged with fullDraw:."

	self fullDraw: aMorph! !

!Canvas methodsFor: 'private' stamp: 'ar 2/12/2000 18:12'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Note: The public use of this protocol is deprecated. It will become private. Nobody in the outside world must assume that a thing like a combination rule has any specific effect."
	^self subclassResponsibility! !

!Canvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:21'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule alpha: sourceAlpha
	"Privately used for blending forms w/ constant alpha. Fall back to simpler case by defaul."
	^self image: aForm at: aPoint sourceRect: sourceRect rule: rule! !

!Canvas methodsFor: 'testing' stamp: 'ar 11/13/1998 13:19'!
isBalloonCanvas
	^false! !

!Canvas methodsFor: 'testing' stamp: 'IgorStasenko 7/18/2011 18:53'!
isShadowDrawing
	^false! !

!Canvas methodsFor: 'testing' stamp: 'ar 6/22/1999 14:10'!
isVisible: aRectangle
	"Return true if the given rectangle is (partially) visible"
	^self clipRect intersects: aRectangle
! !

!Canvas methodsFor: 'drawing' stamp: 'jm 8/2/97 13:54'!
line: pt1 to: pt2 color: c

	self line: pt1 to: pt2 width: 1 color: c.
! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:31'!
line: pt1 to: pt2 width: w color: c
	"Draw a line using the given width and color"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing' stamp: 'aoy 2/15/2003 21:41'!
line: pt1 to: pt2 width: width color: color1 dashLength: s1 secondColor: color2 secondDashLength: s2 startingOffset: startingOffset 
	"Draw a line using the given width, colors and dash lengths.
	Originally written by Stephan Rudlof; tweaked by Dan Ingalls
	to use startingOffset for sliding offset as in 'ants' animations.
	Returns the sum of the starting offset and the length of this line."

	| dist deltaBig colors nextPhase segmentOffset phase segmentLength startPoint distDone endPoint segLens |
	dist := pt1 dist: pt2.
	dist = 0 ifTrue: [^startingOffset].
	s1 = 0 & (s2 = 0) ifTrue: [^startingOffset].
	deltaBig := pt2 - pt1.
	colors := { 
				color1.
				color2}.
	segLens := { 
				s1 asFloat.
				s2 asFloat}.
	nextPhase := { 
				2.
				1}.

	"Figure out what phase we are in and how far, given startingOffset."
	segmentOffset := startingOffset \\ (s1 + s2).
	segmentLength := segmentOffset < s1 
		ifTrue: 
			[phase := 1.
			s1 - segmentOffset]
		ifFalse: 
			[phase := 2.
			 s1 + s2 - segmentOffset].
	startPoint := pt1.
	distDone := 0.0.
	[distDone < dist] whileTrue: 
			[segmentLength := segmentLength min: dist - distDone.
			endPoint := startPoint + (deltaBig * segmentLength / dist).
			self 
				line: startPoint truncated
				to: endPoint truncated
				width: width
				color: (colors at: phase).
			distDone := distDone + segmentLength.
			phase := nextPhase at: phase.
			startPoint := endPoint.
			segmentLength := segLens at: phase].
	^startingOffset + dist! !

!Canvas methodsFor: 'drawing' stamp: 'sr 4/27/2000 03:31'!
line: pt1 to: pt2 width: w1 color: c1 stepWidth: s1 secondWidth: w2 secondColor: c2 secondStepWidth: s2 
	"Draw a line using the given width, colors and steps; both steps can  
	have different stepWidths (firstStep, secondStep), draw widths and  
	colors."
	| bigSteps offsetPoint dist p1p2Vec deltaBig delta1 delta2 lastPoint bigStep |
	s1 = 0 & (s2 = 0) ifTrue: [^ self].
	dist := pt1 dist: pt2.
	dist = 0 ifTrue: [^ self].
	bigStep := s1 + s2.
	bigSteps := dist / bigStep.
	p1p2Vec := pt2 - pt1.
	deltaBig := p1p2Vec / bigSteps.
	delta1 := deltaBig * (s1 / bigStep).
	delta2 := deltaBig * (s2 / bigStep).
	dist <= s1
		ifTrue: 
			[self
				line: pt1 rounded
				to: pt2 rounded
				width: w1
				color: c1.
			^ self].
	0 to: bigSteps truncated - 1 do: 
		[:bigStepIx | 
		self
			line: (pt1 + (offsetPoint := deltaBig * bigStepIx)) rounded
			to: (pt1 + (offsetPoint := offsetPoint + delta1)) rounded
			width: w1
			color: c1.
		self
			line: (pt1 + offsetPoint) rounded
			to: (pt1 + (offsetPoint + delta2)) rounded
			width: w2
			color: c2].
	"if there was no loop, offsetPoint is nil"
	lastPoint := pt1 + ((offsetPoint ifNil: [0 @ 0])
					+ delta2).
	(lastPoint dist: pt2)
		<= s1
		ifTrue: [self
				line: lastPoint rounded
				to: pt2 rounded
				width: w1
				color: c1]
		ifFalse: 
			[self
				line: lastPoint rounded
				to: (lastPoint + delta1) rounded
				width: w1
				color: c1.
			self
				line: (lastPoint + delta1) rounded
				to: pt2
				width: w1
				color: c2]! !

!Canvas methodsFor: 'drawing' stamp: 'gvc 6/17/2006 10:42'!
line: pt1 to: pt2 width: width colors: colors dashes: dashes startingOffset: startingOffset 
	"Draw a line using the given width, colors and dash lengths.
	Dash lengths are considered as multiples of width."

	|dist deltaBig segmentOffset phase segmentLength startPoint distDone endPoint segLens lens l ep|
	width = 0 ifTrue: [^startingOffset].
	dist := pt1 dist: pt2.
	dist = 0 ifTrue: [^startingOffset].
	(dashes allSatisfy: [:d | d = 0]) ifTrue: [^startingOffset].
	deltaBig := pt2 - pt1.
	segLens := dashes collect: [:d | d * width].
	
	"Figure out what phase we are in and how far, given startingOffset."
	segmentOffset := startingOffset \\ segLens sum.
	lens := segLens readStream.
	l := 0.
	[lens atEnd or: [segmentOffset <= (l := l + lens next)]] whileFalse: [].
	segmentLength := lens atEnd
		ifTrue: [phase := segLens size.
				segLens sum - segmentOffset]
		ifFalse: [phase := lens position.
				(segLens first: phase) sum - segmentOffset.].
	startPoint := pt1.
	distDone := 0.0.
	segmentLength < (segLens at: phase)
		ifTrue: [startPoint := startPoint + (deltaBig * segmentLength / dist).
				distDone := distDone + segmentLength.
				phase := phase \\ segLens size + 1.
				segmentLength :=  (segLens at: phase)].
	[distDone < dist] whileTrue: 
			[segmentLength := segmentLength min: dist - distDone.
			endPoint := startPoint + (deltaBig * segmentLength / dist).
			ep := startPoint + (deltaBig * (segmentLength - width max: 0) / dist).
			self 
				line: startPoint truncated
				to: ep truncated
				width: width
				color: (colors at: phase).
			distDone := distDone + segmentLength.
			phase := phase \\ segLens size + 1.
			startPoint := endPoint.
			segmentLength := segLens at: phase].
	^startingOffset + distDone! !

!Canvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:11'!
origin
	"Return the current origin for drawing operations"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/16/2000 23:48'!
paintImage: aForm at: aPoint
	"Draw the given Form, which is assumed to be a Form or ColorForm following the convention that zero is the transparent pixel value."

	self paintImage: aForm
		at: aPoint
		sourceRect: aForm boundingBox
! !

!Canvas methodsFor: 'drawing-images' stamp: 'IgorStasenko 7/18/2011 18:08'!
paintImage: aForm at: aPoint sourceRect: sourceRect
	"Draw the given Form, which is assumed to be a Form or ColorForm following the convention that zero is the transparent pixel value."

	^self image: aForm
		at: aPoint
		sourceRect: sourceRect
		rule: Form paint! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:31'!
paragraph: paragraph bounds: bounds color: c
	"Draw the given paragraph"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 01:43'!
preserveStateDuring: aBlock
	"Preserve the full canvas state during the execution of aBlock"
	^aBlock value: self copy! !

!Canvas methodsFor: 'drawing' stamp: 'ar 2/5/1999 18:28'!
render: anObject
	"Do some 3D operations with the object if possible"! !

!Canvas methodsFor: 'initialization' stamp: 'MarcusDenker 10/21/2013 14:21'!
reset
	"Reset the canvas."

	self initWithTarget:self class defaultTarget.
! !

!Canvas methodsFor: 'drawing-general' stamp: 'ar 12/30/2001 18:47'!
roundCornersOf: aMorph during: aBlock
	^self roundCornersOf: aMorph in: aMorph bounds during: aBlock! !

!Canvas methodsFor: 'drawing-general' stamp: 'ar 12/30/2001 18:47'!
roundCornersOf: aMorph in: bounds during: aBlock
	^aBlock value! !

!Canvas methodsFor: 'drawing-general' stamp: 'GaryChambers 9/8/2011 14:47'!
roundShadowCornersOf: aMorph during: aBlock
	^self roundShadowCornersOf: aMorph in: aMorph bounds during: aBlock! !

!Canvas methodsFor: 'drawing-general' stamp: 'GaryChambers 9/8/2011 14:47'!
roundShadowCornersOf: aMorph in: bounds during: aBlock
	^aBlock value! !

!Canvas methodsFor: 'testing' stamp: 'di 9/24/2000 16:10'!
seesNothingOutside: aRectangle
	"Return true if this canvas will not touch anything outside aRectangle"
	^ aRectangle containsRect: self clipRect
! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:46'!
shadowColor
	"Return the current override color or nil if no such color exists"
	^nil! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:46'!
shadowColor: aColor
	"Set a shadow color. If set this color overrides any client-supplied color."! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 6/25/1999 12:17'!
stencil: stencilForm at: aPoint color: aColor
	"Flood this canvas with aColor wherever stencilForm has non-zero pixels"
	^self stencil: stencilForm
		at: aPoint
		sourceRect: stencilForm boundingBox
		color: aColor! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 6/25/1999 12:17'!
stencil: stencilForm at: aPoint sourceRect: sourceRect color: aColor
	"Flood this canvas with aColor wherever stencilForm has non-zero pixels"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-support' stamp: 'di 10/16/1999 16:02'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock
	"Transform the receiver by the given display transformation during the execution of aBlock. The given clip rectangle defines the *global* (e.g., outer) rectangle against which the receiver should clip (which would be equivalent to 'self clipRect: aClipRect; transformBy: aDisplayTransform')."
	^ self transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: 1
! !

!Canvas methodsFor: 'drawing-support' stamp: 'di 10/16/1999 15:56'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: cellSize
	"Transform the receiver by the given display transformation during the execution of aBlock. The given clip rectangle defines the *global* (e.g., outer) rectangle against which the receiver should clip (which would be equivalent to 'self clipRect: aClipRect; transformBy: aDisplayTransform')."
	^ self subclassResponsibility! !

!Canvas methodsFor: 'other' stamp: ''!
translateBy:aPoint clippingTo:aRect during:aBlock
	^aBlock value:(self copyOffset:aPoint clipRect:aRect).! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 03:00'!
translateBy: delta during: aBlock
	"Set a translation only during the execution of aBlock."
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 14:08'!
translateTo: newOrigin clippingTo: aRectangle during: aBlock
	"Set a new origin and clipping rectangle only during the execution of aBlock."
	self translateBy: newOrigin - self origin 
		clippingTo: (aRectangle translateBy: self origin negated) 
		during: aBlock! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/17/2000 14:05'!
translucentImage: aForm at: aPoint
	"Draw a translucent image using the best available way of representing translucency."
	self translucentImage: aForm
		at: aPoint
		sourceRect: aForm boundingBox! !

!Canvas methodsFor: 'drawing-images' stamp: 'IgorStasenko 7/18/2011 18:08'!
translucentImage: aForm at: aPoint sourceRect: sourceRect
	"Draw a translucent image using the best available way of representing translucency.
	Note: This will be fixed in the future."

	(self depth < 32 or:[aForm isTranslucent not]) 
		ifTrue:[^self paintImage: aForm at: aPoint sourceRect: sourceRect].
	self image: aForm
		at: aPoint
		sourceRect: sourceRect
		rule: Form blend! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 12/28/2001 23:44'!
warpImage: aForm transform: aTransform
	"Warp the given form using aTransform"
	^self warpImage: aForm transform: aTransform at: 0@0! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 12/28/2001 23:54'!
warpImage: aForm transform: aTransform at: extraOffset
	"Warp the given form using aTransform.
	TODO: Use transform to figure out appropriate cell size"
	^self warpImage: aForm transform: aTransform at: extraOffset sourceRect: aForm boundingBox cellSize: 1! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 12/29/2001 00:20'!
warpImage: aForm transform: aTransform at: extraOffset sourceRect: sourceRect cellSize: cellSize
	"Warp the given using the appropriate transform and offset."
	^self subclassResponsibility! !

Canvas subclass: #ColorMappingCanvas
	instanceVariableNames: 'myCanvas'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Canvas'!

!ColorMappingCanvas commentStamp: 'LaurentLaffont 2/23/2011 20:17' prior: 0!
I'm an abstract class which introduce a filter between a drawing request and the final output, handled by #mapColor:

For each potential pixel operation like: source -> op -> output
it introducing a color mapping stage: source -> op -> mapping -> output

Then #mapColor: can be redefined in subclasses to implement more specific behavior. For example:
- ShadowDrawingCanvas will replace the rendered color with the color of shadow (if not transparent).
- AlphaBlendingCanvas will add or intensify alpha of rendered color.

For an easy to understand example see  #drawPolygon:color:borderWidth:borderColor:!

!ColorMappingCanvas class methodsFor: 'instance creation' stamp: 'ar 6/22/1999 18:23'!
on: aCanvas
	^self new on: aCanvas! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 18:19'!
clipBy: aRectangle during: aBlock
	"Set a clipping rectangle active only during the execution of aBlock.
	Note: In the future we may want to have more general clip shapes - not just rectangles"
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas clipBy: aRectangle during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:40'!
clipRect
	^myCanvas clipRect! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:39'!
depth
	^myCanvas depth! !

!ColorMappingCanvas methodsFor: 'drawing-polygons' stamp: 'mir 9/12/2001 14:24'!
drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc
	"Draw the given polygon."
	^myCanvas
		drawPolygon: vertices
		color: aColor
		borderWidth: bw
		borderColor: (self mapColor: bc)! !

!ColorMappingCanvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:28'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c
	"Draw the given string in the given font and color clipped to the given rectangle. If the font is nil, the default font is used."
	myCanvas
		drawString: s from: firstIndex to: lastIndex 
		in: boundsRect 
		font: fontOrNil 
		color: (self mapColor: c)! !

!ColorMappingCanvas methodsFor: 'drawing-text' stamp: 'tween 3/10/2009 07:45'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c underline: underline underlineColor: uc strikethrough: strikethrough strikethroughColor: sc
	"Draw the given string in the given font and color clipped to the given rectangle. If the font is nil, the default font is used."
	myCanvas
		drawString: s 
		from: firstIndex 
		to: lastIndex 
		in: boundsRect 
		font: fontOrNil 
		color: (self mapColor: c) 
		underline: underline 
		underlineColor: (self mapColor: uc) 
		strikethrough: strikethrough 
		strikethroughColor: (self mapColor: sc)! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:39'!
extent
	^myCanvas extent! !

!ColorMappingCanvas methodsFor: 'drawing-ovals' stamp: 'ar 6/22/1999 17:59'!
fillOval: r color: c borderWidth: borderWidth borderColor: borderColor
	"Fill the given oval."
	myCanvas
		fillOval: r
		color: (self mapColor: c)
		borderWidth: borderWidth
		borderColor: (self mapColor: borderColor)! !

!ColorMappingCanvas methodsFor: 'initialization' stamp: 'ar 6/22/1999 18:24'!
flush
	myCanvas flush.! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/24/1999 17:54'!
form
	^myCanvas form! !

!ColorMappingCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/22/1999 17:59'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor
	"Draw the rectangle using the given attributes"
	myCanvas
		frameAndFillRectangle: r
		fillColor: (self mapColor: fillColor)
		borderWidth: borderWidth
		borderColor: (self mapColor: borderColor)! !

!ColorMappingCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/22/1999 18:01'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor
	"Draw the rectangle using the given attributes"
	myCanvas
		frameAndFillRectangle: r 
		fillColor: (self mapColor: fillColor) 
		borderWidth: borderWidth 
		topLeftColor: (self mapColor: topLeftColor)
		bottomRightColor: (self mapColor: bottomRightColor)! !

!ColorMappingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:15'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Draw the given form. For the 'paint' combination rule use stenciling otherwise simply fill the source rectangle."
	^myCanvas image: aForm at: aPoint sourceRect: sourceRect rule: rule.! !

!ColorMappingCanvas methodsFor: 'testing' stamp: 'ar 8/8/2001 14:16'!
isShadowDrawing
	^myCanvas isShadowDrawing! !

!ColorMappingCanvas methodsFor: 'drawing' stamp: 'ar 6/22/1999 18:15'!
line: pt1 to: pt2 width: w color: c
	"Draw a line using the given width and color"
	myCanvas
		line: pt1
		to: pt2
		width: w
		color: (self mapColor: c).! !

!ColorMappingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:15'!
mapColor: aColor
	^aColor! !

!ColorMappingCanvas methodsFor: 'initialization' stamp: 'ar 8/8/2001 14:14'!
on: aCanvas
	myCanvas := aCanvas.! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:39'!
origin
	^myCanvas origin! !

!ColorMappingCanvas methodsFor: 'drawing' stamp: 'ar 6/22/1999 18:16'!
paragraph: paragraph bounds: bounds color: c
	"Draw the given paragraph"
	myCanvas
		paragraph: paragraph
		bounds: bounds
		color: (self mapColor: c)! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'nice 12/26/2009 19:22'!
preserveStateDuring: aBlock
	"Preserve the full canvas state during the execution of aBlock"
	| oldCanvas result |
	oldCanvas := myCanvas.
	result := myCanvas preserveStateDuring:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.
	^result! !

!ColorMappingCanvas methodsFor: 'initialization' stamp: 'ar 6/22/1999 18:23'!
reset
	myCanvas reset.! !

!ColorMappingCanvas methodsFor: 'drawing-images' stamp: 'ar 6/24/1999 18:26'!
stencil: aForm at: aPoint color: aColor
	myCanvas
		stencil: aForm
		at: aPoint
		color: (self mapColor: aColor)! !

!ColorMappingCanvas methodsFor: 'drawing-images' stamp: 'ar 6/24/1999 18:26'!
stencil: aForm at: aPoint sourceRect: aRect color: aColor
	myCanvas
		stencil: aForm
		at: aPoint
		sourceRect: aRect
		color: (self mapColor: aColor)! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'di 10/16/1999 16:01'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock	 smoothing: cellSize

	"Transform the receiver by the given display transformation during the execution of aBlock. The given clip rectangle defines the *global* (e.g., outer) rectangle against which the receiver should clip (which would be equivalent to 'self clipRect: aClipRect; transformBy: aDisplayTransform')."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas transformBy: aDisplayTransform
		clippingTo: aClipRect
		during: [:newCanvas |
				myCanvas := newCanvas.
				aBlock value: self]
		smoothing: cellSize.
	myCanvas := oldCanvas.! !

!ColorMappingCanvas methodsFor: 'other' stamp: 'ar 6/22/1999 18:21'!
translateBy: delta clippingTo: aRectangle during: aBlock
	"Set a translation and clipping rectangle only during the execution of aBlock."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas translateBy: delta clippingTo: aRectangle during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 18:22'!
translateBy: delta during: aBlock
	"Set a translation only during the execution of aBlock."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas translateBy: delta during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 18:22'!
translateTo: newOrigin clippingTo: aRectangle during: aBlock
	"Set a new origin and clipping rectangle only during the execution of aBlock."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas translateTo: newOrigin clippingTo: aRectangle during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.! !

ColorMappingCanvas subclass: #ShadowDrawingCanvas
	instanceVariableNames: 'shadowColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Canvas'!

!ShadowDrawingCanvas commentStamp: 'LaurentLaffont 2/23/2011 20:17' prior: 0!
see ColorMappingCanvas comment.!

!ShadowDrawingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:13'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Draw the given form. For the 'paint' combination rule use stenciling otherwise simply fill the source rectangle."
	rule = Form paint ifTrue:[
		^myCanvas
			stencil: aForm
			at: aPoint
			sourceRect: sourceRect
			color: shadowColor
	] ifFalse:[
		^myCanvas
			fillRectangle: (sourceRect translateBy: aPoint)
			color: shadowColor
	].! !

!ShadowDrawingCanvas methodsFor: 'testing' stamp: 'IgorStasenko 7/18/2011 18:53'!
isShadowDrawing
	^true! !

!ShadowDrawingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:14'!
mapColor: aColor
	aColor isColor ifFalse:[^aColor]. "Should not happen but who knows..."
	^aColor isTransparent
		ifTrue:[aColor]
		ifFalse:[shadowColor]! !

!ShadowDrawingCanvas methodsFor: 'initialization' stamp: 'ar 8/8/2001 14:14'!
on: aCanvas
	myCanvas := aCanvas.
	shadowColor := Color black.! !

!ShadowDrawingCanvas methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:14'!
shadowColor
	^shadowColor! !

!ShadowDrawingCanvas methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:14'!
shadowColor: aColor
	shadowColor := aColor! !

Canvas subclass: #FormCanvas
	instanceVariableNames: 'origin clipRect form port'
	classVariableNames: 'TranslucentPatterns'
	poolDictionaries: ''
	category: 'Morphic-Core-Canvas'!

!FormCanvas commentStamp: '<historical>' prior: 0!
Note that when shadowDrawing is true, shadowStipple may be either a color, for a solid shadow of the given color, or it may be a stipple used to simulate gray shading when the display cannot support alpha blending.!

!FormCanvas class methodsFor: 'instance creation' stamp: ''!
extent: aPoint

	^ self extent: aPoint depth: Display depth
! !

!FormCanvas class methodsFor: 'instance creation' stamp: ''!
extent: extent depth: depth

	^ self new setForm: (Form extent: extent depth: depth)! !

!FormCanvas class methodsFor: 'instance creation' stamp: 'nk 7/4/2003 10:11'!
extent: extent depth: depth origin: aPoint clipRect: aRectangle

	^ self new
		setForm: (Form extent: extent depth: depth);
		setOrigin: aPoint clipRect: aRectangle;
		yourself! !

!FormCanvas class methodsFor: 'caching' stamp: 'FernandoOlivero 9/12/2013 13:35'!
initializeTranslucentPatterns
	
	TranslucentPatterns := Array new: 8. 
	#(1 2 4 8 ) do: 
		[ :d | | mask bits pattern patternList | 
		patternList := Array new: 5.
		mask := (1 bitShift: d) - 1.
		bits := 2 * d.
		[ bits >= 32 ] whileFalse: 
			[ mask := mask bitOr: (mask bitShift: bits).	"double the length of mask"
			bits := bits + bits ].
		"0% pattern"
		pattern := Bitmap 
			with: 0
			with: 0.
		patternList 
			at: 1
			put: pattern.
		"25% pattern"
		pattern := Bitmap 
			with: mask
			with: 0.
		patternList 
			at: 2
			put: pattern.
		"50% pattern"
		pattern := Bitmap 
			with: mask
			with: mask bitInvert32.
		patternList 
			at: 3
			put: pattern.
		"75% pattern"
		pattern := Bitmap 
			with: mask
			with: 4294967295.
		patternList 
			at: 4
			put: pattern.
		"100% pattern"
		pattern := Bitmap 
			with: 4294967295
			with: 4294967295.
		patternList 
			at: 5
			put: pattern.
		TranslucentPatterns 
			at: d
			put: patternList ]! !

!FormCanvas class methodsFor: 'instance creation' stamp: 'jm 8/2/97 13:54'!
on: aForm

	^ self new setForm: aForm
! !

!FormCanvas class methodsFor: 'caching' stamp: 'FernandoOlivero 9/12/2013 13:36'!
translucentPatterns
	TranslucentPatterns isNil ifTrue:[ self initializeTranslucentPatterns ].
	^ 	TranslucentPatterns  ! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 5/28/2000 17:11'!
allocateForm: extentPoint
	"Allocate a new form which is similar to the receiver"
	^form allocateForm: extentPoint! !

!FormCanvas methodsFor: 'Morphic-Base-Balloon' stamp: 'ar 11/11/1998 22:57'!
asBalloonCanvas
	^(BalloonCanvas on: form) setOrigin: origin clipRect: clipRect! !

!FormCanvas methodsFor: 'drawing-ovals' stamp: 'RAA 11/6/2000 15:21'!
balloonFillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc

	self asBalloonCanvas
		fillOval: aRectangle
		fillStyle: aFillStyle
		borderWidth: bw
		borderColor: bc! !

!FormCanvas methodsFor: 'Morphic-Base-Balloon' stamp: 'gvc 3/16/2009 13:36'!
balloonFillRectangle: aRectangle fillStyle: aFillStyle

	self asBalloonCanvas fillRectangle: aRectangle basicFillStyle: aFillStyle! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 03:02'!
clipBy: aRectangle during: aBlock
	"Set a clipping rectangle active only during the execution of aBlock.
	Note: In the future we may want to have more general clip shapes - not just rectangles"
	^aBlock value: (self copyClipRect: aRectangle)! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:06'!
clipRect
	"Return the currently active clipping rectangle"
	^ clipRect translateBy: origin negated! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 12/31/2001 03:26'!
contentsOfArea: aRectangle into: aForm
	| bb |
	self flush.
	bb := BitBlt toForm: aForm.
	bb sourceForm: form; combinationRule: Form over;
		sourceX: (aRectangle left + origin x); sourceY: (aRectangle top + origin y);
		width: aRectangle width; height: aRectangle height;
		copyBits.
	^aForm! !

!FormCanvas methodsFor: 'copying' stamp: 'ar 6/17/1999 02:51'!
copyClipRect: aRectangle
	^ self copyOrigin: origin clipRect: (aRectangle translateBy: origin)
! !

!FormCanvas methodsFor: 'copying' stamp: 'ar 6/17/1999 02:52'!
copyOffset: aPoint
	^ self copyOrigin: origin + aPoint clipRect: clipRect! !

!FormCanvas methodsFor: 'copying' stamp: 'IgorStasenko 12/22/2012 03:26'!
copyOffset: aPoint clipRect: sourceClip
	"Make a copy of me offset by aPoint, and further clipped
	by sourceClip, a rectangle in the un-offset coordinates"
	^ self copyOrigin: aPoint + origin
		clipRect: ((sourceClip translateBy: origin) intersect: clipRect ifNone: [ 0@0 corner: 0@0 ])! !

!FormCanvas methodsFor: 'copying' stamp: 'IgorStasenko 12/22/2012 03:26'!
copyOrigin: aPoint clipRect: aRectangle
	"Return a copy of this canvas with the given origin. The clipping rectangle of this canvas is the intersection of the given rectangle and the receiver's current clipping rectangle. This allows the clipping rectangles of nested clipping morphs to be composed."
	^ self copy
		setOrigin: aPoint
		clipRect: (clipRect intersect: aRectangle ifNone: ["well, now we will clip everything" 0@0 corner: 0@0])! !

!FormCanvas methodsFor: 'accessing' stamp: ''!
depth

	^ form depth
! !

!FormCanvas methodsFor: 'drawing-polygons' stamp: 'ar 6/18/1999 08:57'!
drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc
	"Generalize for the BalloonCanvas"
	^self drawPolygon: vertices fillStyle: aColor borderWidth: bw borderColor: bc! !

!FormCanvas methodsFor: 'drawing-polygons' stamp: 'IgorStasenko 7/18/2011 18:11'!
drawPolygon: vertices fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Use a BalloonCanvas"
	self asBalloonCanvas 
		drawPolygon: vertices asArray
		fillStyle: aFillStyle
		borderWidth: bw 
		borderColor: bc! !

!FormCanvas methodsFor: 'drawing-text' stamp: 'IgorStasenko 7/18/2011 18:12'!
drawString: aString from: firstIndex to: lastIndex at: aPoint font: fontOrNil color: c
	| font |
	port colorMap: nil.
	font := fontOrNil ifNil: [TextStyle defaultFont].
	port combinationRule: Form paint.
	font installOn: port
		foregroundColor: c
		backgroundColor: Color transparent.
	font displayString: aString on: port 
		from: firstIndex to: lastIndex at: (origin + aPoint) kern: 0.! !

!FormCanvas methodsFor: 'drawing-text' stamp: 'IgorStasenko 7/18/2011 18:12'!
drawString: aString from: firstIndex to: lastIndex autoBoundAt: aPoint font: fontOrNil color: c

	| font textStyle portRect bounds character width carriageReturn lineWidth lineHeight |

	carriageReturn := Character cr.
	width := lineWidth := 0.
	font := StandardFonts codeFont.
	textStyle := font textStyle.
	lineHeight := textStyle lineGrid.
	1 to: aString size do: 
		[:i | 
		character := aString at: i.
		character = carriageReturn
		  ifTrue: 
			[lineWidth := lineWidth max: width.
			lineHeight := lineHeight + textStyle lineGrid.
			width := 0]
		  ifFalse: [width := width + (font widthOf: character)]].
	lineWidth := lineWidth max: width.
	bounds := aPoint extent: (lineWidth  @ lineHeight).
	
	port colorMap: nil.
	portRect := port clipRect.
	port clipByX1: bounds left + origin x 
		y1: bounds top + origin y 
		x2: bounds right + origin x 
		y2: bounds bottom + origin y.
	port combinationRule: Form paint.
	port fill: bounds fillColor: Color white rule: Form paint.
	font installOn: port
		foregroundColor: c
		backgroundColor: Color white.
	aString lines doWithIndex: [:line :index |
		font displayString: line on: port 
			from: 1 to: line size at: (bounds topLeft + origin + (0@(index-1)*textStyle lineGrid)) kern: 0.
	].
	port clipRect: portRect.! !

!FormCanvas methodsFor: 'drawing-text' stamp: 'IgorStasenko 7/18/2011 18:12'!
drawString: aString from: firstIndex to: lastIndex in: bounds font: fontOrNil color: c
	| font portRect |
	port colorMap: nil.
	portRect := port clipRect.
	port clipByX1: bounds left + origin x 
		y1: bounds top + origin y 
		x2: bounds right + origin x 
		y2: bounds bottom + origin y.
	font := fontOrNil ifNil: [TextStyle defaultFont].
	port combinationRule: Form paint.
	font installOn: port
		foregroundColor: c
		backgroundColor: Color transparent.
	font displayString: aString asString on: port 
		from: firstIndex to: lastIndex at: (bounds topLeft + origin) kern: 0.
	port clipRect: portRect.! !

!FormCanvas methodsFor: 'drawing-text' stamp: 'IgorStasenko 7/18/2011 18:12'!
drawString: aString from: firstIndex to: lastIndex in: bounds font: fontOrNil color: c underline: underline underlineColor: uc strikethrough: strikethrough strikethroughColor: sc
	| font portRect endPoint |
	port colorMap: nil.
	portRect := port clipRect.
	port clipByX1: bounds left + origin x 
		y1: bounds top + origin y 
		x2: bounds right + origin x 
		y2: bounds bottom + origin y.
	font := fontOrNil ifNil: [TextStyle defaultFont].
	port combinationRule: Form paint.
	font installOn: port
		foregroundColor: c
		backgroundColor: Color transparent.
	endPoint := font displayString: aString asString on: port 
		from: firstIndex to: lastIndex at: (bounds topLeft + origin) kern: 0.
	underline ifTrue:[
		font installOn: port
			foregroundColor: uc
			backgroundColor: Color transparent.
		font displayUnderlineOn: port from: (bounds topLeft + origin + (0@font ascent)) to: endPoint.	
		].
	port clipRect: portRect.! !

!FormCanvas methodsFor: 'accessing' stamp: ''!
extent

	^ form extent! !

!FormCanvas methodsFor: 'drawing' stamp: 'ar 5/14/2000 15:50'!
fillColor: c
	"Note: This always fills, even if the color is transparent."
	self setClearColor: c.
	port fillRect: form boundingBox offset: origin.! !

!FormCanvas methodsFor: 'drawing-ovals' stamp: 'di 5/25/2001 01:40'!
fillOval: r color: fillColor borderWidth: borderWidth borderColor: borderColor
	| rect |
	"draw the border of the oval"
	rect := (r translateBy: origin) truncated.
	(borderWidth = 0 or: [borderColor isTransparent]) ifFalse:[
		self setFillColor: borderColor.
		(r area > 10000 or: [fillColor isTranslucent]) 
			ifTrue: [port frameOval: rect borderWidth: borderWidth]
			ifFalse: [port fillOval: rect]]. "faster this way"
	"fill the inside"
	fillColor isTransparent ifFalse:
		[self setFillColor: fillColor.
		port fillOval: (rect insetBy: borderWidth)].
! !

!FormCanvas methodsFor: 'drawing-ovals' stamp: 'IgorStasenko 7/18/2011 18:13'!
fillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given oval."

	self flag: #bob.		"this and its siblings could be moved up to Canvas with the
						right #balloonFillOval:..."

	(aFillStyle isBitmapFill and:[aFillStyle isKindOf: InfiniteForm]) ifTrue:[
		self flag: #fixThis.
		^self fillOval: aRectangle color: aFillStyle borderWidth: bw borderColor: bc].
	(aFillStyle isSolidFill) ifTrue:[
		^self fillOval: aRectangle color: aFillStyle asColor borderWidth: bw borderColor: bc].
	"Use a BalloonCanvas instead"
	self balloonFillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'IgorStasenko 7/18/2011 18:14'!
fillRectangle: aRectangle basicFillStyle: aFillStyle
	"Fill the given rectangle with the given, non-composite, fill style."
	
	| pattern |

	(aFillStyle isKindOf: InfiniteForm) ifTrue: [
		^self infiniteFillRectangle: aRectangle fillStyle: aFillStyle
	].

	(aFillStyle isSolidFill) 
		ifTrue:[^self fillRectangle: aRectangle color: aFillStyle asColor].
	"We have a very special case for filling with infinite forms"
	(aFillStyle isBitmapFill and:[aFillStyle origin = (0@0)]) ifTrue:[
		pattern := aFillStyle form.
		(aFillStyle direction = (pattern width @ 0) 
			and:[aFillStyle normal = (0@pattern height)]) ifTrue:[
				"Can use an InfiniteForm"
				^self fillRectangle: aRectangle color: (InfiniteForm with: pattern)].
	].
	"Use a BalloonCanvas instead"
	self balloonFillRectangle: aRectangle fillStyle: aFillStyle.! !

!FormCanvas methodsFor: 'initialization' stamp: 'ar 5/27/2000 21:51'!
finish
	"If there are any pending operations on the receiver complete them. Do not return before all modifications have taken effect."
	form finish! !

!FormCanvas methodsFor: 'other' stamp: ''!
flushDisplay
		Display deferUpdates: false; forceDisplayUpdate.! !

!FormCanvas methodsFor: 'other' stamp: 'StephaneDucasse 2/9/2011 14:51'!
forceToScreen: rect
	^Display forceToScreen: rect.
! !

!FormCanvas methodsFor: 'accessing' stamp: ''!
form

	^ form! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 5/14/2000 15:50'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor
	| rect |
	rect := r translateBy: origin.
	"draw the border of the rectangle"
	borderColor isTransparent ifFalse:[
		self setFillColor: borderColor.
		(r area > 10000 or: [fillColor isTranslucent]) ifTrue: [
			port frameRect: rect borderWidth: borderWidth.
		] ifFalse: ["for small rectangles, it's faster to fill the entire outer rectangle
					than to compute and fill the border rects"
					port fillRect: rect offset: origin]].

	"fill the inside"
	fillColor isTransparent ifFalse:
		[self setFillColor: fillColor.
		port fillRect: (rect insetBy: borderWidth) offset: origin].! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 2/16/2000 22:07'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor

	| w h rect |
	"First use quick code for top and left borders and fill"
	self frameAndFillRectangle: r
		fillColor: fillColor
		borderWidth: borderWidth
		borderColor: topLeftColor.

	"Now use slow code for bevelled bottom and right borders"
	bottomRightColor isTransparent ifFalse: [
		borderWidth isNumber
			ifTrue: [w := h := borderWidth]
			ifFalse: [w := borderWidth x.   h := borderWidth y].
		rect := r translateBy: origin.
		self setFillColor: bottomRightColor.
		port 
			 frameRectRight: rect width: w;
			 frameRectBottom: rect height: h].
! !

!FormCanvas methodsFor: 'private' stamp: 'ar 5/14/2001 23:34'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule 
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."
	port colorMap: (aForm colormapIfNeededFor: form); fillColor: nil.
	port image: aForm at: aPoint + origin sourceRect: sourceRect rule: rule.! !

!FormCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:21'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule alpha: sourceAlpha
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."
	port colorMap: (aForm colormapIfNeededFor: form); fillColor: nil.
	port image: aForm at: aPoint + origin sourceRect: sourceRect rule: rule alpha: sourceAlpha.! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'IgorStasenko 12/22/2012 03:25'!
infiniteFillRectangle: aRectangle fillStyle: aFillStyle

	| additionalOffset rInPortTerms clippedPort targetTopLeft clipOffset ex |

	"this is a bit of a kludge to get the form to be aligned where I *think* it should be.
	something better is needed, but not now"

	additionalOffset := 0@0.
	ex := aFillStyle form extent.
	rInPortTerms := (aRectangle intersect: aFillStyle boundingBox ifNone: ["nothing to draw" ^ self ]) translateBy: origin.
	clippedPort := port clippedBy: rInPortTerms.
	targetTopLeft := clippedPort clipRect topLeft truncateTo: ex.
	clipOffset := rInPortTerms topLeft - targetTopLeft.
	additionalOffset := (clipOffset \\ ex) - ex.
	^aFillStyle
		displayOnPort: clippedPort
		offsetBy: additionalOffset
! !

!FormCanvas methodsFor: 'initialization' stamp: 'FernandoOlivero 9/11/2013 13:29'!
initializeTranslucentPatterns
	
	TranslucentPatterns := Array new: 8. 
	#(1 2 4 8 ) do: 
		[ :d | | mask bits pattern patternList | 
		patternList := Array new: 5.
		mask := (1 bitShift: d) - 1.
		bits := 2 * d.
		[ bits >= 32 ] whileFalse: 
			[ mask := mask bitOr: (mask bitShift: bits).	"double the length of mask"
			bits := bits + bits ].
		"0% pattern"
		pattern := Bitmap 
			with: 0
			with: 0.
		patternList 
			at: 1
			put: pattern.
		"25% pattern"
		pattern := Bitmap 
			with: mask
			with: 0.
		patternList 
			at: 2
			put: pattern.
		"50% pattern"
		pattern := Bitmap 
			with: mask
			with: mask bitInvert32.
		patternList 
			at: 3
			put: pattern.
		"75% pattern"
		pattern := Bitmap 
			with: mask
			with: 4294967295.
		patternList 
			at: 4
			put: pattern.
		"100% pattern"
		pattern := Bitmap 
			with: 4294967295
			with: 4294967295.
		patternList 
			at: 5
			put: pattern.
		TranslucentPatterns 
			at: d
			put: patternList ]! !

!FormCanvas methodsFor: 'testing' stamp: 'ar 6/22/1999 14:08'!
isVisible: aRectangle
	"Optimization"
	(aRectangle right + origin x) < clipRect left	ifTrue: [^ false].
	(aRectangle left + origin x) > clipRect right	ifTrue: [^ false].
	(aRectangle bottom + origin y) < clipRect top	ifTrue: [^ false].
	(aRectangle top + origin y) > clipRect bottom	ifTrue: [^ false].
	^ true
! !

!FormCanvas methodsFor: 'drawing' stamp: 'ar 2/16/2000 22:07'!
line: pt1 to: pt2 width: w color: c
	| offset |
	offset := origin - (w // 2) asPoint.
	self setFillColor: c.
	port width: w; height: w;
		drawFrom: (pt1 + offset) to: (pt2 + offset)! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:10'!
origin
	"Return the current origin for drawing operations"
	^ origin! !

!FormCanvas methodsFor: 'drawing' stamp: 'IgorStasenko 7/18/2011 18:15'!
paragraph: para bounds: bounds color: c

	| scanner |
	self setPaintColor: c.
	scanner := (port clippedBy: (bounds translateBy: origin)) displayScannerFor: para
		foreground: c background: Color transparent
		ignoreColorChanges: false.
	para displayOn: (self copyClipRect: bounds) using: scanner at: origin+ bounds topLeft.
! !

!FormCanvas methodsFor: 'copying' stamp: 'nice 1/13/2010 21:21'!
postCopy
	"The copy share same underlying Form but with its own grafPort."

	super postCopy.
	self resetGrafPort! !

!FormCanvas methodsFor: 'printing' stamp: 'ar 5/28/2000 17:07'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll:' on: '; print: form.! !

!FormCanvas methodsFor: 'private' stamp: 'RAA 12/17/2000 13:24'!
privateClipRect

	^clipRect! !

!FormCanvas methodsFor: 'private' stamp: 'RAA 12/17/2000 13:25'!
privatePort

	^port! !

!FormCanvas methodsFor: 'private' stamp: 'yo 6/18/2004 15:11'!
privateWarp: aForm transform: aTransform at: extraOffset sourceRect: sourceRect cellSize: cellSize
	"Warp the given using the appropriate transform and offset."
	| globalRect sourceQuad warp tfm |
	tfm := aTransform.
	globalRect := tfm localBoundsToGlobal: sourceRect.
	sourceQuad := (tfm sourceQuadFor: globalRect) collect:[:p| p - sourceRect topLeft].
	extraOffset ifNotNil:[globalRect := globalRect translateBy: extraOffset].
     warp := (WarpBlt current toForm: port destForm)
                combinationRule: Form paint;
                sourceQuad: sourceQuad destRect: (globalRect origin corner: globalRect corner+(1@1));
                clipRect: port clipRect.
	warp cellSize: cellSize.
	warp sourceForm: aForm.
	warp warpBits! !

!FormCanvas methodsFor: 'drawing' stamp: 'ar 9/9/2000 22:18'!
render: anObject
	"Do some 3D operations with the object if possible"
	^self asBalloonCanvas render: anObject! !

!FormCanvas methodsFor: 'initialization' stamp: 'ar 2/17/2000 00:21'!
reset

	origin := 0@0.							"origin of the top-left corner of this cavas"
	clipRect := (0@0 corner: 10000@10000).		"default clipping rectangle"
	self shadowColor: nil.! !

!FormCanvas methodsFor: 'private' stamp: 'MarcusDenker 3/24/2012 21:31'!
resetGrafPort
	"Private!! Create a new grafPort for a new copy."

	port := GrafPort toForm: form.
	port clipRect: clipRect.
! !

!FormCanvas methodsFor: 'drawing-general' stamp: 'ar 12/30/2001 18:58'!
roundCornersOf: aMorph in: bounds during: aBlock
	aMorph wantsRoundedCorners ifFalse:[^aBlock value].
	(self seesNothingOutside: (CornerRounder rectWithinCornersOf: bounds))
		ifTrue: ["Don't bother with corner logic if the region is inside them"
				^ aBlock value].
	CornerRounder roundCornersOf: aMorph on: self in: bounds
		displayBlock: aBlock
		borderWidth: aMorph borderWidthForRounding
		corners: aMorph roundedCorners! !

!FormCanvas methodsFor: 'drawing-general' stamp: 'GaryChambers 9/8/2011 14:48'!
roundShadowCornersOf: aMorph in: bounds during: aBlock
	aMorph wantsRoundedCorners ifFalse:[^aBlock value].
	(self seesNothingOutside: (CornerRounder rectWithinCornersOf: bounds))
		ifTrue: ["Don't bother with corner logic if the region is inside them"
				^ aBlock value].
	CornerRounder roundShadowCornersOf: aMorph on: self in: bounds
		displayBlock: aBlock
		borderWidth: aMorph borderWidthForRounding
		corners: aMorph roundedCorners! !

!FormCanvas methodsFor: 'private' stamp: 'tpr 9/15/2004 10:28'!
setClearColor: aColor
	"Install a new clear color - e.g., a color is used for clearing the background"
	| clearColor |
	clearColor := aColor ifNil:[Color transparent].
	clearColor isColor ifFalse:[
		(clearColor isKindOf: InfiniteForm) ifFalse:[^self error:'Cannot install color'].
		^port fillPattern: clearColor; combinationRule: Form over].
	"Okay, so clearColor really *is* a color"
	port sourceForm: nil.
	port combinationRule: Form over.
	port fillPattern: clearColor.
	self depth = 8 ifTrue:[
		"Use a stipple pattern"
		port fillColor: (form balancedPatternFor: clearColor)].
! !

!FormCanvas methodsFor: 'private' stamp: 'FernandoOlivero 9/9/2013 11:51'!
setFillColor: aColor
	"Install a new color used for filling."
	| screen patternWord fillColor |
	fillColor := aColor.
	fillColor ifNil:[fillColor := Color transparent].
	fillColor isColor ifFalse:[
		(fillColor isKindOf: InfiniteForm) ifFalse:[^self error:'Cannot install color'].
		^port fillPattern: fillColor; combinationRule: Form over].
	"Okay, so fillColor really *is* a color"
	port sourceForm: nil.
	fillColor isTranslucent ifFalse:[
		port combinationRule: Form over.
		port fillPattern: fillColor.
		self depth = 8 ifTrue:[
			"In 8 bit depth it's usually a good idea to use a stipple pattern"
			port fillColor: (form balancedPatternFor: fillColor)].
		^self].
	"fillColor is some translucent color"

	self depth > 8 ifTrue:[
		"BitBlt setup for alpha masked transfer"
		port fillPattern: fillColor.
		self depth = 16
			ifTrue:[port alphaBits: fillColor privateAlpha; combinationRule: 30]
			ifFalse:[port combinationRule: Form blend].
		^self].
	"Can't represent actual transparency -- use stipple pattern"
	screen := self translucentMaskFor: fillColor alpha depth: self depth.
	patternWord := form pixelWordFor: fillColor.
	port fillPattern: (screen collect: [:maskWord | maskWord bitAnd: patternWord]).
	port combinationRule: Form paint.
! !

!FormCanvas methodsFor: 'private' stamp: 'MarcusDenker 3/24/2012 21:31'!
setForm: aForm

	self reset.
	form := aForm.
	port := GrafPort toForm: form.
! !

!FormCanvas methodsFor: 'private' stamp: 'ar 6/22/1999 14:06'!
setOrigin: aPoint clipRect: aRectangle

	origin := aPoint.
	clipRect := aRectangle.
	port clipRect: aRectangle.
! !

!FormCanvas methodsFor: 'private' stamp: 'FernandoOlivero 9/9/2013 11:52'!
setPaintColor: aColor
	"Install a new color used for filling."
	
	| paintColor screen patternWord |
	paintColor := aColor.
	paintColor ifNil: [paintColor := Color transparent].
	paintColor isColor ifFalse: [ 
		(paintColor isKindOf: InfiniteForm) 
			ifFalse: [^self error:'Cannot install color'].
		^port fillPattern: paintColor; combinationRule: Form paint].
	"Okay, so paintColor really *is* a color"
	port sourceForm: nil.
	(paintColor isTranslucent) ifFalse: [
		port fillPattern: paintColor.
		port combinationRule: Form paint.
		self depth = 8 ifTrue: [
			port fillColor: (form balancedPatternFor: paintColor)].
		^self].
	"paintColor is translucent color"

	self depth > 8 ifTrue: [
		"BitBlt setup for alpha mapped transfer"
		port fillPattern: paintColor.
		self depth = 16
			ifTrue: [port alphaBits: paintColor privateAlpha; combinationRule: 31]
			ifFalse: [port combinationRule: Form blend].
		^self].

	"Can't represent actual transparency -- use stipple pattern"
	screen := self translucentMaskFor: paintColor alpha depth: self depth.
	patternWord := form pixelWordFor: paintColor.
	port fillPattern: (screen collect: [:maskWord | maskWord bitAnd: patternWord]).
	port combinationRule: Form paint
! !

!FormCanvas methodsFor: 'other' stamp: ''!
showAt: pt

	^ form displayAt: pt! !

!FormCanvas methodsFor: 'other' stamp: 'StephaneDucasse 10/25/2013 16:17'!
showAt: pt invalidRects: updateRects
	| blt |
	blt := (BitBlt toForm: Display)
			sourceForm: form;
			combinationRule: Form over.
	updateRects do:
		[:rect |
		blt sourceRect: rect;
			destOrigin: rect topLeft + pt;
			copyBits]! !

!FormCanvas methodsFor: 'drawing-images' stamp: 'tpr 9/15/2004 10:27'!
stencil: stencilForm at: aPoint sourceRect: sourceRect color: aColor
	"Flood this canvas with aColor wherever stencilForm has non-zero pixels"
	self setPaintColor: aColor.
	port colorMap: stencilForm maskingMap.
	port stencil: stencilForm
		at: aPoint + origin
		sourceRect: sourceRect.! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'IgorStasenko 7/18/2011 18:21'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock	 smoothing: cellSize

	"Note: This method has been originally copied from TransformationMorph."
	| innerRect patchRect sourceQuad warp start subCanvas rule|
	(aDisplayTransform isPureTranslation) ifTrue:[
		^aBlock value: (self copyOffset: aDisplayTransform offset negated truncated
							clipRect: aClipRect)
	].
	"Prepare an appropriate warp from patch to innerRect"
	innerRect := aClipRect.
	patchRect := (aDisplayTransform globalBoundsToLocal: innerRect) truncated.
	sourceQuad := (aDisplayTransform sourceQuadFor: innerRect)
					collect: [:p | p - patchRect topLeft].
	warp := self warpFrom: sourceQuad toRect: innerRect.
	warp cellSize: cellSize.

	"Render the submorphs visible in the clipping rectangle, as patchForm"
	start := (self depth = 1)
		"If this is true B&W, then we need a first pass for erasure."
		ifTrue: [1] ifFalse: [2].
	"If my depth has alpha, do blending rather than paint"
	rule := self depth = 32 ifTrue: [Form blend] ifFalse: [Form paint].
	start to: 2 do:
		[:i | "If i=1 we first make a shadow and erase it for opaque whites in B&W"
		subCanvas := self class extent: patchRect extent depth: self depth.
		i=1	ifTrue: [ warp combinationRule: Form erase ]
			ifFalse: [ warp combinationRule: rule].
		subCanvas translateBy: patchRect topLeft negated
			during:[:offsetCanvas| aBlock value: offsetCanvas].
		warp sourceForm: subCanvas form; warpBits.
		warp sourceForm: nil.  subCanvas := nil "release space for next loop"]
! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 03:02'!
translateBy: delta during: aBlock
	"Set a translation only during the execution of aBlock."
	^aBlock value: (self copyOffset: delta)! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 02:55'!
translateTo: newOrigin clippingTo: aRectangle during: aBlock
	"Set a new origin and clipping rectangle only during the execution of aBlock."
	aBlock value: (self copyOrigin: newOrigin clipRect: aRectangle)! !

!FormCanvas methodsFor: 'private' stamp: 'FernandoOlivero 9/12/2013 13:36'!
translucentMaskFor: alphaValue depth: d
	"Return a pattern representing a mask usable for stipple transparency"
	^(self translucentPatterns at: d) at: ((alphaValue min: 1.0 max: 0.0) * 4) rounded + 1 ! !

!FormCanvas methodsFor: 'private' stamp: 'FernandoOlivero 9/12/2013 13:36'!
translucentPatterns
	^ self class translucentPatterns! !

!FormCanvas methodsFor: 'other' stamp: 'ar 5/28/2000 12:12'!
warpFrom: sourceQuad toRect: destRect
        ^ (WarpBlt current toForm: port destForm)
                combinationRule: Form paint;
                sourceQuad: sourceQuad destRect: (destRect translateBy: origin);
                clipRect: clipRect! !

!FormCanvas methodsFor: 'drawing-images' stamp: 'ar 12/30/2001 16:36'!
warpImage: aForm transform: aTransform at: extraOffset sourceRect: sourceRect cellSize: cellSize
	"Warp the given using the appropriate transform and offset."
	| tfm |
	tfm := (MatrixTransform2x3 withOffset: origin) composedWithLocal: aTransform.
	^self privateWarp: aForm transform: tfm at: extraOffset sourceRect: sourceRect cellSize: cellSize! !

Object subclass: #DamageRecorder
	instanceVariableNames: 'invalidRects totalRepaint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Support'!

!DamageRecorder commentStamp: 'TorstenBergmann 2/20/2014 18:27' prior: 0!
Recording damages to repair by repainting!

!DamageRecorder class methodsFor: 'instance creation' stamp: ''!
new

	^ super new reset
! !

!DamageRecorder methodsFor: 'recording' stamp: ''!
doFullRepaint
	"Record that a full redisplay is needed. No further damage rectangles will be recorded until after the next reset."

	^ totalRepaint := true.
! !

!DamageRecorder methodsFor: 'recording' stamp: 'StephaneDucasse 2/7/2011 22:25'!
invalidRectsFullBounds: aRectangle
	"Return a collection of damaged rectangles for the given canvas. If a total repaint has been requested, return the given rectangle."

	^ totalRepaint
		ifTrue: [ Array with: aRectangle]
		ifFalse: [ invalidRects copy].

! !

!DamageRecorder methodsFor: 'recording' stamp: 'IgorStasenko 12/22/2012 03:41'!
recordInvalidRect: newRect
	"Record the given rectangle in my damage list, a list of rectangular areas of the display that should be redraw on the next display cycle."
	"Details: Damaged rectangles are often identical or overlap significantly. In these cases, we merge them to reduce the number of damage rectangles that must be processed when the display is updated. Moreover, above a certain threshold, we ignore the individual rectangles completely, and simply do a complete repaint on the next cycle."

	| mergeRect zeroRect |
	totalRepaint ifTrue: [^ self].  "planning full repaint; don't bother collecting damage"
	zeroRect := 0@0 corner: 0@0. 
	invalidRects do:
		[:rect | | a |
		((a := (rect intersect: newRect ifNone: [ zeroRect ]) area) > 40
			and: ["Avoid combining a vertical and horizontal rects.
				  Can make a big diff and we only test when likely."
				  a > (newRect area // 4) or: [a > (rect area // 4)]])
			ifTrue:
			["merge rectangle in place (see note below) if there is significant overlap"
			rect setOrigin: (rect origin min: newRect origin) truncated
				corner: (rect corner max: newRect corner) truncated.
			^ self]].


	invalidRects size >= 50 ifTrue:
		["if there are too many separate areas, merge them all"
		mergeRect := Rectangle merging: invalidRects.
		self reset.
		invalidRects addLast: mergeRect].

	"add the given rectangle to the damage list"
	"Note: We make a deep copy of all rectangles added to the damage list,
		since rectangles in this list may be extended in place."
	newRect hasPositiveExtent ifTrue: [ 
		invalidRects addLast:
			(newRect topLeft truncated corner: newRect bottomRight truncated).
	].! !

!DamageRecorder methodsFor: 'initialization' stamp: 'sma 6/5/2000 11:55'!
reset
	"Clear the damage list."

	invalidRects := OrderedCollection new: 15.
	totalRepaint := false
! !

!DamageRecorder methodsFor: 'testing' stamp: 'dgd 2/22/2003 14:43'!
updateIsNeeded
	"Return true if the display needs to be updated."

	^totalRepaint or: [invalidRects notEmpty]! !

Object subclass: #EllipseMidpointTracer
	instanceVariableNames: 'rect x y a b aSquared bSquared d1 d2 inFirstRegion'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Support'!

!EllipseMidpointTracer commentStamp: 'TorstenBergmann 2/20/2014 18:32' prior: 0!
Utility class for calculating!

!EllipseMidpointTracer methodsFor: 'initialize' stamp: 'ar 6/28/1999 15:33'!
on: aRectangle
	rect := aRectangle.
	a := rect width // 2.
	b := rect height // 2.
	x := 0.
	y := b.
	aSquared := a * a.
	bSquared := b * b.
	d1 := bSquared - (aSquared * b) + (0.25 * aSquared).
	d2 := nil.
	inFirstRegion := true.! !

!EllipseMidpointTracer methodsFor: 'computing' stamp: 'ar 6/28/1999 15:35'!
stepInY
	"Step to the next y value"
	inFirstRegion ifTrue:[
		"In the upper region we must step until we reach the next y value"
		[(aSquared * (y-0.5)) > (bSquared * (x+1))] whileTrue:[
			d1 < 0.0
				ifTrue:[d1 := d1 + (bSquared * (2*x+3)).
						x := x + 1]
				ifFalse:[d1 := d1 + (bSquared * (2*x+3)) + (aSquared * (-2*y+2)).
						y := y - 1.
						^x := x + 1]].
		"Stepping into second region"
		d2 := (bSquared * (x + 0.5) squared) + (aSquared * (y-1) squared) - (aSquared * bSquared).
		inFirstRegion := false.
	].
	"In the lower region each step is a y-step"
	d2 < 0.0
		ifTrue:[d2 := d2 + (bSquared * (2*x+2)) + (aSquared * (-2*y+3)).
				x := x + 1]
		ifFalse:[d2 := d2 + (aSquared * (-2*y+3))].
	y := y - 1.
	^x! !

Object subclass: #Morph
	instanceVariableNames: 'bounds owner submorphs fullBounds color extension'
	classVariableNames: 'CmdGesturesEnabled CycleHalosBothDirections DefaultYellowButtonMenuEnabled EmptyArray HalosEnabled'
	poolDictionaries: ''
	category: 'Morphic-Core-Kernel'!

Morph class
	instanceVariableNames: 'announcer'!

!Morph commentStamp: 'efc 2/26/2003 20:01' prior: 0!
A Morph (from the Greek "shape" or "form") is an interactive graphical object. General information on the Morphic system can be found at http://minnow.cc.gatech.edu/squeak/30. 

Morphs exist in a tree, rooted at a World (generally a PasteUpMorph). The morphs owned by a morph are its submorphs. Morphs are drawn recursively; if a Morph has no owner it never gets drawn. To hide a Morph and its submorphs, set its #visible property to false using the #visible: method. 

The World (screen) coordinate system is used for most coordinates, but can be changed if there is a TransformMorph somewhere in the owner chain. 

My instance variables have accessor methods (e.g., #bounds, #bounds:). Most users should use the accessor methods instead of using the instance variables directly.

Structure:
instance var 	Type 			Description 
bounds 			Rectangle 		A Rectangle indicating my position and a size that will enclose 									me. 
owner 			Morph		 	My parent Morph, or nil for the top-level Morph, which is a
 				or nil			world, typically a PasteUpMorph.
submorphs 		Array 			My child Morphs. 
fullBounds 		Rectangle 		A Rectangle minimally enclosing me and my submorphs. 
color 			Color 			My primary color. Subclasses can use this in different ways. 
extension 		MorphExtension Allows extra properties to be stored without adding a
				or nil  				storage burden to all morphs. 

By default, Morphs do not position their submorphs. Morphs may position their submorphs directly or use a LayoutPolicy to automatically control their submorph positioning.

Although Morph has some support for BorderStyle, most users should use BorderedMorph if they want borders.!

Morph class
	instanceVariableNames: 'announcer'!

Morph subclass: #BorderedMorph
	instanceVariableNames: 'borderWidth borderColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Kernel'!

!BorderedMorph commentStamp: 'kfr 10/27/2003 11:17' prior: 0!
BorderedMorph introduce borders to morph. Borders have the instanceVariables borderWidth and borderColor.
 
BorderedMorph new borderColor: Color red; borderWidth: 10; openInWorld.

BorderedMorph also have a varaity of border styles: simple, inset, raised, complexAltFramed, complexAltInset, complexAltRaised, complexFramed, complexInset, complexRaised.
These styles are set using the classes BorderStyle, SimpleBorder, RaisedBorder, InsetBorder and ComplexBorder.

BorderedMorph new borderStyle: (SimpleBorder width: 1 color: Color white); openInWorld.
BorderedMorph new borderStyle: (BorderStyle inset width: 2); openInWorld.


!

!BorderedMorph class methodsFor: 'examples' stamp: 'MarcusDenker 7/8/2012 12:09'!
exampleGradient
	"self exampleGradient"
	
	| morph fs |
	morph := BorderedMorph new.
	fs := GradientFillStyle ramp: {0.0 -> Color red. 1.0 -> Color green}.
	fs origin: morph bounds center.
	fs direction: (morph bounds width // 2) @ 0.
	fs radial: true.
	morph fillStyle: fs.
	World primaryHand attachMorph: morph.! !

!BorderedMorph methodsFor: 'geometry' stamp: 'StephaneDucasse 5/28/2011 13:31'!
acquireBorderWidth: aBorderWidth
	"Gracefully acquire the new border width, keeping the interior area intact and not seeming to shift"

	| delta |
	(delta := aBorderWidth- self borderWidth) = 0 ifTrue: [^ self].
	self bounds: ((self bounds origin - (delta @ delta)) corner: (self bounds corner + (delta @ delta))).
	self borderWidth: aBorderWidth.
	self layoutChanged! !

!BorderedMorph methodsFor: 'menu' stamp: 'ClementBera 10/25/2013 13:45'!
addBorderStyleMenuItems: aMenu hand: aHandMorph
	"Add border-style menu items"

	| subMenu |
	subMenu := UIManager default 
					newMenuIn: self for: self.
	subMenu addStayUpItemSpecial.
	subMenu addList: 
		{{'border width...' translated. #changeBorderWidth:}}.
	subMenu addLine.
	BorderStyle borderStyleChoices do:
		[:sym | (self borderStyleForSymbol: sym)
			ifNotNil:
				[subMenu add: sym translated target: self selector: #setBorderStyle: argument: sym]].
	aMenu add: 'border style' translated subMenu: subMenu
! !

!BorderedMorph methodsFor: 'lookenhancements' stamp: 'StephaneDucasse 5/30/2013 21:34'!
addCornerGrips
	self addMorphBack: (TopLeftGripMorph new target: self; position: self position).
	self addMorphBack: (TopRightGripMorph new target: self; position: self position).
	self addMorphBack: (BottomLeftGripMorph new target: self; position: self position).
	self addMorphBack: (BottomRightGripMorph new target: self; position: self position)! !

!BorderedMorph methodsFor: 'lookenhancements' stamp: 'IgorStasenko 12/19/2012 17:19'!
addPaneSplitters
	| splitter remaining target targetX sameX minY maxY targetY sameY minX maxX |
	self removePaneSplitters.
	self removeCornerGrips.

	remaining := submorphs reject: [:each | each layoutFrame rightFraction = 1].
	[remaining notEmpty] whileTrue:
		[target := remaining first.
		targetX := target layoutFrame rightFraction.
		sameX := submorphs select: [:each | each layoutFrame rightFraction = targetX].
		minY := (sameX detectMin: [:each | each layoutFrame topFraction]) layoutFrame topFraction.
		maxY := (sameX detectMax: [:each | each layoutFrame bottomFraction]) layoutFrame bottomFraction.
		splitter := ProportionalSplitterMorph new.
		splitter layoutFrame: (
		
		 (targetX @ minY corner: targetX @ maxY) asLayoutFrame 
			leftOffset: target layoutFrame rightOffset;
			topOffset: target layoutFrame topOffset;
			rightOffset: 4 + target layoutFrame rightOffset;
			bottomOffset: target layoutFrame bottomOffset).
			
		self addMorphBack: (splitter position: self position).
		remaining := remaining copyWithoutAll: sameX].

	remaining := submorphs copy reject: [:each | each layoutFrame bottomFraction = 1].
	[remaining notEmpty]
		whileTrue: [target := remaining first.
			targetY := target layoutFrame bottomFraction.
			sameY := submorphs select: [:each | each layoutFrame bottomFraction = targetY].
			minX := (sameY detectMin: [:each | each layoutFrame leftFraction]) layoutFrame leftFraction.
			maxX := (sameY detectMax: [:each | each layoutFrame rightFraction]) layoutFrame rightFraction.
			splitter := ProportionalSplitterMorph new beSplitsTopAndBottom; yourself.
			splitter layoutFrame: (
				(minX @ targetY corner: maxX @ targetY) asLayoutFrame
					leftOffset: target layoutFrame leftOffset;
					topOffset: target layoutFrame bottomOffset;
					rightOffset: target layoutFrame rightOffset;
					bottomOffset: 4 + target layoutFrame bottomOffset ).
			self addMorphBack: (splitter position: self position).
			remaining := remaining copyWithoutAll: sameY].

	self linkSubmorphsToSplitters.
	self splitters do: [:each | each comeToFront].
! !

!BorderedMorph methodsFor: 'lookenhancements' stamp: 'IgorStasenko 12/19/2012 17:23'!
addPaneVSplitterBetween: leftMorph and: rightMorphs 

	| targetX minY maxY splitter |
	targetX := leftMorph layoutFrame rightFraction.
	minY := (rightMorphs detectMin: [:each | each layoutFrame topFraction]) layoutFrame topFraction.
	maxY := (rightMorphs detectMax: [:each | each layoutFrame bottomFraction]) layoutFrame bottomFraction.
	
	splitter := ProportionalSplitterMorph new.
	splitter layoutFrame: ( (targetX @ minY corner: targetX @ maxY) asLayoutFrame 
		leftOffset: leftMorph layoutFrame rightOffset;
		rightOffset: 4+ leftMorph layoutFrame rightOffset;
		topOffset:  leftMorph layoutFrame topOffset;
		bottomOffset: leftMorph layoutFrame bottomOffset).

	self addMorphBack: (splitter position: self position).! !

!BorderedMorph methodsFor: 'accessing' stamp: 'sw 8/6/97 14:34'!
borderColor
	^ borderColor! !

!BorderedMorph methodsFor: 'accessing' stamp: 'ar 8/17/2001 16:52'!
borderColor: colorOrSymbolOrNil
	self doesBevels ifFalse:[
		colorOrSymbolOrNil isColor ifFalse:[^self]].
	borderColor = colorOrSymbolOrNil ifFalse: [
		borderColor := colorOrSymbolOrNil.
		self changed].
! !

!BorderedMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:53'!
borderInitialize
	"initialize the receiver state related to border"
	borderColor:= self defaultBorderColor.
	borderWidth := self defaultBorderWidth! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:24'!
borderInset
	self borderColor: #inset! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:25'!
borderRaised
	self borderColor: #raised! !

!BorderedMorph methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:17'!
borderStyle
	"Work around the borderWidth/borderColor pair"

	| style |
	borderColor ifNil: [^BorderStyle default].
	borderWidth isZero ifTrue: [^BorderStyle default].
	self assureExtension.
	style := extension borderStyle ifNil: [BorderStyle default].
	(borderWidth = style width and: 
			["Hah!! Try understanding this..."

			borderColor == style style or: 
					["#raised/#inset etc"

					#simple == style style and: [borderColor = style color]]]) 
		ifFalse: 
			[style := borderColor isColor 
				ifTrue: [BorderStyle width: borderWidth color: borderColor]
				ifFalse: [(BorderStyle perform: borderColor) width: borderWidth	"argh."].
			extension borderStyle: style].
	^style trackColorFrom: self! !

!BorderedMorph methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:16'!
borderStyle: aBorderStyle 
	"Work around the borderWidth/borderColor pair"

	aBorderStyle = self borderStyle ifTrue: [^self].
	self assureExtension.
	"secure against invalid border styles"
	(self canDrawBorder: aBorderStyle) 
		ifFalse: 
			["Replace the suggested border with a simple one"

			^self borderStyle: (BorderStyle width: aBorderStyle width
						color: (aBorderStyle trackColorFrom: self) color)].
	aBorderStyle width = self borderStyle width ifFalse: [self changed].
	(aBorderStyle isNil or: [aBorderStyle == BorderStyle default]) 
		ifTrue: 
			[extension borderStyle: nil.
			borderWidth := 0.
			^self changed].
	extension borderStyle: aBorderStyle.
	borderWidth := aBorderStyle width.
	borderColor := aBorderStyle style == #simple 
				ifTrue: [aBorderStyle color]
				ifFalse: [aBorderStyle style].
	self changed! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:09'!
borderWidth
	^ borderWidth! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/4/1999 09:42'!
borderWidth: anInteger
	borderColor ifNil: [borderColor := Color black].
	borderWidth := anInteger max: 0.
	self changed! !

!BorderedMorph methodsFor: 'menu' stamp: 'MarcusDenker 9/13/2013 15:55'!
changeBorderWidth: evt
	| aHand origin handle newWidth |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	handle := HandleMorph new
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			newWidth := (newPoint - origin) r asInteger // 5.
			self borderWidth: newWidth]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNil: [:halo | halo addHandles]].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor farther from
this point to increase border width.
Click when done.' translated hand: evt hand.
	handle startStepping! !

!BorderedMorph methodsFor: 'geometry' stamp: 'nk 4/5/2001 14:24'!
closestPointTo: aPoint
	"account for round corners. Still has a couple of glitches at upper left and right corners"
	| pt |
	pt := self bounds pointNearestTo: aPoint.
	self wantsRoundedCorners ifFalse: [ ^pt ].
	self bounds corners with: (self bounds insetBy: 6) corners do: [ :out :in |
		(pt - out) abs < (6@6)
			ifTrue: [ ^(in + (Point r: 5.0 degrees: (pt - in) degrees)) asIntegerPoint ].
	].
	^pt.! !

!BorderedMorph methodsFor: 'accessing' stamp: 'gvc 5/9/2006 15:50'!
colorForInsets
	"Return the color to be used for shading inset borders."
	
	self owner isSystemWindow
		ifTrue: [^self owner colorForInsets].
	^super colorForInsets! !

!BorderedMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:33'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ Color black! !

!BorderedMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:36'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:19'!
doesBevels
	"To return true means that this object can show bevelled borders, and
	therefore can accept, eg, #raised or #inset as valid borderColors.
	Must be overridden by subclasses that do not support bevelled borders."

	^ true! !

!BorderedMorph methodsFor: 'initialization' stamp: 'MarcusDenker 12/11/2009 23:56'!
initialize
	"initialize the state of the receiver"
	super initialize.
	
	self borderInitialize! !

!BorderedMorph methodsFor: 'testing' stamp: 'FernandoOlivero 9/5/2013 10:55'!
isTranslucentButNotTransparent
	"Answer true if this any of this morph is translucent but not transparent."

	(color isColor and: [color isTranslucentButNotTransparent]) ifTrue: [^ true].
	(borderColor isColor and: [borderColor isTranslucentButNotTransparent]) ifTrue: [^ true].
	^ false
! !

!BorderedMorph methodsFor: 'lookenhancements' stamp: 'jrp 7/23/2005 22:16'!
linkSubmorphsToSplitters

	self splitters do:
		[:each |
		each splitsTopAndBottom
			ifTrue:
				[self submorphsDo:
					[:eachMorph |
					(eachMorph ~= each and: [eachMorph layoutFrame bottomFraction = each layoutFrame topFraction]) ifTrue: [each addLeftOrTop: eachMorph].
					(eachMorph ~= each and: [eachMorph layoutFrame topFraction = each layoutFrame bottomFraction]) ifTrue: [each addRightOrBottom: eachMorph]]]
			ifFalse:
				[self submorphsDo:
					[:eachMorph |
					(eachMorph ~= each and: [eachMorph layoutFrame rightFraction = each layoutFrame leftFraction]) ifTrue: [each addLeftOrTop: eachMorph].
					(eachMorph ~= each and: [eachMorph layoutFrame leftFraction = each layoutFrame rightFraction]) ifTrue: [each addRightOrBottom: eachMorph]]]]! !

!BorderedMorph methodsFor: 'lookenhancements' stamp: 'jrp 7/23/2005 00:03'!
removeCornerGrips

	| corners |
	corners := self submorphsSatisfying: [:each | each isKindOf: CornerGripMorph].
	corners do: [:each | each delete]! !

!BorderedMorph methodsFor: 'lookenhancements' stamp: 'jrp 7/23/2005 22:28'!
removePaneSplitters

	self splitters do: [:each | each delete]! !

!BorderedMorph methodsFor: 'lookenhancements' stamp: 'jrp 7/23/2005 22:16'!
splitters

	^ self submorphsSatisfying: [:each | each isKindOf: ProportionalSplitterMorph]! !

!BorderedMorph methodsFor: 'accessing' stamp: 'sw 11/29/1999 17:37'!
useSquareCorners
	self cornerStyle: #square! !

BorderedMorph subclass: #PasteUpMorph
	instanceVariableNames: 'backgroundMorph worldState griddingOn'
	classVariableNames: 'WindowEventHandler WorldAnnouncer'
	poolDictionaries: ''
	category: 'Morphic-Core-Worlds'!

!PasteUpMorph commentStamp: '<historical>' prior: 0!
A morph whose submorphs comprise a paste-up of rectangular subparts which "show through".  Anything called a 'Playfield' is a PasteUpMorph.

Facilities commonly needed on pages of graphical presentations and on simulation playfields, such as the painting of new objects, turtle trails, gradient fills, background paintings, parts-bin behavior, collision-detection, etc., are (or will be) provided.




!

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'StephaneDucasse 7/2/2013 17:26'!
acceptDroppingMorph: dropped event: evt
	"The supplied morph, known to be acceptable to the receiver, is now to be assimilated; the precipitating event is supplied"

	| aMorph |
	aMorph := self morphToDropFrom: dropped.
	super acceptDroppingMorph: aMorph event: evt.
	aMorph
		submorphsDo: [ :m | 
			(m isKindOf: HaloMorph)
				ifTrue: [ m delete ] ].
	self bringTopmostsToFront! !

!PasteUpMorph methodsFor: 'world state' stamp: 'di 7/15/1999 09:51'!
addMorph: aMorph centeredNear: aPoint
	"Add the given morph to this world, attempting to keep its center as close to the given point possible while also keeping the it entirely within the bounds of this world."

	| trialRect delta |
	trialRect := Rectangle center: aPoint extent: aMorph fullBounds extent.
	delta := trialRect amountToTranslateWithin: bounds.
	aMorph position: trialRect origin + delta.
	self addMorph: aMorph.
! !

!PasteUpMorph methodsFor: 'submorphs-add/remove' stamp: 'RAA 12/16/2000 18:37'!
addMorphFront: aMorph

	^self addMorphInFrontOfLayer: aMorph
! !

!PasteUpMorph methodsFor: 'wiw support' stamp: 'dgd 8/31/2004 16:25'!
addMorphInLayer: aMorph
	super addMorphInLayer: aMorph.
	aMorph wantsToBeTopmost ifFalse:[self bringTopmostsToFront].! !

!PasteUpMorph methodsFor: 'world state' stamp: 'MarcusDenker 9/13/2013 15:50'!
addMorphs: aMorphOrList 
	"Dump in submorphs, and stepList from aMorphOrList.  Used to bring a world, paste-up, or other morph in from an object file."

	aMorphOrList isMorph 
		ifTrue:  
			[
			aMorphOrList privateOwner: nil.
			aMorphOrList isWorldMorph 
					ifFalse:  
						[ 
						self firstHand attachMorph: aMorphOrList.
						self startSteppingSubmorphsOf: aMorphOrList]
					ifTrue: 
						[
						self addMorph: aMorphOrList ].
					aMorphOrList privateSubmorphs reverseDo: 
						[:m | 
							m privateOwner: nil.
							self addMorph: m.
							m changed].
				(aMorphOrList instVarNamed: 'stepList') 
							do: [:entry | entry first startSteppingIn: self]]
		ifFalse: 
			["list, add them all"
			aMorphOrList reverseDo: 
					[:m | 
					m privateOwner: nil.
					self addMorph: m.
					self startSteppingSubmorphsOf: m.	"It may not want this!!"
					m changed]]! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'StephaneDucasse 3/3/2010 15:49'!
addWorldHaloMenuItemsTo: aMenu hand: aHandMorph
	"Add standard halo items to the menu, given that the receiver is a World"

	self addFillStyleMenuItems: aMenu hand: aHandMorph.
	self addLayoutMenuItems: aMenu hand: aHandMorph.

	aMenu addLine.
	self addWorldToggleItemsToHaloMenu: aMenu.
	aMenu addLine.
	self addExportMenuItems: aMenu hand: aHandMorph.
	self addMiscExtrasTo: aMenu.
	self addDebuggingItemsTo: aMenu hand: aHandMorph.

	aMenu addLine.
	aMenu defaultTarget: aHandMorph.
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 4/20/2002 01:38'!
addWorldToggleItemsToHaloMenu: aMenu
	"Add toggle items for the world to the halo menu"

	#(
	(hasDragAndDropEnabledString changeDragAndDrop 'whether I am open to having objects dropped into me')
	(roundedCornersString toggleCornerRounding 'whether the world should have rounded corners')) do:

		[:trip | aMenu addUpdating: trip first action: trip second.
			aMenu balloonTextForLastItem: trip third]! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'ClementBera 10/3/2013 10:30'!
becomeActiveDuring: aBlock
	"Make the receiver the ActiveWorld during the evaluation of aBlock.
	Note that this method does deliberately *not* use #ensure: to prevent
	re-installation of the world on project switches."
	| priorWorld priorHand priorEvent |
	priorWorld := ActiveWorld.
	priorHand := ActiveHand.
	priorEvent := ActiveEvent.
	ActiveWorld := self.
	ActiveHand := self hands first. "default"
	ActiveEvent := nil. "not in event cycle"
	aBlock
		on: Error
		do: [:ex | 
			ActiveWorld := priorWorld.
			ActiveEvent := priorEvent.
			ActiveHand := priorHand.
			ex pass]! !

!PasteUpMorph methodsFor: 'viewing' stamp: 'dgd 4/4/2006 13:58'!
bringTopmostsToFront
	submorphs
		select:[:m| m wantsToBeTopmost]
		thenDo:[:m| self addMorphInLayer: m].! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'MarcusDenker 9/13/2013 16:28'!
bringWindowsFullOnscreen
	"Make ever SystemWindow on the desktop be totally on-screen, whenever possible."

	| r |
	bounds
		allAreasOutsideList: (self taskbars collect: [ :t | t bounds ])
		do: [ :e | 
			r isNil
				ifTrue: [ r := e ]
				ifFalse: [ 
					r area > e area
						ifTrue: [ r ]
						ifFalse: [ e ] ] ].
	(self windowsSatisfying: [ :w | true ])
		do: [ :aWindow | 
			aWindow right: (aWindow right min: r right).
			aWindow bottom: (aWindow bottom min: r bottom).
			aWindow left: (aWindow left max: r left).
			aWindow top: (aWindow top max: r top) ]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'alain.plantec 2/6/2009 11:39'!
checkCurrentHandForObjectToPaste

	| response |
	self primaryHand pasteBuffer ifNil: [^self].
	response := self confirm: ('Hand is holding a Morph in its paste buffer:' translated, '\') withCRs,
			self primaryHand pasteBuffer printString, 
			('\', 'Delete it ?' translated) withCRs.
	response = 1 ifTrue: [self primaryHand pasteBuffer: nil].
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'IgorStasenko 1/2/2012 19:04'!
chooseClickTarget
	Cursor crossHair showWhile:
		[self activeHand waitButton].
	Cursor down showWhile:
		[self activeHand anyButtonPressed].
	^ (self morphsAt: self activeHand cursorPoint) first! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 11/7/2011 22:43'!
cleanseOtherworldlySteppers
	"If the current project is a morphic one, then remove from its steplist
	those morphs that are not really in the world"
	"Utilities cleanseOtherworldlySteppers"
	| old delta |
	old := self currentWorld stepListSize.
	self currentWorld steppingMorphsNotInWorld
		do: [:m | m delete].
	self currentWorld cleanseStepList.
	(delta := old - self currentWorld stepListSize) > 0
		ifTrue: [ self crTrace: delta asString , ' morphs removed from steplist' ]! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'MarianoMartinezPeck 5/2/2012 23:49'!
closeAllWindowsDiscardingChanges
	World systemWindows do: [:w | [w delete] valueSupplyingAnswer: false]! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'alain.plantec 2/6/2009 17:17'!
closeUnchangedWindows
	"Present a menu of window titles for all windows with changes,
	and activate the one that gets chosen."
	(self confirm:
'Do you really want to close all windows
except those with unaccepted edits?' translated)
		ifFalse: [^ self].

	(self  windowsSatisfying: [:w | w model canDiscardEdits])
		do: [:w | w delete]! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'AlainPlantec 1/9/2010 06:22'!
collapseAll
	"Collapse all windows"
	(self windowsSatisfying: [:w | w isCollapsed not])
		reverseDo: [:w | w collapseOrExpand.  self displayWorld].
	self collapseNonWindows! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'AlainPlantec 1/9/2010 06:07'!
collapseNonWindows
	self nonWindows reject: [:m | m isSticky] thenDo: [:m | m collapse]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 13:52'!
commandKeySelectors
	"Answer my command-key table"

	| aDict |
	aDict := self valueOfProperty: #commandKeySelectors ifAbsentPut: [self initializeDesktopCommandKeySelectors].
	^ aDict! !

!PasteUpMorph methodsFor: 'taskbar' stamp: 'MarcusDenker 3/23/2011 15:30'!
createTaskbarIfNecessary
	"Private - create a new taskbar if not present."

	self taskbars ifEmpty: [
		TaskbarMorph new openInWorld: self.
		self moveCollapsedWindowsToTaskbar]! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:35'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ Color
		r: 0.861
		g: 1.0
		b: 0.722! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:38'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:29'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color
		r: 0.8
		g: 1.0
		b: 0.6! !

!PasteUpMorph methodsFor: 'halos and balloon help' stamp: 'sw 1/10/2000 16:44'!
defersHaloOnClickTo: aSubMorph
	"If a cmd-click on aSubMorph would make it a preferred recipient of the halo, answer true"
	^ true
	! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'dgd 9/11/2004 20:45'!
delayedInvokeWorldMenu: evt 
	self
		addAlarm: #invokeWorldMenu:
		with: evt
		after: 200! !

!PasteUpMorph methodsFor: 'world state' stamp: 'StephaneDucasse 2/9/2011 17:59'!
deleteAllHalos
	"self halt.
	self haloMorphs
		do: [:each | (each target isKindOf: SelectionMorph)
				ifTrue: [each target delete]].
	self hands
		do: [:each | each removeHalo]"! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:20'!
discoveredWorldMenu
	^ owner discoveredWorldMenu
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 12:19'!
dispatchCommandKeyInWorld: aChar event: evt
	"Dispatch the desktop command key if possible.  Answer whether handled"

	| aMessageSend |
	aMessageSend := self commandKeySelectors at: aChar ifAbsent: [^ false].
	aMessageSend selector numArgs = 0
		ifTrue:
			[aMessageSend value]
		ifFalse:
			[aMessageSend valueWithArguments: (Array with: evt)].
	^ true
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'StephaneDucasse 7/2/2013 17:29'!
dispatchKeystroke: anEvent
	anEvent keyCharacter == Character tab
		ifTrue: [ 
			self tabAmongFields
				ifTrue: [ ^ self tabHitWithEvent: anEvent ] ]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'di 9/19/2000 22:17'!
dragThroughOnDesktop: evt
	"Draw out a selection rectangle"
	| selection |
	selection := SelectionMorph newBounds: (evt cursorPoint extent: 8@8).
	self addMorph: selection.
	^ selection extendByHand: evt hand
! !

!PasteUpMorph methodsFor: 'drawing' stamp: 'MarcusDenker 10/10/2013 23:47'!
drawBackgroundSketchOn: aCanvas
	backgroundMorph ifNil: [ ^ self ].
	self clipSubmorphs
		ifTrue: [ aCanvas clipBy: self clippingBounds during: [ :canvas | canvas fullDrawMorph: backgroundMorph ] ]
		ifFalse: [ aCanvas fullDrawMorph: backgroundMorph ]! !

!PasteUpMorph methodsFor: 'drawing' stamp: 'StephaneDucasse 6/27/2013 22:54'!
drawGridOn: aCanvas

	(self griddingOn and: [ self gridVisible ])
		ifTrue: [ 
			aCanvas
				fillRectangle: self bounds
				fillStyle:
					(self
						gridFormOrigin: self gridOrigin
						grid: self gridModulus
						background: nil
						line: Color lightGray) ]! !

!PasteUpMorph methodsFor: 'drawing' stamp: 'StephaneDucasse 6/27/2013 22:54'!
drawOn: aCanvas 
	"Draw in order:
	- background color
	- grid, if any
	- background sketch, if any
	Later (in drawSubmorphsOn:) I will skip drawing the background sketch."

	super drawOn: aCanvas.	
	self drawGridOn: aCanvas.
	self drawBackgroundSketchOn: aCanvas! !

!PasteUpMorph methodsFor: 'drawing' stamp: 'StephaneDucasse 6/27/2013 19:19'!
drawSubmorphsOn: aCanvas
	"Display submorphs back to front, but skip my background sketch."

	| drawBlock |
	submorphs isEmpty
		ifTrue: [ ^ self ].
	drawBlock := [ :canvas | 
	submorphs
		reverseDo: [ :m | 
			m ~~ backgroundMorph
				ifTrue: [ canvas fullDrawMorph: m ] ] ].
	self clipSubmorphs
		ifTrue: [ aCanvas clipBy: (aCanvas clipRect intersect: self clippingBounds ifNone: [ ^ self ]) during: drawBlock ]
		ifFalse: [ drawBlock value: aCanvas ]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 3/3/2010 15:32'!
drawingClass

	^ ImageMorph! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'sw 2/4/2001 00:54'!
dropEnabled
	"Get this morph's ability to add and remove morphs via drag-n-drop."

	^ (self valueOfProperty: #dropEnabled) ~~ false
! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'BenjaminVanRyseghem 5/4/2013 12:55'!
dropFiles: anEvent
	"Handle a number of dropped files from the OS.
	TODO:
		- use a more general mechanism for figuring out what to do with the file (perhaps even offering a choice from a menu)
		- remember the resource location or (when in browser) even the actual file handle"
	| numFiles stream handler |
	
	Smalltalk tools userManager canDropOSFile ifFalse: [ ^ self ].
	
	numFiles := anEvent contents.
	1 to: numFiles do: [:i |
		(stream := FileStream requestDropStream: i) ifNotNil: [
			handler := ExternalDropHandler lookupExternalDropHandler: stream.
			[handler ifNotNil: [handler handle: stream in: self dropEvent: anEvent]]
				ensure: [stream close]]].! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'stephane.ducasse 9/25/2008 13:33'!
expandAll
	"Expand all windows"
	(self  windowsSatisfying: [:w | w isCollapsed])
		reverseDo: [:w | w collapseOrExpand.  self displayWorld]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'nk 1/6/2004 12:38'!
extractScreenRegion: poly andPutSketchInHand: hand
	"The user has specified a polygonal area of the Display.
	Now capture the pixels from that region, and put in the hand as a Sketch."
	| screenForm outline topLeft innerForm exterior |
	outline := poly shadowForm.
	topLeft := outline offset.
	exterior := (outline offset: 0@0) anyShapeFill reverse.
	screenForm := Form fromDisplay: (topLeft extent: outline extent).
	screenForm eraseShape: exterior.
	innerForm := screenForm trimBordersOfColor: Color transparent.
	innerForm isAllWhite ifFalse:
		[hand attachMorph: (self drawingClass withForm: innerForm)]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'MarcusDenker 9/29/2013 15:30'!
findAChangeSorter: evt
	"Locate a change sorter, open it, and bring it to the front.  Create one if necessary"
	| changeSorterClass |
	
	changeSorterClass := Smalltalk tools changeSorter.
	
	self 
		findAWindowSatisfying: [:aWindow | aWindow model isKindOf: changeSorterClass]
		orMakeOneUsing: [changeSorterClass open]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'nice 1/5/2010 15:59'!
findAWindowSatisfying: qualifyingBlock orMakeOneUsing: makeBlock 
	"Locate a window satisfying a block, open it, and bring it to the front.  Create one if necessary, by using the makeBlock"

	
	submorphs do: 
			[:aMorph | | aWindow | 
			(((aWindow := aMorph renderedMorph) isSystemWindow) 
				and: [qualifyingBlock value: aWindow]) 
					ifTrue: 
						[aWindow isCollapsed ifTrue: [aWindow expand].
						aWindow activateAndForceLabelToShow.
						^self]].
	"None found, so create one"
	makeBlock value openInWorld! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'CamilleTeruel 12/6/2013 14:44'!
findWindow: evt
	"Present a menu names of windows and naked morphs, and activate the one that gets chosen.  Collapsed windows appear below line, expand if chosen; naked morphs appear below second line; if any of them has been given an explicit name, that is what's shown, else the class-name of the morph shows; if a naked morph is chosen, bring it to front and have it don a halo."
	| menu expanded collapsed nakedMorphs |
	menu := UIManager default newMenuIn: self for: self.
	expanded := self  windowsSatisfying: [:w | w isCollapsed not].
	collapsed := self  windowsSatisfying: [:w | w isCollapsed].
	nakedMorphs := self submorphsSatisfying:[:m | m isSystemWindow not].
	(expanded isEmpty and: [collapsed isEmpty and: [nakedMorphs isEmpty]]) ifTrue: [^ self inform: 'No morph in world'].
	(expanded asSortedCollection: [:w1 :w2 | w1 label caseInsensitiveLessOrEqual: w2 label]) do:
		[:w | menu add: w label target: w action: #activateAndForceLabelToShow.
			w model canDiscardEdits ifFalse: [menu lastItem color: Color red]].
	(expanded isEmpty or: [ collapsed isEmpty and: [ nakedMorphs isEmpty ]]) ifFalse: [menu addLine].
	(collapsed asSortedCollection: [:w1 :w2 | w1 label caseInsensitiveLessOrEqual: w2 label]) do: 
		[:w | menu add: w label target: w action: #collapseOrExpand.
		w model canDiscardEdits ifFalse: [menu lastItem color: Color red]].
	nakedMorphs isEmpty ifFalse: [menu addLine].
	(nakedMorphs asSortedCollection: [:w1 :w2 | w1 class name caseInsensitiveLessOrEqual: w2 class name]) do:
		[:w | menu add: w class name target: w action: #comeToFrontAndAddHalo].
	menu addTitle: 'find window' translated.
	
	menu popUpEvent: evt in: self.! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'janniklaval 3/8/2011 12:31'!
fitAll
	"Fit all windows as visible"
	|allowedArea|
	allowedArea := RealEstateAgent maximumUsableAreaInWorld: World.
	(self  windowsSatisfying: [:w | w isCollapsed not])
		reverseDo:[:w | 
			w extent: w initialExtent.  
			w bounds:((w position extent: w initialExtent) translatedAndSquishedToBeWithin: allowedArea).
		].
	self displayWorld! !

!PasteUpMorph methodsFor: 'world state' stamp: 'StephaneDucasse 10/25/2013 16:17'!
flashRects: rectangleList color: aColor
	"For testing. Flashes the given list of rectangles on the Display so you can watch incremental redisplay at work."
	"Details: Uses two reverses so that the display is restored to its original state. This is necessary when in deferred update mode."

	| blt |
	blt := (BitBlt toForm: Display)
		sourceForm: nil;
		sourceOrigin: 0@0;
		clipRect: self viewBox;
		combinationRule: Form reverse.
	rectangleList do: [:r | | screenRect |
		screenRect := r translateBy: self viewBox origin.
		blt destRect: screenRect; copyBits.
		Display forceToScreen: screenRect; forceDisplayUpdate.
		(Delay forMilliseconds: 15) wait.
		blt destRect: screenRect; copyBits.
		Display forceToScreen: screenRect; forceDisplayUpdate].
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'IgorStasenko 1/2/2012 19:05'!
grabDrawingFromScreen: evt
	"Allow the user to specify a rectangular area of the Display, capture the pixels from that area, and use them to create a new drawing morph. Attach the result to the hand."
	| m |
	m := self drawingClass  withForm: Form fromUser.
	evt hand position: self activeHand cursorPoint.  "update hand pos after Sensor loop in fromUser"
	evt hand attachMorph: m.! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'di 10/18/2001 02:58'!
grabLassoFromScreen: evt
	"Allow the user to specify a polygonal area of the Display, capture the pixels from that area, and use them to create a new drawing morph. Attach the result to the hand."

	self extractScreenRegion: (PolygonMorph fromHandFreehand: evt hand)
		andPutSketchInHand: evt hand
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'di 10/18/2001 01:13'!
grabRubberBandFromScreen: evt
	"Allow the user to specify a polygonal area of the Display, capture the pixels from that area, and use them to create a new drawing morph. Attach the result to the hand."

	self extractScreenRegion: (PolygonMorph fromHand: evt hand)
		andPutSketchInHand: evt hand! !

!PasteUpMorph methodsFor: 'drawing' stamp: 'AlainPlantec 5/8/2010 00:31'!
gridFormOrigin: origin grid: smallGrid background: backColor line: lineColor

	| bigGrid gridForm gridOrigin |
	gridOrigin := origin \\ smallGrid.
	bigGrid := (smallGrid asPoint x) @ (smallGrid asPoint y).
	gridForm := Form extent: bigGrid depth: Display depth.
	backColor ifNotNil: [gridForm fillWithColor: backColor].
	gridOrigin x to: gridForm width by: smallGrid x do:
		[:x | gridForm fill: (x@0 extent: 1@gridForm height) fillColor: lineColor].
	gridOrigin y to: gridForm height by: smallGrid y do:
		[:y | gridForm fill: (0@y extent: gridForm width@1) fillColor: lineColor].
	^ InfiniteForm with: gridForm
! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:29'!
gridModulus

	^ self gridSpec extent! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:47'!
gridModulus: newModulus

	self gridSpecPut: (self gridOrigin extent: newModulus).
	self changed! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:28'!
gridOrigin

	^ self gridSpec origin! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:29'!
gridOrigin: newOrigin

	^ self gridSpecPut: (newOrigin extent: self gridModulus)! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:26'!
gridSpec
	"Gridding rectangle provides origin and modulus"

	^ self valueOfProperty: #gridSpec ifAbsent: [0@0 extent: 8@8]! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:28'!
gridSpecPut: newSpec
	"Gridding rectangle provides origin and modulus"

	^ self setProperty: #gridSpec toValue: newSpec! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:11'!
gridVisible

	^ self hasProperty: #gridVisible! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:47'!
gridVisibleOnOff

	self setProperty: #gridVisible toValue: self gridVisible not.
	self changed! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'dgd 12/13/2003 19:30'!
gridVisibleString
	"Answer a string to be used in a menu offering the opportunity 
	to show or hide the grid"
	^ (self gridVisible
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'show grid when gridding' translated! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:15'!
griddingOn

	^ griddingOn ifNil: [false]! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:48'!
griddingOnOff

	griddingOn := self griddingOn not.
	self changed! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'dgd 12/13/2003 19:30'!
griddingString
	"Answer a string to use in a menu offering the user the 
	opportunity to start or stop using gridding"
	^ (self griddingOn
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'use gridding' translated! !

!PasteUpMorph methodsFor: 'world state' stamp: 'alain.plantec 5/30/2008 14:07'!
handleFatalDrawingError: errMsg
	"Handle a fatal drawing error."
	Display deferUpdates: false. "Just in case"
	self primitiveError: errMsg.

	"Hm... we should jump into a 'safe' worldState here, but how do we find it?!!"! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'StephaneDucasse 7/2/2013 17:27'!
handlesKeyboard: evt
	^ evt keyCharacter == Character tab and: [ self tabAmongFields ]! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'ar 10/3/2000 22:46'!
handlesMouseDown: evt
	^true! !

!PasteUpMorph methodsFor: 'thumbnail' stamp: 'EstebanLorenzano 5/14/2013 09:44'!
icon
	^ Smalltalk ui icons homeIcon! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'StephaneDucasse 7/18/2010 15:20'!
initialize
	"initialize the state of the receiver"
	
	super initialize.
	self enableDragNDrop.
	self clipSubmorphs: true! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'nice 1/5/2010 15:59'!
initializeDesktopCommandKeySelectors
	"Provide the starting settings for desktop command key selectors.  Answer the dictionary."

	"ActiveWorld initializeDesktopCommandKeySelectors"
	| dict |
	dict := IdentityDictionary new.
	self defaultDesktopCommandKeyTriplets do:
		[:trip | | messageSend |
			messageSend := MessageSend receiver: trip second selector: trip third.
			dict at: trip first put: messageSend].
	self setProperty: #commandKeySelectors toValue: dict.
	^ dict

! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'AlainPlantec 2/17/2010 09:35'!
invokeWorldMenu: evt
	"Put up the world menu, triggered by the passed-in event."
	| menu |
	self bringTopmostsToFront.
	"put up screen menu"
	(menu := self worldMenu) popUpEvent: evt in: self.
	^ menu! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'AlainPlantec 2/17/2010 01:53'!
invokeWorldMenuFromEscapeKey
	self invokeWorldMenu: ActiveEvent! !

!PasteUpMorph methodsFor: 'testing' stamp: 'StephaneDucasse 6/28/2013 13:17'!
isEasySelecting
	"This is to isolate easySelection predicate. 
	Selectors in holders make no sense so we are limiting easy
	selection to the worldMorph.
	It would also make sense in playfield so feel free to adjust this
	predicate. Selection can always be forced by using the shift
	before mouse down."
	^ false! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'StephaneDucasse 7/2/2013 17:28'!
keyStroke: anEvent
	"A keystroke has been made.  Service event handlers and, if it's a keystroke presented to the world, dispatch it to #unfocusedKeystroke:"

	super keyStroke: anEvent.	"Give event handlers a chance"
	self selectedObject ifNotNil: [ :selected | selected moveOrResizeFromKeystroke: anEvent ].
	self dispatchKeystroke: anEvent! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'tbn 11/4/2008 09:15'!
keyboardNavigationHandler
	"Answer the receiver's existing keyboardNavigationHandler, or nil if none."

	| aHandler |
	aHandler := self valueOfProperty: #keyboardNavigationHandler ifAbsent: [^ nil].
	(aHandler hasProperty: #moribund) ifTrue:  "got clobbered in another project"
		[self removeProperty: #keyboardNavigationHandler.
		^ nil].
	^ aHandler! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/18/2003 23:10'!
keyboardNavigationHandler: aHandler
	"Set the receiver's keyboard navigation handler as indicated.  A nil argument means to remove the handler"

	aHandler
		ifNil:
			[self removeProperty: #keyboardNavigationHandler]
		ifNotNil:
			[self setProperty: #keyboardNavigationHandler toValue: aHandler]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'GuillermoPolito 9/4/2010 17:37'!
keystrokeInWorld: evt
	"A keystroke was hit when no keyboard focus was set, so it is sent here to the world instead."

	|  aChar ascii |
	aChar := evt keyCharacter.
	(ascii := aChar asciiValue) = 27 ifTrue: "escape key"
		[^ self invokeWorldMenuFromEscapeKey].
	
	(self navigationKey: evt) ifTrue: [^self].
	
	(evt controlKeyPressed not
		and: [(#(1 4 8 28 29 30 31 32) includes: ascii)  "home, end, backspace, arrow keys, space"
			and: [self keyboardNavigationHandler notNil]])
				ifTrue: [self keyboardNavigationHandler navigateFromKeystroke: aChar].

	(evt commandKeyPressed and: [Editor cmdKeysInText])
		ifTrue: [^ self dispatchCommandKeyInWorld: aChar event: evt].

	"It was unhandled. Remember the keystroke."
	self lastKeystroke: evt keyString.
	self triggerEvent: #keyStroke! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'AlainPlantec 12/13/2009 08:11'!
lastKeystroke	
	"Answer the last keystroke fielded by the receiver"	
	^ self valueOfProperty: #lastKeystroke ifAbsent: ['']! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'AlainPlantec 12/13/2009 08:12'!
lastKeystroke: aString	
	"Remember the last keystroke fielded by the receiver"	
	^ self setProperty: #lastKeystroke toValue: aString! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'SeanDeNigris 2/16/2014 20:33'!
makeAScreenshot
	| filePrefix |
	filePrefix :=  'PharoScreenshot'.
	(UIManager default
		chooseFrom: {'The entire world' translated. 'A selected area' translated}
		values: #(#world #area )
		message: 'What do you want to shoot? (File will be saved in image directory)' translated
		title: 'Make a screenshot')
		ifNotNil: [:choice | 
				|form |
				form := choice = #world
							ifTrue: [World imageForm]
							ifFalse: [Form fromUser].		
				PNGReadWriter
						putForm: form
						onFileNamed: (FileSystem workingDirectory / filePrefix , 'png') nextVersion] 
							
		! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'StephaneDucasse 6/27/2013 19:34'!
modalWindow: aMorph
	(self valueOfProperty: #modalWindow) ifNotNil: [ :morph | morph doCancel ].
	self setProperty: #modalWindow toValue: aMorph.
	aMorph ifNotNil: [ self when: #aboutToLeaveWorld send: #removeModalWindow to: self ]! !

!PasteUpMorph methodsFor: 'user interface' stamp: 'dgd 2/22/2003 14:11'!
modelWakeUp
	"I am the model of a SystemWindow, that has just been activated"

	| aWindow |
	owner isNil ifTrue: [^self].	"Not in Morphic world"
	(owner isKindOf: TransformMorph) ifTrue: [^self viewBox: self fullBounds].
	(aWindow := self containingWindow) ifNotNil: 
			[self viewBox = aWindow panelRect 
				ifFalse: [self viewBox: aWindow panelRect]]! !

!PasteUpMorph methodsFor: 'Morphic-Base-Basic' stamp: 'adrian_lienhard 7/19/2009 20:50'!
morphToDropFrom: aMorph 
	"Given a morph being carried by the hand, which the hand is about to drop, answer the actual morph to be deposited.  Normally this would be just the morph itself, but several unusual cases arise, which this method is designed to service."

	^aMorph ! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'ar 2/23/2001 16:44'!
morphToGrab: event
	"Return the morph to grab from a mouse down event. If none, return nil."
	self submorphsDo:[:m|
		((m rejectsEvent: event) not and:[m fullContainsPoint: event cursorPoint]) ifTrue:[^m].
	].
	^nil! !

!PasteUpMorph methodsFor: 'submorphs-accessing' stamp: 'nk 7/4/2003 16:49'!
morphsInFrontOf: someMorph overlapping: aRectangle do: aBlock
	"Include hands if the receiver is the World"
	self handsDo:[:m|
		m == someMorph ifTrue:["Try getting out quickly"
			owner ifNil:[^self].
			^owner morphsInFrontOf: self overlapping: aRectangle do: aBlock].
		"The hand only overlaps if it's not the hardware cursor"
		m needsToBeDrawn ifTrue:[
			(m fullBoundsInWorld intersects: aRectangle)
				ifTrue:[aBlock value: m]]].
	^super morphsInFrontOf: someMorph overlapping: aRectangle do: aBlock! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'StephaneDucasse 7/2/2013 17:40'!
mouseDown: evt
	"Handle a mouse down event."
	| grabbedMorph handHadHalos |
	(self defaultYellowButtonMenuEnabled and: [evt yellowButtonPressed])
		ifTrue: [  (self yellowButtonActivity: evt shiftPressed) ifTrue: [ ^ self ]].

	grabbedMorph := self morphToGrab: evt.
	grabbedMorph ifNotNil: [
		grabbedMorph isSticky ifTrue: [^self].
		^evt hand grabMorph: grabbedMorph].

	(super handlesMouseDown: evt)
		ifTrue: [^super mouseDown: evt].

	handHadHalos := evt hand halo notNil.
	evt hand removeHalo. "shake off halos"
	self isWorldMorph ifTrue: [
		self currentWindow ifNotNil: [:topWindow |
			SystemWindow passivateTopWindow]]. "since pasteup will release keyboard focus now"
	evt hand releaseKeyboardFocus. "shake of keyboard foci"
 
	(evt shiftPressed not
			and: [ self isWorldMorph not 
				and: [ self wantsEasySelection not ]])
		ifTrue: [
		"explicitly ignore the event if we're not the world and we'll not select,
		so that we could be picked up if need be"
		evt wasHandled: false.
		^ self.
	].

	( evt shiftPressed or: [ self wantsEasySelection ] )
		 ifTrue: [
			"We'll select on drag, let's decide what to do on click"
			| clickSelector |
			clickSelector := nil.
			evt shiftPressed 
				ifTrue: [ clickSelector := #findWindow:.]
				ifFalse: [self isWorldMorph 
							ifTrue: [clickSelector := handHadHalos
										ifTrue: [ #delayedInvokeWorldMenu: ]
										ifFalse: [ #invokeWorldMenu: ]]].
		evt hand 
				waitForClicksOrDrag: self 
				event: evt 
				selectors: { clickSelector. nil. nil. #dragThroughOnDesktop: }
				threshold: 5.
	]
	ifFalse: [
		"We wont select, just bring world menu if I'm the world"
		self isWorldMorph ifTrue: [
			handHadHalos
				ifTrue: [ self delayedInvokeWorldMenu: evt ]
				ifFalse: [ self invokeWorldMenu: evt ]
		]
	].
! !

!PasteUpMorph methodsFor: 'taskbar' stamp: 'MarcusDenker 3/23/2011 15:29'!
moveCollapsedWindowsToTaskbar
	"Move collapsed windows to the taskbar."
	
	(self systemWindows select: [:w | w isCollapsed]) do: [:w |
		w restore; minimize]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'AlainPlantec 10/17/2009 17:32'!
nonWindows
	^ (self submorphs select: [:m | (m isSystemWindow) not and: [m wantsToBeTopmost not]])! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'sw 7/6/1999 13:26'!
originAtCenter
	^ self hasProperty: #originAtCenter! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'stephane.ducasse 5/21/2009 15:15'!
positionNear: aPoint forExtent: anExtent adjustmentSuggestion: adjustmentPoint
	"Compute a plausible positioning for adding a subpart of size anExtent, somewhere near aPoint, using adjustmentPoint as the unit of adjustment"

	| adjustedPosition |
	adjustedPosition := aPoint.
	[((self morphsAt: (adjustedPosition + (anExtent // 2))) size > 1) and:  "that 1 is self here"
		[bounds containsPoint: adjustedPosition]]
	whileTrue:
		[adjustedPosition := adjustedPosition + adjustmentPoint].

	^ adjustedPosition! !

!PasteUpMorph methodsFor: 'private' stamp: 'nk 7/8/2003 09:18'!
privateRemoveMorph: aMorph
	backgroundMorph == aMorph ifTrue: [ backgroundMorph := nil ].
	^super privateRemoveMorph: aMorph.
! !

!PasteUpMorph methodsFor: 'caching' stamp: 'StephaneDucasse 7/2/2013 17:33'!
releaseCachedState
	super releaseCachedState.
	self removeModalWindow.
! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'tak 3/15/2005 17:31'!
removeModalWindow
	self modalWindow: nil! !

!PasteUpMorph methodsFor: 'taskbar' stamp: 'MarcusDenker 3/23/2011 15:37'!
removeTaskbar
	"Remove the receiver's taskbars."
	
	self taskbars do: [:each | each removeFromWorld]! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:22'!
repelsMorph: aMorph event: ev
	(aMorph wantsToBeDroppedInto: self) ifFalse: [^ false].
	self dropEnabled ifFalse: [^ true].
	(self wantsDroppedMorph: aMorph event: ev) ifFalse: [^ true].
	^ super repelsMorph: aMorph event: ev "consults #repelling flag"! !

!PasteUpMorph methodsFor: 'options' stamp: 'ar 11/9/2000 12:49'!
resizeToFit
	^self vResizing == #shrinkWrap! !

!PasteUpMorph methodsFor: 'options' stamp: 'dgd 12/13/2003 19:30'!
resizeToFitString
	"Answer a string, to be used in a self-updating menu, to 
	represent whether the receiver is currently using resize-to-fit 
	or not"
	^ (self resizeToFit
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'resize to fit' translated! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 13:56'!
respondToCommand: aCharacter bySending: aSelector to: aReceiver
	"Respond to the command-key use of the given character by sending the given selector to the given receiver.  If the selector is nil, retract any prior such setting"

	aSelector
		ifNil:
			[self commandKeySelectors removeKey: aCharacter]
		ifNotNil:
			[self commandKeySelectors at: aCharacter put: (MessageSend receiver: aReceiver selector: aSelector)]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 3/18/2001 00:35'!
restoreDisplay

	World restoreMorphicDisplay.	"I don't actually expect this to be called"! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 9/5/2004 19:46'!
restoreMainDockingBarDisplay
	"Restore the display of docking bars"
	self dockingBars
		do: [:each | each updateBounds]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'AlainPlantec 7/9/2013 12:39'!
restoreMorphicDisplay

	DisplayScreen startUp.

	ThumbnailMorph recursionReset.

	self
		extent: Display extent;
		viewBox: Display boundingBox;
		handsDo: [:h | h visible: true; showTemporaryCursor: nil];
		resizeBackgroundMorph;
		restoreMainDockingBarDisplay;
		fullRepaintNeeded.
		
	self defer: [Cursor normal show].
		
	self layoutChanged.
! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'MarcusDenker 5/2/2013 11:35'!
setGridSpec
	"Gridding rectangle provides origin and modulus"
	| response result |
	response := UIManager default
			request: 'New grid origin (usually 0@0):' translated
			initialAnswer: self gridOrigin printString.
	response isEmpty ifTrue: [^ self].
	result := [self class compiler evaluate: response] ifError: [^ self].
	(result isPoint and: [(result >= (0@0))])
		ifTrue: [self gridOrigin: result]
		ifFalse: [self inform: ('Must be a Point with coordinates (for example 10@10)' translated )].

	response := UIManager default
			request: 'New grid spacing:' translated
			initialAnswer: self gridModulus printString.
	response isEmptyOrNil ifTrue: [^ self].
	result := [self class compiler evaluate: response] ifError: [^ self].
	(result isPoint and: [(result > (0@0)) ])
		ifTrue: [self gridModulus: result]
		ifFalse: [self inform: ('Must be a Point with coordinates (for example 10@10)' translated )].

! !

!PasteUpMorph methodsFor: 'wiw support' stamp: 'RAA 8/14/2000 12:10'!
shouldGetStepsFrom: aWorld

	(self isWorldMorph and: [owner notNil]) ifTrue: [
		^self outermostWorldMorph == aWorld
	].
	^super shouldGetStepsFrom: aWorld! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 2/13/2010 11:38'!
showImage: aForm 
	"Show an image, possibly attached to the pointer for positioning"

       HandMorph attach: (World drawingClass withForm: aForm)! !

!PasteUpMorph methodsFor: 'taskbar' stamp: 'MarcusDenker 3/23/2011 15:37'!
showWorldTaskbar: aBoolean
	"Change the receiver to show the taskbar."
	
	aBoolean
		ifTrue: [self createTaskbarIfNecessary]
		ifFalse: [self removeTaskbar]! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:39'!
startStepping: aMorph
	"Add the given morph to the step list. Do nothing if it is already being stepped."
	^self startStepping: aMorph at: Time millisecondClockValue selector: #stepAt: arguments: nil stepTime: nil! !

!PasteUpMorph methodsFor: 'project state' stamp: 'MarcusDenker 9/13/2013 15:51'!
steppingMorphsNotInWorld
	| all |
	all := self allMorphs.
	^ self listOfSteppingMorphs reject: [ :m | all includes: m ]	"self currentWorld steppingMorphsNotInWorld do: [:m | m delete]"! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 7/6/1999 13:36'!
toggleOriginAtCenter
	| hasIt |
	hasIt := self hasProperty: #originAtCenter.
	hasIt
		ifTrue:
			[self removeProperty: #originAtCenter]
		ifFalse:
			[self setProperty: #originAtCenter toValue: true]! !

!PasteUpMorph methodsFor: 'options' stamp: 'tk 10/30/2001 18:41'!
toggleResizeToFit
	"Toggle whether the receiver is set to resize-to-fit"

	self vResizeToFit: self resizeToFit not! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'ar 4/25/2001 17:15'!
useRoundedCorners
	"Somewhat special cased because we do have to fill Display for this"
	super useRoundedCorners.
	self == World ifTrue:[Display bits primFill: 0]. "done so that we *don't* get a flash"! !

!PasteUpMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/28/2013 13:14'!
viewBox: newViewBox 
	"I am now displayed within newViewBox; react."

	super position: newViewBox topLeft.
	fullBounds := bounds := newViewBox.

	! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'StephaneDucasse 6/27/2013 19:32'!
wantsDroppedMorph: aMorph event: evt
	
	self visible ifFalse: [^ false].  "will be a call to #hidden again very soon"
	self dropEnabled ifFalse: [^ false].
	^ true! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'AlainPlantec 12/10/2009 11:38'!
wantsEasySelection
	"Answer if the receiver want easy selection mode"
	^ self isEasySelecting ! !

!PasteUpMorph methodsFor: 'halos and balloon help' stamp: 'StephaneDucasse 2/19/2010 16:02'!
wantsHaloFromClick
	(owner isSystemWindow) ifTrue: [^ false].
	^ true.
! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'StephaneDucasse 7/18/2010 16:37'!
wantsKeyboardFocusFor: aSubmorph

	aSubmorph wouldAcceptKeyboardFocus ifTrue: [ ^ true].
	^ super wantsKeyboardFocusFor: aSubmorph! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'StephaneDucasse 7/2/2013 17:28'!
wantsWindowEvent: anEvent
	^ self windowEventHandler notNil! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'HenrikSperreJohansen 4/7/2010 09:58'!
windowEvent: anEvent
	self windowEventHandler
		ifNotNil: [^self windowEventHandler windowEvent: anEvent].

	anEvent type == #windowClose
		ifTrue: [WorldState quitSession].
! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'tbn 3/12/2010 01:47'!
windowEventHandler
	"This is a class variable so it is global to all projects and does not get saved"
	^WindowEventHandler! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'tbn 3/12/2010 01:47'!
windowEventHandler: anObject
	"This is a class variable so it is global to all projects and does not get saved"
	WindowEventHandler := anObject
! !

!PasteUpMorph methodsFor: 'Morphic-Base-Windows' stamp: 'nice 1/5/2010 15:59'!
windowsSatisfying: windowBlock

	| windows |
	windows := OrderedCollection new.
	self submorphs do:
		[:m | | s |
		((m isSystemWindow) and: [windowBlock value: m])
			ifTrue: [windows addLast: m]
			ifFalse: [((m isKindOf: TransformationMorph) and: [m submorphs size = 1])
					ifTrue: [s := m firstSubmorph.
							((s isSystemWindow) and: [windowBlock value: s])
								ifTrue: [windows addLast: s]]]].
	^ windows! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:04'!
worldMenu
	^ owner worldMenu
	! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:04'!
worldMenuAt: aMenuItemName
	^ owner worldMenuAt: aMenuItemName
		! !

PasteUpMorph subclass: #WorldMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Worlds'!

!WorldMorph commentStamp: '<historical>' prior: 0!
A WorldMorph is a special morph that represents the world.
It is a paste up and it holds active hand.

A World, the entire Smalltalk screen, is a PasteUpMorph.  A World responds true to isWorld.!

!WorldMorph class methodsFor: 'system startup' stamp: 'StephaneDucasse 6/27/2013 17:00'!
cleanUp
	World cleanseOtherworldlySteppers.! !

!WorldMorph class methodsFor: 'setting' stamp: 'StephaneDucasse 6/27/2013 19:29'!
defaultWorldColor

	^ Color r: 0.937 g: 0.937 b: 0.937.
! !

!WorldMorph class methodsFor: 'world creation' stamp: 'StephaneDucasse 7/2/2013 16:07'!
installNewWorld
	"let us play God."
	
	 [  |  man |
      
        man := MorphicUIManager new.
        "we do not use UIManager default because in headless mode there is not such process."
        man uiProcess ifNotNil: [ man terminateUIProcess  ].
       self new installForUIProcessReinstall.
        man spawnNewProcess.
     
    ] forkAt:  Processor userInterruptPriority! !

!WorldMorph class methodsFor: 'system startup' stamp: 'MarcusDenker 10/15/2013 10:32'!
shutDown
	World ifNotNil: [:world | world triggerEvent: #aboutToLeaveWorld ]! !

!WorldMorph class methodsFor: 'system startup' stamp: 'MarcusDenker 10/15/2013 10:31'!
startUp
	World ifNotNil: [:world | world restoreMorphicDisplay]! !

!WorldMorph methodsFor: 'submorphs-add/remove' stamp: 'CamilleTeruel 12/6/2013 14:45'!
acceptDroppingMorph: dropped event: evt
	"The supplied morph, known to be acceptable to the receiver, is now to be assimilated; the precipitating event is supplied"

	| aMorph |
	aMorph := self morphToDropFrom: dropped.	"Add the given morph to this world and start stepping it if it wants to be."
	self addMorphFront: aMorph.
	(aMorph fullBounds intersects: self viewBox)
		ifFalse: [ 
			self inform: 'Dropped morph not in bounds'.
			aMorph position: self bounds center ].
	aMorph
		submorphsDo: [ :m | 
			(m isKindOf: HaloMorph)
				ifTrue: [ m delete ] ].
	self bringTopmostsToFront! !

!WorldMorph methodsFor: 'structure' stamp: 'StephaneDucasse 6/27/2013 17:46'!
activeHand

	^ worldState activeHand! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 18:57'!
activeHand: aHandMorph
	"temporarily retained for old main event loops"
	worldState activeHand: aHandMorph.

! !

!WorldMorph methodsFor: 'alarms-scheduler' stamp: 'StephaneDucasse 6/27/2013 17:23'!
addAlarm: aSelector withArguments: argArray for: aTarget at: scheduledTime
	"Add a new alarm with the given set of parameters"
	worldState addAlarm: aSelector withArguments: argArray for: aTarget at: scheduledTime.! !

!WorldMorph methodsFor: 'submorphs-add/remove' stamp: 'StephaneDucasse 7/2/2013 16:30'!
addAllMorphs: array

	super addAllMorphs: array.
	array do: [:m | self startSteppingSubmorphsOf: m].
! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 18:56'!
addHand: aHandMorph
	"Add the given hand to the list of hands for this world."

	aHandMorph owner ifNotNil: [ aHandMorph owner removeHand: aHandMorph ].
	worldState addHand: aHandMorph.
	aHandMorph privateOwner: self! !

!WorldMorph methodsFor: 'submorphs-accessing' stamp: 'StephaneDucasse 6/27/2013 17:51'!
allMorphsDo: aBlock
	"Enumerate all morphs in the world, including those held in hands."

	super allMorphsDo: aBlock.
	worldState handsReverseDo: [ :h | h allMorphsDo: aBlock ]! !

!WorldMorph methodsFor: 'accessing' stamp: 'StephaneDucasse 6/27/2013 17:22'!
announcer
	WorldAnnouncer ifNil: [ WorldAnnouncer := Announcer new ].
	^ WorldAnnouncer! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 18:57'!
assuredCanvas
	^ worldState assuredCanvas! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/27/2013 17:50'!
canvas

	^ worldState canvas! !

!WorldMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 6/28/2013 13:12'!
cleanseStepList
	"Remove morphs from the step list that are not in this World."

	worldState cleanseStepListForWorld: self! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 17:31'!
defaultWorldColor
	^ self class defaultWorldColor! !

!WorldMorph methodsFor: 'deferred message' stamp: 'AlainPlantec 7/9/2013 13:40'!
defer: aValuable
	worldState defer: aValuable! !

!WorldMorph methodsFor: 'menu & halo' stamp: 'StephaneDucasse 6/27/2013 20:07'!
deleteBalloonTarget: aMorph
	"Delete the balloon help targeting the given morph"

	self handsDo: [ :h | h deleteBalloonTarget: aMorph ]! !

!WorldMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:20'!
discoveredWorldMenu
	^ worldState discoveredWorldMenu! !

!WorldMorph methodsFor: 'menu & halo' stamp: 'ClementBera 10/25/2013 14:00'!
dispatchKeystroke: anEvent
	"A keystroke was hit when no keyboard focus was set, so it is sent here to the world instead."

	|  aChar ascii |
	aChar := anEvent keyCharacter.
	(ascii := aChar asciiValue) = 27 ifTrue: "escape key"
		[^ self invokeWorldMenuFromEscapeKey].
	
	(self navigationKey: anEvent) ifTrue: [^self].

	(anEvent commandKeyPressed and: [Editor cmdKeysInText])
		ifTrue: [^ self dispatchCommandKeyInWorld: aChar event: anEvent].

	"It was unhandled. Remember the keystroke."
	self lastKeystroke: anEvent keyString.
	self triggerEvent: #keyStroke.
	super dispatchKeystroke: anEvent! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 18:55'!
displayWorld
	worldState displayWorld: self submorphs: submorphs! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 18:57'!
displayWorldSafely
	worldState displayWorldSafely: self.
! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 17:57'!
doOneCycle
	"see the comment in doOneCycleFor:"

	worldState doOneCycleFor: self! !

!WorldMorph methodsFor: 'interaction loop' stamp: 'StephaneDucasse 6/27/2013 19:00'!
doOneCycleNow
	"see the comment in doOneCycleNowFor:"
	worldState doOneCycleNowFor: self.
! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/28/2013 13:12'!
doOneSubCycle
	"Like doOneCycle, but preserves activeHand."

	worldState doOneSubCycleFor: self! !

!WorldMorph methodsFor: 'geometry' stamp: 'StephaneDucasse 6/27/2013 18:51'!
extent: aPoint
	super extent: aPoint.
	worldState viewBox
		ifNotNil: [ 
			worldState canvas: nil.
			worldState viewBox: bounds ]! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/27/2013 18:51'!
firstHand
	^ worldState hands first! !

!WorldMorph methodsFor: 'geometry testing' stamp: 'StephaneDucasse 6/28/2013 13:23'!
fullContainsPoint: pt
	^bounds containsPoint: pt
! !

!WorldMorph methodsFor: 'Morphic-Base-Windows' stamp: 'StephaneDucasse 6/28/2013 13:25'!
fullRepaintNeeded

	worldState doFullRepaint.
	self  windowsSatisfying: [:w | w makeMeVisible. false].

! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 18:53'!
haloMorphs

	^ self hands collect: [ :h | h halo ] thenSelect: [ :halo | halo notNil ]! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/27/2013 18:47'!
hands

	^ worldState hands! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/27/2013 18:57'!
handsDo: aBlock

	^  worldState handsDo: aBlock ! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/27/2013 18:58'!
handsReverseDo: aBlock
	^ worldState handsReverseDo: aBlock! !

!WorldMorph methodsFor: 'initialization' stamp: 'StephaneDucasse 7/1/2013 10:09'!
initialize 
	worldState := WorldState new.
 	super initialize! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 7/2/2013 16:10'!
install
	"this method was used to be invoked from onPrimitiveError: 
	the problem is that it cannot be used to recreate the world instance.
	For this have a look class side installNewWorld."
	owner := nil.	"since we may have been inside another world previously"
	ActiveWorld := self.
	World := self.
	ActiveHand := self hands first.	"default"
	ActiveEvent := nil.
	submorphs do: [:ss | ss owner isNil ifTrue: [ss privateOwner: self]].
	"Transcript that was in outPointers and then got deleted."
	self viewBox: Display boundingBox.
	Sensor flushAllButDandDEvents.
	worldState handsDo: [:h | h initForEvents].
	self borderWidth: 0.	"default"
	SystemWindow noteTopWindowIn: self.
	self displayWorldSafely! !

!WorldMorph methodsFor: 'initialization' stamp: 'StephaneDucasse 7/2/2013 16:07'!
installForUIProcessReinstall
	"This method should be changed because it cannot be invoked from onPrimitiveError: as the previous PasteUpMorph 
	method was used to."
	
	| morphs  |
	owner := nil.
	morphs := World submorphs.
	"we should decide if we get morph or not. Now submorphs returns an array and this is 
	strange that this is not an OrderedCollection"
	
	Sensor
		flushNonKbdEvents;
		flushKeyboard;
		flushEvents;
		shutDown.
	self class shutDown.
 
	self addHand: HandMorph new.	
	self borderWidth: 0.
	self viewBox: Display boundingBox.
	self borderWidth: 0.	"default"
	self color: self class defaultWorldColor.
	self addAllMorphs: morphs.
	SystemWindow noteTopWindowIn: self.
	World  := self.
	ActiveWorld := self.
	ActiveHand := self hands first.	"default"
	ActiveEvent := nil.
	worldState handsDo: [ :h | h releaseCachedState ].
	Sensor startUp.
	self class startUp.
	self displayWorldSafely.! !

!WorldMorph methodsFor: 'change reporting' stamp: 'StephaneDucasse 6/27/2013 18:52'!
invalidRect: damageRect from: aMorph
	"Clip damage reports to my bounds, since drawing is clipped to my bounds."

	worldState recordDamagedRect: (damageRect intersect: self bounds ifNone: [ ^ self ])! !

!WorldMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:17'!
isEasySelecting
	"This is to isolate easySelection predicate. 
	Selectors in holders make no sense so we are limiting easy
	selection to the worldMorph.
	It would also make sense in playfield so feel free to adjust this
	predicate. Selection can always be forced by using the shift
	before mouse down."
	^ worldState isEasySelecting! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/28/2013 13:11'!
isStepping: aMorph
	^ worldState isStepping: aMorph! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/28/2013 13:10'!
isStepping: aMorph selector: aSelector
	^ worldState isStepping: aMorph selector: aSelector! !

!WorldMorph methodsFor: 'classification' stamp: 'StephaneDucasse 6/27/2013 16:48'!
isWorldMorph

	^ true! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/28/2013 13:09'!
listOfSteppingMorphs
	^ worldState listOfSteppingMorphs

"self currentWorld listOfSteppingMorphs"! !

!WorldMorph methodsFor: 'event handling' stamp: 'StephaneDucasse 6/27/2013 19:43'!
mouseUp: evt
	self removeAlarm: #invokeWorldMenu:.
	super mouseUp: evt.! !

!WorldMorph methodsFor: 'geometry' stamp: 'MarcusDenker 10/10/2013 23:48'!
position: aPoint
	"Prevent moving a world (e.g. via HandMorph>>specialGesture:)"

	super position: aPoint.
	self viewBox ifNotNil: [:viewBox | self viewBox: (aPoint extent:  viewBox extent)].

! !

!WorldMorph methodsFor: 'printing' stamp: 'StephaneDucasse 6/27/2013 19:02'!
printOn: aStream
	"Reimplemented to add a tag showing that the receiver is currently functioning as a 'world', if it is"

	super printOn: aStream.
	aStream nextPutAll: ' [world]'! !

!WorldMorph methodsFor: 'private' stamp: 'StephaneDucasse 6/27/2013 19:24'!
privateMoveBy: delta
	super privateMoveBy: delta.
	worldState viewBox ifNotNil: [ worldState viewBox: bounds ]! !

!WorldMorph methodsFor: 'event handling' stamp: 'StephaneDucasse 7/2/2013 17:33'!
releaseCachedState
	super releaseCachedState.
	self cleanseStepList.! !

!WorldMorph methodsFor: 'alarms-scheduler' stamp: 'StephaneDucasse 6/27/2013 18:51'!
removeAlarm: aSelector for: aTarget
	"Remove the alarm with the given selector"
	
	worldState removeAlarm: aSelector for: aTarget! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/27/2013 18:51'!
removeHand: aHandMorph
	"Remove the given hand from the list of hands for this world."

	(worldState hands includes: aHandMorph)
		ifFalse: [ ^ self ].
	aHandMorph dropMorphs.
	self invalidRect: aHandMorph fullBounds.
	worldState removeHand: aHandMorph! !

!WorldMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:08'!
resetWorldMenu
	
	worldState resetWorldMenu
! !

!WorldMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 6/28/2013 13:08'!
runLocalStepMethods

	worldState runLocalStepMethodsIn: self
! !

!WorldMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 6/28/2013 13:07'!
runStepMethods

	worldState runStepMethodsIn: self
! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/28/2013 13:15'!
sleep

	worldState canvas ifNil: [^ self  "already called (clean this up)"].
	Cursor normal show.	"restore the normal cursor"
	worldState canvas: nil.		"free my canvas to save space"
	self fullReleaseCachedState.
! !

!WorldMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 6/28/2013 13:07'!
startStepping: aMorph at: scheduledTime selector: aSelector arguments: args stepTime: stepTime
	worldState startStepping: aMorph at: scheduledTime selector: aSelector arguments: args stepTime: stepTime.! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 7/2/2013 17:21'!
startSteppingSubmorphsOf: aMorph
	"Ensure that all submorphs of the given morph that want to be stepped are added to the step list.   Typically used after adding a morph to the world."

	aMorph allMorphsDo: [:m |
		m wantsSteps ifTrue: [m arrangeToStartSteppingIn: m world].
	]

! !

!WorldMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 7/2/2013 16:50'!
step
	owner ifNotNil: [ ^ self runLocalStepMethods ].
	^ super step! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/28/2013 13:07'!
stepListSize
	^ worldState stepListSize

"Transcript cr; show: self currentWorld stepListSize printString, ' items on steplist as of ', Date dateAndTimeNow printString"! !

!WorldMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 6/28/2013 13:07'!
stopStepping: aMorph
	"Remove the given morph from the step list."

	worldState stopStepping: aMorph
! !

!WorldMorph methodsFor: 'stepping' stamp: 'StephaneDucasse 6/28/2013 13:06'!
stopStepping: aMorph selector: aSelector
	"Remove the given morph from the step list."

	worldState stopStepping: aMorph selector: aSelector
! !

!WorldMorph methodsFor: 'copying' stamp: 'StephaneDucasse 6/27/2013 18:51'!
veryDeepCopyWith: deepCopier
	"never copy the World"

	^ self! !

!WorldMorph methodsFor: 'project state' stamp: 'StephaneDucasse 6/27/2013 18:58'!
viewBox

	^ worldState viewBox! !

!WorldMorph methodsFor: 'world state' stamp: 'StephaneDucasse 6/28/2013 13:15'!
viewBox: newViewBox
	"I am now displayed within newViewBox; react."

	(self viewBox isNil or: [ self viewBox extent ~= newViewBox extent ])
		ifTrue: [ worldState canvas: nil ].
	worldState viewBox: newViewBox.
	super viewBox: newViewBox.
	worldState handsDo: [ :hand | hand releaseKeyboardFocus ].
	self fullRepaintNeeded! !

!WorldMorph methodsFor: 'halos and balloon help' stamp: 'StephaneDucasse 7/2/2013 17:19'!
wantsDirectionHandles

	^ false! !

!WorldMorph methodsFor: 'event handling' stamp: 'StephaneDucasse 7/2/2013 17:13'!
wantsDropFiles: anEvent
	^ true! !

!WorldMorph methodsFor: 'private' stamp: 'StephaneDucasse 6/27/2013 19:33'!
wantsDroppedMorph: aMorph event: evt
	
	^ true! !

!WorldMorph methodsFor: 'menu & halo' stamp: 'StephaneDucasse 7/2/2013 17:30'!
wantsWindowEvent: anEvent
	^ true! !

!WorldMorph methodsFor: 'structure' stamp: 'StephaneDucasse 6/27/2013 19:03'!
world
	
	^self! !

!WorldMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:05'!
worldMenu
	^ worldState worldMenu
	! !

!WorldMorph methodsFor: 'world menu' stamp: 'StephaneDucasse 6/28/2013 13:04'!
worldMenuAt: aMenuItemName
	
	^ worldState worldMenuAt: aMenuItemName! !

Morph subclass: #HandMorph
	instanceVariableNames: 'mouseFocus keyboardFocus eventListeners mouseListeners keyboardListeners mouseClickState mouseOverHandler lastMouseEvent targetOffset damageRecorder cacheCanvas cachedCanvasHasHoles temporaryCursor temporaryCursorOffset hardwareCursor hasChanged savedPatch lastEventBuffer lastKeyScanCode combinedChar eventSource lastSystemEvent captureBlock recentModifiers'
	classVariableNames: 'DoubleClickTime EventSource EventStats NormalCursor PasteBuffer ShowEvents UpperHandLimit'
	poolDictionaries: 'EventSensorConstants'
	category: 'Morphic-Core-Kernel'!

!HandMorph commentStamp: '<historical>' prior: 0!
The cursor may be thought of as the HandMorph.  The hand's submorphs hold anything being carried by dragging.  

There is some minimal support for multiple hands in the same world.!

!HandMorph class methodsFor: 'utilities' stamp: 'sma 4/30/2000 10:34'!
attach: aMorph
	"Attach aMorph the current world's primary hand."

	self currentWorld primaryHand attachMorph: aMorph! !

!HandMorph class methodsFor: 'accessing' stamp: ''!
doubleClickTime

	^ DoubleClickTime
! !

!HandMorph class methodsFor: 'accessing' stamp: ''!
doubleClickTime: milliseconds

	DoubleClickTime := milliseconds.
! !

!HandMorph class methodsFor: 'class initialization' stamp: 'kfr 7/13/2003 14:15'!
initialize
	"HandMorph initialize"

	PasteBuffer := nil.
	DoubleClickTime := 350.
	NormalCursor := CursorWithMask normal asCursorForm.
! !

!HandMorph class methodsFor: 'utilities' stamp: 'MarcusDenker 7/28/2013 13:03'!
showEvents: aBool
	"HandMorph showEvents: true"
	"HandMorph showEvents: false"
	ShowEvents := aBool.
	aBool ifFalse: [ World invalidRect: (0@0 extent: 500@120) ].! !

!HandMorph class methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 6/25/2013 17:43'!
upperHandLimit

	^ UpperHandLimit ifNil: [ UpperHandLimit := 0 ]! !

!HandMorph class methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 6/25/2013 17:45'!
upperHandLimit: anInteger

	UpperHandLimit := anInteger! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/24/2000 20:43'!
addEventListener: anObject
	"Make anObject a listener for all events. All events will be reported to the object."
	self eventListeners: (self addListener: anObject to: self eventListeners)! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/24/2000 20:40'!
addListener: anObject to: aListenerGroup
	"Add anObject to the given listener group. Return the new group."
	| listeners |
	listeners := aListenerGroup.
	(listeners notNil and:[listeners includes: anObject]) ifFalse:[
		listeners
			ifNil:[listeners := WeakArray with: anObject]
			ifNotNil:[listeners := listeners copyWith: anObject]].
	listeners := listeners copyWithout: nil. "obsolete entries"
	^listeners! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/24/2000 20:42'!
addMouseListener: anObject
	"Make anObject a listener for mouse events. All mouse events will be reported to the object."
	self mouseListeners: (self addListener: anObject to: self mouseListeners)! !

!HandMorph methodsFor: 'accessing' stamp: 'tk 10/20/2004 15:54'!
anyButtonPressed
	^lastMouseEvent anyButtonPressed! !

!HandMorph methodsFor: 'accessing' stamp: 'IgorStasenko 1/22/2012 18:24'!
anyModifierKeyPressed
	^recentModifiers anyMask: 16r0E	"cmd | opt | ctrl"! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 10/8/2000 23:42'!
attachMorph: m
	"Position the center of the given morph under this hand, then grab it.
	This method is used to grab far away or newly created morphs."
	| delta |
	self releaseMouseFocus. "Break focus"
	delta := m bounds extent // 2.
	m position: (self position - delta).
	m formerPosition: m position.
	targetOffset := m position - self position.
	self addMorphBack: m.! !

!HandMorph methodsFor: 'balloon help' stamp: 'ar 10/3/2000 16:49'!
balloonHelp
	"Return the balloon morph associated with this hand"
	^self valueOfProperty: #balloonHelpMorph! !

!HandMorph methodsFor: 'balloon help' stamp: 'MarcusDenker 10/10/2013 23:45'!
balloonHelp: aBalloonMorph
	"Return the balloon morph associated with this hand"
	self balloonHelp ifNotNil:[:oldHelp |oldHelp delete].
	aBalloonMorph
		ifNil:[self removeProperty: #balloonHelpMorph]
		ifNotNil:[self setProperty: #balloonHelpMorph toValue: aBalloonMorph]! !

!HandMorph methodsFor: 'events-processing' stamp: 'IgorStasenko 1/2/2012 18:58'!
captureEventsUntil: aBlock
"
	Capture all input events, bypassing normal processing flow and redirect all events into block instead.
	Repeat until block will answer true.

	World activeHand captureEventsUntil: [:evt |
		evt isKeyboard and: [ evt keyCharacter = $a ] ]
"

	| release |
	
	release := false.
	
	captureBlock := [:evt | release := aBlock value: evt ].
	
	[
		[ self world doOneCycle. release ] whileFalse.
	] ensure: [
		captureBlock := nil.
	]! !

!HandMorph methodsFor: 'events-processing' stamp: 'IgorStasenko 1/2/2012 18:40'!
captureEventsWhile: aBlock
"
	Capture all input events, bypassing normal processing flow and redirect all events into block instead.
	Repeat until block will answer false.

"

	^ self captureEventsUntil: [:evt | (aBlock value: evt) not ]! !

!HandMorph methodsFor: 'updating' stamp: 'jm 2/20/98 19:54'!
changed

	hasChanged := true.
! !

!HandMorph methodsFor: 'accessing' stamp: 'sw 2/11/98 18:00'!
colorForInsets
	"Morphs being dragged by the hand use the world's color"
	^ owner colorForInsets! !

!HandMorph methodsFor: 'meta-actions' stamp: 'ar 11/6/2000 13:07'!
copyToPasteBuffer: aMorph
	"Save this morph in the paste buffer. This is mostly useful for copying morphs between projects."
	aMorph ifNil:[^PasteBuffer := nil].
	Cursor wait showWhile:[
		PasteBuffer := aMorph topRendererOrSelf veryDeepCopy.
		PasteBuffer privateOwner: nil].

! !

!HandMorph methodsFor: 'accessing' stamp: 'MarcusDenker 9/13/2013 16:31'!
cursorBounds

	temporaryCursor
		ifNil: [^ self position extent: NormalCursor extent]
		ifNotNil: [^ self position + temporaryCursorOffset
								extent: temporaryCursor extent]! !

!HandMorph methodsFor: 'event handling' stamp: 'dgd 2/21/2003 22:43'!
cursorPoint
	"Implemented for allowing embedded worlds in an event cycle to query a hand's position and get it in its coordinates. The same can be achieved by #point:from: but this is simply much more convenient since it will look as if the hand is in the lower world."

	| pos |
	pos := self position.
	(ActiveWorld isNil or: [ActiveWorld == owner]) ifTrue: [^pos].
	^ActiveWorld point: pos from: owner! !

!HandMorph methodsFor: 'balloon help' stamp: 'sw 10/15/2002 20:01'!
deleteBalloonTarget: aMorph
	"Delete any existing balloon help.  This is now done unconditionally, whether or not the morph supplied is the same as the current balloon target"
	
	self balloonHelp: nil

"	| h |
	h := self balloonHelp ifNil: [^ self].
	h balloonOwner == aMorph ifTrue: [self balloonHelp: nil]"! !

!HandMorph methodsFor: 'drawing' stamp: 'ClementBera 7/30/2013 11:06'!
drawOn: aCanvas 
	"Draw the hand itself (i.e., the cursor)."

	temporaryCursor 
		ifNil: [aCanvas paintImage: NormalCursor at: bounds topLeft]
		ifNotNil: [aCanvas paintImage: temporaryCursor at: bounds topLeft].
	! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 8/13/2003 11:39'!
dropMorph: aMorph event: anEvent
	"Drop the given morph which was carried by the hand"
	| event dropped |
	(anEvent isMouseUp and:[aMorph shouldDropOnMouseUp not]) ifTrue:[^self].

	"Note: For robustness in drag and drop handling we remove the morph BEFORE we drop him, but we keep his owner set to the hand. This prevents system lockups when there is a problem in drop handling (for example if there's an error in #wantsToBeDroppedInto:). THIS TECHNIQUE IS NOT RECOMMENDED FOR CASUAL USE."
	self privateRemove: aMorph.
	aMorph privateOwner: self.

	dropped := aMorph.
	(dropped hasProperty: #addedFlexAtGrab) 
		ifTrue:[dropped := aMorph removeFlexShell].
	event := DropEvent new setPosition: self position contents: dropped hand: self.
	self sendEvent: event focus: nil.
	event wasHandled ifFalse:[aMorph rejectDropMorphEvent: event].
	aMorph owner == self ifTrue:[aMorph delete].
	self mouseOverHandler processMouseOver: anEvent.! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 10/5/2000 16:23'!
dropMorphs
	"Drop the morphs at the hands position"
	self dropMorphs: lastMouseEvent.! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 9/14/2000 11:22'!
dropMorphs: anEvent
	"Drop the morphs at the hands position"
	self submorphsReverseDo:[:m|
		"Drop back to front to maintain z-order"
		self dropMorph: m event: anEvent.
	].! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/26/2000 01:27'!
eventListeners
	^eventListeners! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/26/2000 01:27'!
eventListeners: anArrayOrNil
	eventListeners := anArrayOrNil! !

!HandMorph methodsFor: 'layout' stamp: 'jm 2/20/98 18:55'!
fullBounds
	"Extend my bounds by the shadow offset when carrying morphs."

	| bnds |
	bnds := super fullBounds.
	submorphs isEmpty
		ifTrue: [^ bnds ]
		ifFalse: [^ bnds topLeft corner: bnds bottomRight + self shadowOffset].
! !

!HandMorph methodsFor: 'private events' stamp: 'dgd 3/31/2003 18:22'!
generateDropFilesEvent: evtBuf 
	"Generate the appropriate mouse event for the given raw event buffer"

	"Note: This is still in an experimental phase and will need more work"

	| position buttons modifiers stamp numFiles dragType |
	stamp := evtBuf second.
	stamp = 0 ifTrue: [stamp := Time millisecondClockValue].
	dragType := evtBuf third.
	position := evtBuf fourth @ evtBuf fifth.
	buttons := 0.
	modifiers := evtBuf sixth.
	buttons := buttons bitOr: (modifiers bitShift: 3).
	numFiles := evtBuf seventh.
	dragType = 4 
		ifTrue: 
			["e.g., drop"

			owner borderWidth: 0.
			^DropFilesEvent new 
				setPosition: position
				contents: numFiles
				hand: self].
	"the others are currently not handled by morphs themselves"
	dragType = 1 
		ifTrue: 
			["experimental drag enter"

			owner
				borderWidth: 4;
				borderColor: owner color asColor negated].
	dragType = 2 
		ifTrue: 
			["experimental drag move"

			].
	dragType = 3 
		ifTrue: 
			["experimental drag leave"

			owner borderWidth: 0].
	^nil! !

!HandMorph methodsFor: 'private events' stamp: 'GuillermoPolito 6/28/2013 13:15'!
generateKeyboardEvent: evtBuf
	"Generate the appropriate mouse event for the given raw event buffer"

	| buttons modifiers type pressType stamp charCode keyValue keyEvent |
	stamp := evtBuf second.
	stamp = 0 ifTrue: [stamp := Time millisecondClockValue].
	pressType := evtBuf fourth.
	pressType = EventKeyDown
		ifTrue: [
			type := #keyDown.
			lastKeyScanCode := evtBuf third].
	pressType = EventKeyUp ifTrue: [type := #keyUp].
	pressType = EventKeyChar ifTrue: [
		type := #keystroke].
	modifiers := evtBuf fifth.
	buttons := modifiers bitShift: 3.
	keyValue := evtBuf third. 
	charCode := evtBuf sixth.

	type = #keystroke
		ifTrue: [combinedChar
			ifNil: [
				| peekedEvent |
				peekedEvent := Sensor peekEvent.
				(peekedEvent notNil
					and: [peekedEvent fourth = EventKeyDown])
					ifTrue: [
						(CombinedChar isCompositionCharacter: charCode)
							ifTrue: [
								combinedChar := CombinedChar new.
								combinedChar simpleAdd: charCode asCharacter.
								(combinedChar combinesWith: peekedEvent third asCharacter)
									ifTrue: [^nil].
								]]]
			ifNotNil: [
				(combinedChar simpleAdd: charCode asCharacter)
					ifTrue: [charCode := combinedChar combined charCode].
				combinedChar := nil]].

	self flag: #fixme.
	"This piece of code handles the creation of scrolling events. When a scroll is done by the user, the VM forwards a keystroke event with the up/down key. So we reconvert it to a MouseWheelEvent in that case."
	
	(type = #keystroke and: [(buttons anyMask: 16) 
			and: [charCode = 30 or: [charCode = 31]]])
		ifTrue: [^MouseWheelEvent new 
					setType: #mouseWheel
					position: lastMouseEvent cursorPoint
					direction: (charCode = 30 ifTrue: [#up] ifFalse: [#down])
					buttons: buttons
					hand: self
					stamp: stamp].

	keyEvent := KeyboardEvent new
		setType: type
		buttons: buttons
		position: self position
		keyValue: keyValue
		charCode: charCode
		hand: self
		stamp: stamp.
	keyEvent scanCode: lastKeyScanCode.
	^keyEvent
! !

!HandMorph methodsFor: 'private events' stamp: 'BenjaminVanRyseghem 7/6/2012 14:00'!
generateMouseEvent: evtBuf 
	"Generate the appropriate mouse event for the given raw event buffer"

	| position buttons modifiers type trail stamp oldButtons evtChanged |
	KMBuffer uniqueInstance clearBuffer.
	evtBuf first = lastEventBuffer first 
		ifTrue: 
			["Workaround for Mac VM bug, *always* generating 3 events on clicks"

			evtChanged := false.
			3 to: evtBuf size
				do: [:i | (lastEventBuffer at: i) = (evtBuf at: i) ifFalse: [evtChanged := true]].
			evtChanged ifFalse: [^nil]].
	stamp := evtBuf second.
	stamp = 0 ifTrue: [stamp := Time millisecondClockValue].
	position := evtBuf third @ evtBuf fourth.
	buttons := evtBuf fifth.
	modifiers := evtBuf sixth.
	type := buttons = 0 
		ifTrue: 
			[lastEventBuffer fifth = 0 ifTrue: [#mouseMove] ifFalse: [#mouseUp]]
		ifFalse: 
			[lastEventBuffer fifth = 0 
						ifTrue: [#mouseDown]
						ifFalse: [#mouseMove]].
	buttons := buttons bitOr: (modifiers bitShift: 3).
	oldButtons := lastEventBuffer fifth 
				bitOr: (lastEventBuffer sixth bitShift: 3).
	lastEventBuffer := evtBuf.
	type == #mouseMove 
		ifTrue: 
			[trail := self mouseTrailFrom: evtBuf.
			^MouseMoveEvent basicNew 
				setType: type
				startPoint: (self position)
				endPoint: trail last
				trail: trail
				buttons: buttons
				hand: self
				stamp: stamp].
	^MouseButtonEvent basicNew 
		setType: type
		position: position
		which: (oldButtons bitXor: buttons)
		buttons: buttons
		hand: self
		stamp: stamp! !

!HandMorph methodsFor: 'private events' stamp: 'tbn 3/12/2010 01:38'!
generateWindowEvent: evtBuf 
	"Generate the appropriate window event for the given raw event buffer"

	| evt |
	evt := WindowEvent new.
	evt setTimeStamp: evtBuf second.
	evt timeStamp = 0 ifTrue: [evt setTimeStamp: Time millisecondClockValue].
	evt action: evtBuf third.
	evt rectangle: (Rectangle origin: evtBuf fourth @ evtBuf fifth corner: evtBuf sixth @ evtBuf seventh ).
	
	^evt
! !

!HandMorph methodsFor: 'meta-actions' stamp: 'adrian_lienhard 3/5/2009 22:44'!
grabMorph: aMorph
	"Grab the given morph (i.e., add it to this hand and remove it from its current owner) without changing its position. This is used to pick up a morph under the hand's current position, versus attachMorph: which is used to pick up a morph that may not be near this hand."
	
	| grabbed |
	aMorph = World ifTrue: [^ self].
	self releaseMouseFocus.
	grabbed := aMorph aboutToBeGrabbedBy: self.
	grabbed ifNil: [^self].
	grabbed := grabbed topRendererOrSelf.
	^self grabMorph: grabbed from: grabbed owner! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'AlainPlantec 5/7/2010 22:28'!
grabMorph: aMorph from: formerOwner 
	"Grab the given morph (i.e., add it to this hand and remove it from its
	current owner) without changing its position. This is used to pick up a
	morph under the hand's current position, versus attachMorph: which
	is used to pick up a morph that may not be near this hand."
	| grabbed offset targetPoint grabTransform fullTransform |
	self releaseMouseFocus.
	"Break focus"
	grabbed := aMorph.
	aMorph keepsTransform
		ifTrue: [grabTransform := fullTransform := IdentityTransform new]
		ifFalse: ["Compute the transform to apply to the grabbed morph"
			grabTransform := formerOwner
						ifNil: [IdentityTransform new]
						ifNotNil: [formerOwner grabTransform].
			fullTransform := formerOwner
						ifNil: [IdentityTransform new]
						ifNotNil: [formerOwner transformFrom: owner]].
	"targetPoint is point in aMorphs reference frame"
	targetPoint := fullTransform globalPointToLocal: self position.
	"but current position will be determined by grabTransform, so
	compute offset"
	offset := targetPoint
				- (grabTransform globalPointToLocal: self position).
	"apply the transform that should be used after grabbing"
	grabbed := grabbed transformedBy: grabTransform.
	grabbed == aMorph
		ifFalse: [grabbed setProperty: #addedFlexAtGrab toValue: true].
	"offset target to compensate for differences in transforms"
	grabbed position: grabbed position - offset asIntegerPoint.
	"And compute distance from hand's position" 
	targetOffset := grabbed position - self position.
	self addMorphBack: grabbed.
	grabbed justGrabbedFrom: formerOwner! !

!HandMorph methodsFor: 'halos and balloon help' stamp: 'ar 10/4/2000 13:40'!
halo
	"Return the halo associated with this hand, if any"
	^self valueOfProperty: #halo! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 10/4/2000 13:40'!
halo: newHalo
	"Set halo associated with this hand"
	| oldHalo |
	oldHalo := self halo.
	(oldHalo isNil or:[oldHalo == newHalo]) ifFalse:[oldHalo delete].
	newHalo
		ifNil:[self removeProperty: #halo]
		ifNotNil:[self setProperty: #halo toValue: newHalo]! !

!HandMorph methodsFor: 'events-processing' stamp: 'BenjaminVanRyseghem 6/25/2013 17:43'!
handleEvent: anEvent
	| evt  |
	owner ifNil:[^self].
	evt := anEvent.
	self logEventStats: evt.
	
	evt isMouse ifTrue:[
		 "just for record, to be used by capture block" 
		lastMouseEvent := evt].

	captureBlock ifNotNil: [ ^ captureBlock value: anEvent ].
	evt isMouseOver ifTrue:[^self sendMouseEvent: evt].
	
	self showDebugEvent: evt.
	
	"Notify listeners"
	self sendListenEvent: evt to: self eventListeners.

	evt isWindowEvent ifTrue: [
		self sendEvent: evt focus: nil.
		^self mouseOverHandler processMouseOver: lastMouseEvent].
	

	evt isKeyboard ifTrue:[
		self sendListenEvent: evt to: self keyboardListeners.
		self sendKeyboardEvent: evt.
		^self mouseOverHandler processMouseOver: lastMouseEvent].

	evt isDropEvent ifTrue:[
		self sendEvent: evt focus: nil.
		^self mouseOverHandler processMouseOver: lastMouseEvent].

	evt isMouse ifTrue:[
		self sendListenEvent: evt to: self mouseListeners.
		lastMouseEvent := evt].

	"Check for pending drag or double click operations."
	mouseClickState ifNotNil:[
		(mouseClickState handleEvent: evt from: self) ifFalse:[
			"Possibly dispatched #click: or something and will not re-establish otherwise"
			^self mouseOverHandler processMouseOver: lastMouseEvent]].

	evt isMove ifTrue:[
		| pos |
		pos := evt position.
		evt isDraggingEvent ifTrue: [ 
			| treshold |
			treshold := 0.
			(self submorphs at: 1 ifAbsent: [ nil ])
				ifNotNil: [ :first | treshold := self top-first top ].
			pos y < (self class upperHandLimit+treshold) ifTrue: [ pos := pos x @ (self class upperHandLimit + treshold)] ].
		
		self position: pos.
		self sendMouseEvent: evt.
	] ifFalse:[
		"Issue a synthetic move event if we're not at the position of the event"
		(evt position = self position) ifFalse:[self moveToEvent: evt].
		"Drop submorphs on button events"
		(self hasSubmorphs) 
			ifTrue:[self dropMorphs: evt]
			ifFalse:[self sendMouseEvent: evt].
	].
	self showMouseFocusEvent: evt.
	self mouseOverHandler processMouseOver: lastMouseEvent.
! !

!HandMorph methodsFor: 'drawing' stamp: 'ls 4/3/2000 20:30'!
hasChanged
	"Return true if this hand has changed, either because it has moved or because some morph it is holding has changed."

	^ hasChanged ifNil: [ true ]
! !

!HandMorph methodsFor: 'initialization' stamp: 'IgorStasenko 1/22/2012 18:19'!
initForEvents
	mouseOverHandler := nil.
	lastMouseEvent := MouseEvent basicNew setType: #mouseMove position: 0@0 buttons: 0 hand: self.
	lastEventBuffer := {1. 0. 0. 0. 0. 0. nil. nil}.
	recentModifiers := 0.
	self resetClickState.! !

!HandMorph methodsFor: 'initialization' stamp: 'StephaneDucasse 12/26/2011 10:18'!
initialize
	super initialize.
	self initForEvents.
	keyboardFocus := nil.
	mouseFocus := nil.
	bounds := 0@0 extent: Cursor normal extent.
	damageRecorder := DamageRecorder new.
	cachedCanvasHasHoles := false.
	temporaryCursor := temporaryCursorOffset := nil.
	self initForEvents.! !

!HandMorph methodsFor: 'initialization' stamp: 'nk 2/14/2004 18:28'!
interrupted
	"Something went wrong - we're about to bring up a debugger. 
	Release some stuff that could be problematic."
	self releaseAllFoci. "or else debugger might not handle clicks"
! !

!HandMorph methodsFor: 'change reporting' stamp: 'ar 12/30/2001 17:32'!
invalidRect: damageRect from: aMorph
	"Note that a change has occurred and record the given damage rectangle relative to the origin this hand's cache."
	hasChanged := true.
	aMorph == self ifTrue:[^self].
	damageRecorder recordInvalidRect: damageRect.
! !

!HandMorph methodsFor: 'classification' stamp: ''!
isHandMorph

	^ true! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:09'!
keyboardFocus 
	^ keyboardFocus! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/26/2000 01:30'!
keyboardFocus: aMorphOrNil
	keyboardFocus := aMorphOrNil! !

!HandMorph methodsFor: 'listeners' stamp: 'MarcusDenker 9/13/2013 15:53'!
keyboardListeners
	^nil! !

!HandMorph methodsFor: 'accessing' stamp: 'ar 10/5/2000 23:17'!
lastEvent
	^ lastMouseEvent! !

!HandMorph methodsFor: 'events-processing' stamp: 'StephaneDucasse 12/26/2011 10:45'!
logEventStats: evt
	EventStats ifNil:[EventStats := IdentityDictionary new].
	EventStats at: #count put: (EventStats at: #count ifAbsent:[0]) + 1.
	EventStats at: evt type put: (EventStats at: evt type ifAbsent:[0]) + 1.! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:10'!
mouseFocus
	^mouseFocus! !

!HandMorph methodsFor: 'focus handling' stamp: 'nk 2/14/2004 18:44'!
mouseFocus: aMorphOrNil
	mouseFocus := aMorphOrNil! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/26/2000 01:28'!
mouseListeners
	^mouseListeners! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/26/2000 01:27'!
mouseListeners: anArrayOrNil
	mouseListeners := anArrayOrNil! !

!HandMorph methodsFor: 'accessing' stamp: 'ar 9/25/2000 14:24'!
mouseOverHandler
	^mouseOverHandler ifNil:[mouseOverHandler := MouseOverHandler new].! !

!HandMorph methodsFor: 'private events' stamp: 'PavelKrivanek 6/21/2011 13:05'!
mouseTrailFrom: currentBuf 
  "Current event, a mouse event buffer, is about to be processed.  If there are other similar mouse events queued up, then drop them from the queue, and report the positions inbetween."

  | nextEvent trail |
  trail := (Array new: 1) writeStream.
  trail nextPut: currentBuf third @ currentBuf fourth.
  [(nextEvent := Sensor peekEvent) isNil] whileFalse: 
      [nextEvent first = currentBuf first 
        ifFalse: [^trail contents "different event type"].
      nextEvent fifth = currentBuf fifth 
        ifFalse: [^trail contents "buttons changed"].
      nextEvent sixth = currentBuf sixth 
        ifFalse: [^trail contents "modifiers changed"].
      "nextEvent is similar.  Remove it from the queue, and check the next."
      nextEvent := Sensor nextEvent.
      nextEvent 
        ifNotNil: [ trail nextPut: nextEvent third @ nextEvent fourth ]].
  ^trail contents! !

!HandMorph methodsFor: 'private events' stamp: 'marcus.denker 8/24/2008 21:40'!
moveToEvent: anEvent
	"Issue a mouse move event to make the receiver appear at the given position"
	self handleEvent: (MouseMoveEvent basicNew
		setType: #mouseMove 
		startPoint: self position 
		endPoint: anEvent position 
		trail: (Array with: self position with: anEvent position)
		buttons: anEvent buttons
		hand: self
		stamp: anEvent timeStamp)! !

!HandMorph methodsFor: 'drawing' stamp: 'StephaneDucasse 12/26/2011 10:19'!
needsToBeDrawn
	"Return true if this hand must be drawn explicitely instead of being drawn via the hardware cursor. This is the case if it (a) it is a remote hand, (b) it is showing a temporary cursor, or (c) it is not empty and there are any visible submorphs. If using the software cursor, ensure that the hardware cursor is hidden."
	"Details:  Return true if this hand has a saved patch to ensure that is is processed by the world. This saved patch will be deleted after one final display pass when it becomes possible to start using the hardware cursor again. This trick gives us one last display cycle to allow us to remove the software cursor and shadow from the display."
	| cursor |
	(savedPatch notNil
		or: [ (submorphs anySatisfy: [ :ea | ea visible ])
			or: [ (temporaryCursor notNil and: [hardwareCursor isNil])
				]])
		ifTrue: [
			"using the software cursor; hide the hardware one"
			Cursor blank isCurrent ifFalse: [Cursor blank show].
			^ true].
	"Switch from one hardware cursor to another, if needed."
	cursor := hardwareCursor ifNil: [Cursor normal].
	cursor isCurrent ifFalse: [cursor show].
	^ false
! !

!HandMorph methodsFor: 'focus handling' stamp: 'BenjaminVanRyseghem 7/3/2012 12:25'!
newKeyboardFocus: aMorphOrNil
	"Make the given morph the new keyboard focus, canceling the previous keyboard focus if any. If the argument is nil, the current keyboard focus is cancelled."
	| oldFocus |
	oldFocus := self keyboardFocus.
	
	self keyboardFocus: aMorphOrNil.
	oldFocus ifNotNil: [oldFocus == aMorphOrNil ifFalse: [oldFocus keyboardFocusChange: false]].
	aMorphOrNil ifNotNil: [ aMorphOrNil keyboardFocusChange: true ].
! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/26/2000 01:32'!
newMouseFocus: aMorphOrNil
	"Make the given morph the new mouse focus, canceling the previous mouse focus if any. If the argument is nil, the current mouse focus is cancelled."
	self mouseFocus: aMorphOrNil.
! !

!HandMorph methodsFor: 'focus handling' stamp: 'ClementBera 7/30/2013 11:06'!
newMouseFocus: aMorph event: event 
	aMorph ifNotNil: [targetOffset := event cursorPoint - aMorph position].
	^self newMouseFocus: aMorph! !

!HandMorph methodsFor: 'accessing' stamp: 'tk 10/20/2004 15:54'!
noButtonPressed
	"Answer whether any mouse button is not being pressed."

	^self anyButtonPressed not! !

!HandMorph methodsFor: 'drawing' stamp: 'nice 1/5/2010 15:59'!
nonCachingFullDrawOn: aCanvas
	
	"A HandMorph has unusual drawing requirements:
		1. the hand itself (i.e., the cursor) appears in front of its submorphs
		2. morphs being held by the hand cast a shadow on the world/morphs below
	The illusion is that the hand plucks up morphs and carries them above the world."
	"Note: This version does not cache an image of the morphs being held by the hand.
	 Thus, it is slower for complex morphs, but consumes less space."

	submorphs isEmpty ifTrue: [^ self drawOn: aCanvas].  "just draw the hand itself"
	aCanvas asShadowDrawingCanvas
		translateBy: self shadowOffset during:[:shadowCanvas| | shadowForm |
		"Note: We use a shadow form here to prevent drawing
		overlapping morphs multiple times using the transparent
		shadow color."
		shadowForm := self shadowForm.
"
shadowForm displayAt: shadowForm offset negated. Display forceToScreen: (0@0 extent: shadowForm extent).
"
		shadowCanvas paintImage: shadowForm at: shadowForm offset.  "draw shadows"
	].
	"draw morphs in front of shadows"
	self drawSubmorphsOn: aCanvas.
	self drawOn: aCanvas.  "draw the hand itself in front of morphs"
! !

!HandMorph methodsFor: 'event handling' stamp: 'ar 9/25/2000 14:27'!
noticeMouseOver: aMorph event: anEvent
	mouseOverHandler ifNil:[^self].
	mouseOverHandler noticeMouseOver: aMorph event: anEvent.! !

!HandMorph methodsFor: 'paste buffer' stamp: 'ar 10/5/2000 19:10'!
objectToPaste
	"It may need to be sent #startRunning by the client"
	^ Cursor wait showWhile: [PasteBuffer veryDeepCopy]

	"PasteBuffer usableDuplicateIn: self world"
! !

!HandMorph methodsFor: 'halo handling' stamp: 'GabrielOmarCotelli 11/30/2013 16:46'!
obtainHalo: aHalo
	"Used for transfering halos between hands"

	self halo == aHalo
		ifTrue: [ ^ self ].	
	"Find former owner"
	self world hands detect: [ :hand | hand halo == aHalo ] ifFound: [ :formerOwner | formerOwner releaseHalo: aHalo ].
	self halo: aHalo! !

!HandMorph methodsFor: 'paste buffer' stamp: 'ar 10/5/2000 19:10'!
pasteBuffer
	"Return the paste buffer associated with this hand"
	^ PasteBuffer! !

!HandMorph methodsFor: 'paste buffer' stamp: 'ar 10/5/2000 19:11'!
pasteBuffer: aMorphOrNil
	"Set the contents of the paste buffer."
	PasteBuffer := aMorphOrNil.

! !

!HandMorph methodsFor: 'geometry' stamp: 'ar 3/20/2001 20:34'!
position

	^temporaryCursor
		ifNil: [bounds topLeft]
		ifNotNil: [bounds topLeft - temporaryCursorOffset]! !

!HandMorph methodsFor: 'geometry' stamp: 'nk 8/20/2003 17:39'!
position: aPoint
	"Overridden to align submorph origins to the grid if gridding is on."
	| adjustedPosition delta box |
	adjustedPosition := aPoint.
	temporaryCursor ifNotNil: [adjustedPosition := adjustedPosition + temporaryCursorOffset].

	"Copied from Morph to avoid owner layoutChanged"
	"Change the position of this morph and and all of its submorphs."
	delta := adjustedPosition - bounds topLeft.
	(delta x = 0 and: [delta y = 0]) ifTrue: [^ self].  "Null change"
	box := self fullBounds.
	(delta dotProduct: delta) > 100 ifTrue:[
		"e.g., more than 10 pixels moved"
		self invalidRect: box.
		self invalidRect: (box translateBy: delta).
	] ifFalse:[
		self invalidRect: (box merge: (box translateBy: delta)).
	].
	self privateFullMoveBy: delta.
! !

!HandMorph methodsFor: 'private events' stamp: 'IgorStasenko 1/22/2012 18:30'!
processEvents
	"Process user input events from the local input devices."

	| evt evtBuf type hadAny |
	ActiveEvent ifNotNil: 
			["Meaning that we were invoked from within an event response.
		Make sure z-order is up to date"

			self mouseOverHandler processMouseOver: lastMouseEvent].
	hadAny := false.
	[(evtBuf := Sensor nextEvent) isNil] whileFalse: 
			[evt := nil.	"for unknown event types"
			type := evtBuf first.
			type = EventTypeMouse ifTrue: [recentModifiers := evtBuf sixth. evt := self generateMouseEvent: evtBuf].
			type = EventTypeKeyboard 
				ifTrue: [recentModifiers := evtBuf fifth. evt := self generateKeyboardEvent: evtBuf].
			type = EventTypeDragDropFiles 
				ifTrue: [evt := self generateDropFilesEvent: evtBuf].
			type = EventTypeWindow
				ifTrue:[evt := self generateWindowEvent: evtBuf].				
			"All other events are ignored"
			(type ~= EventTypeDragDropFiles and: [evt isNil]) ifTrue: [^self].
			evt isNil 
				ifFalse: 
					["Finally, handle it"

					self handleEvent: evt.
					hadAny := true.

					"For better user feedback, return immediately after a mouse event has been processed."
					(evt isMouse and: [evt isMouseWheel not]) ifTrue: [^self]]].
	"note: if we come here we didn't have any mouse events"
	mouseClickState notNil 
		ifTrue: 
			["No mouse events during this cycle. Make sure click states time out accordingly"

			mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	hadAny 
		ifFalse: 
			["No pending events. Make sure z-order is up to date"

			self mouseOverHandler processMouseOver: lastMouseEvent]! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:09'!
releaseAllFoci
	mouseFocus := nil.
	keyboardFocus := nil.
! !

!HandMorph methodsFor: 'caching' stamp: 'StephaneDucasse 12/26/2011 10:39'!
releaseCachedState
	| oo |
	super releaseCachedState.
	cacheCanvas := nil.
	oo := owner.
	self removeAllMorphs.
	self initialize.	"nuke everything"
	self privateOwner: oo.
	self releaseAllFoci.! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 10/24/2000 18:40'!
releaseHalo: aHalo
	"Used for transfering halos between hands"
	self removeProperty: #halo! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:09'!
releaseKeyboardFocus
	"Release the current keyboard focus unconditionally"
	self newKeyboardFocus: nil.
! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/26/2000 01:31'!
releaseKeyboardFocus: aMorph
	"If the given morph had the keyboard focus before, release it"
	self keyboardFocus == aMorph ifTrue:[self releaseKeyboardFocus].! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:10'!
releaseMouseFocus
	"Release the current mouse focus unconditionally."
	self newMouseFocus: nil.! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:10'!
releaseMouseFocus: aMorph
	"If the given morph had the mouse focus before, release it"
	self mouseFocus == aMorph ifTrue:[self releaseMouseFocus].! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/24/2000 20:41'!
removeEventListener: anObject
	"Remove anObject from the current event listeners."
	self eventListeners: (self removeListener: anObject from: self eventListeners).! !

!HandMorph methodsFor: 'halo handling' stamp: 'dgd 4/4/2006 16:14'!
removeHalo
	"remove the receiver's halo (if any)"
	| halo |
	halo := self halo.
	halo
		ifNil: [^ self].
	halo delete.
	self removeProperty: #halo! !

!HandMorph methodsFor: 'halo handling' stamp: 'dgd 9/9/2004 22:44'!
removeHaloFromClick: anEvent on: aMorph 
	| halo |
	halo := self halo
				ifNil: [^ self].
	(halo target hasOwner: self)
		ifTrue: [^ self].
	(halo staysUpWhenMouseIsDownIn: aMorph)
		ifFalse: [self removeHalo]! !

!HandMorph methodsFor: 'listeners' stamp: 'dgd 2/21/2003 22:48'!
removeListener: anObject from: aListenerGroup 
	"Remove anObject from the given listener group. Return the new group."

	| listeners |
	aListenerGroup ifNil: [^nil].
	listeners := aListenerGroup.
	listeners := listeners copyWithout: anObject.
	listeners := listeners copyWithout: nil.	"obsolete entries"
	listeners isEmpty ifTrue: [listeners := nil].
	^listeners! !

!HandMorph methodsFor: 'listeners' stamp: 'ar 10/24/2000 20:41'!
removeMouseListener: anObject
	"Remove anObject from the current mouse listeners."
	self mouseListeners: (self removeListener: anObject from: self mouseListeners).! !

!HandMorph methodsFor: 'balloon help' stamp: 'ar 10/6/2000 00:14'!
removePendingBalloonFor: aMorph
	"Get rid of pending balloon help."
	self removeAlarm: #spawnBalloonFor:.
	self deleteBalloonTarget: aMorph.! !

!HandMorph methodsFor: 'double click support' stamp: 'ar 9/18/2000 17:16'!
resetClickState
	"Reset the double-click detection state to normal (i.e., not waiting for a double-click)."
	mouseClickState := nil.! !

!HandMorph methodsFor: 'drawing' stamp: 'StephaneDucasse 12/26/2011 10:19'!
restoreSavedPatchOn: aCanvas 
	"Clear the changed flag and restore the part of the given canvas under this hand from the previously saved patch. If necessary, handle the transition to using the hardware cursor."
	| cursor |

	hasChanged := false.
	savedPatch ifNotNil: 
			[aCanvas drawImage: savedPatch at: savedPatch offset.
			submorphs notEmpty ifTrue: [^self].
			(temporaryCursor notNil and: [hardwareCursor isNil]) ifTrue: [^self].

			"Make the transition to using hardware cursor. Clear savedPatch and
		 report one final damage rectangle to erase the image of the software cursor."
			super invalidRect: (savedPatch offset 
						extent: savedPatch extent + self shadowOffset)
				from: self.
			cursor := hardwareCursor ifNil: [Cursor normal].
			cursor isCurrent ifFalse: [cursor show].	"show hardware cursor"
			savedPatch := nil]! !

!HandMorph methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:49'!
savePatchFrom: aCanvas 
	"Save the part of the given canvas under this hand as a Form and return its bounding rectangle."

	"Details: The previously used patch Form is recycled when possible to reduce the burden on storage management."

	| damageRect myBnds |
	damageRect := myBnds := self fullBounds.
	savedPatch ifNotNil: 
			[damageRect := myBnds merge: (savedPatch offset extent: savedPatch extent)].
	(savedPatch isNil or: [savedPatch extent ~= myBnds extent]) 
		ifTrue: 
			["allocate new patch form if needed"

			savedPatch := aCanvas form allocateForm: myBnds extent].
	aCanvas contentsOfArea: (myBnds translateBy: aCanvas origin)
		into: savedPatch.
	savedPatch offset: myBnds topLeft.
	^damageRect! !

!HandMorph methodsFor: 'selected object' stamp: 'ClementBera 7/30/2013 11:06'!
selectedObject
	"answer the selected object for the hand or nil is none"
	| halo |
	halo := self halo.
	halo ifNil: [^ nil].
	^ halo target renderedMorph! !

!HandMorph methodsFor: 'private events' stamp: 'ar 3/18/2001 01:43'!
sendEvent: anEvent focus: focusHolder
	"Send the event to the morph currently holding the focus, or if none to the owner of the hand."
	^self sendEvent: anEvent focus: focusHolder clear:[nil]! !

!HandMorph methodsFor: 'private events' stamp: 'ar 3/18/2001 01:42'!
sendEvent: anEvent focus: focusHolder clear: aBlock
	"Send the event to the morph currently holding the focus, or if none to the owner of the hand."
	| result |
	focusHolder ifNotNil:[^self sendFocusEvent: anEvent to: focusHolder clear: aBlock].
	ActiveEvent := anEvent.
	result := owner processEvent: anEvent.
	ActiveEvent := nil.
	^result! !

!HandMorph methodsFor: 'private events' stamp: 'ar 3/18/2001 01:42'!
sendFocusEvent: anEvent to: focusHolder clear: aBlock
	"Send the event to the morph currently holding the focus"
	| result w |
	w := focusHolder world ifNil:[^ aBlock value].
	w becomeActiveDuring:[
		ActiveHand := self.
		ActiveEvent := anEvent.
		result := focusHolder handleFocusEvent: 
			(anEvent transformedBy: (focusHolder transformedFrom: self)).
	].
	^result! !

!HandMorph methodsFor: 'private events' stamp: 'AlainPlantec 1/7/2010 19:44'!
sendKeyboardEvent: anEvent 
	"Send the event to the morph currently holding the focus, or if none to
	the owner of the hand."
	^ self
		sendEvent: anEvent
		focus: self keyboardFocus
		clear: [self keyboardFocus: nil]! !

!HandMorph methodsFor: 'private events' stamp: 'ar 10/26/2000 01:43'!
sendListenEvent: anEvent to: listenerGroup
	"Send the event to the given group of listeners"
	listenerGroup ifNil:[^self].
	listenerGroup do:[:listener| 
		listener ifNotNil:[listener handleListenEvent: anEvent copy]].! !

!HandMorph methodsFor: 'private events' stamp: 'ar 3/18/2001 01:45'!
sendMouseEvent: anEvent
	"Send the event to the morph currently holding the focus, or if none to the owner of the hand."
	^self sendEvent: anEvent focus: self mouseFocus clear:[self mouseFocus: nil]! !

!HandMorph methodsFor: 'drawing' stamp: 'JW 7/12/2005 20:13'!
shadowForm
	"Return a 1-bit shadow of my submorphs.  Assumes submorphs is not empty"
	| bnds canvas |
	bnds := Rectangle merging: (submorphs collect: [:m | m fullBounds]).
	canvas := (Display defaultCanvasClass extent: bnds extent depth: 1) 
		asShadowDrawingCanvas: Color black.
	canvas translateBy: bnds topLeft negated
		during:[:tempCanvas| self drawSubmorphsOn: tempCanvas].
	^ canvas form offset: bnds topLeft! !

!HandMorph methodsFor: 'drop shadows' stamp: ''!
shadowOffset

	^ 6@8! !

!HandMorph methodsFor: 'accessing' stamp: 'ar 12/22/2008 12:04'!
shiftPressed
	^lastMouseEvent shiftPressed! !

!HandMorph methodsFor: 'events-processing' stamp: 'MarcusDenker 7/28/2013 13:03'!
showDebugEvent: evt

	ShowEvents == true ifTrue: [
		| ofs| 
		Display fill: (0@0 extent: 500@120) rule: Form over fillColor: Color white.
		ofs := (owner hands indexOf: self) - 1 * 60.
		evt printString displayAt: (0@ofs) + (evt isKeyboard ifTrue: [0@30] ifFalse: [0@0]).
		self keyboardFocus printString displayAt: (0@ofs)+(0@45).
		].! !

!HandMorph methodsFor: 'events-processing' stamp: 'StephaneDucasse 12/26/2011 10:49'!
showMouseFocusEvent: evt

	ShowEvents == true ifTrue: [
		| ofs| 
		ofs := (owner hands indexOf: self) - 1 * 60.
		self mouseFocus printString displayAt: (0@ofs) + (0@15)].
		! !

!HandMorph methodsFor: 'cursor' stamp: 'di 3/6/1999 23:52'!
showTemporaryCursor: cursorOrNil
	"Set the temporary cursor to the given Form. If the argument is nil, revert to the normal cursor."

	self showTemporaryCursor: cursorOrNil hotSpotOffset: 0@0
! !

!HandMorph methodsFor: 'cursor' stamp: 'ClementBera 7/30/2013 11:06'!
showTemporaryCursor: cursorOrNil hotSpotOffset: hotSpotOffset 
	"Set the temporary cursor to the given Form.
	If the argument is nil, revert to the normal hardware cursor."

	self changed.
	temporaryCursorOffset 
		ifNotNil: [bounds := bounds translateBy: temporaryCursorOffset negated].
	cursorOrNil 
		ifNil: [temporaryCursor := temporaryCursorOffset := hardwareCursor := nil]
		ifNotNil: 
			[temporaryCursor := cursorOrNil asCursorForm.
			temporaryCursorOffset := temporaryCursor offset - hotSpotOffset.
			(cursorOrNil isKindOf: Cursor) ifTrue: [hardwareCursor := cursorOrNil]].
	bounds := self cursorBounds.
	self
		layoutChanged;
		changed! !

!HandMorph methodsFor: 'balloon help' stamp: 'ar 10/3/2000 17:15'!
spawnBalloonFor: aMorph
	aMorph showBalloon: aMorph balloonText hand: self.! !

!HandMorph methodsFor: 'accessing' stamp: ''!
targetOffset
	"Return the offset of the last mouseDown location relative to the origin of the recipient morph. During menu interactions, this is the absolute location of the mouse down event that invoked the menu."

	^ targetOffset
! !

!HandMorph methodsFor: 'accessing' stamp: 'wiz 12/4/2006 00:16'!
targetPoint
	"Return the new position of the target.
	I.E. return the position of the hand less 
	the original distance between hand and target position"

	^ self position - targetOffset
! !

!HandMorph methodsFor: 'cursor' stamp: 'NS 2/17/2001 11:01'!
temporaryCursor
	^ temporaryCursor! !

!HandMorph methodsFor: 'balloon help' stamp: 'ar 10/3/2000 17:14'!
triggerBalloonFor: aMorph after: timeOut
	"Trigger balloon help after the given time out for some morph"
	self addAlarm: #spawnBalloonFor: with: aMorph after: timeOut.! !

!HandMorph methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:49'!
updateCacheCanvas: aCanvas 
	"Update the cached image of the morphs being held by this hand."

	"Note: The following is an attempt to quickly get out if there's no change"

	| subBnds rectList nPix |
	subBnds := Rectangle merging: (submorphs collect: [:m | m fullBounds]).
	rectList := damageRecorder invalidRectsFullBounds: subBnds.
	damageRecorder reset.
	(rectList isEmpty 
		and: [cacheCanvas notNil and: [cacheCanvas extent = subBnds extent]]) 
			ifTrue: [^self].

	"Always check for real translucency -- can't be cached in a form"
	self submorphsDo: 
			[:m | 
			m wantsToBeCachedByHand 
				ifFalse: 
					[cacheCanvas := nil.
					cachedCanvasHasHoles := true.
					^self]].
	(cacheCanvas isNil or: [cacheCanvas extent ~= subBnds extent]) 
		ifTrue: 
			[cacheCanvas := (aCanvas allocateForm: subBnds extent) getCanvas.
			cacheCanvas translateBy: subBnds origin negated
				during: [:tempCanvas | self drawSubmorphsOn: tempCanvas].
			self submorphsDo: 
					[:m | 
					(m areasRemainingToFill: subBnds) isEmpty 
						ifTrue: [^cachedCanvasHasHoles := false]].
			nPix := cacheCanvas form tallyPixelValues first.
			"--> begin rounded corners hack <---"
			cachedCanvasHasHoles := (nPix = 48 
						and: [submorphs size = 1 and: [submorphs first wantsRoundedCorners]]) 
							ifTrue: [false]
							ifFalse: [nPix > 0].
			"--> end rounded corners hack <---"
			^self].

	"incrementally update the cache canvas"
	cacheCanvas translateBy: subBnds origin negated
		during: 
			[:cc | 
			rectList do: 
					[:r | 
					cc clipBy: r
						during: 
							[:c | 
							c fillColor: Color transparent.
							self drawSubmorphsOn: c]]]! !

!HandMorph methodsFor: 'copying' stamp: 'ar 10/6/2000 00:11'!
veryDeepCopyWith: deepCopier
	"Return self.  Do not copy hands this way."
	^ self! !

!HandMorph methodsFor: 'drawing' stamp: 'nk 10/24/2003 22:12'!
visible: aBoolean
	self needsToBeDrawn ifFalse: [ ^self ].
	super visible: aBoolean! !

!HandMorph methodsFor: 'events-processing' stamp: 'IgorStasenko 1/2/2012 18:41'!
waitButton
	self captureEventsUntil: [:evt | self anyButtonPressed ]
! !

!HandMorph methodsFor: 'double click support' stamp: 'nk 7/26/2004 10:29'!
waitForClicksOrDrag: aMorph event: evt
	"Wait for mouse button and movement events, informing aMorph about events interesting to it via callbacks.
	This message is typically sent to the Hand by aMorph when it first receives a mouse-down event.
	The callback methods invoked on aMorph (which are passed a copy of evt) are:
		#click:	sent when the mouse button goes up within doubleClickTime.
		#doubleClick:	sent when the mouse goes up, down, and up again all within DoubleClickTime.
		#doubleClickTimeout:  sent when the mouse does not have a doubleClick within DoubleClickTime.
		#startDrag:	sent when the mouse moves more than 10 pixels from evt's position within DoubleClickTime.
	Note that mouseMove: and mouseUp: events are not sent to aMorph until it becomes the mouse focus,
	which is typically done by aMorph in its click:, doubleClick:, or drag: methods."
	
	^self waitForClicksOrDrag: aMorph event: evt selectors: #( #click: #doubleClick: #doubleClickTimeout: #startDrag:) threshold: 10
! !

!HandMorph methodsFor: 'double click support' stamp: 'nk 7/26/2004 10:32'!
waitForClicksOrDrag: aMorph event: evt selectors: clickAndDragSelectors threshold: threshold

	"Wait for mouse button and movement events, informing aMorph about events interesting to it via callbacks.
	This message is typically sent to the Hand by aMorph when it first receives a mouse-down event.
	The callback methods, named in clickAndDragSelectors and passed a copy of evt, are:
		1 	(click) sent when the mouse button goes up within doubleClickTime.
		2	(doubleClick) sent when the mouse goes up, down, and up again all within DoubleClickTime.
		3	(doubleClickTimeout) sent when the mouse does not have a doubleClick within DoubleClickTime.
		4	(startDrag) sent when the mouse moves more than threshold pixels from evt's position within DoubleClickTime.
	Note that mouseMove: and mouseUp: events are not sent to aMorph until it becomes the mouse focus,
	which is typically done by aMorph in its click:, doubleClick:, or drag: methods."
	
	mouseClickState := 
		MouseClickState new
			client: aMorph 
			click: clickAndDragSelectors first 
			dblClick: clickAndDragSelectors second 
			dblClickTime: DoubleClickTime 
			dblClickTimeout: clickAndDragSelectors third
			drag: clickAndDragSelectors fourth 
			threshold: threshold 
			event: evt.
! !

!Morph class methodsFor: 'accessing' stamp: 'BenjaminVanRyseghem 6/25/2013 18:27'!
announcer

	^ announcer ifNil: [ announcer := Announcer new ].! !

!Morph class methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:12'!
cmdGesturesEnabled
	^ CmdGesturesEnabled ifNil: [CmdGesturesEnabled := true]! !

!Morph class methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:12'!
cmdGesturesEnabled: aBoolean
	CmdGesturesEnabled := aBoolean! !

!Morph class methodsFor: 'settings' stamp: 'SeanDeNigris 4/24/2012 19:51'!
cycleHalosBothDirections

	^ CycleHalosBothDirections ifNil: [ CycleHalosBothDirections := false ].! !

!Morph class methodsFor: 'settings' stamp: 'SeanDeNigris 4/24/2012 19:52'!
cycleHalosBothDirections: aBoolean

	CycleHalosBothDirections := aBoolean.! !

!Morph class methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 13:09'!
defaultYellowButtonMenuEnabled
	^ DefaultYellowButtonMenuEnabled ifNil: [DefaultYellowButtonMenuEnabled := false]! !

!Morph class methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 13:02'!
defaultYellowButtonMenuEnabled: aBoolean
	DefaultYellowButtonMenuEnabled := aBoolean! !

!Morph class methodsFor: 'settings' stamp: 'usmanbhatti 3/6/2012 19:23'!
halosEnabled

	^ HalosEnabled ifNil: [ HalosEnabled := true ]! !

!Morph class methodsFor: 'settings' stamp: 'BenjaminVanRyseghem 3/6/2012 19:38'!
halosEnabled: aBoolean

	HalosEnabled := aBoolean! !

!Morph class methodsFor: 'initialize-release' stamp: ''!
initialize
	"Morph initialize"

	"this empty array object is shared by all morphs with no submorphs:"
	EmptyArray := Array new.
! !

!Morph class methodsFor: 'settings' stamp: 'GuillermoPolito 3/19/2013 19:12'!
morphNavigationShortcutsOn: aBuilder
	"Basic, general navigation shortcut among morphs. #MorphNoCtrl will not work for TextMorphs."

	<keymap>
	(aBuilder shortcut: #navigateFocusForwardCtrl)
		category: #MorphFocusCtrlNavigation
		default: Character tab ctrl asKeyCombination
		do: [ :target :morph :event | morph navigateFocusForward ].
	(aBuilder shortcut: #navigateFocusBackwardCtrl)
		category: #MorphFocusCtrlNavigation
		default: Character tab shift ctrl asKeyCombination
		do: [ :target :morph :event | morph navigateFocusBackward ].
	(aBuilder shortcut: #navigateFocusForward)
		category: #MorphFocusNavigation
		default: Character tab asKeyCombination
		do: [ :target :morph :event | morph navigateFocusForward ].
	(aBuilder shortcut: #navigateFocusBackward)
		category: #MorphFocusNavigation
		default: Character tab shift asKeyCombination
		do: [ :target :morph :event | morph navigateFocusBackward ]! !

!Morph class methodsFor: 'misc' stamp: 'MarcusDenker 10/21/2013 14:22'!
morphsUnknownToTheirOwners
	"Return a list of all morphs (other than HandMorphs) whose owners do not contain them in their submorph lists"

	"Morph morphsUnknownToTheirOwners"

	| problemMorphs |
	problemMorphs := OrderedCollection new.
	self allSubInstances
		do: [ :m | 
			| itsOwner |
			(m isHandMorph not and: [ (itsOwner := m owner) notNil and: [ (itsOwner submorphs includes: m) not ] ])
				ifTrue: [ problemMorphs add: m ] ].
	^ problemMorphs! !

!Morph class methodsFor: 'instance creation' stamp: ''!
newBounds: bounds

	^ self new privateBounds: bounds! !

!Morph class methodsFor: 'instance creation' stamp: 'jm 5/29/1998 21:28'!
newBounds: bounds color: color

	^ (self new privateBounds: bounds) privateColor: color
! !

!Morph class methodsFor: 'instance creation' stamp: 'sw 8/4/97 12:05'!
newSticky

	^ self new beSticky! !

!Morph class methodsFor: 'misc' stamp: 'PeterHugossonMiller 9/2/2009 16:13'!
obtainArrowheadFor: aPrompt defaultValue: defaultPoint
	"Allow the user to supply a point to serve as an arrowhead size.  Answer nil if we fail to get a good point"

	| result  |
	result := UIManager default request: aPrompt initialAnswer: defaultPoint asString.
	result isEmptyOrNil ifTrue: [^ nil].
	^ [(Point readFrom: result readStream)]
		on: Error do: [:ex |  nil].! !

!Morph class methodsFor: 'initialize-release' stamp: 'IgorStasenko 4/15/2011 17:22'!
unload

	Smalltalk tools fileList unregisterFileReader: self ! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'tk 12/15/1998 14:23'!
abandon
	"Like delete, but we really intend not to use this morph again.  Clean up a few things."

	self delete! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'AlainPlantec 12/1/2010 09:32'!
aboutToBeGrabbedBy: aHand
	"The receiver is being grabbed by a hand.
	Perform necessary adjustments (if any) and return the actual morph
	that should be added to the hand."
	| extentToHandToHand |
	self formerOwner: owner.
	self formerPosition: self position.
	(extentToHandToHand := self valueOfProperty: #expandedExtent)
		ifNotNil:	[self removeProperty: #expandedExtent.
			self extent: extentToHandToHand].
	^self "Grab me"! !

!Morph methodsFor: 'layout' stamp: 'ar 11/12/2000 17:33'!
acceptDroppingMorph: aMorph event: evt
	"This message is sent when a morph is dropped onto a morph that has agreed to accept the dropped morph by responding 'true' to the wantsDroppedMorph:Event: message. This default implementation just adds the given morph to the receiver."
	| layout |
	layout := self layoutPolicy.
	layout ifNil:[^self addMorph: aMorph].
	self privateAddMorph: aMorph 
		atIndex: (layout indexForInserting: aMorph at: evt position in: self).! !

!Morph methodsFor: 'updating' stamp: 'BenjaminVanRyseghem 7/31/2013 14:10'!
actAsExecutor
	"Prepare the receiver to act as executor for any resources associated with it"
	self breakDependents! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/28/2001 08:39'!
actWhen
	"Answer when the receiver, probably being used as a button, should have its action triggered"

	^ self valueOfProperty: #actWhen ifAbsentPut: [#buttonDown]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/25/2001 10:23'!
actWhen: aButtonPhase
	"Set the receiver's actWhen trait"

	self setProperty: #actWhen toValue: aButtonPhase! !

!Morph methodsFor: 'events-accessing' stamp: 'gvc 9/11/2009 17:43'!
actionMap
	"Answer an action map"

	^self updateableActionMap! !

!Morph methodsFor: 'structure' stamp: 'ar 3/18/2001 00:11'!
activeHand
	^ActiveHand! !

!Morph methodsFor: 'menus' stamp: 'sw 11/27/2001 14:36'!
addAddHandMenuItemsForHalo: aMenu hand: aHandMorph
	"The former charter of this method was to add halo menu items that pertained specifically to the hand.  Over time this charter has withered, and most morphs reimplement this method simply to add their morph-specific menu items.  So in the latest round, all other implementors in the standard image have been removed.  However, this is left here as a hook for the benefit of existing code in client uses."

! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: #() after: delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: #() at: scheduledTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1) after: delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1) at: scheduledTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 with: arg2 after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1 with: arg2) after: delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 with: arg2 at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1 with: arg2) at: scheduledTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector withArguments: args after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: args at: Time millisecondClockValue + delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/14/2000 12:15'!
addAlarm: aSelector withArguments: args at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	| scheduler |
	scheduler := self alarmScheduler.
	scheduler ifNotNil:[scheduler addAlarm: aSelector withArguments: args for: self at: scheduledTime].! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 8/12/2003 23:28'!
addAllMorphs: aCollection
	^self privateAddAllMorphs: aCollection atIndex: submorphs size! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 8/12/2003 23:29'!
addAllMorphs: aCollection after: anotherMorph
	^self privateAddAllMorphs: aCollection 
			atIndex: (submorphs indexOf: anotherMorph ifAbsent: [submorphs size])! !

!Morph methodsFor: 'menu' stamp: 'sw 11/27/2001 15:21'!
addBorderStyleMenuItems: aMenu hand: aHandMorph
	"Probably one could offer border-style items even if it's not a borderedMorph, so this remains a loose end for the moment"
! !

!Morph methodsFor: 'layout-menu' stamp: 'GuillermoPolito 5/29/2011 14:48'!
addCellLayoutMenuItems: aMenu hand: aHand
	"Cell (e.g., child) related items"
	| menu sub |
	menu := UIManager default newMenuIn: self for: self.
		menu addUpdating: #hasDisableTableLayoutString action: #changeDisableTableLayout.
		menu addLine.

		sub := UIManager default newMenuIn: self for: self.
		#(rigid shrinkWrap spaceFill) do:[:sym|
			sub addUpdating: #hResizingString: target: self selector: #hResizing: argumentList: (Array with: sym)].
		menu add:'horizontal resizing' translated subMenu: sub.

		sub := UIManager default newMenuIn: self for: self.
		#(rigid shrinkWrap spaceFill) do:[:sym|
			sub addUpdating: #vResizingString: target: self selector: #vResizing: argumentList: (Array with: sym)].
		menu add:'vertical resizing' translated subMenu: sub.

	aMenu ifNotNil:[aMenu add: 'child layout' translated subMenu: menu].
	^menu! !

!Morph methodsFor: 'menus' stamp: 'sw 4/27/1998 03:44'!
addCustomHaloMenuItems: aMenu hand: aHandMorph
	"Add morph-specific items to the given menu which was invoked by the given hand from the halo.  To get started, we defer to the counterpart method used with the option-menu, but in time we can have separate menu choices for halo-menus and for option-menus"

	self addCustomMenuItems: aMenu hand: aHandMorph! !

!Morph methodsFor: 'menus' stamp: 'sw 11/27/2001 07:17'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	"Add morph-specific items to the given menu which was invoked by the given hand.  This method provides is invoked both from the halo-menu and from the control-menu regimes."
! !

!Morph methodsFor: 'debug and other' stamp: 'EstebanLorenzano 5/14/2013 09:44'!
addDebuggingItemsTo: aMenu hand: aHandMorph
	aMenu add: 'debug...' translated subMenu:  (self buildDebugMenu: aHandMorph).
	aMenu lastItem icon: Smalltalk ui icons smallDebugIcon! !

!Morph methodsFor: 'updating' stamp: 'IgorStasenko 8/27/2013 16:29'!
addDependent: anObject

	self announcer weak
		on: MorphChanged , MorphChangedWithArguments
		send: #handleUpdate:
		to: anObject.
			
	^ anObject! !

!Morph methodsFor: 'drop shadows' stamp: 'RAA 1/19/2001 07:51'!
addDropShadow

	self hasDropShadow ifTrue:[^self].
	self changed.
	self hasDropShadow: true.
	self shadowOffset: 3@3.
	self layoutChanged.
	self changed.! !

!Morph methodsFor: 'drop shadows' stamp: 'GuillermoPolito 5/29/2011 14:48'!
addDropShadowMenuItems: aMenu hand: aHand
	| menu |
	menu := UIManager default newMenuIn: self for: self.
	menu
		addUpdating: #hasDropShadowString
		action: #toggleDropShadow.
	menu addLine.
	menu add: 'shadow color...' translated target: self selector: #changeShadowColor.
	menu add: 'shadow offset...' translated target: self selector: #setShadowOffset:.
	aMenu add: 'drop shadow' translated subMenu: menu.! !

!Morph methodsFor: 'meta-actions' stamp: 'GuillermoPolito 5/29/2011 14:48'!
addEmbeddingMenuItemsTo: aMenu hand: aHandMorph
	"Construct a menu offerring embed targets for the receiver.  If the incoming menu is is not degenerate, add the constructed menu as a submenu; in any case, answer the embed-target menu"

	| menu potentialEmbeddingTargets |

	potentialEmbeddingTargets := self potentialEmbeddingTargets.
	potentialEmbeddingTargets size > 1 ifFalse:[^ self].

	menu := UIManager default newMenuIn: self for: self.

	potentialEmbeddingTargets reverseDo: [:m | 
			menu
				add: (m class name asString)
				target: m
				selector: #addMorphFrontFromWorldPosition:
				argument: self topRendererOrSelf.

			menu lastItem icon: (m iconOrThumbnailOfSize: 16).

			self owner == m ifTrue:[menu lastItem emphasis: 1].
		].

	aMenu add:'embed into' translated subMenu: menu.

	^ menu! !

!Morph methodsFor: 'menus' stamp: 'EstebanLorenzano 5/14/2013 09:44'!
addExportMenuItems: aMenu hand: aHandMorph
	"Add export items to the menu"

	aMenu ifNotNil:
		[ | aSubMenu |
		aSubMenu := UIManager default newMenuIn: self for: self.
		aSubMenu add: 'BMP file' translated action: #exportAsBMP.
		aSubMenu add: 'GIF file' translated action: #exportAsGIF.
		aSubMenu add: 'JPEG file' translated action: #exportAsJPEG.
		aSubMenu add: 'PNG file' translated action: #exportAsPNG.
		aMenu add: 'export...' translated icon: Smalltalk ui icons smallExportIcon subMenu: aSubMenu.
		aMenu lastItem icon: Smalltalk ui icons smallExportIcon ].
! !

!Morph methodsFor: 'menus' stamp: 'MarcusDenker 10/26/2011 14:49'!
addFillStyleMenuItems: aMenu hand: aHand
	"Add the items for changing the current fill style of the Morph"
	| menu |
	menu := UIManager default newMenuIn: self for: self.
	self fillStyle addFillStyleMenuItems: menu hand: aHand from: self.
	menu addLine.
	menu add: 'solid fill' translated action: #useSolidFill.
	menu add: 'gradient fill' translated action: #useGradientFill.
	menu add: 'bitmap fill' translated action: #useBitmapFill.
	menu add: 'default fill' translated action: #useDefaultFill.
	aMenu add: 'fill style' translated subMenu: menu.
! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'sw 3/30/2005 03:44'!
addFlexShell
	"Wrap a rotating and scaling shell around this morph."

	| oldHalo flexMorph myWorld anIndex |

	myWorld := self world.
	oldHalo := self halo.
	anIndex := self owner submorphIndexOf: self.
	self owner addMorph: (flexMorph := self newTransformationMorph asFlexOf: self)
		asElementNumber: anIndex.
	self transferStateToRenderer: flexMorph.
	oldHalo ifNotNil: [oldHalo setTarget: flexMorph].
	myWorld ifNotNil: [myWorld startSteppingSubmorphsOf: flexMorph].

	^ flexMorph! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 11/7/1999 18:57'!
addHalo
	"Invoke a halo programatically (e.g., not from a meta gesture)"
	^self addHalo: nil! !

!Morph methodsFor: 'halos and balloon help' stamp: 'BenjaminVanRyseghem 5/4/2013 12:54'!
addHalo: evt
	| halo prospectiveHaloClass |
	
	Smalltalk tools userManager canShowMorphHalo
		ifFalse: [ ^ self ].
	prospectiveHaloClass := Smalltalk globals at: self haloClass ifAbsent: [ HaloMorph ].
	halo := prospectiveHaloClass new.
	halo bounds: (halo worldBoundsForMorph: self).
	halo popUpFor: self event: evt.
	^ halo! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 11/7/1999 21:55'!
addHalo: evt from: formerHaloOwner
	"Transfer a halo from the former halo owner to the receiver"
	^self addHalo: evt! !

!Morph methodsFor: 'menus' stamp: 'GuillermoPolito 5/29/2011 14:48'!
addHaloActionsTo: aMenu
	"Add items to aMenu representing actions requestable via halo"

	| subMenu |
	subMenu := UIManager default newMenuIn: self for: self.
	subMenu addTitle: self externalName.
	subMenu addStayUpItemSpecial.
	subMenu addLine.
	subMenu add: 'delete' translated action: #dismissViaHalo.
	subMenu balloonTextForLastItem: 'Delete this object -- warning -- can be destructive!!' translated.

	self maybeAddCollapseItemTo: subMenu.
	subMenu add: 'grab' translated action: #openInHand.
	subMenu balloonTextForLastItem: 'Pick this object up -- warning, since this removes it from its container, it can have adverse effects.' translated.

	subMenu addLine.

	subMenu add: 'resize' translated action: #resizeFromMenu.
	subMenu balloonTextForLastItem: 'Change the size of this object' translated.

	subMenu add: 'duplicate' translated action: #maybeDuplicateMorph.
	subMenu balloonTextForLastItem: 'Hand me a copy of this object' translated.

	subMenu addLine.

	subMenu add: 'set color' translated target: self renderedMorph action: #changeColor.
	subMenu balloonTextForLastItem: 'Change the color of this object' translated.

	subMenu addLine.

	subMenu add: 'inspect' translated target: self action: #inspect.
	subMenu balloonTextForLastItem: 'Open an Inspector on this object' translated.

	aMenu add: 'halo actions...' translated subMenu: subMenu
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'SeanDeNigris 1/29/2013 13:10'!
addHandlesTo: aHaloMorph box: box
	"Add halo handles to the halo.  Apply the halo filter if appropriate"

	
	aHaloMorph haloBox: box.
	HaloMorph  currentHaloSpecifications do:
		[:aSpec | | wantsIt aSelector | 
			aSelector := aSpec addHandleSelector.
			(wantsIt := self wantsHaloHandleWithSelector: aSelector inHalo: aHaloMorph)
				ifTrue: [(#(addDupHandle:) includes: aSelector) 
					ifTrue: [wantsIt := self preferredDuplicationHandleSelector = aSelector]].
			wantsIt 
				ifTrue: [aHaloMorph perform: aSelector with: aSpec]].
	aHaloMorph innerTarget addOptionalHandlesTo: aHaloMorph box: box! !

!Morph methodsFor: 'layout-menu' stamp: 'GuillermoPolito 5/29/2011 14:48'!
addLayoutMenuItems: topMenu hand: aHand
	| aMenu |
	aMenu := UIManager default newMenuIn: self for: self.
	aMenu addUpdating: #hasNoLayoutString action: #changeNoLayout.
	aMenu addUpdating: #hasProportionalLayoutString action: #changeProportionalLayout.
	aMenu addUpdating: #hasTableLayoutString action: #changeTableLayout.
	aMenu addLine.
	aMenu add: 'change layout inset...' translated action: #changeLayoutInset:.
	aMenu addLine.
	self addCellLayoutMenuItems: aMenu hand: aHand.
	self addTableLayoutMenuItems: aMenu hand: aHand.
	topMenu ifNotNil:[topMenu add: 'layout' translated subMenu: aMenu].
	^aMenu! !

!Morph methodsFor: 'menus' stamp: 'GuillermoPolito 5/29/2011 14:48'!
addMiscExtrasTo: aMenu
	"Add a submenu of miscellaneous extra items to the menu."

	| subMenu |
	subMenu := UIManager default newMenuIn: self for: self.
	(self isWorldMorph not and: [(self renderedMorph isSystemWindow) not])
		ifTrue: [subMenu add: 'put in a window' translated action: #embedInWindow].

	self isWorldMorph ifFalse:
		[subMenu add: 'adhere to edge...' translated action: #adhereToEdge.
		subMenu addLine].

	subMenu
		add: 'add mouse up action' translated action: #addMouseUpAction;
		add: 'remove mouse up action' translated action: #removeMouseUpAction.
	subMenu addLine.
	aMenu add: 'extras...' translated subMenu: subMenu! !

!Morph methodsFor: 'menu' stamp: 'marcus.denker 11/10/2008 10:04'!
addModelYellowButtonItemsTo: aCustomMenu event: evt 
	"Give my models a chance to add their context-menu items to  
	aCustomMenu."
	self model
		ifNotNil: [:mod |
			mod
				addModelYellowButtonMenuItemsTo: aCustomMenu
				forMorph: self
				hand: evt hand]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: ''!
addMorph: aMorph

	self addMorphFront: aMorph.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:54'!
addMorph: newMorph after: aMorph
	"Add the given morph as one of my submorphs, inserting it after anotherMorph"
	^self privateAddMorph: newMorph atIndex: (submorphs indexOf: aMorph)+1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/7/2000 08:29'!
addMorph: aMorph asElementNumber: aNumber
	"Add the given morph so that it becomes the aNumber'th element of my submorph list.  If aMorph is already one of my submorphs, reposition it"

	(submorphs includes: aMorph) ifTrue:
		[aMorph privateDelete].
	(aNumber <= submorphs size)
		ifTrue:
			[self addMorph: aMorph inFrontOf: (submorphs at: aNumber)]
		ifFalse:
			[self addMorphBack: aMorph]
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:44'!
addMorph: newMorph behind: aMorph
	"Add a morph to the list of submorphs behind the specified morph"
	^self privateAddMorph: newMorph atIndex: (submorphs indexOf: aMorph) + 1.
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'StephaneDucasse 12/21/2012 18:11'!
addMorph: aMorph fullFrame: aLayoutFrame

	aMorph layoutFrame: aLayoutFrame asLayoutFrame.
	aMorph hResizing: #spaceFill; vResizing: #spaceFill.
	self addMorph: aMorph.

! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:45'!
addMorph: newMorph inFrontOf: aMorph
	"Add a morph to the list of submorphs in front of the specified morph"
	^self privateAddMorph: newMorph atIndex: ((submorphs indexOf: aMorph) max: 1).! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:55'!
addMorphBack: aMorph
	^self privateAddMorph: aMorph atIndex: submorphs size+1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'RAA 12/15/2000 19:34'!
addMorphCentered: aMorph

	aMorph position: bounds center - (aMorph extent // 2).
	self addMorphFront: aMorph.
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:54'!
addMorphFront: aMorph
	^self privateAddMorph: aMorph atIndex: 1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 11/15/1998 23:42'!
addMorphFront: aMorph fromWorldPosition: wp

	self addMorphFront: aMorph.
	aMorph position: (self transformFromWorld globalPointToLocal: wp)! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 12/16/2001 21:08'!
addMorphFrontFromWorldPosition: aMorph
	^self addMorphFront: aMorph fromWorldPosition: aMorph positionInWorld.! !

!Morph methodsFor: 'wiw support' stamp: 'nice 1/5/2010 15:59'!
addMorphInFrontOfLayer: aMorph

	| targetLayer |

	targetLayer := aMorph morphicLayerNumberWithin: self.
	submorphs do: [ :each | | layerHere |
		each == aMorph ifTrue: [^self].
		layerHere := each morphicLayerNumberWithin: self.
		"the <= is the difference - it insures we go to the front of our layer"
		targetLayer <= layerHere ifTrue: [
			^self addMorph: aMorph inFrontOf: each
		].
	].
	self addMorphBack: aMorph.
! !

!Morph methodsFor: 'wiw support' stamp: 'RAA 6/29/2000 10:49'!
addMorphInLayer: aMorph

	submorphs do: [ :each |
		each == aMorph ifTrue: [^self].
		aMorph morphicLayerNumber < each morphicLayerNumber ifTrue: [
			^self addMorph: aMorph inFrontOf: each
		].
	].
	self addMorphBack: aMorph
! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 1/19/2001 07:51'!
addMouseActionIndicatorsWidth: anInteger color: aColor

	self deleteAnyMouseActionIndicators.

	self changed.
	self hasRolloverBorder: true.
	self setProperty: #rolloverWidth toValue: anInteger@anInteger.
	self setProperty: #rolloverColor toValue: aColor.
	self layoutChanged.
	self changed.

! !

!Morph methodsFor: 'debug and other' stamp: 'alain.plantec 2/6/2009 15:31'!
addMouseUpAction
	| codeToRun oldCode |
	oldCode := self
				valueOfProperty: #mouseUpCodeToRun
				ifAbsent: [''].
	codeToRun := UIManager default request: 'MouseUp expression:' translated initialAnswer: oldCode.
	self addMouseUpActionWith: codeToRun! !

!Morph methodsFor: 'debug and other' stamp: 'gm 2/22/2003 13:41'!
addMouseUpActionWith: codeToRun 
	((codeToRun isMessageSend) not and: [codeToRun isEmptyOrNil]) 
		ifTrue: [^self].
	self setProperty: #mouseUpCodeToRun toValue: codeToRun.
	self 
		on: #mouseUp
		send: #programmedMouseUp:for:
		to: self.
	self 
		on: #mouseDown
		send: #programmedMouseDown:for:
		to: self.
	self 
		on: #mouseEnter
		send: #programmedMouseEnter:for:
		to: self.
	self 
		on: #mouseLeave
		send: #programmedMouseLeave:for:
		to: self! !

!Morph methodsFor: 'menu' stamp: 'nk 3/10/2004 19:49'!
addMyYellowButtonMenuItemsToSubmorphMenus
	"Answer true if I have items to add to the context menus of my submorphs"

	^true! !

!Morph methodsFor: 'menu' stamp: 'GuillermoPolito 5/29/2011 14:49'!
addNestedYellowButtonItemsTo: aMenu event: evt 
	"Add items to aMenu starting with me and proceeding down 
	through my submorph chain, 
	letting any submorphs that include the event position 
	contribute their items to the bottom of the menu, separated by 
	a line."
	| underMouse |

	self addYellowButtonMenuItemsTo: aMenu event: evt.

	underMouse := self
				submorphThat: [:each | each containsPoint: evt position]
				ifNone: [^ self].

	(underMouse addMyYellowButtonMenuItemsToSubmorphMenus
			and: [underMouse hasYellowButtonMenu])
		ifTrue: [| submenu |
			aMenu addLine.
			submenu := UIManager default newMenuIn: underMouse for: underMouse.
			underMouse addNestedYellowButtonItemsTo: submenu event: evt.
			aMenu
				add: underMouse externalName
				icon: (underMouse iconOrThumbnailOfSize: 16)
				subMenu: submenu
		]
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 5/7/2010 23:29'!
addOptionalHandlesTo: aHalo box: box
! !

!Morph methodsFor: 'menus' stamp: 'StephaneDucasse 3/3/2010 15:49'!
addStandardHaloMenuItemsTo: aMenu hand: aHandMorph
	"Add standard halo items to the menu"

	self isWorldMorph ifTrue:
		[^ self addWorldHaloMenuItemsTo: aMenu hand: aHandMorph].

	aMenu add: 'send to back' translated action: #goBehind.
	aMenu add: 'bring to front' translated action: #comeToFront.
	self addEmbeddingMenuItemsTo: aMenu hand: aHandMorph.
	aMenu addLine.

	self addFillStyleMenuItems: aMenu hand: aHandMorph.
	self addBorderStyleMenuItems: aMenu hand: aHandMorph.
	self addDropShadowMenuItems: aMenu hand: aHandMorph.
	self addLayoutMenuItems: aMenu hand: aHandMorph.
	self addHaloActionsTo: aMenu.
	owner isTextMorph ifTrue:[self addTextAnchorMenuItems: aMenu hand: aHandMorph].
	aMenu addLine.
	self addToggleItemsToHaloMenu: aMenu.
	aMenu addLine.
	self addExportMenuItems: aMenu hand: aHandMorph.
	self addMiscExtrasTo: aMenu.
	self addDebuggingItemsTo: aMenu hand: aHandMorph.
	aMenu addLine.
	aMenu defaultTarget: aHandMorph.
! !

!Morph methodsFor: 'layout-menu' stamp: 'GuillermoPolito 5/29/2011 14:50'!
addTableLayoutMenuItems: aMenu hand: aHand
	| menu sub |
	menu := UIManager default newMenuIn: self for: self.
	menu addUpdating: #hasReverseCellsString action: #changeReverseCells.
	menu addUpdating: #hasClipLayoutCellsString action: #changeClipLayoutCells.
	menu addUpdating: #hasRubberBandCellsString action: #changeRubberBandCells.
	menu addLine.
	menu add: 'change cell inset...' translated action: #changeCellInset:.
	menu add: 'change min cell size...' translated action: #changeMinCellSize:.
	menu add: 'change max cell size...' translated action: #changeMaxCellSize:.
	menu addLine.

	sub := UIManager default newMenuIn: self for: self.
	#(leftToRight rightToLeft topToBottom bottomToTop) do:[:sym|
		sub addUpdating: #listDirectionString: target: self selector: #changeListDirection: argumentList: (Array with: sym)].
	menu add: 'list direction' translated subMenu: sub.

	sub := UIManager default newMenuIn: self for: self.
	#(none leftToRight rightToLeft topToBottom bottomToTop) do:[:sym|
		sub addUpdating: #wrapDirectionString: target: self selector: #wrapDirection: argumentList: (Array with: sym)].
	menu add: 'wrap direction' translated subMenu: sub.

	sub := UIManager default newMenuIn: self for: self.
	#(center topLeft topRight bottomLeft bottomRight topCenter leftCenter rightCenter bottomCenter) do:[:sym|
		sub addUpdating: #cellPositioningString: target: self selector: #cellPositioning: argumentList: (Array with: sym)].
	menu add: 'cell positioning' translated subMenu: sub.

	sub := UIManager default newMenuIn: self for: self.
	#(topLeft bottomRight center justified) do:[:sym|
		sub addUpdating: #listCenteringString: target: self selector: #listCentering: argumentList: (Array with: sym)].
	menu add: 'list centering' translated subMenu: sub.

	sub := UIManager default newMenuIn: self for: self.
	#(topLeft bottomRight center justified) do:[:sym|
		sub addUpdating: #wrapCenteringString: target: self selector: #wrapCentering: argumentList: (Array with: sym)].
	menu add: 'wrap centering' translated subMenu: sub.

	sub := UIManager default newMenuIn: self for: self.
	#(none equal) do:[:sym|
		sub addUpdating: #listSpacingString: target: self selector: #listSpacing: argumentList: (Array with: sym)].
	menu add: 'list spacing' translated subMenu: sub.

	sub := UIManager default newMenuIn: self for: self.
	#(none localRect localSquare globalRect globalSquare) do:[:sym|
		sub addUpdating: #cellSpacingString: target: self selector: #cellSpacing: argumentList: (Array with: sym)].
	menu add: 'cell spacing' translated subMenu: sub.

	aMenu ifNotNil:[aMenu add: 'table layout' translated subMenu: menu].
	^menu! !

!Morph methodsFor: 'text-anchor' stamp: 'GuillermoPolito 5/29/2011 14:50'!
addTextAnchorMenuItems: topMenu hand: aHand
	| aMenu |
	aMenu := UIManager default newMenuIn: self for: self.
	aMenu addUpdating: #hasInlineAnchorString action: #changeInlineAnchor.
	aMenu addUpdating: #hasParagraphAnchorString action: #changeParagraphAnchor.
	aMenu addUpdating: #hasDocumentAnchorString action: #changeDocumentAnchor.
	topMenu ifNotNil:[topMenu add: 'text anchor' subMenu: aMenu].
	^aMenu! !

!Morph methodsFor: 'menu' stamp: 'MarcusDenker 9/7/2010 17:41'!
addTitleForHaloMenu: aMenu 
	aMenu
		addTitle: self externalName
		icon: (self iconOrThumbnailOfSize: 28)! !

!Morph methodsFor: 'menus' stamp: 'marcus.denker 11/19/2008 13:44'!
addToggleItemsToHaloMenu: aMenu
	"Add standard true/false-checkbox items to the memu"

	#(
		(resistsRemovalString toggleResistsRemoval 'whether I should be reistant to easy deletion via the pink X handle' true)
		(stickinessString toggleStickiness 'whether I should be resistant to a drag done by mousing down on me' true)
		(lockedString lockUnlockMorph 'when "locked", I am inert to all user interactions' true)
		(hasClipSubmorphsString changeClipSubmorphs 'whether the parts of objects within me that are outside my bounds should be masked.' false)
		(hasDirectionHandlesString changeDirectionHandles 'whether direction handles are shown with the halo' false)
		(hasDragAndDropEnabledString changeDragAndDrop 'whether I am open to having objects dropped into me' false)
	) do: [:each |
			aMenu addUpdating: each first action: each second.
			aMenu balloonTextForLastItem: each third translated].

	self couldHaveRoundedCorners ifTrue:
		[aMenu addUpdating: #roundedCornersString action: #toggleCornerRounding.
		aMenu balloonTextForLastItem: 'whether my corners should be rounded' translated]! !

!Morph methodsFor: 'utilities' stamp: 'sw 10/23/1998 12:00'!
addTransparentSpacerOfSize: aPoint
	self addMorphBack: (self transparentSpacerOfSize: aPoint)! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 12/12/2009 09:54'!
addWorldHandlesTo: aHaloMorph box: box
	aHaloMorph haloBox: box.
	HaloMorph haloSpecificationsForWorld 
		do: [:aSpec | aHaloMorph perform: aSpec addHandleSelector with: aSpec].
	aHaloMorph innerTarget addOptionalHandlesTo: aHaloMorph box: box! !

!Morph methodsFor: 'menu' stamp: 'AlainPlantec 12/19/2009 23:13'!
addYellowButtonMenuItemsTo: aMenu event: evt 
	"Populate aMenu with appropriate menu items for a  
	yellow-button (context menu) click."
	aMenu defaultTarget: self.
	self defaultYellowButtonMenuEnabled ifFalse: [^ self].
	aMenu addStayUpItem.
	self addModelYellowButtonItemsTo: aMenu event: evt.
	self cmdGesturesEnabled ifTrue: [
			aMenu addLine.
			aMenu add: 'inspect' translated action: #inspect].
	aMenu addLine.
	self isWorldMorph ifFalse: [aMenu add: 'delete' translated action: #delete].
	self world selectedObject == self
		ifTrue: [aMenu add: 'halo off' translated action: #removeHalo]
		ifFalse: [aMenu add: 'halo on' translated action: #addHalo].
	(self isWorldMorph
			or: [self wantsToBeTopmost])
		ifFalse: [
			aMenu addLine.
			aMenu add: 'send to back' translated action: #goBehind.
			aMenu add: 'bring to front' translated action: #comeToFront.
			self addEmbeddingMenuItemsTo: aMenu hand: evt hand].
	self isWorldMorph ifFalse: [
			self isFullOnScreen ifFalse: [aMenu add: 'move onscreen' translated action: #goHome]].
			self addLayoutMenuItems: aMenu hand: evt hand.
			(owner notNil
					and: [owner isTextMorph])
				ifTrue: [self addTextAnchorMenuItems: aMenu hand: evt hand].
	self isWorldMorph ifFalse: [
			aMenu addLine.
			self addToggleItemsToHaloMenu: aMenu].
	aMenu addLine.
	self isWorldMorph
		ifFalse: [aMenu add: 'copy to paste buffer' translated action: #copyToPasteBuffer:].
	(self allStringsAfter: nil) isEmpty ifFalse: [aMenu add: 'copy text' translated action: #clipText].
	self addExportMenuItems: aMenu hand: evt hand.
	aMenu addLine.
	aMenu add: 'adhere to edge...' translated action: #adhereToEdge.
	self addCustomMenuItems: aMenu hand: evt hand! !

!Morph methodsFor: 'change reporting' stamp: 'ar 8/12/2003 21:50'!
addedMorph: aMorph
	"Notify the receiver that the given morph was just added."
! !

!Morph methodsFor: 'menus' stamp: 'GuillermoPolito 5/29/2011 14:50'!
adhereToEdge
	| menu |
	menu := UIManager default newMenuIn: self for: self.
	#(top right bottom left - center - topLeft topRight bottomRight bottomLeft - none)
		do: [:each |
			each == #-
				ifTrue: [menu addLine]
				ifFalse: [menu add: each asString translated selector: #setToAdhereToEdge: argument: each]].
	menu popUpEvent: self currentEvent in: self world! !

!Morph methodsFor: 'menus' stamp: 'wiz 10/19/2006 00:35'!
adhereToEdge: edgeSymbol 
	| edgeMessage |
	(owner isNil or: [owner isHandMorph]) ifTrue: [^self].
	(owner class canUnderstand:  edgeSymbol) ifFalse:  [^self].
	(self class canUnderstand: ( edgeMessage := (edgeSymbol , ':') asSymbol ))
		 ifFalse:  [^self].
	
	self perform: edgeMessage
		withArguments: (Array with: (owner perform: edgeSymbol))! !

!Morph methodsFor: 'layout' stamp: 'ar 11/12/2000 17:34'!
adjustLayoutBounds
	"Adjust the receivers bounds depending on the resizing strategy imposed"
	| hFit vFit box myExtent extent |
	hFit := self hResizing.
	vFit := self vResizing.
	(hFit == #shrinkWrap or:[vFit == #shrinkWrap]) ifFalse:[^self]. "not needed"
	box := self layoutBounds.
	myExtent := box extent.
	extent := self submorphBounds corner - box origin.
	hFit == #shrinkWrap ifTrue:[myExtent := extent x @ myExtent y].
	vFit == #shrinkWrap ifTrue:[myExtent := myExtent x @ extent y].
	"Make sure we don't get smaller than minWidth/minHeight"
	myExtent x < self minWidth ifTrue:[
		myExtent := (myExtent x max: 
			(self minWidth - self bounds width + self layoutBounds width)) @ myExtent y].
	myExtent y < self minHeight ifTrue:[
		myExtent := myExtent x @ (myExtent y max:
			(self minHeight - self bounds height + self layoutBounds height))].
	self layoutBounds: (box origin extent: myExtent).! !

!Morph methodsFor: 'menus' stamp: 'sw 2/3/2000 00:14'!
adjustedCenter
	"Provides a hook for objects to provide a reference point other than the receiver's center,for the purpose of centering a submorph under special circumstances, such as BalloonMorph"

	^ self center! !

!Morph methodsFor: 'menus' stamp: 'sw 2/3/2000 00:12'!
adjustedCenter: c
	"Set the receiver's position based on the #adjustedCenter protocol for adhereToEdge.  By default this simply sets the receiver's center.   Though there are (at its inception anyway) no other implementors of this method, it is required in use with the #adhereToEdge when the centering of a submorph is to be with reference to a rectangle  other than the receiver's center."

	self center: c! !

!Morph methodsFor: 'accessing' stamp: 'ar 12/18/2001 20:09'!
adoptPaneColor: paneColor
	self submorphsDo:[:m| m adoptPaneColor: paneColor].! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:34'!
alarmScheduler
	"Return the scheduler being responsible for triggering alarms"
	^self world! !

!Morph methodsFor: 'geometry' stamp: 'di 7/24/97 11:55'!
align: aPoint1 with: aPoint2
	"Translate by aPoint2 - aPoint1."

	^ self position: self position + (aPoint2 - aPoint1)! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'AlainPlantec 2/17/2010 14:53'!
allMorphs
	"Return a collection containing all morphs in this composite morph (including the receiver)."

	| all |
	all := (Array new: submorphs size) writeStream.
	self allMorphsDo: [:m | all nextPut: m].
	^ all contents! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:27'!
allMorphsDo: aBlock 
	"Evaluate the given block for all morphs in this composite morph (including the receiver)."

	submorphs do: [:m | m allMorphsDo: aBlock].
	aBlock value: self! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'alain.plantec 6/8/2009 23:41'!
allMorphsInto: aSet
	"Return a set of all submorphs.  Don't forget the hidden ones. Consider only objects that are in memory (see allNonSubmorphMorphs)." 

	submorphs do: [:m | m allMorphsInto: aSet].
	self allNonSubmorphMorphs do: [:m | 
			(aSet includes: m) ifFalse: ["Stop infinite recursion"
				m allMorphsInto: aSet]].
	aSet add: self.
	^ aSet! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'alain.plantec 6/8/2009 23:44'!
allNonSubmorphMorphs
	"Return a collection containing all morphs in this morph which are not currently in the submorph containment hierarchy"

	^ OrderedCollection new! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 01:00'!
allOwners
	"Return the owners of the reciever"

	^ Array streamContents: [:strm | self allOwnersDo: [:m | strm nextPut: m]]! !

!Morph methodsFor: 'structure' stamp: 'ar 9/14/2000 16:47'!
allOwnersDo: aBlock
	"Evaluate aBlock with all owners of the receiver"
	owner ifNotNil:[^owner withAllOwnersDo: aBlock].! !

!Morph methodsFor: 'debug and other' stamp: 'nice 1/5/2010 15:59'!
allStringsAfter: aSubmorph 
	"return an OrderedCollection of strings of text in my submorphs.  If aSubmorph is non-nil, begin with that container."

	| list ok |
	list := OrderedCollection new.
	ok := aSubmorph isNil.
	self allMorphsDo: 
			[:sub | | string | 
			ok ifFalse: [ok := sub == aSubmorph].	"and do this one too"
			ok 
				ifTrue: 
					[(string := sub userString) ifNotNil: 
							[string isString ifTrue: [list add: string] ifFalse: [list addAll: string]]]].
	^list! !

!Morph methodsFor: 'announcements' stamp: 'ThierryGoubier 1/8/2014 14:35'!
announceDeleted
	self doAnnounce: (MorphDeleted morph: self).
	self submorphs do: #announceDeleted! !

!Morph methodsFor: 'announcements' stamp: 'ThierryGoubier 1/8/2014 14:36'!
announceKeyboardFocusChange: gotFocus
	| announcement |
	announcement := gotFocus 
						ifTrue: [ MorphGotFocus morph: self ]
						ifFalse: [ MorphLostFocus morph: self ].
	self doAnnounce: announcement.
	Morph announcer announce: announcement.! !

!Morph methodsFor: 'announcements' stamp: 'ThierryGoubier 1/8/2014 14:36'!
announceOpened
	self doAnnounce: (MorphOpened morph: self).
	self submorphs do: #announceOpened! !

!Morph methodsFor: 'announcements' stamp: 'GuillermoPolito 5/1/2012 17:01'!
announcer
	^self valueOfProperty: #announcer ifAbsentPut: [ Announcer new ]! !

!Morph methodsFor: 'drawing' stamp: 'Henrik Sperre Johansen 3/15/2009 00:03'!
areasRemainingToFill: aRectangle 
	"Pushed up from BorderedMorph, all cases tested for there are
	supported by basic Morph."
	"Morphs which achieve translucency by other means than fillStyle will have
	to reimplement this"
	"Fixed here to test the fillStyle rather than color for translucency.
	Since can have a translucent fillStyle while the (calculated) color is not."
	self fillStyle isTranslucent
		ifTrue: [^ Array with: aRectangle].
	self wantsRoundedCorners
		ifTrue: [(self borderWidth > 0
					and: [self borderColor isColor
							and: [self borderColor isTranslucent]])
				ifTrue: [^ aRectangle
						areasOutside: (self innerBounds intersect: self boundsWithinCorners)]
				ifFalse: [^ aRectangle areasOutside: self boundsWithinCorners]]
		ifFalse: [(self borderWidth > 0
					and: [self borderColor isColor
							and: [self borderColor isTranslucent]])
				ifTrue: [^ aRectangle areasOutside: self innerBounds]
				ifFalse: [^ aRectangle areasOutside: self bounds]]! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 3/22/2000 14:27'!
arrangeToStartStepping
	"Arrange to start getting sent the 'step' message, but don't do that initial #step call that startStepping does"

	self arrangeToStartSteppingIn: self world! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 3/22/2000 14:26'!
arrangeToStartSteppingIn: aWorld
	"Start getting sent the 'step' message in aWorld.  Like startSteppingIn:, but without the initial one to get started'"
	aWorld ifNotNil:
		[aWorld startStepping: self.
		self changed]! !

!Morph methodsFor: 'converting' stamp: ''!
asDraggableMorph
	^self! !

!Morph methodsFor: 'creation' stamp: 'tk 2/6/1999 22:43'!
asMorph
	^ self! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:46'!
assureExtension
	"creates an extension for the receiver if needed"
	extension ifNil: [self initializeExtension].
	^ extension! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:56'!
assureLayoutProperties
	| props |
	props := self layoutProperties.
	props == self ifTrue:[props := nil].
	props ifNil:[
		props := LayoutProperties new initializeFrom: self.
		self layoutProperties: props].
	^props! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:57'!
assureTableProperties
	| props |
	props := self layoutProperties.
	props == self ifTrue:[props := nil].
	props ifNil:[
		props := TableLayoutProperties new initializeFrom: self.
		self layoutProperties: props].
	props includesTableProperties 
		ifFalse:[self layoutProperties: (props := props asTableLayoutProperties)].
	^props! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 11/11/2000 14:54'!
balloonColor
	^ self
		valueOfProperty: #balloonColor
		ifAbsent: [self defaultBalloonColor]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 11/11/2000 14:55'!
balloonColor: aColor
	^ self
		setProperty: #balloonColor
		toValue: aColor! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sd 12/5/2001 20:29'!
balloonFont
	^ self
		valueOfProperty: #balloonFont
		ifAbsent: [self defaultBalloonFont]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sd 12/5/2001 20:30'!
balloonFont: aFont 
	^ self setProperty: #balloonFont toValue: aFont! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 2/7/2000 11:27'!
balloonHelpAligner
	"Answer the morph to which the receiver's balloon help should point"
	^ (self valueOfProperty: #balloonTarget) ifNil: [self]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 12/19/2009 23:30'!
balloonHelpDelayTime
	"Return the number of milliseconds before a balloon help should be put up on the receiver. The balloon help will only be put up if the receiver responds to #wantsBalloon by returning true."
	^ self theme settings balloonHelpDelayTime! !

!Morph methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:16'!
balloonHelpEnabled
	^ self theme settings balloonHelpEnabled! !

!Morph methodsFor: 'halos and balloon help' stamp: 'SeanDeNigris 1/29/2013 13:57'!
balloonHelpTextForHandle: aHandle 
	"Answer a string providing balloon help for the
	given halo handle"
	(aHandle eventHandler mouseSelectorsInclude: #doRecolor:with:) ifTrue: [^ 'Change color'].
	(aHandle eventHandler mouseSelectorsInclude: #mouseDownInDimissHandle:with:) ifTrue: [^ 'Remove from screen' translated].
	#(#(#addFullHandles 'More halo handles')  #(#chooseEmphasisOrAlignment 'Emphasis & alignment') #(#chooseFont 'Change font') #(#chooseNewGraphicFromHalo 'Choose a new graphic') #(#chooseStyle 'Change style') #(#doDebug:with: 'Debug') #(#doDirection:with: 'Choose forward direction') #(#doDup:with: 'Duplicate')  #(#doMenu:with: 'Menu') #(#doGrab:with: 'Pick up')  #(#mouseDownInCollapseHandle:with: 'Collapse') #(#mouseDownOnHelpHandle: 'Help')  #(#prepareToTrackCenterOfRotation:with: 'Move object or set center of rotation') #(#startDrag:with: 'Move') #(#startGrow:with: 'Change size') #(#startRot:with: 'Rotate') #(#startScale:with: 'Change scale')#(#trackCenterOfRotation:with: 'Set center of rotation') )
		do: [:pair | (aHandle eventHandler mouseSelectorsInclude: pair first)
				ifTrue: [^ pair last]].
	^ 'unknown halo handle' translated! !

!Morph methodsFor: 'accessing' stamp: 'MarcusDenker 9/7/2010 16:59'!
balloonText
	"Answer balloon help text or nil, if no help is available.  
	NB: subclasses may override such that they programatically  
	construct the text, for economy's sake, such as model phrases in 
	a Viewer"

	extension ifNil: [^nil].
	^extension balloonText ifNotNil: [:text | 
				text asString withNoLineLongerThan: self theme settings maxBalloonHelpLineLength]! !

!Morph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:57'!
beSticky
	"make the receiver sticky"
	self assureExtension sticky: true! !

!Morph methodsFor: 'accessing' stamp: 'sw 10/23/1998 12:01'!
beTransparent
	self color: Color transparent! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:52'!
beUnsticky
	"If the receiver is marked as sticky, make it now be unsticky"
	extension ifNotNil: [extension sticky: false]! !

!Morph methodsFor: 'user interface' stamp: 'tak 3/15/2005 17:36'!
becomeModal
	self currentWorld
		ifNotNil: [self currentWorld modalWindow: self]! !

!Morph methodsFor: 'Morphic-Base-Pluggable Widgets' stamp: 'BenjaminVanRyseghem 5/7/2012 15:03'!
beginsWith: aString fromList: aMorph
	| string |
	string := self userString ifNil: [(self submorphs collect: [:m | m userString]) detect: [:us | us notNil] ifNone: ['']].
	
	^ string asString beginsWith: aString fromList: aMorph! !

!Morph methodsFor: 'meta-actions' stamp: 'yo 3/15/2005 14:45'!
blueButtonDown: anEvent
	"Special gestures (cmd-mouse on the Macintosh; Alt-mouse on Windows and Unix) allow a mouse-sensitive morph to be moved or bring up a halo for the morph."
	| h tfm doNotDrag |
	h := anEvent hand halo.
	"Prevent wrap around halo transfers originating from throwing the event back in"
	doNotDrag := false.
	h ifNotNil:[
		(h innerTarget == self) ifTrue:[doNotDrag := true].
		(h innerTarget hasOwner: self) ifTrue:[doNotDrag := true].
		(self hasOwner: h target) ifTrue:[doNotDrag := true]].

	tfm := (self transformedFrom: nil) inverseTransformation.

	"cmd-drag on flexed morphs works better this way"
	h := self addHalo: (anEvent transformedBy: tfm).
	h ifNil: [^ self].
	doNotDrag ifTrue:[^self].
	"Initiate drag transition if requested"
	anEvent hand 
		waitForClicksOrDrag: h
		event: (anEvent transformedBy: tfm)
		selectors: { nil. nil. nil. #dragTarget:. }
		threshold: 5.
	"Pass focus explicitly here"
	anEvent hand newMouseFocus: h.! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 9/15/2000 20:25'!
blueButtonUp: anEvent
	"Ignored. Theoretically we should never get here since control is transferred to the halo on #blueButtonDown: but subclasses may implement this differently."! !

!Morph methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:28'!
borderColor
	^self borderStyle color! !

!Morph methodsFor: 'accessing' stamp: 'nk 4/15/2004 10:55'!
borderColor: aColorOrSymbolOrNil 
	"Unfortunately, the argument to borderColor could be more than 	just a color. 
	It could also be a symbol, in which case it is to be interpreted as a style identifier.
	But I might not be able to draw that kind of border, so it may have to be ignored.
	Or it could be nil, in which case I should revert to the default border."

	| style newStyle |
	style := self borderStyle.
	style baseColor = aColorOrSymbolOrNil
		ifTrue: [^ self].

	aColorOrSymbolOrNil isColor
		ifTrue: [style style = #none "default border?"
				ifTrue: [self borderStyle: (SimpleBorder width: 0 color: aColorOrSymbolOrNil)]
				ifFalse: [style baseColor: aColorOrSymbolOrNil.
					self changed].
			^ self].

	self
		borderStyle: ( ({ nil. #none } includes: aColorOrSymbolOrNil)
				ifTrue: [BorderStyle default]
				ifFalse: [ "a symbol"
					self doesBevels ifFalse: [ ^self ].
					newStyle := (BorderStyle perform: aColorOrSymbolOrNil)
								color: style color;
								width: style width;
								yourself.
					(self canDrawBorder: newStyle)
						ifTrue: [newStyle]
						ifFalse: [style]])! !

!Morph methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:09'!
borderStyle

	extension ifNil: [^BorderStyle default trackColorFrom: self].
	^(extension borderStyle ifNil: [BorderStyle default]) trackColorFrom: self! !

!Morph methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:09'!
borderStyle: newStyle
	newStyle = self borderStyle ifFalse:[
		(self canDrawBorder: newStyle) ifFalse:[
			"Replace the suggested border with a simple one"
			^self borderStyle: (BorderStyle width: newStyle width color: (newStyle trackColorFrom: self) color)].
		self assureExtension.
		self extension borderStyle: newStyle.
		self changed].! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/26/2001 16:18'!
borderStyleForSymbol: aStyleSymbol
	"Answer a suitable BorderStyle for me of the type represented by a given symbol"

	| aStyle existing |
	aStyle := BorderStyle borderStyleForSymbol: aStyleSymbol asSymbol.
	aStyle ifNil: [self error: 'bad style'].
	existing := self borderStyle.
	aStyle width: existing width;
		baseColor: existing baseColor.
	^ (self canDrawBorder: aStyle)
		ifTrue:
			[aStyle]
		ifFalse:
			[nil]! !

!Morph methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:28'!
borderWidth
	^self borderStyle width! !

!Morph methodsFor: 'accessing' stamp: 'nk 4/14/2004 17:48'!
borderWidth: aNumber
	| style |
	style := self borderStyle.
	style width = aNumber ifTrue: [ ^self ].

	style style = #none
		ifTrue: [ self borderStyle: (SimpleBorder width: aNumber color: Color transparent) ]
		ifFalse: [ style width: aNumber. self changed ].
! !

!Morph methodsFor: 'accessing' stamp: 'di 2/6/2001 14:02'!
borderWidthForRounding

	^ self borderWidth! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:17'!
bottom
	" Return the y-coordinate of my bottom side "

	^ bounds bottom! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:14'!
bottom: aNumber
	" Move me so that my bottom is at the y-coordinate aNumber. My extent (width & height) are unchanged "

	self position: (bounds left @ (aNumber - self height))! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
bottomCenter

	^ bounds bottomCenter! !

!Morph methodsFor: 'geometry' stamp: 'tk 9/8/97 10:44'!
bottomLeft

	^ bounds bottomLeft! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:08'!
bottomLeft: aPoint
	" Move me so that my bottom left corner is at aPoint. My extent (width & height) are unchanged "

	self position: ((aPoint x) @ (aPoint y - self height)).
! !

!Morph methodsFor: 'geometry' stamp: 'di 6/12/97 11:17'!
bottomRight

	^ bounds bottomRight! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:09'!
bottomRight: aPoint
	" Move me so that my bottom right corner is at aPoint. My extent (width & height) are unchanged "

	self position: ((aPoint x - bounds width) @ (aPoint y - self height))
! !

!Morph methodsFor: 'drawing' stamp: 'sw 6/4/2000 22:02'!
boundingBoxOfSubmorphs
	| aBox |
	aBox := bounds origin extent: self minimumExtent.  "so won't end up with something empty"
	submorphs do:
		[:m | m visible ifTrue: [aBox := aBox quickMerge: m fullBounds]].
	^ aBox
! !

!Morph methodsFor: 'geometry' stamp: 'jm 8/3/97 15:50'!
bounds
	"Return the bounds of this morph."
	"Note: It is best not to override this method because many methods in Morph and its subclasses use the instance variable directly rather than 'self bounds'. Instead, subclasses should be sure that the bounds instance variable is correct."

	^ bounds
! !

!Morph methodsFor: 'geometry' stamp: 'ar 12/14/2000 13:48'!
bounds: newBounds
	| oldExtent newExtent |
	oldExtent := self extent.
	newExtent := newBounds extent.
	(oldExtent dotProduct: oldExtent) <= (newExtent dotProduct: newExtent) ifTrue:[
		"We're growing. First move then resize."
		self position: newBounds topLeft; extent: newExtent.
	] ifFalse:[
		"We're shrinking. First resize then move."
		self extent: newExtent; position: newBounds topLeft.
	].! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:04'!
bounds: aRectangle from: referenceMorph
	"Return the receiver's bounds as seen by aMorphs coordinate frame"
	owner ifNil: [^ aRectangle].
	^(owner transformFrom: referenceMorph) globalBoundsToLocal: aRectangle
! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:04'!
bounds: aRectangle in: referenceMorph
	"Return the receiver's bounds as seen by aMorphs coordinate frame"
	owner ifNil: [^ aRectangle].
	^(owner transformFrom: referenceMorph) localBoundsToGlobal: aRectangle
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'RAA 7/21/2000 11:10'!
boundsForBalloon

	"some morphs have bounds that are way too big"
	^self boundsInWorld! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:04'!
boundsIn: referenceMorph
	"Return the receiver's bounds as seen by aMorphs coordinate frame"
	^self bounds: self bounds in: referenceMorph! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:05'!
boundsInWorld
	^self bounds: self bounds in: self world! !

!Morph methodsFor: 'updating' stamp: 'StephaneDucasse 8/8/2013 21:03'!
breakDependents

	self flag: #GSoC. "Flag added by Benjamin Van Ryseghem the July 22, 2013 to remember to remove this line in a while"
	[ super breakDependents ]
		on: Error
		do: [].
	
	self removeProperty: #announcer! !

!Morph methodsFor: 'debug and other' stamp: 'BenjaminVanRyseghem 9/29/2013 14:54'!
buildDebugMenu: aHand
	"Answer a debugging menu for the receiver. 
	 The hand argument is seemingly historical and plays no role presently"

	| aMenu |
	aMenu := UIManager default newMenuIn: self for: self.
	aMenu addStayUpItem.
	(self hasProperty: #errorOnDraw) ifTrue:
		[aMenu add: 'start drawing again' translated action: #resumeAfterDrawError.].
	(self hasProperty: #drawError) ifTrue:
		[aMenu add: 'debug drawing error' translated action: #debugDrawError.
		aMenu addLine].
	(self hasProperty: #errorOnStep) ifTrue:
		[aMenu add: 'start stepping again' translated action: #resumeAfterStepError.
		aMenu addLine].

	aMenu add: 'inspect morph' translated action: #inspectInMorphic:.
	aMenu lastItem icon: Smalltalk ui icons smallInspectItIcon.
	aMenu add: 'inspect owner chain' translated action: #inspectOwnerChain.
	aMenu lastItem icon: Smalltalk ui icons smallInspectItIcon.
	self isMorphicModel ifTrue: [
		aMenu add: 'inspect model' translated target: self model action: #inspect.
		aMenu lastItem icon: Smalltalk ui icons smallInspectItIcon ].
     aMenu add: 'explore morph' translated target: self selector: #explore.
	aMenu lastItem icon: Smalltalk ui icons smallInspectItIcon.
	aMenu addLine.
	aMenu add: 'browse morph class' translated target: self selector: #browseHierarchy.
	(self isMorphicModel)
		ifTrue: [aMenu
				add: 'browse model class'
				target: self model
				selector: #browseHierarchy].
	^ aMenu! !

!Morph methodsFor: 'meta-actions' stamp: 'GuillermoPolito 5/29/2011 14:50'!
buildHandleMenu: aHand
	"Build the morph menu for the given morph's halo's menu handle. This menu has two sections. The first section contains commands that are interpreted by the hand; the second contains commands provided by the target morph. This method allows the morph to decide which items should be included in the hand's section of the menu."

	| menu |
	menu := UIManager default newMenuIn: self for: self.
	menu addStayUpItem.
	menu addLine.
	self addStandardHaloMenuItemsTo: menu hand: aHand.
	menu defaultTarget: aHand.
	self addAddHandMenuItemsForHalo: menu  hand: aHand.
	menu defaultTarget: self.
	self addCustomHaloMenuItems: menu hand: aHand.
	menu defaultTarget: aHand.
	^ menu
! !

!Morph methodsFor: 'meta-actions' stamp: 'EstebanLorenzano 5/14/2013 09:44'!
buildMetaMenu: evt
	"Build the morph menu. This menu has two sections. The first section contains commands that are handled by the hand; the second contains commands handled by the argument morph."
	| menu |
	menu := UIManager default newMenuIn: self for: self.
	menu addStayUpItem.
	menu add: 'grab' translated action: #grabMorph:.
	menu add: 'copy to paste buffer' translated action: #copyToPasteBuffer:.
	self maybeAddCollapseItemTo: menu.
	menu add: 'delete' translated action: #dismissMorph:.
	menu addLine.
	menu add: 'copy text' translated action: #clipText.
	menu addLine.
	menu add: 'go behind' translated action: #goBehind.
	menu add: 'add halo' translated action: #addHalo:.
	menu add: 'duplicate' translated action: #maybeDuplicateMorph:.

	self addEmbeddingMenuItemsTo: menu hand: evt hand.

	menu add: 'resize' translated action: #resizeMorph:.
	"Give the argument control over what should be done about fill styles"
	self addFillStyleMenuItems: menu hand: evt hand.
	self addDropShadowMenuItems: menu hand: evt hand.
	self addLayoutMenuItems: menu hand: evt hand.
	menu addUpdating: #hasClipSubmorphsString target: self selector: #changeClipSubmorphs argumentList: #().
	menu addLine.
	menu add: 'inspect' translated selector: #inspectAt:event: argument: evt position.
	menu add: 'explore' translated action: #explore.
	menu lastItem icon: Smalltalk ui icons smallInspectItIcon.
	menu addLine..
	menu add: 'show actions' translated action: #showActions.
	menu addLine.
	self addDebuggingItemsTo: menu hand: evt hand.

	self addCustomMenuItems: menu hand: evt hand.
	^ menu
! !

!Morph methodsFor: 'menu' stamp: 'AlainPlantec 11/5/2011 14:27'!
buildYellowButtonMenu: aHand 
	"build the morph menu for the yellow button"
	| menu |
	menu := UIManager default newMenuIn: self for: self.
	self addNestedYellowButtonItemsTo: menu event: ActiveEvent.
	^ menu! !

!Morph methodsFor: 'testing' stamp: 'ar 8/25/2001 19:14'!
canDrawBorder: aBorderStyle
	"Return true if the receiver can be drawn with the given border style."
	^true! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:54'!
cellInset
	"Layout specific. This property specifies an extra inset for each cell in the layout."
	| props |
	props := self layoutProperties.
	^props ifNil:[0] ifNotNil:[props cellInset].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:54'!
cellInset: aNumber
	"Layout specific. This property specifies an extra inset for each cell in the layout."
	self assureTableProperties cellInset: aNumber.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:54'!
cellPositioning
	"Layout specific. This property describes how the receiver should be layed out in its owner when the bounds of the cell assigned to the receiver do not exactly match its bounds. Possible values are:
		#topLeft, #topRight, #bottomLeft, #bottomRight, #topCenter, #leftCenter, #rightCenter, #bottomCenter, #center 
	which align the receiver's bounds with the cell at the given point."
	| props |
	props := self layoutProperties.
	^props ifNil:[#center] ifNotNil:[props cellPositioning].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:39'!
cellPositioning: aSymbol
	"Layout specific. This property describes how the receiver should be layed out in its owner when the bounds of the cell assigned to the receiver do not exactly match its bounds. Possible values are:
		#topLeft, #topRight, #bottomLeft, #bottomRight, #topCenter, #leftCenter, #rightCenter, #bottomCenter, #center 
	which align the receiver's bounds with the cell at the given point."
	self assureTableProperties cellPositioning: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:48'!
cellPositioningString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self cellPositioning! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:55'!
cellSpacing
	"Layout specific. This property describes how the cell size for each element in a list should be computed.
		#globalRect - globally equal rectangular cells
		#globalSquare - globally equal square cells
		#localRect - locally (e.g., per row/column) equal rectangular cells
		#localSquare - locally (e.g., per row/column) equal square cells
		#none - cells are sized based on available row/column constraints
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#none] ifNotNil:[props cellSpacing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:56'!
cellSpacing: aSymbol
	"Layout specific. This property describes how the cell size for each element in a list should be computed.
		#globalRect - globally equal rectangular cells
		#globalSquare - globally equal square cells
		#localRect - locally (e.g., per row/column) equal rectangular cells
		#localSquare - locally (e.g., per row/column) equal square cells
		#none - cells are sized based on available row/column constraints
	"
	self assureTableProperties cellSpacing: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
cellSpacingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self cellSpacing! !

!Morph methodsFor: 'geometry' stamp: ''!
center

	^ bounds center! !

!Morph methodsFor: 'geometry' stamp: 'sw 6/11/1999 18:48'!
center: aPoint
	self position: (aPoint - (self extent // 2))! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:08'!
changeCellInset: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self cellInset: (newPoint - evt cursorPoint) asIntegerPoint // 5].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 18:54'!
changeClipLayoutCells
	self invalidRect: self fullBounds.
	self clipLayoutCells: self clipLayoutCells not.
	self invalidRect: self fullBounds.! !

!Morph methodsFor: 'drawing' stamp: 'ar 11/4/2000 23:39'!
changeClipSubmorphs
	self clipSubmorphs: self clipSubmorphs not.! !

!Morph methodsFor: 'menus' stamp: 'SeanDeNigris 1/23/2014 00:23'!
changeColor
	
	| dialog |
	dialog := ColorSelectorDialogWindow new
		title: 'Choose color';
		selectedColor: self color.
	self openModal: dialog.
	dialog cancelled
		ifFalse: [self fillStyle: dialog selectedColor]
	
	! !

!Morph methodsFor: 'meta-actions' stamp: 'HenrikSperreJohansen 5/21/2010 13:33'!
changeColorTarget: anObject selector: aSymbol originalColor: aColor hand: aHand
	"Put up a color picker for changing some kind of color."
	
	(UIManager default chooseColor: aColor)
		ifNotNil: [:nc | anObject perform: aSymbol with: nc].
! !

!Morph methodsFor: 'menus' stamp: 'ar 11/29/2001 19:57'!
changeDirectionHandles
	^self wantsDirectionHandles: self wantsDirectionHandles not! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:19'!
changeDisableTableLayout
	self disableTableLayout: self disableTableLayout not.
	self layoutChanged.! !

!Morph methodsFor: 'text-anchor' stamp: 'aoy 2/15/2003 21:47'!
changeDocumentAnchor
	"Change the anchor from/to document anchoring"

	| newType |
	newType := self textAnchorType == #document 
		ifTrue: [#paragraph]
		ifFalse: [ #document].
	owner isTextMorph 
		ifTrue: 
			[owner 
				anchorMorph: self
				at: self position
				type: newType]! !

!Morph methodsFor: 'menus' stamp: 'ar 11/2/2000 15:04'!
changeDragAndDrop
	^self enableDragNDrop: self dragNDropEnabled not! !

!Morph methodsFor: 'text-anchor' stamp: 'aoy 2/15/2003 21:48'!
changeInlineAnchor
	"Change the anchor from/to line anchoring"

	| newType |
	newType := self textAnchorType == #inline 
				ifTrue: [#paragraph]
				ifFalse: [#inline]. 
	owner isTextMorph 
		ifTrue: 
			[owner 
				anchorMorph: self
				at: self position
				type: newType]! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:09'!
changeLayoutInset: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self layoutInset: (newPoint - evt cursorPoint) asIntegerPoint // 5].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeListDirection: aSymbol
	| listDir wrapDir |
	self listDirection: aSymbol.
	(self wrapDirection == #none) ifTrue:[^self].
	"otherwise automatically keep a valid table layout"
	listDir := self listDirection.
	wrapDir := self wrapDirection.
	(listDir == #leftToRight or:[listDir == #rightToLeft]) ifTrue:[
		wrapDir == #leftToRight ifTrue:[^self wrapDirection: #topToBottom].
		wrapDir == #rightToLeft ifTrue:[^self wrapDirection: #bottomToTop].
	] ifFalse:[
		wrapDir == #topToBottom ifTrue:[^self wrapDirection: #leftToRight].
		wrapDir == #bottomToTop ifTrue:[^self wrapDirection: #rightToLeft].
	].
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeMaxCellSize: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self maxCellSize: (newPoint - evt cursorPoint) asIntegerPoint].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeMinCellSize: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self minCellSize: (newPoint - evt cursorPoint) asIntegerPoint].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:19'!
changeNoLayout
	self layoutPolicy ifNil:[^self]. "already no layout"
	self layoutPolicy: nil.
	self layoutChanged.! !

!Morph methodsFor: 'text-anchor' stamp: 'aoy 2/15/2003 21:48'!
changeParagraphAnchor
	"Change the anchor from/to paragraph anchoring"

	| newType |
	newType := self textAnchorType == #paragraph 
		ifTrue: [#document]
		ifFalse: [#paragraph].
	owner isTextMorph 
		ifTrue: 
			[owner 
				anchorMorph: self
				at: self position
				type: newType]! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:19'!
changeProportionalLayout
	| layout |
	((layout := self layoutPolicy) notNil and:[layout isProportionalLayout])
		ifTrue:[^self]. "already proportional layout"
	self layoutPolicy: ProportionalLayout new.
	self layoutChanged.! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeReverseCells
	self reverseTableCells: self reverseTableCells not.! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeRubberBandCells
	self rubberBandCells: self rubberBandCells not.! !

!Morph methodsFor: 'drop shadows' stamp: 'HenrikSperreJohansen 5/21/2010 13:34'!
changeShadowColor
	"Change the shadow color of the receiver -- triggered, e.g. from a menu"

	(UIManager default chooseColor: self shadowColor)
		ifNotNil: [:nc | self shadowColor: nc].
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:20'!
changeTableLayout
	| layout |
	((layout := self layoutPolicy) notNil and:[layout isTableLayout])
		ifTrue:[^self]. "already table layout"
	self layoutPolicy: TableLayout new.
	self layoutChanged.! !

!Morph methodsFor: 'updating' stamp: 'BenjaminVanRyseghem 7/22/2013 15:32'!
changed: anAspect

	self flag: #GSoC. "Flag added by Benjamin Van Ryseghem the July 22, 2013 to remember to remove this line in a while"
	[ super changed: anAspect ]
		on: Exception
		do: [].
	
	self announcer announce: (MorphChanged new
		morph: self;
		selector: anAspect)! !

!Morph methodsFor: 'updating' stamp: 'BenjaminVanRyseghem 7/23/2013 12:33'!
changed: anAspect with: anObject

	self flag: #GSoC. "Flag added by Benjamin Van Ryseghem the July 22, 2013 to remember to remove this line in a while"
	[ super changed: anAspect ]
		on: Exception
		do: [].
	
	self announcer announce: (MorphChangedWithArguments new
		morph: self;
		selector: anAspect;
		arguments: anObject)! !

!Morph methodsFor: 'Morphic-Base-Worlds' stamp: 'dgd 9/1/2004 16:10'!
clearArea
	"Answer the clear area of the receiver. It means the area free  
	of docking bars."
	| visTop visBottom visLeft visRight |

	visTop := self top.
	visBottom := self bottom.
	visLeft := self left.
	visRight := self right.

	self dockingBars
		do: [:each | 
			(each isAdheringToTop and: [each bottom > visTop])
				ifTrue: [visTop := each bottom].

			(each isAdheringToBottom and: [each top < visBottom])
				ifTrue: [visBottom := each top].

			(each isAdheringToLeft and: [each right > visLeft])
				ifTrue: [visLeft := each right].

			(each isAdheringToRight and: [each left < visRight])
				ifTrue: [visRight := each left]
		].

	^ Rectangle
		left: visLeft
		right: visRight
		top: visTop
		bottom: visBottom
! !

!Morph methodsFor: 'event handling' stamp: 'tk 9/6/2000 12:42'!
click
	"Pretend the user clicked on me."

	(self handlesMouseDown: nil) ifTrue: [
		self mouseDown: nil.
		self mouseUp: nil].! !

!Morph methodsFor: 'event handling' stamp: 'CamilloBruni 8/1/2012 16:02'!
click: evt
	"Handle a single-click event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing."

	^ self eventHandler ifNotNil:
		[self eventHandler click: evt fromMorph: self].! !

!Morph methodsFor: 'drawing' stamp: 'dgd 2/16/2003 20:02'!
clipLayoutCells
	"Drawing/layout specific. If this property is set, clip the  
	submorphs of the receiver by its cell bounds."
	^ self
		valueOfProperty: #clipLayoutCells
		ifAbsent: [false]! !

!Morph methodsFor: 'drawing' stamp: 'ar 10/29/2000 19:22'!
clipLayoutCells: aBool
	"Drawing/layout specific. If this property is set, clip the submorphs of the receiver by its cell bounds."
	aBool == false
		ifTrue:[self removeProperty: #clipLayoutCells]
		ifFalse:[self setProperty: #clipLayoutCells toValue: aBool].
	self changed.! !

!Morph methodsFor: 'drawing' stamp: 'gvc 9/11/2009 17:59'!
clipSubmorphs
	"Drawing specific. If this property is set, clip the receiver's  
	submorphs to the receiver's clipping bounds."
	
	extension ifNil: [^false].
	^extension clipSubmorphs ifNil: [false]! !

!Morph methodsFor: 'drawing' stamp: 'gvc 9/11/2009 17:59'!
clipSubmorphs: aBool
	"Drawing specific. If this property is set, clip the receiver's submorphs to the receiver's clipping bounds."
	self invalidRect: self fullBounds.
	
	aBool == self clipSubmorphs ifFalse:[
		self assureExtension.
		extension clipSubmorphs: aBool.
		self invalidRect: self fullBounds]! !

!Morph methodsFor: 'printing' stamp: 'bf 7/17/2003 12:53'!
clipText
	"Copy the text in the receiver or in its submorphs to the clipboard"
	| content |
	"My own text"
	content := self userString.
	"Or in my submorphs"
	content ifNil: [
		| list |
		list := self allStringsAfter: nil.
		list notEmpty ifTrue: [
			content := String streamContents: [:stream |
				list do: [:each | stream nextPutAll: each; cr]]]].
	"Did we find something?"
	content
		ifNil: [self flash "provide feedback"]
		ifNotNil: [Clipboard clipboardText: content].! !

!Morph methodsFor: 'drawing' stamp: 'ar 10/29/2000 19:16'!
clippingBounds
	"Return the bounds to which any submorphs should be clipped if the property is set"
	^self innerBounds! !

!Morph methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:12'!
cmdGesturesEnabled
	^ self class cmdGesturesEnabled! !

!Morph methodsFor: 'menus' stamp: 'sw 2/21/2000 15:21'!
collapse
	CollapsedMorph new beReplacementFor: self! !

!Morph methodsFor: 'accessing' stamp: 'tk 2/15/2001 15:55'!
color

	^ color 	"has already been set to ((self valueOfProperty: #fillStyle) asColor)"! !

!Morph methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:52'!
color: aColor
	"Set the receiver's color.  Directly set the color if appropriate, else go by way of fillStyle"

	(aColor isColor or: [aColor isKindOf: InfiniteForm]) ifFalse:[^ self fillStyle: aColor].
	color = aColor ifFalse:
		[self assureExtension.
		extension fillStyle: nil.
		color := aColor.
		self changed]! !

!Morph methodsFor: 'change reporting' stamp: 'sw 9/10/1998 08:18'!
colorChangedForSubmorph: aSubmorph
	"The color associated with aSubmorph was changed through the UI; react if needed"! !

!Morph methodsFor: 'accessing' stamp: 'ar 8/15/2001 22:40'!
colorForInsets
	"Return the color to be used for shading inset borders.  The default is my own color, but it might want to be, eg, my owner's color.  Whoever's color ends up prevailing, the color itself gets the last chance to determine, so that when, for example, an InfiniteForm serves as the color, callers won't choke on some non-Color object being returned"
	(color isColor and:[color isTransparent and:[owner notNil]]) ifTrue:[^owner colorForInsets].
	^ color colorForInsets
! !

!Morph methodsFor: 'printing' stamp: 'StephaneDucasse 8/2/2011 22:46'!
colorString: aColor

	aColor ifNil: [ ^'nil' ].
	^aColor name
		ifNil: [ aColor storeString ]
		ifNotNil: [ :colorName | 'Color ', colorName ]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'dgd 2/22/2003 14:30'!
comeToFront
	| outerMorph |
	outerMorph := self topRendererOrSelf.
	(outerMorph owner isNil or: [outerMorph owner hasSubmorphs not]) 
		ifTrue: [^self].
	outerMorph owner firstSubmorph == outerMorph 
		ifFalse: [outerMorph owner addMorphFront: outerMorph]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 3/1/2000 11:39'!
comeToFrontAndAddHalo
	self comeToFront.
	self addHalo! !

!Morph methodsFor: 'layout' stamp: 'StephaneDucasse 5/31/2013 19:18'!
computeBounds
	[ self doLayoutIn: self layoutBounds ]
		on: Error
		do: [ :ex | 
			"This should do it unless you don't screw up the bounds"
			fullBounds := bounds.
			ex pass ]! !

!Morph methodsFor: 'layout' stamp: 'StephaneDucasse 6/2/2013 14:18'!
computeFullBounds
	[ self doLayoutIn: self layoutBounds ]
		on: Error
		do: [ :ex | 
			"This should do it unless you don't screw up the bounds"
			fullBounds := bounds.
			ex pass ]! !

!Morph methodsFor: 'structure' stamp: 'alain.plantec 6/19/2008 09:34'!
containingWindow
	"Answer a window that contains the receiver"

	^ self ownerThatIsA: SystemWindow! !

!Morph methodsFor: 'geometry testing' stamp: ''!
containsPoint: aPoint

	^ self bounds containsPoint: aPoint! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/13/2000 17:58'!
containsPoint: aPoint event: anEvent
	"Return true if aPoint is considered to be inside the receiver for the given event.
	The default implementation treats locked children as integral part of their owners."
	(self fullBounds containsPoint: aPoint) ifFalse:[^false].
	(self containsPoint: aPoint) ifTrue:[^true].
	self submorphsDo:[:m|
		(m isLocked and:[m fullContainsPoint: 
			((m transformedFrom: self) globalPointToLocal: aPoint)]) ifTrue:[^true]].
	^false! !

!Morph methodsFor: 'copying' stamp: 'tk 2/19/2001 18:21'!
copy

	^ self veryDeepCopy! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 11/7/2009 18:52'!
copyToPasteBuffer: evt
	^evt hand copyToPasteBuffer: self.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'pmm 3/13/2010 11:32'!
copyWithoutSubmorph: sub
	"Needed to get a morph to draw without one of its submorphs.
	NOTE:  This must be thrown away immediately after use."
	^ self shallowCopy privateSubmorphs: (submorphs copyWithout: sub)! !

!Morph methodsFor: 'visual properties' stamp: 'gvc 9/11/2009 17:27'!
cornerStyle
	"Returns one of the following symbols:
		#square
		#rounded
	according to the current corner style."

	self assureExtension.
	^extension cornerStyle ifNil: [#square]! !

!Morph methodsFor: 'rounding' stamp: 'gvc 9/11/2009 17:28'!
cornerStyle: aSymbol
	"This method makes it possible to set up desired corner style. aSymbol has to be one of:
		#square
		#rounded"

	aSymbol == self cornerStyle ifFalse:[
		self assureExtension.
		extension cornerStyle: aSymbol.
		self changed]! !

!Morph methodsFor: 'accessing' stamp: 'ar 12/27/2001 17:56'!
couldHaveRoundedCorners
	^ true! !

!Morph methodsFor: 'event handling' stamp: 'sw 3/8/1999 00:17'!
cursorPoint
	^ self currentHand lastEvent cursorPoint! !

!Morph methodsFor: 'debug and other' stamp: 'CamilloBruni 5/29/2012 17:17'!
debugDrawError
	
	(self valueOfProperty: #drawError) debug.! !

!Morph methodsFor: 'copying' stamp: 'tk 2/14/2001 12:47'!
deepCopy

	self error: 'Please use veryDeepCopy'.
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 11/11/2000 16:15'!
defaultBalloonColor
	^ Display depth <= 2
		ifTrue: [Color white]
		ifFalse: [BalloonMorph balloonColor]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sd 12/5/2001 20:23'!
defaultBalloonFont
	^ BalloonMorph balloonFont! !

!Morph methodsFor: 'visual properties' stamp: 'nk 8/28/2003 15:56'!
defaultBitmapFillForm
	^ImageMorph defaultForm.
! !

!Morph methodsFor: 'initialize' stamp: 'dgd 3/7/2003 15:06'!
defaultBounds
"answer the default bounds for the receiver"
	^ 0 @ 0 corner: 50 @ 40! !

!Morph methodsFor: 'initialize' stamp: 'dgd 2/14/2003 22:28'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color blue! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/13/2000 14:51'!
defaultEventDispatcher
	"Return the default event dispatcher to use with events that are directly sent to the receiver"
	^MorphicEventDispatcher new! !

!Morph methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 13:03'!
defaultYellowButtonMenuEnabled
	^ self class defaultYellowButtonMenuEnabled! !

!Morph methodsFor: 'deferred message' stamp: 'AlainPlantec 7/9/2013 11:26'!
defer: aValuable
	self owner ifNotNil: [:o | o defer: aValuable] ifNil: [ UIManager default defer: aValuable ]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 1/11/2000 18:24'!
defersHaloOnClickTo: aSubMorph
	"If a cmd-click on aSubMorph would make it a preferred recipient of the halo, answer true"
	"May want to add a way (via a property) for morphs to assert true here -- this would let certain kinds of morphs that are unusually reluctant to take the halo on initial click"

	^ false
	! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'AlainPlantec 5/7/2010 23:44'!
degreesOfFlex
	"Return any rotation due to flexing"
	"NOTE: because renderedMorph, which is used by the halo to set heading, goes down through dropShadows as well as transformations, we need this method (and its other implems) to come back up through such a chain."
	^ 0.0! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'GuillermoPolito 5/3/2012 21:05'!
delete
	"Remove the receiver as a submorph of its owner and make its 
	new owner be nil."
	
	self removeHalo.
	self activeHand 
		releaseKeyboardFocus: self;
		releaseMouseFocus: self.
	owner ifNotNil:[ 
		self privateDelete.
		self announceDeleted.
	].! !

!Morph methodsFor: 'debug and other' stamp: 'ClementBera 10/25/2013 13:58'!
deleteAnyMouseActionIndicators

	self changed.
	self removeProperty: #mouseActionIndicatorMorphs.
	self hasRolloverBorder: false.
	self removeProperty: #rolloverWidth.
	self removeProperty: #rolloverColor.
	self layoutChanged.
	self changed.

! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 10/3/2000 17:03'!
deleteBalloon
	"If I am showing a balloon, delete it."
	| w |
	w := self world ifNil:[^self].
	w deleteBalloonTarget: self.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'dgd 9/1/2004 16:26'!
deleteDockingBars
	"Delete the receiver's docking bars"
	self dockingBars
		do: [:each | each delete]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:41'!
disableDragNDrop
	self enableDragNDrop: false! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:56'!
disableTableLayout
	"Layout specific. Disable laying out the receiver in table layout"
	| props |
	props := self layoutProperties.
	^props ifNil:[false] ifNotNil:[props disableTableLayout].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:06'!
disableTableLayout: aBool
	"Layout specific. Disable laying out the receiver in table layout"
	self assureLayoutProperties disableTableLayout: aBool.
	self layoutChanged.! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 7/9/2012 15:31'!
dismissMorph
	"This is called from an explicit halo destroy/delete action."

	| w |
	w := self world ifNil:[^self].
	w deleteAllHalos; stopStepping: self.
	self delete! !

!Morph methodsFor: 'meta-actions' stamp: 'mir 3/17/2006 18:01'!
dismissMorph: evt
	self dismissMorph! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'AlainPlantec 12/1/2010 09:39'!
dismissViaHalo
	"The user has clicked in the delete halo-handle.  This provides a hook in case some concomitant action should be taken, or if the particular morph is not one which should be put in the trash can, for example."

	self setProperty: #lastPosition toValue: self positionInWorld.
	self dismissMorph! !

!Morph methodsFor: 'announcements' stamp: 'ThierryGoubier 1/8/2014 14:40'!
doAnnounce: anAnnouncement
	"Take care of not creating the announcer when announcing. If the announcer doesn't exist then this means nobody has expressed an interest in the message."

	"Do not override announce: for now, there is a need to refactor the announcements code in at least SystemWindow and ExpanderMorph."

	(self valueOfProperty: #announcer ifAbsent: [ ^ self ]) announce: anAnnouncement! !

!Morph methodsFor: 'button' stamp: 'sw 2/6/2001 23:09'!
doButtonAction
	"If the receiver has a button-action defined, do it now.  The default button action of any morph is, well, to do nothing.  Note that there are several ways -- too many ways -- for morphs to have button-like actions.  This one refers not to the #mouseUpCodeToRun feature, nor does it refer to the Player-scripting mechanism.  Instead it is intended for morph classes whose very nature is to be buttons -- this method provides glue so that arbitrary buttons on the UI can be 'fired' programatticaly from user scripts"! !

!Morph methodsFor: 'user interface' stamp: 'tak 3/15/2005 17:10'!
doCancel
	self delete! !

!Morph methodsFor: 'user interface' stamp: 'StephaneDucasse 5/13/2012 20:49'!
doFastReframe: ptName

	| newBounds |
	"For fast display, only higlight the rectangle during loop"
	newBounds := self boundsInWorld newRectButtonPressedDo: [:f :pt | 
		f 
			withSideOrCorner: ptName
			setToPoint: pt
			minExtent: self minimumExtent].
	Display deferUpdatesIn: Display boundingBox while: [
		self bounds: newBounds].

	^newBounds.! !

!Morph methodsFor: 'layout' stamp: 'MarcusDenker 10/3/2013 23:34'!
doLayoutIn: layoutBounds 
	"Compute a new layout based on the given layout bounds."

	"Note: Testing for #bounds or #layoutBounds would be sufficient to
	figure out if we need an invalidation afterwards but #outerBounds
	is what we need for all leaf nodes so we use that."

	| box priorBounds |
	priorBounds := self outerBounds.
	submorphs isEmpty ifTrue: [^fullBounds := priorBounds].
	"Send #ownerChanged to our children"
	submorphs do: [:m | m ownerChanged].
	self layoutPolicy ifNotNil: [:layout | layout layout: self in: layoutBounds].
	self adjustLayoutBounds.
	fullBounds := self privateFullBounds.
	box := self outerBounds.
	box = priorBounds 
		ifFalse: [self invalidRect: (priorBounds quickMerge: box)]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 8/31/2004 16:53'!
dockingBars
	"Answer the receiver's dockingBars"
	^ self submorphs
		select: [:each | each isDockingBar]
! !

!Morph methodsFor: 'accessing' stamp: 'nk 4/15/2004 07:50'!
doesBevels
	"To return true means that this object can show bevelled borders, and
	therefore can accept, eg, #raised or #inset as valid borderColors.
	Must be overridden by subclasses that do not support bevelled borders."

	^ false! !

!Morph methodsFor: 'drawing' stamp: 'tk 8/2/1998 14:33'!
doesOwnRotation
	"Some morphs don't want to TransformMorph to rotate their images, but we do"
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'CamilloBruni 8/1/2012 16:04'!
doubleClick: evt
	"Handle a double-click event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing."

	^ self eventHandler ifNotNil:
		[self eventHandler doubleClick: evt fromMorph: self].! !

!Morph methodsFor: 'event handling' stamp: 'BenjaminVanRyseghem 1/18/2012 18:46'!
doubleClickTimeout: evt
	"Handle a double-click timeout event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing."

	^ self eventHandler ifNotNil:
		[self eventHandler doubleClickTimeout: evt fromMorph: self].! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:50'!
dragEnabled
	"Get this morph's ability to add and remove morphs via drag-n-drop."
	^(self valueOfProperty: #dragEnabled) == true
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:18'!
dragEnabled: aBool
	^self enableDrag: aBool! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:20'!
dragNDropEnabled
	"Note: This method is only useful for dragEnabled == dropEnabled at all times"
	self separateDragAndDrop.
	^self dragEnabled and:[self dropEnabled]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 18:36'!
dragSelectionColor
	^ Color magenta! !

!Morph methodsFor: 'drawing' stamp: 'panda 4/28/2000 11:59'!
drawDropHighlightOn: aCanvas
	self highlightedForDrop ifTrue: [
		aCanvas frameRectangle: self fullBounds color: self dropHighlightColor].! !

!Morph methodsFor: 'drawing' stamp: 'GaryChambers 9/8/2011 14:55'!
drawDropShadowOn: aCanvas

	aCanvas 
		translateBy: self shadowOffset 
		during: [ :shadowCanvas |
			shadowCanvas roundShadowCornersOf: self during: [
				(shadowCanvas isVisible: self bounds) ifTrue: [
					shadowCanvas fillRectangle: self bounds fillStyle: self shadowColor
				]
			]
		].
! !

!Morph methodsFor: 'drawing' stamp: 'CamilloBruni 5/29/2012 17:13'!
drawErrorOn: aCanvas
	"The morph (or one of its submorphs) had an error in its drawing method."
	aCanvas
		frameAndFillRectangle: bounds
		fillColor: Color red
		borderWidth: 1
		borderColor: Color yellow.
	aCanvas line: bounds topLeft to: bounds bottomRight width: 1 color: Color yellow.
	aCanvas line: bounds topRight to: bounds bottomLeft width: 1 color: Color yellow.
	
	self valueOfProperty: #drawError ifPresentDo: [ :error| | trace stringBounds |
		trace := String streamContents: [ :s| error signalerContext shortDebugStackOn: s].
		stringBounds := bounds insetBy: 5.
		trace linesDo: [ :aString|
			aCanvas drawString: aString in: stringBounds.
			stringBounds := stringBounds top: stringBounds top + (TextStyle defaultFont pixelSize * 1.2) ]]! !

!Morph methodsFor: 'drawing' stamp: '   9/3/2000 13:55'!
drawMouseDownHighlightOn: aCanvas
	self highlightedForMouseDown ifTrue: [
		aCanvas frameRectangle: self fullBounds color: self color darker darker].! !

!Morph methodsFor: 'drawing' stamp: 'IgorStasenko 7/18/2011 17:48'!
drawOn: aCanvas

	aCanvas fillRectangle: self bounds fillStyle: self fillStyle borderStyle: self borderStyle! !

!Morph methodsFor: 'drawing' stamp: 'IgorStasenko 7/18/2011 18:44'!
drawRolloverBorderOn: aCanvas 
	| colorToUse offsetToUse myShadow newForm f |
	colorToUse := self
				valueOfProperty: #rolloverColor
				ifAbsent: [Color blue alpha: 0.5].
	offsetToUse := self
				valueOfProperty: #rolloverWidth
				ifAbsent: [10 @ 10].
	self hasRolloverBorder: false.
	myShadow := self shadowForm.
	self hasRolloverBorder: true.
	myShadow offset: 0 @ 0.
	f := ColorForm extent: myShadow extent depth: 1.
	myShadow displayOn: f.
	f colors: {Color transparent. colorToUse}.
	newForm := Form extent: offsetToUse * 2 + myShadow extent depth: 32.
	(WarpBlt current toForm: newForm) sourceForm: f;
		 cellSize: 1;
		 combinationRule: 3;
		 copyQuad: f boundingBox innerCorners toRect: newForm boundingBox.
	aCanvas
		translateBy: offsetToUse negated
		during: [:shadowCanvas | 
			(shadowCanvas asShadowDrawingCanvas: colorToUse)	
				paintImage: newForm at: self position]
			! !

!Morph methodsFor: 'drawing' stamp: 'IgorStasenko 12/22/2012 03:14'!
drawSubmorphsOn: aCanvas 
	"Display submorphs back to front"

	| drawBlock |
	submorphs isEmpty ifTrue: [^self].
	drawBlock := [:canvas | submorphs reverseDo: [:m | canvas fullDrawMorph: m]].
	self clipSubmorphs 
		ifTrue: [aCanvas clipBy: (aCanvas clipRect intersect: self clippingBounds ifNone: [ ^ self ]) during: drawBlock]
		ifFalse: [drawBlock value: aCanvas]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:51'!
dropEnabled
	"Get this morph's ability to add and remove morphs via drag-n-drop."
	^(self valueOfProperty: #dropEnabled) == true
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:18'!
dropEnabled: aBool
	^self enableDrop: aBool! !

!Morph methodsFor: 'event handling' stamp: 'ar 1/10/2001 21:28'!
dropFiles: anEvent
	"Handle a number of files dropped from the OS"
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 10:52'!
dropHighlightColor
	^ Color blue! !

!Morph methodsFor: 'copying' stamp: 'MarcusDenker 9/13/2013 16:20'!
duplicate
	"Make and return a duplicate of the receiver"

	| newMorph topRend |
	((topRend := self topRendererOrSelf) ~~ self) ifTrue: [^ topRend duplicate].
	newMorph := self veryDeepCopy.
	newMorph arrangeToStartStepping.
		newMorph privateOwner: nil. "no longer in world"
	^newMorph! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 11/4/2000 17:56'!
duplicateMorph: evt
	"Make and return a duplicate of the receiver's argument"
	| dup |
	dup := self duplicate.
	evt hand grabMorph: dup from: owner. "duplicate was ownerless so use #grabMorph:from: here"
	^dup! !

!Morph methodsFor: 'halos and balloon help' stamp: 'StephaneDucasse 9/7/2011 21:23'!
editBalloonHelpContent: aString
	| reply |
	reply := UIManager default
		multiLineRequest: 'Edit the balloon help text for ' translated, self externalName
		initialAnswer: (aString ifNil: [self noHelpString] ifNotNil: [aString])
		answerHeight: 200.
	reply ifNil: [^ self].  "User cancelled out of the dialog"
	(reply isEmpty or: [reply asString = self noHelpString])
		ifTrue: [self setBalloonText: nil]
		ifFalse: [self setBalloonText: reply]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 12/23/1999 13:24'!
editBalloonHelpText
	"Modify the receiver's balloon help text."

	self editBalloonHelpContent: self balloonText! !

!Morph methodsFor: 'utilities' stamp: 'cb 6/25/2013 13:24'!
embedInWindow

	| window worldToUse |

	worldToUse := self world.		"I'm assuming we are already in a world"
	window := (SystemWindow labelled: self defaultLabel) model: nil.
	window bounds: ((self position - ((0@window labelHeight) + window borderWidth))
						corner: self bottomRight + window borderWidth).
	window addMorph: self frame: (0@0 extent: 1@1).
	window updatePaneColors.
	worldToUse addMorph: window.
	window activate! !

!Morph methodsFor: 'utilities' stamp: 'AlainPlantec 12/16/2009 12:27'!
embeddedInMorphicWindowLabeled: labelString
	| window |
	window := (SystemWindow labelled: labelString) model: nil.
	window setStripeColorsFrom:  (self theme windowColorFor: self).
	window addMorph: self frame: (0@0 extent: 1@1).
	^ window! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:50'!
enableDrag: aBoolean
	self setProperty: #dragEnabled toValue: aBoolean! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:41'!
enableDragNDrop
	self enableDragNDrop: true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:21'!
enableDragNDrop: aBoolean
	"Set both properties at once"
	self separateDragAndDrop.
	self enableDrag: aBoolean.
	self enableDrop: aBoolean.! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:51'!
enableDrop: aBoolean
	self setProperty: #dropEnabled toValue: aBoolean! !

!Morph methodsFor: 'accessing' stamp: 'gvc 1/11/2007 12:21'!
enabled
	"Answer whether the receiver is enabled."

	^true! !

!Morph methodsFor: 'accessing' stamp: 'StephaneDucasse 5/31/2013 17:45'!
enabled: aBoolean
	"does nothing"

	! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:53'!
eventHandler
	"answer the receiver's eventHandler"
	^ extension ifNotNil: [extension eventHandler] ! !

!Morph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 19:25'!
eventHandler: anEventHandler 
	"Note that morphs can share eventHandlers and all is OK. "
	self assureExtension eventHandler: anEventHandler! !

!Morph methodsFor: 'drawing' stamp: 'ar 11/8/2000 19:29'!
expandFullBoundsForRolloverBorder: aRectangle
	| delta |
	delta := self valueOfProperty: #rolloverWidth ifAbsent: [10@10].
	^aRectangle expandBy: delta.

! !

!Morph methodsFor: 'menus' stamp: 'DamienCassou 9/29/2009 13:01'!
exportAsBMP
	| fName |
	fName := UIManager default request:'Please enter the name' translated initialAnswer: self externalName,'.bmp'.
	fName isEmptyOrNil ifTrue:[^self].
	self imageForm writeBMPfileNamed: fName.! !

!Morph methodsFor: 'menus' stamp: 'DamienCassou 9/29/2009 13:01'!
exportAsGIF
	| fName |
	fName := UIManager default request:'Please enter the name' translated initialAnswer: self externalName,'.gif'.
	fName isEmptyOrNil ifTrue:[^self].
	GIFReadWriter putForm: self imageForm onFileNamed: fName.! !

!Morph methodsFor: 'menus' stamp: 'DamienCassou 9/29/2009 13:01'!
exportAsJPEG
	"Export the receiver's image as a JPEG"

	| fName |
	fName := UIManager default request: 'Please enter the name' translated initialAnswer: self externalName,'.jpeg'.
	fName isEmptyOrNil ifTrue: [^ self].
	self imageForm writeJPEGfileNamed: fName! !

!Morph methodsFor: 'menus' stamp: 'DamienCassou 9/29/2009 13:01'!
exportAsPNG
	| fName |
	fName := UIManager default request:'Please enter the name' translated initialAnswer: self externalName,'.png'.
	fName isEmptyOrNil ifTrue:[^self].
	PNGReadWriter putForm: self imageForm onFileNamed: fName.! !

!Morph methodsFor: 'accessing - extension' stamp: 'dgd 2/16/2003 19:22'!
extension
	"answer the recevier's extension"
	^ extension! !

!Morph methodsFor: 'geometry' stamp: ''!
extent

	^ bounds extent! !

!Morph methodsFor: 'geometry' stamp: 'MarianoMartinezPeck 8/24/2012 15:26'!
extent: aPoint
	|newExtent|	
	newExtent := aPoint rounded.
	(bounds extent closeTo: newExtent) ifTrue: [^ self].
	self changed.
	bounds := (bounds topLeft extent: newExtent).
	self layoutChanged.
	self changed.
! !

!Morph methodsFor: 'viewer' stamp: 'AlainPlantec 1/14/2010 09:41'!
externalName
	^ self assureExtension externalName ifNil: [super externalName]! !

!Morph methodsFor: 'visual properties' stamp: 'gvc 9/11/2009 16:35'!
fillStyle
	"Return the current fillStyle of the receiver."
	
	^extension
		ifNil: [^color]
		ifNotNil: [extension fillStyle ifNil: [color]]! !

!Morph methodsFor: 'visual properties' stamp: 'gvc 9/11/2009 16:39'!
fillStyle: aFillStyle
	"Set the current fillStyle of the receiver.
	Optimized for no change."
	
	self assureExtension.
	extension fillStyle = aFillStyle
		ifTrue: [^self]. "no change optimization"
	extension fillStyle: aFillStyle.
	color := aFillStyle asColor.
	self changed! !

!Morph methodsFor: 'visual properties' stamp: 'dgd 1/7/2005 19:31'!
fillWithRamp: rampSpecsOrColor oriented: aRatio 
	rampSpecsOrColor isColor
		ifTrue: [self color: rampSpecsOrColor".
			self borderColor: rampSpecsOrColor muchDarker"]
		ifFalse: [| fill | 
			fill := GradientFillStyle ramp: rampSpecsOrColor.
			fill origin: self bounds topLeft.
			fill direction: (self bounds extent * aRatio) truncated.
			fill radial: false.
			self fillStyle: fill.
			self borderColor: (rampSpecsOrColor first value mixed: 0.5 with: rampSpecsOrColor last value) muchDarker]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'rhi 9/10/2000 12:12'!
findA: aClass
	"Return the first submorph of the receiver that is descended from the given class. Return nil if there is no such submorph. Clients of this code should always check for a nil return value so that the code will be robust if the user takes the morph apart."

	^self submorphs
		detect: [:p | p isKindOf: aClass]
		ifNone: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'LC 9/28/1999 19:12'!
findDeepSubmorphThat: block1 ifAbsent: block2 
	self
		allMorphsDo: [:m | (block1 value: m)
				== true ifTrue: [^ m]].
	^ block2 value! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 1/9/2001 12:30'!
findDeeplyA: aClass
	"Return a morph in the submorph tree of the receiver that is descended from the given class. Return nil if there is no such morph. Clients of this code should always check for a nil return value so that the code will be robust if the user takes the morph apart."

	^ (self allMorphs copyWithout: self)
		detect: [:p | p isKindOf: aClass]
		ifNone: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'BenjaminVanRyseghem 3/19/2011 22:09'!
findSubmorphBinary: aBlock
	"Use binary search for finding a specific submorph of the receiver. Caller must be certain that the ordering holds for the submorphs."
	^submorphs findBinary: aBlock do: [ :found | found ] ifNone: [:a :b | ]! !

!Morph methodsFor: 'button' stamp: 'marcus.denker 8/24/2008 21:42'!
firedMouseUpCode
	"If the user has special mouseUpCodeToRun, then fire it once right now and return true, else return false"

	| evt |
	(self world isNil or: [self mouseUpCodeOrNil isNil]) ifTrue: [^false].
	evt := MouseEvent basicNew 
				setType: nil
				position: self center
				buttons: 0
				hand: self world activeHand.
	self programmedMouseUp: evt for: self.
	^true! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 00:48'!
firstOwnerSuchThat: conditionBlock

	self allOwnersDo: [:m | (conditionBlock value: m) ifTrue: [^ m]].
	^ nil
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:31'!
firstSubmorph
	^submorphs first! !

!Morph methodsFor: 'geometry' stamp: 'SeanDeNigris 5/4/2013 11:17'!
fitInWorld

	self bounds: (self bounds translatedAndSquishedToBeWithin: self world bounds).! !

!Morph methodsFor: 'macpal' stamp: 'GaryChambers 12/21/2011 11:10'!
flash
	| fs w |
	fs := self fillStyle.
	self fillStyle: Color black.
	(w := self world) ifNotNil: [w displayWorldSafely].
	self fillStyle: fs! !

!Morph methodsFor: 'drawing' stamp: 'sw 11/26/2003 17:43'!
flashBounds
	"Flash the receiver's bounds  -- does not use the receiver's color, thus works with StringMorphs and SketchMorphs, etc., for which #flash is useless.  No senders initially, but useful to send this from a debugger or inspector"

	5 timesRepeat:
		[Display flash: self boundsInWorld  andWait: 120]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 18:13'!
formerOwner
	^self valueOfProperty: #formerOwner! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'GuillermoPolito 9/1/2010 18:40'!
formerOwner: aMorphOrNil 
	aMorphOrNil ifNil: [self removeProperty: #formerOwner]
		ifNotNil: [self setProperty: #formerOwner toValue: aMorphOrNil]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 18:13'!
formerPosition
	^self valueOfProperty: #formerPosition! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'GuillermoPolito 9/1/2010 18:41'!
formerPosition: formerPosition 
	formerPosition ifNil: [self removeProperty: #formerPosition]
		ifNotNil: [self setProperty: #formerPosition toValue: formerPosition]! !

!Morph methodsFor: 'layout' stamp: 'StephaneDucasse 6/2/2013 14:18'!
fullBounds
	"Return the bounding box of the receiver and all its children. Recompute the layout if necessary."

	fullBounds ifNotNil: [ ^ fullBounds ].	"Errors at this point can be critical so make sure we catch 'em all right"
	self computeFullBounds.	"This should do it unless you don't screw up the bounds"
	^ fullBounds! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:06'!
fullBoundsInWorld
	^self bounds: self fullBounds in: self world! !

!Morph methodsFor: 'geometry testing' stamp: 'MarcusDenker 10/21/2013 14:22'!
fullContainsPoint: aPoint
	(self fullBounds containsPoint: aPoint)
		ifFalse: [ ^ false ].	"quick elimination"
	(self containsPoint: aPoint)
		ifTrue: [ ^ true ].	"quick acceptance"
	^ submorphs anySatisfy: [ :m | m fullContainsPoint: aPoint ]! !

!Morph methodsFor: 'drawing' stamp: 'CamilloBruni 5/29/2012 16:51'!
fullDrawOn: aCanvas
	"Draw the full Morphic structure on the given Canvas"

	self visible ifFalse: [^ self].
	(aCanvas isVisible: self fullBounds) ifFalse:[^self].
	
	(self hasProperty: #errorOnDraw) ifTrue:[^self drawErrorOn: aCanvas].
	
	[
		"Note: At some point we should generalize this into some sort of 
		multi-canvas so that we can cross-optimize some drawing operations."
		"Pass 1: Draw eventual drop-shadow"
		self hasDropShadow ifTrue: [ self drawDropShadowOn: aCanvas ].
		(self hasRolloverBorder and: [(aCanvas seesNothingOutside: self bounds) not])
			ifTrue: [self drawRolloverBorderOn: aCanvas].

		"Pass 2: Draw receiver itself"
		
		aCanvas roundCornersOf: self during:[
			(aCanvas isVisible: self bounds) ifTrue:[aCanvas drawMorph: self].
			self drawSubmorphsOn: aCanvas.
			self drawDropHighlightOn: aCanvas.
			self drawMouseDownHighlightOn: aCanvas]
		
	] on: Error do: [:err |
		self setProperty: #errorOnDraw toValue: true.
		self setProperty: #drawError toValue: err freeze.
		^ self drawErrorOn: aCanvas
	]! !

!Morph methodsFor: 'caching' stamp: 'jm 11/13/97 16:34'!
fullReleaseCachedState
	"Release the cached state of the receiver and its full submorph tree."

	self allMorphsDo: [:m | m releaseCachedState].
! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:06'!
globalPointToLocal: aPoint
	^self point: aPoint from: nil! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'michael.rueger 3/9/2009 18:48'!
goBehind

	owner addMorphBack: self.
! !

!Morph methodsFor: 'geometry' stamp: 'ClementBera 7/30/2013 11:03'!
goHome
	| box fb |
	owner ifNil: [^ self].
	self visible ifFalse: [^ self].

	box := owner visibleClearArea.
	fb := self fullBounds.

	fb left < box left
		ifTrue: [self left: box left - fb left + self left].
	fb right > box right
		ifTrue: [self right: box right - fb right + self right].

	fb top < box top
		ifTrue: [self top: box top - fb top + self top].
	fb bottom > box bottom
		ifTrue: [self bottom: box bottom - fb bottom + self bottom].
! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/6/2000 16:37'!
grabMorph: evt

	evt hand grabMorph: self! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/6/2000 15:13'!
grabTransform
	"Return the transform for the receiver which should be applied during grabbing"
	^owner ifNil:[IdentityTransform new] ifNotNil:[owner grabTransform]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:57'!
hResizing
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#rigid] ifNotNil:[props hResizing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:06'!
hResizing: aSymbol
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	self assureLayoutProperties hResizing: aSymbol.
	self layoutChanged.
! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/31/2000 20:45'!
hResizingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self hResizing! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 5/7/2010 21:44'!
halo
	^ self outermostWorldMorph 
		ifNotNil: [:w | w haloMorphs detect: [:h | h target == self] ifNone: []]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 9/15/2000 16:13'!
haloClass
	"Answer the name of the desired kind of HaloMorph to launch on behalf of the receiver"

	^ #HaloMorph
! !

!Morph methodsFor: 'accessing' stamp: 'usmanbhatti 3/6/2012 19:23'!
halosEnabled

	^ self class halosEnabled! !

!Morph methodsFor: 'events-processing' stamp: 'ar 1/10/2001 21:35'!
handleDropFiles: anEvent
	"Handle a drop from the OS."
	anEvent wasHandled ifTrue:[^self]. "not interested"
	(self wantsDropFiles: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	self dropFiles: anEvent.
! !

!Morph methodsFor: 'events-processing' stamp: 'AlainPlantec 12/1/2010 09:33'!
handleDropMorph: anEvent
	"Handle a dropping morph."
	| aMorph localPt |
	aMorph := anEvent contents.
	"Do a symmetric check if both morphs like each other"
	((self wantsDroppedMorph: aMorph event: anEvent)	"I want her"
		and: [aMorph wantsToBeDroppedInto: self])		"she wants me"
		ifFalse: [^ self].
	anEvent wasHandled: true.
	"Transform the morph into the receiver's coordinate frame. This is currently incomplete since it only takes the offset into account where it really should take the entire transform."
	localPt := (self transformedFrom: anEvent hand world) "full transform down"
				globalPointToLocal: aMorph referencePosition.
	aMorph referencePosition: localPt.
	self acceptDroppingMorph: aMorph event: anEvent.
	aMorph justDroppedInto: self event: anEvent.
! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/15/2000 21:13'!
handleEvent: anEvent
	"Handle the given event"
	^anEvent sentTo: self.! !

!Morph methodsFor: 'events-processing' stamp: 'ar 10/4/2000 18:48'!
handleFocusEvent: anEvent
	"Handle the given event. This message is sent if the receiver currently has the focus and is therefore receiving events directly from some hand."
	^self handleEvent: anEvent! !

!Morph methodsFor: 'events-processing' stamp: 'GuillermoPolito 4/22/2012 17:20'!
handleKeyDown: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self].
	(self handlesKeyDown: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	^self keyDown: anEvent! !

!Morph methodsFor: 'events-processing' stamp: 'GuillermoPolito 4/22/2012 17:19'!
handleKeyUp: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self].
	(self handlesKeyUp: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	^self keyUp: anEvent! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 11/7/2013 16:57'!
handleKeystroke: anEvent 
	"System level event handling."
	
	anEvent wasHandled
		ifTrue: [^ self].
	
	Smalltalk tools shortcuts handleKeystroke: anEvent inMorph: self.
	anEvent wasHandled ifTrue: [^ self].
	
	(self handlesKeyStroke: anEvent)
		ifFalse: [^ self].
	anEvent wasHandled: true.
	self keyStroke: anEvent.
	^ self eventHandler ifNotNil: [:handler | handler keyStroke: anEvent fromMorph: self ].! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/16/2000 14:22'!
handleListenEvent: anEvent
	"Handle the given event. This message is sent if the receiver is a registered listener for the given event."
	^anEvent sentTo: self.! !

!Morph methodsFor: 'events-processing' stamp: 'BenjaminVanRyseghem 6/28/2012 12:58'!
handleMouseDown: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self]. "not interested"
	anEvent hand removePendingBalloonFor: self.
	anEvent wasHandled: true.
	
	(anEvent controlKeyPressed
			and: [self cmdGesturesEnabled and: [ anEvent shiftPressed]])
		ifTrue: [
			self invokeMetaMenu: anEvent.
			^ self eventHandler ifNotNil: [:handler | handler mouseDown: anEvent fromMorph: self ].].

	"Make me modal during mouse transitions"
	anEvent hand newMouseFocus: self event: anEvent.
	anEvent blueButtonChanged ifTrue:[^self blueButtonDown: anEvent].
	
	self mouseDown: anEvent.
	
	anEvent hand removeHaloFromClick: anEvent on: self.
	(self handlesMouseStillDown: anEvent) ifTrue:[
		self startStepping: #handleMouseStillDown: 
			at: Time millisecondClockValue + self mouseStillDownThreshold
			arguments: {anEvent copy resetHandlerFields}
			stepTime: self mouseStillDownStepRate ].! !

!Morph methodsFor: 'events-processing' stamp: 'BenjaminVanRyseghem 1/18/2012 19:06'!
handleMouseEnter: anEvent
	"System level event handling."
	(anEvent isDraggingEvent) ifTrue:[
		(self handlesMouseOverDragging: anEvent) ifTrue:[
			anEvent wasHandled: true.
			self mouseEnterDragging: anEvent].
		^ self eventHandler ifNotNil: [:handler | handler mouseEnterDragging: anEvent fromMorph: self ]].
	self wantsBalloon
		ifTrue:[anEvent hand triggerBalloonFor: self after: self balloonHelpDelayTime].
	(self handlesMouseOver: anEvent) ifTrue:[
		anEvent wasHandled: true.
		self mouseEnter: anEvent ].
	^ self eventHandler ifNotNil: [:handler | handler mouseEnter: anEvent fromMorph: self ].! !

!Morph methodsFor: 'events-processing' stamp: 'BenjaminVanRyseghem 1/18/2012 18:58'!
handleMouseLeave: anEvent
	"System level event handling."
	anEvent hand removePendingBalloonFor: self.
	anEvent isDraggingEvent ifTrue:[
		(self handlesMouseOverDragging: anEvent) ifTrue:[
			anEvent wasHandled: true.
			self mouseLeaveDragging: anEvent].
		^ self eventHandler ifNotNil: [:handler | handler mouseLeave: anEvent fromMorph: self ]].
	(self handlesMouseOver: anEvent) ifTrue:[
		anEvent wasHandled: true.
		self mouseLeave: anEvent ].
	^ self eventHandler ifNotNil: [:handler | handler mouseLeave: anEvent fromMorph: self ]! !

!Morph methodsFor: 'events-processing' stamp: 'BenjaminVanRyseghem 1/18/2012 18:58'!
handleMouseMove: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self]. "not interested"
	"Rules say that by default a morph gets #mouseMove iff
		* the hand is not dragging anything,
			+ and some button is down,
			+ and the receiver is the current mouse focus."
	(anEvent hand hasSubmorphs) ifTrue:[^self].
	(anEvent anyButtonPressed and:[anEvent hand mouseFocus == self]) ifFalse:[^self].
	anEvent wasHandled: true.
	self mouseMove: anEvent.
	(self handlesMouseStillDown: anEvent) ifTrue:[
		"Step at the new location"
		self startStepping: #handleMouseStillDown: 
			at: Time millisecondClockValue
			arguments: {anEvent copy resetHandlerFields}
			stepTime: self mouseStillDownStepRate ].
	^ self eventHandler ifNotNil: [:handler | handler mouseMove: anEvent fromMorph: self ]
! !

!Morph methodsFor: 'events-processing' stamp: 'ar 4/23/2001 17:24'!
handleMouseOver: anEvent
	"System level event handling."
	anEvent hand mouseFocus == self ifTrue:[
		"Got this directly through #handleFocusEvent: so check explicitly"
		(self containsPoint: anEvent position event: anEvent) ifFalse:[^self]].
	anEvent hand noticeMouseOver: self event: anEvent! !

!Morph methodsFor: 'events-processing' stamp: 'BenjaminVanRyseghem 1/18/2012 18:59'!
handleMouseStillDown: anEvent
	"Called from the stepping mechanism for morphs wanting continuously repeated 'yes the mouse is still down, yes it is still down, yes it has not changed yet, no the mouse is still not up, yes the button is down' etc messages"
	(anEvent hand mouseFocus == self) 
		ifFalse:[^self stopSteppingSelector: #handleMouseStillDown:].
	self mouseStillDown: anEvent.
	^ self eventHandler ifNotNil: [:handler | handler mouseStillDown: anEvent fromMorph: self ]
! !

!Morph methodsFor: 'events-processing' stamp: 'SeanDeNigris 1/31/2014 18:38'!
handleMouseUp: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue: [^self]. "not interested"
	anEvent hand mouseFocus == self ifFalse: [^self]. "Not interested in other parties"
	anEvent hand releaseMouseFocus: self.
	anEvent wasHandled: true.
	^ anEvent blueButtonChanged
		ifTrue: [
			self blueButtonUp: anEvent.
			self eventHandler ifNotNil: [:handler | handler mouseUp: anEvent fromMorph: self]]
		ifFalse: [ | result |
			result := self mouseUp: anEvent.
			self stopSteppingSelector: #handleMouseStillDown:.
			result ].! !

!Morph methodsFor: 'events-processing' stamp: 'CamilleTeruel 12/6/2013 14:36'!
handleUnknownEvent: anEvent
	"An event of an unknown type was sent to the receiver. What shall we do?!!"
	self inform: 'Unknown event: ', anEvent printString. 
	anEvent printString displayAt: 0@0.
	anEvent wasHandled: true.! !

!Morph methodsFor: 'updating' stamp: 'IgorStasenko 8/27/2013 16:25'!
handleUpdate: aMorphChangedAnnouncement
	^ aMorphChangedAnnouncement deliverTo: self! !

!Morph methodsFor: 'event handling' stamp: 'tbn 3/12/2010 01:55'!
handleWindowEvent: anEvent
	"Handle an event concerning our host window"
	anEvent wasHandled ifTrue:[^self]. "not interested"
	(self wantsWindowEvent: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	self windowEvent: anEvent.
! !

!Morph methodsFor: 'meta-actions' stamp: 'SeanDeNigris 4/24/2012 19:49'!
handlerForBlueButtonDown: anEvent
	"Return the (prospective) handler for a mouse down event. The handler is temporarily installed and can be used for morphs further down the hierarchy to negotiate whether the inner or the outer morph should finally handle the event.
	Note: Halos handle blue button events themselves so we will only be asked if there is currently no halo on top of us.
	Check whtehr halods are enabled (for deployment)."
	
	self wantsHaloFromClick ifFalse:[^nil].
	
	self class cycleHalosBothDirections ifTrue: [
		anEvent handler ifNil:[^self]. 
		(anEvent handler isKindOf: PasteUpMorph) ifTrue:[^self] ].

	(anEvent shiftPressed)
		ifFalse:[^nil] "let outer guy have it"
		ifTrue:[^self] "let me have it"
! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/12/2000 17:07'!
handlerForMetaMenu: evt
	"Return the prospective handler for invoking the meta menu. By default, the top-most morph in the innermost world gets this menu"
	self isWorldMorph ifTrue:[^self].
	evt handler ifNotNil:[evt handler isWorldMorph ifTrue:[^self]].
	^nil! !

!Morph methodsFor: 'event handling' stamp: 'nk 3/10/2004 19:47'!
handlerForMouseDown: anEvent 
	"Return the (prospective) handler for a mouse down event. The handler is temporarily 
	installed and can be used for morphs further down the hierarchy to negotiate whether 
	the inner or the outer morph should finally handle the event."

	anEvent blueButtonPressed
		ifTrue: [^ self handlerForBlueButtonDown: anEvent].
	anEvent yellowButtonPressed
		ifTrue: [^ self handlerForYellowButtonDown: anEvent].
	anEvent controlKeyPressed
		ifTrue: [^ self handlerForMetaMenu: anEvent].
	(self handlesMouseDown: anEvent)
		ifFalse: [^ nil].	"not interested"

	anEvent handler
		ifNil: [^ self ].	"Same priority but I am innermost"

	"Nobody else was interested"
	^self mouseDownPriority >= anEvent handler mouseDownPriority
		ifTrue: [ self]
		ifFalse: [ nil]! !

!Morph methodsFor: 'event handling' stamp: 'nk 3/10/2004 19:48'!
handlerForYellowButtonDown: anEvent 
	"Return the (prospective) handler for a mouse down event with the yellow button pressed.
	The 	handler is temporarily installed and can be used for morphs further 
	down the hierarchy to negotiate whether the inner or the outer 
	morph should finally handle the event."

	(self hasYellowButtonMenu or: [ self handlesMouseDown: anEvent ])
		ifFalse: [ ^ nil].	"Not interested."

	anEvent handler
		ifNil: [^ self].	"Nobody else was interested"

	"Same priority but I am innermost."
	^ self mouseDownPriority >= anEvent handler mouseDownPriority
		ifFalse: [nil ]
		ifTrue: [self]! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 4/22/2012 17:19'!
handlesKeyDown: evt
	^self handlesKeyboard: evt! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 4/22/2012 17:20'!
handlesKeyStroke: evt
	^self handlesKeyboard: evt! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 4/22/2012 17:19'!
handlesKeyUp: evt
	^self handlesKeyboard: evt! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/28/2000 22:18'!
handlesKeyboard: evt
	"Return true if the receiver wishes to handle the given keyboard event"
	self eventHandler ifNotNil: [^ self eventHandler handlesKeyboard: evt].
	^ false
! !

!Morph methodsFor: 'event handling' stamp: 'nk 2/14/2004 18:42'!
handlesMouseDown: evt
	"Do I want to receive mouseDown events (mouseDown:, mouseMove:, mouseUp:)?"
	"NOTE: The default response is false, except if you have added sensitivity to mouseDown events using the on:send:to: mechanism.  Subclasses that implement these messages directly should override this one to return true." 

	self eventHandler ifNotNil: [^ self eventHandler handlesMouseDown: evt].
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/1998 07:31'!
handlesMouseOverDragging: evt
	"Return true if I want to receive mouseEnterDragging: and mouseLeaveDragging: when the hand drags something over me (button up or button down), or when the mouse button is down but there is no mouseDown recipient.  The default response is false, except if you have added sensitivity to mouseEnterLaden: or mouseLeaveLaden:, using the on:send:to: mechanism."
	"NOTE:  If the hand state matters in these cases, it may be tested by constructs such as
		event anyButtonPressed
		event hand hasSubmorphs"

	self eventHandler ifNotNil: [^ self eventHandler handlesMouseOverDragging: evt].
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/22/2000 17:06'!
handlesMouseStillDown: evt
	"Return true if the receiver wants to get repeated #mouseStillDown: messages between #mouseDown: and #mouseUp"
	self eventHandler ifNotNil: [^ self eventHandler handlesMouseStillDown: evt].
	^ false
! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/22/2012 16:49'!
hasClipLayoutCellsString
	^ (self clipLayoutCells) -> 'clip to cell size' translated! !

!Morph methodsFor: 'drawing' stamp: 'StephaneDucasse 4/22/2012 16:50'!
hasClipSubmorphsString
	"Answer a string that represents the clip-submophs checkbox"
	^ (self clipSubmorphs) -> 'provide clipping' translated! !

!Morph methodsFor: 'menus' stamp: 'StephaneDucasse 4/22/2012 16:50'!
hasDirectionHandlesString
	^ (self wantsDirectionHandles) ->  'direction handles' translated! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/22/2012 16:50'!
hasDisableTableLayoutString
	^ (self disableTableLayout) -> 'disable layout in tables' translated! !

!Morph methodsFor: 'text-anchor' stamp: 'StephaneDucasse 4/22/2012 16:50'!
hasDocumentAnchorString
	^ (self textAnchorType == #document) ->  'Document' translated! !

!Morph methodsFor: 'menus' stamp: 'StephaneDucasse 4/22/2012 16:51'!
hasDragAndDropEnabledString
	"Answer a string to characterize the drag & drop status of the  
	receiver"
	^ (self dragNDropEnabled) -> 'accept drops' translated! !

!Morph methodsFor: 'drop shadows' stamp: 'dgd 2/16/2003 21:42'!
hasDropShadow
	"answer whether the receiver has DropShadow"
	^ self
		valueOfProperty: #hasDropShadow
		ifAbsent: [false]! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 19:03'!
hasDropShadow: aBool
	aBool
		ifTrue:[self setProperty: #hasDropShadow toValue: true]
		ifFalse:[self removeProperty: #hasDropShadow]! !

!Morph methodsFor: 'drop shadows' stamp: 'StephaneDucasse 4/22/2012 16:51'!
hasDropShadowString
	^ (self hasDropShadow) ->  'show shadow' translated! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:31'!
hasExtension
	"answer whether the receiver has extention"
	^ extension notNil! !

!Morph methodsFor: 'event handling' stamp: 'sw 4/2/98 14:16'!
hasFocus
	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 9/15/2000 16:16'!
hasHalo
	^self hasProperty: #hasHalo.! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 9/28/2000 17:54'!
hasHalo: aBool
	aBool
		ifTrue:[self setProperty: #hasHalo toValue: true]
		ifFalse:[self removeProperty: #hasHalo]! !

!Morph methodsFor: 'text-anchor' stamp: 'StephaneDucasse 4/22/2012 16:51'!
hasInlineAnchorString
	^ (self textAnchorType == #inline)-> 'Inline' translated! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/22/2012 16:52'!
hasNoLayoutString
	^ (self layoutPolicy isNil) -> 'no layout' translated! !

!Morph methodsFor: 'structure' stamp: 'ar 10/3/2000 15:36'!
hasOwner: aMorph
	"Return true if the receiver has aMorph in its owner chain"
	aMorph ifNil:[^true].
	self allOwnersDo:[:m| m = aMorph ifTrue:[^true]].
	^false! !

!Morph methodsFor: 'text-anchor' stamp: 'StephaneDucasse 4/22/2012 16:52'!
hasParagraphAnchorString
	^ (self textAnchorType == #paragraph) -> 'Paragraph' translated! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:47'!
hasProperty: aSymbol 
	"Answer whether the receiver has the property named aSymbol"
	extension ifNil: [^ false].
	^extension hasProperty: aSymbol! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/22/2012 16:52'!
hasProportionalLayoutString
	| layout |
	^ ((layout := self layoutPolicy) notNil
			and: [layout isProportionalLayout]) -> 'proportional layout' translated! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/22/2012 16:52'!
hasReverseCellsString
	^ (self reverseTableCells) -> 'reverse table cells' translated! !

!Morph methodsFor: 'drop shadows' stamp: 'dgd 2/16/2003 21:58'!
hasRolloverBorder
	"answer whether the receiver has RolloverBorder"
	^ self
		valueOfProperty: #hasRolloverBorder
		ifAbsent: [false]! !

!Morph methodsFor: 'drop shadows' stamp: 'RAA 11/7/2000 15:54'!
hasRolloverBorder: aBool
	aBool
		ifTrue:[self setProperty: #hasRolloverBorder toValue: true]
		ifFalse:[self removeProperty: #hasRolloverBorder]! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/22/2012 16:52'!
hasRubberBandCellsString
	^ (self rubberBandCells) -> 'rubber band cells' translated! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:32'!
hasSubmorphs
	^submorphs notEmpty! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/22/2012 16:53'!
hasTableLayoutString
	| layout |
	^ ((layout := self layoutPolicy) notNil
			and: [layout isTableLayout]) ->   'table layout' translated! !

!Morph methodsFor: 'menu' stamp: 'MarcusDenker 12/2/2013 14:14'!
hasYellowButtonMenu
	"Answer true if I have any items at all for a context (yellow  
	button) menu."
	^ self wantsYellowButtonMenu
			or: [self modelOrNil
				ifNil: [false]
				ifNotNil: [:aModel | aModel hasModelYellowButtonMenuItems]]! !

!Morph methodsFor: 'geometry' stamp: ''!
height

	^ bounds height! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:22'!
height: aNumber
	" Set my height; my position (top-left corner) and width will remain the same "

	self extent: self width@aNumber asInteger.
! !

!Morph methodsFor: 'Morphic-Base-Widgets' stamp: 'BenjaminVanRyseghem 7/25/2012 11:42'!
heightToDisplayInList: aList

	^ self minExtent y! !

!Morph methodsFor: 'drawing' stamp: 'MarcusDenker 10/21/2013 14:22'!
hide
	owner ifNil: [ ^ self ].
	self visible
		ifFalse: [ ^ self ].
	self visible: false.
	self changed! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/30/1998 12:44'!
highlight
	"The receiver is being asked to appear in a highlighted state.  Mostly used for textual morphs"
	self color: self highlightColor! !

!Morph methodsFor: 'accessing' stamp: 'MarcusDenker 10/3/2013 23:47'!
highlightColor
	
	^ (self valueOfProperty: #highlightColor)
		ifNotNil:
			[:val | val ifNil: [self error: 'nil highlightColor']]
		ifNil:
			[owner ifNil: [self color] ifNotNil: [owner highlightColor]]! !

!Morph methodsFor: 'accessing' stamp: 'sw 7/2/1998 13:51'!
highlightColor: aColor
	self setProperty: #highlightColor toValue: aColor! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 11:51'!
highlightForDrop
	self highlightForDrop: true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 12:01'!
highlightForDrop: aBoolean
	self setProperty: #highlightedForDrop toValue: aBoolean.
	self changed! !

!Morph methodsFor: 'drawing' stamp: 'LC 5/18/2000 08:51'!
highlightForMouseDown
	self highlightForMouseDown: true! !

!Morph methodsFor: 'drawing' stamp: 'ar 3/17/2001 15:56'!
highlightForMouseDown: aBoolean
	aBoolean 
		ifTrue:[self setProperty: #highlightedForMouseDown toValue: aBoolean]
		ifFalse:[self removeProperty: #highlightedForMouseDown. self resetExtension].
	self changed! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 10:53'!
highlightedForDrop
	^(self valueOfProperty: #highlightedForDrop) == true! !

!Morph methodsFor: 'drawing' stamp: 'LC 5/18/2000 08:48'!
highlightedForMouseDown
	^(self valueOfProperty: #highlightedForMouseDown) == true! !

!Morph methodsFor: 'thumbnail' stamp: 'dgd 9/12/2004 21:12'!
icon
	"Answer a form with an icon to represent the receiver"
	^ self valueOfProperty: #icon! !

!Morph methodsFor: 'thumbnail' stamp: 'dgd 9/12/2004 20:33'!
iconOrThumbnail
	"Answer an appropiate form to represent the receiver"

	^ self icon
		ifNil: [ | maxExtent fb |maxExtent := 320 @ 240.
			fb := self fullBounds.
			fb area <= (maxExtent x * maxExtent y)
				ifTrue: [self imageForm]
				ifFalse: [self imageFormForRectangle: (fb topLeft extent: maxExtent)]
		]
! !

!Morph methodsFor: 'thumbnail' stamp: 'dgd 9/13/2004 12:43'!
iconOrThumbnailOfSize: aNumberOrPoint 
	"Answer an appropiate form to represent the receiver"

	^ self iconOrThumbnail scaledIntoFormOfSize: aNumberOrPoint
! !

!Morph methodsFor: 'drawing' stamp: 'jm 6/11/97 17:21'!
imageForm

	^ self imageFormForRectangle: self fullBounds
! !

!Morph methodsFor: 'drawing' stamp: 'nk 9/1/2004 15:08'!
imageForm: depth backgroundColor: aColor forRectangle: rect
	| canvas |
	canvas := Display defaultCanvasClass extent: rect extent depth: depth.
	canvas translateBy: rect topLeft negated
		during:[:tempCanvas| 
			tempCanvas fillRectangle: rect color: aColor.
			tempCanvas fullDrawMorph: self].
	^ canvas form offset: rect topLeft! !

!Morph methodsFor: 'drawing' stamp: 'ar 9/1/2000 14:23'!
imageForm: depth forRectangle: rect
	| canvas |
	canvas := Display defaultCanvasClass extent: rect extent depth: depth.
	canvas translateBy: rect topLeft negated
		during:[:tempCanvas| tempCanvas fullDrawMorph: self].
	^ canvas form offset: rect topLeft! !

!Morph methodsFor: 'drawing' stamp: 'di 7/8/1998 12:42'!
imageFormDepth: depth

	^ self imageForm: depth forRectangle: self fullBounds
! !

!Morph methodsFor: 'drawing' stamp: 'di 9/9/1998 22:25'!
imageFormForRectangle: rect

	^ self imageForm: Display depth forRectangle: rect
! !

!Morph methodsFor: 'user interface' stamp: 'MarianoMartinezPeck 8/24/2012 15:28'!
initialExtent
	| ext |
	(ext := self valueOfProperty: #initialExtent)
		ifNotNil:
			[^ ext].
	^700@500! !

!Morph methodsFor: 'initialization' stamp: 'alain.plantec 5/28/2009 10:09'!
initialize
	"initialize the state of the receiver"
	super initialize.
	owner := nil.
	submorphs := EmptyArray.
	bounds := self defaultBounds.
	color := self defaultColor! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:35'!
initializeExtension
	"private - initializes the receiver's extension"
	extension := MorphExtension new! !

!Morph methodsFor: 'geometry' stamp: 'ar 12/22/2001 22:43'!
innerBounds
	"Return the inner rectangle enclosed by the bounds of this morph excluding the space taken by its borders. For an unbordered morph, this is just its bounds."

	^ self bounds insetBy: self borderWidth! !

!Morph methodsFor: 'accessing' stamp: 'tk 1/31/2002 10:25'!
insetColor
	owner ifNil:[^self color].
	^ self colorForInsets! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 11/28/2009 14:34'!
inspectAt: aPoint event: evt
	|  morphs target |
	morphs := self morphsAt: aPoint.
	(morphs includes: self) ifFalse:[morphs := morphs copyWith: self].
	target := UIManager default
				chooseFrom: (morphs
						collect: [:t |t class name asString])
				values: morphs
				title: ('inspect whom? (deepest at top)' translated).
	target ifNil:[^self].
	target inspectInMorphic: evt! !

!Morph methodsFor: 'menus' stamp: 'MarcusDenker 9/28/2013 15:50'!
inspectInMorphic: evt
	Smalltalk tools inspector inspect: self! !

!Morph methodsFor: 'debug and other' stamp: 'sw 11/5/1998 20:31'!
inspectOwnerChain
	self ownerChain inspectWithLabel: 'Owner chain for ', self printString! !

!Morph methodsFor: 'geometry' stamp: 'nk 4/27/2003 16:16'!
intersects: aRectangle
	"Answer whether aRectangle, which is in World coordinates, intersects me."

	^self fullBoundsInWorld intersects: aRectangle! !

!Morph methodsFor: 'initialize' stamp: 'ar 1/31/2001 13:57'!
intoWorld: aWorld
	"The receiver has just appeared in a new world. Note:
		* aWorld can be nil (due to optimizations in other places)
		* owner is already set
		* owner's submorphs may not include receiver yet.
	Important: Keep this method fast - it is run whenever morphs are added."
	aWorld ifNil:[^self].
	self wantsSteps ifTrue:[aWorld startStepping: self].
	self submorphsDo:[:m| m intoWorld: aWorld].
! !

!Morph methodsFor: 'change reporting' stamp: 'ar 11/12/2000 18:50'!
invalidRect: damageRect
	^self invalidRect: damageRect from: self! !

!Morph methodsFor: 'change reporting' stamp: 'IgorStasenko 12/22/2012 03:13'!
invalidRect: aRectangle from: aMorph
	| damageRect |
	aRectangle hasPositiveExtent ifFalse: [ ^self ].
	damageRect := aRectangle.
	aMorph == self ifFalse:[
		"Clip to receiver's clipping bounds if the damage came from a child"
		self clipSubmorphs 
			ifTrue:[damageRect := aRectangle intersect: self clippingBounds ifNone: [ ^ self ]]].
	owner ifNotNil: [owner invalidRect: damageRect from: self].! !

!Morph methodsFor: 'meta-actions' stamp: 'BenjaminVanRyseghem 5/4/2013 12:55'!
invokeMetaMenu: evt
	| menu |
	
	Smalltalk tools userManager canShowMorphHalo
		ifFalse: [ ^ self ].
	menu := self buildMetaMenu: evt.
	menu addTitle: self externalName.
	self world ifNotNil: [
		menu popUpEvent: evt in: self world
	]! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 11/28/2009 14:34'!
invokeMetaMenuAt: aPoint event: evt 
	| morphs target |
	morphs := self morphsAt: aPoint.
	(morphs includes: self)
		ifFalse: [morphs := morphs copyWith: self].
	morphs size = 1
		ifTrue: [morphs anyOne invokeMetaMenu: evt]
		ifFalse: [target := UIManager default
						chooseFrom: (morphs
								collect: [:t | t class name asString])
						values: morphs.
			target 
				ifNil: [^ self].
			target invokeMetaMenu: evt]! !

!Morph methodsFor: 'classification' stamp: 'di 5/7/1998 01:21'!
isAlignmentMorph

	^ false! !

!Morph methodsFor: 'testing' stamp: 'dgd 8/31/2004 15:00'!
isDockingBar
	"Return true if the receiver is a docking bar"
	^ false! !

!Morph methodsFor: 'classification' stamp: 'jm 4/17/1998 00:44'!
isFlexMorph

	^ false
! !

!Morph methodsFor: 'testing' stamp: 'AlainPlantec 5/5/2010 17:55'!
isFlexed
	"Return true if the receiver is currently flexed"
	^ owner notNil and: [owner isFlexMorph]! !

!Morph methodsFor: 'testing' stamp: 'MarcusDenker 3/21/2010 20:16'!
isFullOnScreen
	"Answer if the receiver is full contained in the owner visible  
	area."
	owner ifNil: [^ true].
	self visible ifFalse: [^ true].
	^ owner clearArea containsRect: self fullBounds! !

!Morph methodsFor: 'classification' stamp: ''!
isHandMorph

	^ false! !

!Morph methodsFor: 'structure' stamp: 'dgd 9/1/2004 17:17'!
isInDockingBar
	"answer if the receiver is in a menu bar"
	^ (owner notNil) and: [owner isDockingBar]! !

!Morph methodsFor: 'structure' stamp: 'dgd 9/18/2004 15:56'!
isInSystemWindow
	"answer if the receiver is in a system window"
	^ owner isMorph and:[owner isSystemWindow or:[owner isInSystemWindow]]! !

!Morph methodsFor: 'structure' stamp: 'dgd 2/22/2003 19:05'!
isInWorld
	"Return true if this morph is in a world."

	^self world notNil! !

!Morph methodsFor: 'testing' stamp: 'nk 10/13/2003 18:36'!
isLineMorph
	^false! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:53'!
isLocked
	"answer whether the receiver is Locked"
	extension ifNil: [^ false].
	^ extension locked! !

!Morph methodsFor: 'testing' stamp: ''!
isMorph

	^ true! !

!Morph methodsFor: 'classification' stamp: 'jm 5/7/1998 13:45'!
isRenderer
	"A *renderer* morph transforms the appearance of its submorph in some manner. For example, it might supply a drop shadow or scale and rotate the morph it encases. Answer true if this morph acts as a renderer. This default implementation returns false."
	"Details: A renderer is assumed to have a single submorph. Renderers may be nested to concatenate their transformations. It is useful to be able to find the outer-most renderer. This can be done by ascending the owner chain from the rendered morph. To find the morph being rendered, one can descend through the (singleton) submorph lists of the renderer chain until a non-renderer is encountered."

	^ false
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 3/22/2000 14:28'!
isStepping
	"Return true if the receiver is currently stepping in its world"
	| aWorld |
	^ (aWorld := self world)
		ifNil:		[false]
		ifNotNil:	[aWorld isStepping: self]! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 10/22/2000 16:43'!
isSteppingSelector: aSelector
	"Return true if the receiver is currently stepping in its world"
	| aWorld |
	^ (aWorld := self world)
		ifNil:		[false]
		ifNotNil:	[aWorld isStepping: self selector: aSelector]! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:57'!
isSticky
	"answer whether the receiver is Sticky"
	extension ifNil: [^ false].
	^ extension sticky! !

!Morph methodsFor: 'classification' stamp: 'ar 12/16/2001 18:28'!
isTextMorph
	^false! !

!Morph methodsFor: 'testing' stamp: 'StephaneDucasse 10/13/2013 21:49'!
isTranslucentButNotTransparent
	"Answer true if this any of this morph is translucent but not transparent."

	^ color isColor and: [ color isTranslucentButNotTransparent ]
! !

!Morph methodsFor: 'classification' stamp: ''!
isWorldMorph

	^ false! !

!Morph methodsFor: 'classification' stamp: ''!
isWorldOrHandMorph

	^ self isWorldMorph or: [self isHandMorph]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'BenjaminVanRyseghem 6/21/2013 17:53'!
justDroppedInto: aMorph event: anEvent
	"This message is sent to a dropped morph after it has been dropped on -- and been accepted by -- a drop-sensitive morph"

	| aWindow |
	(self formerOwner notNil and: [self formerOwner ~~ aMorph])
		ifTrue: [self removeHalo].
	self formerOwner: nil.
	self formerPosition: nil.
	(aWindow := aMorph ownerThatIsA: SystemWindow) 
		ifNotNil: [aWindow isActive 
			ifFalse: [aWindow activate]].

	"An object launched by certain parts-launcher mechanisms should end up fully visible..."
	(self hasProperty: #beFullyVisibleAfterDrop) 
		ifTrue: [aMorph == ActiveWorld 
			ifTrue: [self goHome].
			self removeProperty: #beFullyVisibleAfterDrop].

! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 2/6/2001 22:12'!
justGrabbedFrom: formerOwner
	"The receiver was just grabbed from its former owner and is now attached to the hand. By default, we pass this message on if we're a renderer."
	(self isRenderer and:[self hasSubmorphs]) 
		ifTrue:[self firstSubmorph justGrabbedFrom: formerOwner].! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'ar 11/24/1998 14:19'!
keepsTransform
	"Return true if the receiver will keep it's transform while being grabbed by a hand."
	^false! !

!Morph methodsFor: 'event handling' stamp: 'ar 9/14/2000 18:23'!
keyDown: anEvent
	"Handle a key down event. The default response is to do nothing."! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 3/15/2013 16:17'!
keyStroke: anEvent
	"Handle a keystroke event.  The default response is to let my eventHandler, if any, handle it."
	^false! !

!Morph methodsFor: 'event handling' stamp: 'KTT 6/1/2004 11:41'!
keyUp: anEvent
	"Handle a key up event. The default response is to do nothing."! !

!Morph methodsFor: 'event handling' stamp: 'CamilloBruni 10/10/2012 13:37'!
keyboardFocusChange: gotFocus
	"The message is sent to a morph when its keyboard focus change. The given argument indicates that the receiver is gaining keyboard focus (versus losing) the keyboard focus. Morphs that accept keystrokes should change their appearance in some way when they are the current keyboard focus. This default implementation does nothing."
	self announceKeyboardFocusChange: gotFocus! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:32'!
lastSubmorph
	^submorphs last! !

!Morph methodsFor: 'layout' stamp: 'ar 11/12/2000 23:10'!
layoutBounds
	"Return the bounds for laying out children of the receiver"
	| inset box |
	inset := self layoutInset.
	box := self innerBounds.
	inset isZero ifTrue:[^box].
	^box insetBy: inset.! !

!Morph methodsFor: 'layout' stamp: 'ar 10/31/2000 21:09'!
layoutBounds: aRectangle
	"Set the bounds for laying out children of the receiver.
	Note: written so that #layoutBounds can be changed without touching this method"
	| outer inner |
	outer := self bounds.
	inner := self layoutBounds.
	bounds := aRectangle origin + (outer origin - inner origin) corner:
				aRectangle corner + (outer corner - inner corner).! !

!Morph methodsFor: 'layout-properties' stamp: 'md 2/27/2006 09:59'!
layoutFrame
	"Layout specific. Return the layout frame describing where the  
	receiver should appear in a proportional layout"
	^ extension ifNotNil: [extension layoutFrame]! !

!Morph methodsFor: 'layout-properties' stamp: 'StephaneDucasse 12/21/2012 18:15'!
layoutFrame: aLayoutFrame
	"Layout specific. Return the layout frame describing where the receiver should appear in a proportional layout"
	self layoutFrame == aLayoutFrame ifTrue: [^self].
	self assureExtension layoutFrame: aLayoutFrame asLayoutFrame.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'MarcusDenker 10/3/2013 23:32'!
layoutInset
	"Return the extra inset for layouts"

	^ self layoutProperties ifNil: [ 0 ] ifNotNil: [ :props | props layoutInset ]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 16:38'!
layoutInset: aNumber
	"Return the extra inset for layouts"
	self assureTableProperties layoutInset: aNumber.
	self layoutChanged.! !

!Morph methodsFor: 'layout-menu' stamp: 'StephaneDucasse 4/23/2012 10:21'!
layoutMenuPropertyString: aSymbol from: currentSetting 
	| wording |
	wording := String
				streamContents: [:stream | 
					| index | 
					index := 1.
					aSymbol keysAndValuesDo: [:idx :ch | ch isUppercase
								ifTrue: [stream nextPutAll: (aSymbol copyFrom: index to: idx - 1) asLowercase.
										stream nextPutAll: ' '.
										index := idx]].
					index < aSymbol size
						ifTrue: [stream nextPutAll: (aSymbol copyFrom: index to: aSymbol size) asLowercase]].
	^ (aSymbol == currentSetting) -> wording translated! !

!Morph methodsFor: 'layout-properties' stamp: 'md 2/27/2006 10:00'!
layoutPolicy
	"Layout specific. Return the layout policy describing how children 
	of the receiver should appear."
	^ extension ifNotNil: [ extension layoutPolicy]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:21'!
layoutPolicy: aLayoutPolicy
	"Layout specific. Return the layout policy describing how children of the receiver should appear."
	self layoutPolicy == aLayoutPolicy ifTrue:[^self].
	self assureExtension layoutPolicy: aLayoutPolicy.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'md 2/27/2006 09:58'!
layoutProperties
	"Return the current layout properties associated with the  
	receiver"
	^ extension ifNotNil: [ extension layoutProperties]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:21'!
layoutProperties: newProperties
	"Return the current layout properties associated with the receiver"
	self layoutProperties == newProperties ifTrue:[^self].
	self assureExtension layoutProperties: newProperties.
! !

!Morph methodsFor: 'layout' stamp: 'ar 11/12/2000 17:35'!
layoutProportionallyIn: newBounds
	"Layout specific. Apply the given bounds to the receiver."
	| box frame |
	frame := self layoutFrame ifNil:[^self].
	"before applying the proportional values make sure the receiver's layout is computed"
	self fullBounds. "sigh..."
	"compute the cell size the receiver has given its layout frame"
	box := frame layout: self bounds in: newBounds.
	(box = self bounds) ifTrue:[^self]. "no change"
	^self layoutInBounds: box.! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:16'!
left
	" Return the x-coordinate of my left side "

	^ bounds left! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:15'!
left: aNumber
	" Move me so that my left side is at the x-coordinate aNumber. My extent (width & height) are unchanged "

	self position: (aNumber @ bounds top)! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
leftCenter

	^ bounds leftCenter! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:58'!
listCentering
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of primary direction
		#bottomRight - center at end of primary direction
		#center - center in the middle of primary direction
		#justified - insert extra space inbetween rows/columns
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#topLeft] ifNotNil:[props listCentering].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:05'!
listCentering: aSymbol
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of primary direction
		#bottomRight - center at end of primary direction
		#center - center in the middle of primary direction
		#justified - insert extra space inbetween rows/columns
	"
	self assureTableProperties listCentering: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
listCenteringString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self listCentering! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:59'!
listDirection
	"Layout specific. This property describes the direction in which a list-like layout should be applied. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
	indicating the direction in which any layout should take place"
	| props |
	props := self layoutProperties.
	^props ifNil:[#topToBottom] ifNotNil:[props listDirection].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
listDirection: aSymbol
	"Layout specific. This property describes the direction in which a list-like layout should be applied. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
	indicating the direction in which any layout should take place"
	self assureTableProperties listDirection: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
listDirectionString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self listDirection! !

!Morph methodsFor: 'Morphic-Base-Widgets' stamp: 'BenjaminVanRyseghem 5/21/2013 18:02'!
listRenderOn: aCanvas atRow: aRow bounds: drawBounds color: drawColor backgroundColor: backgroundColor from: aMorph

	self bounds: drawBounds.
	self fullDrawOn: aCanvas.
	(aMorph submorphs includes: self)
		ifFalse: [ aMorph addMorph: self ]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:59'!
listSpacing
	"Layout specific. This property describes how the heights for different rows in a table layout should be handled.
		#equal - all rows have the same height
		#none - all rows may have different heights
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#none] ifNotNil:[props listSpacing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
listSpacing: aSymbol
	"Layout specific. This property describes how the heights for different rows in a table layout should be handled.
		#equal - all rows have the same height
		#none - all rows may have different heights
	"
	self assureTableProperties listSpacing: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
listSpacingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self listSpacing! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:07'!
localPointToGlobal: aPoint
	^self point: aPoint in: nil! !

!Morph methodsFor: 'accessing' stamp: 'sw 8/4/97 12:05'!
lock
	self lock: true! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:58'!
lock: aBoolean 
	"change the receiver's lock property"
	(extension isNil and: [aBoolean not]) ifTrue: [^ self].
	self assureExtension locked: aBoolean! !

!Morph methodsFor: 'menus' stamp: 'sw 9/21/2000 22:50'!
lockUnlockMorph
	"If the receiver is locked, unlock it; if unlocked, lock it"

	self isLocked ifTrue: [self unlock] ifFalse: [self lock]! !

!Morph methodsFor: 'menus' stamp: 'StephaneDucasse 4/22/2012 16:53'!
lockedString
	"Answer the string to be shown in a menu to represent the 'locked' status"
	^ (self isLocked) -> 'be locked' translated! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 9/6/2004 14:17'!
mainDockingBars
	"Answer the receiver's main dockingBars"
	^ self dockingBars
		select: [:each | each hasProperty: #mainDockingBarTimeStamp]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:59'!
maxCellSize
	"Layout specific. This property specifies the maximum size of a table cell."
	| props |
	props := self layoutProperties.
	^props ifNil:[SmallInteger maxVal] ifNotNil:[props maxCellSize].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
maxCellSize: aPoint
	"Layout specific. This property specifies the maximum size of a table cell."
	self assureTableProperties maxCellSize: aPoint.
	self layoutChanged.! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 20:15'!
maybeAddCollapseItemTo: aMenu
	"If appropriate, add a collapse item to the given menu"

	| anOwner |
	(anOwner := self topRendererOrSelf owner) ifNotNil:
			[anOwner isWorldMorph ifTrue:
				[aMenu add: 'collapse' translated target: self action: #collapse]]! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 3/27/2011 17:12'!
maybeDuplicateMorph
	"Maybe duplicate the morph"

	^self duplicate openInHand! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 11/7/2009 18:50'!
maybeDuplicateMorph: evt
	^self duplicateMorph: evt! !

!Morph methodsFor: 'other events' stamp: 'sw 8/1/2001 14:08'!
menuButtonMouseEnter: event
	"The mouse entered a menu-button area; show the menu cursor temporarily"

	event hand showTemporaryCursor: Cursor menu! !

!Morph methodsFor: 'other events' stamp: 'sw 8/1/2001 14:09'!
menuButtonMouseLeave: event
	"The mouse left a menu-button area; restore standard cursor"

	event hand showTemporaryCursor: nil! !

!Morph methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:23'!
menuKeyboardControl
	^ self theme settings menuKeyboardControl! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:00'!
minCellSize
	"Layout specific. This property specifies the minimal size of a table cell."
	| props |
	props := self layoutProperties.
	^props ifNil:[0] ifNotNil:[props minCellSize].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
minCellSize: aPoint
	"Layout specific. This property specifies the minimal size of a table cell."
	self assureTableProperties minCellSize: aPoint.
	self layoutChanged.! !

!Morph methodsFor: 'layout' stamp: 'dgd 2/16/2003 21:52'!
minHeight
	"answer the receiver's minHeight"
	^ self
		valueOfProperty: #minHeight
		ifAbsent: [2]! !

!Morph methodsFor: 'layout' stamp: 'ClementBera 7/30/2013 11:04'!
minHeight: aNumber 
	aNumber 
		ifNil: [self removeProperty: #minHeight]
		ifNotNil: [self setProperty: #minHeight toValue: aNumber].
	self layoutChanged! !

!Morph methodsFor: 'layout' stamp: 'dgd 2/16/2003 21:54'!
minWidth
	"answer the receiver's minWidth"
	^ self
		valueOfProperty: #minWidth
		ifAbsent: [2]! !

!Morph methodsFor: 'layout' stamp: 'ClementBera 7/30/2013 11:04'!
minWidth: aNumber 
	aNumber 
		ifNil: [self removeProperty: #minWidth]
		ifNotNil: [self setProperty: #minWidth toValue: aNumber].
	self layoutChanged! !

!Morph methodsFor: 'geometry' stamp: 'sw 6/4/2000 21:59'!
minimumExtent
	| ext |
	"This returns the minimum extent that the morph may be shrunk to.  Not honored in too many places yet, but respected by the resizeToFit feature, at least.  copied up from SystemWindow 6/00"
	(ext := self valueOfProperty: #minimumExtent)
		ifNotNil:
			[^ ext].
	^ 100 @ 80! !

!Morph methodsFor: 'geometry' stamp: 'sw 6/4/2000 22:00'!
minimumExtent: aPoint
	"Remember a minimumExtent, for possible future use"

	self setProperty: #minimumExtent toValue: aPoint
! !

!Morph methodsFor: 'menus' stamp: 'dgd 9/22/2004 20:30'!
model
	^ nil ! !

!Morph methodsFor: 'accessing' stamp: 'sw 10/23/1999 22:35'!
modelOrNil
	^ nil! !

!Morph methodsFor: 'structure' stamp: 'sw 8/29/2000 14:55'!
morphPreceding: aSubmorph
	"Answer the morph immediately preceding aSubmorph, or nil if none"

	| anIndex |
	anIndex := submorphs indexOf: aSubmorph ifAbsent: [^ nil].
	^ anIndex > 1
		ifTrue:
			[submorphs at: (anIndex - 1)]
		ifFalse:
			[nil]! !

!Morph methodsFor: 'wiw support' stamp: 'RAA 7/19/2000 20:44'!
morphicLayerNumber

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"

	^(owner isNil or: [owner isWorldMorph]) ifTrue: [
		self valueOfProperty: #morphicLayerNumber ifAbsent: [100]
	] ifFalse: [
		owner morphicLayerNumber
	].

	"leave lots of room for special things"! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 10/8/2000 15:40'!
morphsAt: aPoint
	"Return a collection of all morphs in this morph structure that contain the given point, possibly including the receiver itself.  The order is deepest embedding first."
	^self morphsAt: aPoint unlocked: false! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'nice 1/5/2010 15:59'!
morphsAt: aPoint behind: aMorph unlocked: aBool 
	"Return all morphs at aPoint that are behind frontMorph; if aBool is true return only unlocked, visible morphs."

	| isBack all tfm |
	all := (aMorph isNil or: [owner isNil]) 
				ifTrue: 
					["Traverse down"

					(self fullBounds containsPoint: aPoint) ifFalse: [^#()].
					(aBool and: [self isLocked or: [self visible not]]) ifTrue: [^#()].
					nil]
				ifFalse: 
					["Traverse up"

					tfm := self transformedFrom: owner.
					all := owner 
								morphsAt: (tfm localPointToGlobal: aPoint)
								behind: self
								unlocked: aBool.
					WriteStream with: all].
	isBack := aMorph isNil.
	self submorphsDo: 
			[:m | | found | 
			isBack 
				ifTrue: 
					[tfm := m transformedFrom: self.
					found := m 
								morphsAt: (tfm globalPointToLocal: aPoint)
								behind: nil
								unlocked: aBool.
					found notEmpty 
						ifTrue: 
							[all ifNil: [all := Array new  writeStream].
							all nextPutAll: found]].
			m == aMorph ifTrue: [isBack := true]].
	(isBack and: [self containsPoint: aPoint]) 
		ifTrue: 
			[all ifNil: [^Array with: self].
			all nextPut: self].
	^all ifNil: [#()] ifNotNil: [all contents]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'PeterHugossonMiller 9/3/2009 10:06'!
morphsAt: aPoint unlocked: aBool
	"Return a collection of all morphs in this morph structure that contain the given point, possibly including the receiver itself.  The order is deepest embedding first."
	| mList |
	mList := Array new writeStream.
	self morphsAt: aPoint unlocked: aBool do:[:m| mList nextPut: m].
	^mList contents! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'nice 1/5/2010 15:59'!
morphsAt: aPoint unlocked: aBool do: aBlock
	"Evaluate aBlock with all the morphs starting at the receiver which appear at aPoint. If aBool is true take only visible, unlocked morphs into account."
	
	(self fullBounds containsPoint: aPoint) ifFalse:[^self].
	(aBool and:[self isLocked or:[self visible not]]) ifTrue:[^self].
	self submorphsDo:[:m| | tfm |
		tfm := m transformedFrom: self.
		m morphsAt: (tfm globalPointToLocal: aPoint) unlocked: aBool do: aBlock].
	(self containsPoint: aPoint) ifTrue:[aBlock value: self].! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 9/9/2000 17:31'!
morphsInFrontOf: someMorph overlapping: aRectangle do: aBlock
	"Evaluate aBlock with all top-level morphs in front of someMorph that overlap with the given rectangle. someMorph is either an immediate child of the receiver or nil (in which case all submorphs of the receiver are enumerated)."
	self submorphsDo:[:m|
		m == someMorph ifTrue:["Try getting out quickly"
			owner ifNil:[^self].
			^owner morphsInFrontOf: self overlapping: aRectangle do: aBlock].
		(m fullBoundsInWorld intersects: aRectangle)
			ifTrue:[aBlock value: m]].
	owner ifNil:[^self].
	^owner morphsInFrontOf: self overlapping: aRectangle do: aBlock.! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'PeterHugossonMiller 9/3/2009 10:06'!
morphsInFrontOverlapping: aRectangle
	"Return all top-level morphs in front of someMorph that overlap with the given rectangle."
	| morphList |
	morphList := Array new writeStream.
	self morphsInFrontOf: nil overlapping: aRectangle do:[:m | morphList nextPut: m].
	^morphList contents! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 9/9/2000 17:31'!
morphsInFrontOverlapping: aRectangle do: aBlock
	"Evaluate aBlock with all top-level morphs in front of someMorph that overlap with the given rectangle."
	^self morphsInFrontOf: nil overlapping: aRectangle do: aBlock! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 5/3/2013 10:58'!
mouseDown: evt 
	"Handle a mouse down event. The default response is to let my 
	eventHandler, if any, handle it."
	self eventHandler
		ifNotNil: [self eventHandler mouseDown: evt fromMorph: self]
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 10/3/2000 17:05'!
mouseDownOnHelpHandle: anEvent
	"The mouse went down in the show-balloon handle"
	
	| str |
	anEvent shiftPressed ifTrue: [^ self editBalloonHelpText].
	str := self balloonText.
	str ifNil: [str := self noHelpString].
	self showBalloon: str hand: anEvent hand.
! !

!Morph methodsFor: 'events-processing' stamp: 'StephaneDucasse 7/18/2010 16:22'!
mouseDownPriority
	"Return the default mouse down priority for the receiver"

	^ 0
! !

!Morph methodsFor: 'event handling' stamp: 'BenjaminVanRyseghem 1/18/2012 18:53'!
mouseEnter: evt
	"Handle a mouseEnter event, meaning the mouse just entered my bounds with no button pressed. The default response is to let my eventHandler, if any, handle it."

	^ self eventHandler ifNotNil:
		[self eventHandler mouseEnter: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/1998 07:33'!
mouseEnterDragging: evt
	"Handle a mouseEnterDragging event, meaning the mouse just entered my bounds with a button pressed or laden with submorphs.  The default response is to let my eventHandler, if any, handle it, or else to do nothing."

	self eventHandler ifNotNil:
		[^ self eventHandler mouseEnterDragging: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'BenjaminVanRyseghem 1/18/2012 18:54'!
mouseLeave: evt
	"Handle a mouseLeave event, meaning the mouse just left my bounds with no button pressed. The default response is to let my eventHandler, if any, handle it."

	^ self eventHandler ifNotNil:
		[self eventHandler mouseLeave: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/1998 07:38'!
mouseLeaveDragging: evt
	"Handle a mouseLeaveLaden event, meaning the mouse just left my bounds with a button pressed or laden with submorphs. The default response is to let my eventHandler, if any, handle it; else to do nothing."

	self eventHandler ifNotNil:
		[self eventHandler mouseLeaveDragging: evt fromMorph: self]! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:04'!
mouseMove: evt
	"Handle a mouse move event. The default response is to let my eventHandler, if any, handle it."
	self eventHandler ifNotNil:
		[self eventHandler mouseMove: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/22/2000 17:08'!
mouseStillDown: evt
	"Handle a mouse move event. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler mouseStillDown: evt fromMorph: self].
! !

!Morph methodsFor: 'geniestubs' stamp: 'nk 3/11/2004 17:30'!
mouseStillDownStepRate
	"At what rate do I want to receive #mouseStillDown: notifications?"
	^1! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:02'!
mouseStillDownThreshold
	"Return the number of milliseconds after which mouseStillDown: should be sent"
	^200! !

!Morph methodsFor: 'event handling' stamp: 'SeanDeNigris 1/31/2014 18:32'!
mouseUp: evt
	"Handle a mouse up event. The default response is to let my eventHandler, if any, handle it."

	^ self eventHandler ifNotNil:
		[self eventHandler mouseUp: evt fromMorph: self].
! !

!Morph methodsFor: 'debug and other' stamp: 'sw 2/6/2001 22:35'!
mouseUpCodeOrNil
	"If the receiver has a mouseUpCodeToRun, return it, else return nil"

	^ self valueOfProperty: #mouseUpCodeToRun ifAbsent: [nil]! !

!Morph methodsFor: 'event handling' stamp: 'StephaneDucasse 7/2/2013 16:26'!
moveOrResizeFromKeystroke: anEvent
	"move or resize the receiver based on a keystroke"

	| dir |
	anEvent keyValue = 28
		ifTrue: [ dir := -1 @ 0 ].
	anEvent keyValue = 29
		ifTrue: [ dir := 1 @ 0 ].
	anEvent keyValue = 30
		ifTrue: [ dir := 0 @ -1 ].
	anEvent keyValue = 31
		ifTrue: [ dir := 0 @ 1 ].
	dir notNil
		ifTrue: [ 
			anEvent controlKeyPressed
				ifTrue: [ dir := dir * 10 ].
			anEvent shiftPressed
				ifTrue: [ self extent: self extent + dir ]
				ifFalse: [ self position: self position + dir ] ]! !

!Morph methodsFor: 'naming' stamp: 'gm 2/22/2003 13:16'!
name: aName 
	(aName isString) ifTrue: [self setNameTo: aName]! !

!Morph methodsFor: 'structure' stamp: 'di 11/12/2000 16:13'!
nearestOwnerThat: conditionBlock
	"Return the first enclosing morph for which aBlock evaluates to true, or nil if none"

	^ self firstOwnerSuchThat: conditionBlock
! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'ar 2/16/1999 18:59'!
newTransformationMorph
	^TransformationMorph new! !

!Morph methodsFor: 'halos and balloon help' stamp: 'em 3/24/2005 10:05'!
noHelpString
	^ 'Help not yet supplied' translated! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 8/13/2003 11:32'!
noteNewOwner: aMorph
	"I have just been added as a submorph of aMorph"! !

!Morph methodsFor: 'halos and balloon help' stamp: 'MarcusDenker 11/7/2009 21:51'!
okayToAddDismissHandle
	"Answer whether a halo on the receiver should offer a dismiss handle.  This provides a hook for making it harder to disassemble some strucures even momentarily"

	^self resistsRemoval not! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/27/2001 14:50'!
okayToBrownDragEasily
	"Answer whether it it okay for the receiver to be brown-dragged easily -- i.e. repositioned within its container without extracting it.  At present this is just a hook -- nobody declines."

	^ true



"
	^ (self topRendererOrSelf owner isKindOf: PasteUpMorph) and:
		[self layoutPolicy isNil]"! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/29/2001 06:29'!
okayToResizeEasily
	"Answer whether it is appropriate to have the receiver be easily resized by the user from the halo"

	^ true

	"This one was too jarring, not that it didn't most of the time do the right  thing but because some of the time it didn't, such as in a holder.  If we pursue this path, the test needs to be airtight, obviously...
	^ (self topRendererOrSelf owner isKindOf: PasteUpMorph) and:
		[self layoutPolicy isNil]"! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 5/7/2010 22:53'!
okayToRotateEasily
	"Answer whether it is appropriate for a rotation handle to be shown for the receiver.  This is a hook -- at present nobody declines."

	^ self respondsTo: #prepareForRotating! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 7/24/2012 13:07'!
on: eventName send: selector to: recipient
	self eventHandler ifNil: [self eventHandler: MorphicEventHandler new].
	self eventHandler on: eventName send: selector to: recipient! !

!Morph methodsFor: 'event handling' stamp: 'GuillermoPolito 7/24/2012 13:08'!
on: eventName send: selector to: recipient withValue: value
	"NOTE: selector must take 3 arguments, of which value will be the *** FIRST ***"

	self eventHandler ifNil: [self eventHandler: MorphicEventHandler new].
	self eventHandler on: eventName send: selector to: recipient withValue: value
! !

!Morph methodsFor: 'announcements' stamp: 'StephaneDucasse 8/22/2013 23:38'!
onAnnouncement: anAnnouncement do: aValuable
	self announcer when: anAnnouncement do: aValuable.! !

!Morph methodsFor: 'announcements' stamp: 'GuillermoPolito 5/1/2012 19:50'!
onAnnouncement: anAnnouncement send: aMessageSelector to: anObject
	self announcer on: anAnnouncement send: aMessageSelector to: anObject.! !

!Morph methodsFor: 'initialize' stamp: 'RAA 10/18/2000 12:33'!
openCenteredInWorld

	self 
		fullBounds;
		position: Display extent - self extent // 2;
		openInWorld.! !

!Morph methodsFor: 'initialize' stamp: 'sw 3/21/2000 14:46'!
openInHand
	"Attach the receiver to the current hand in the current morphic world"

	self currentHand attachMorph: self! !

!Morph methodsFor: 'initialize' stamp: 'cb 6/25/2013 13:24'!
openInWindow

	^self openInWindowLabeled: self defaultLabel
! !

!Morph methodsFor: 'initialize' stamp: 'sma 4/22/2000 20:28'!
openInWindowLabeled: aString

	^self openInWindowLabeled: aString inWorld: self currentWorld! !

!Morph methodsFor: 'initialize' stamp: 'alain.plantec 6/10/2008 18:35'!
openInWorld
        "Add this morph to the world."

        self openInWorld: self currentWorld! !

!Morph methodsFor: 'initialize' stamp: 'GuillermoPolito 5/1/2012 20:01'!
openInWorld: aWorld
	"Add this morph to the requested World."
	(aWorld visibleClearArea origin ~= (0@0) and: [self position = (0@0)]) ifTrue:
		[self position: aWorld visibleClearArea origin].
	aWorld addMorph: self.
	aWorld startSteppingSubmorphsOf: self.
	
	self announceOpened.! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:47'!
otherProperties
	"answer the receiver's otherProperties"
	^ extension ifNotNil: [extension otherProperties]! !

!Morph methodsFor: 'initialize' stamp: 'ar 1/31/2001 13:58'!
outOfWorld: aWorld
	"The receiver has just appeared in a new world. Notes:
		* aWorld can be nil (due to optimizations in other places)
		* owner is still valid
	Important: Keep this method fast - it is run whenever morphs are removed."
	aWorld ifNil:[^self].
	"ar 1/31/2001: We could explicitly stop stepping the receiver here but for the sake of speed I'm for now relying on the lazy machinery in the world itself."
	"aWorld stopStepping: self."
	self submorphsDo:[:m| m outOfWorld: aWorld].
! !

!Morph methodsFor: 'geometry' stamp: 'ar 11/12/2000 22:06'!
outerBounds
	"Return the 'outer' bounds of the receiver, e.g., the bounds that need to be invalidated when the receiver changes."
	| box |
	box := self bounds.
	self hasDropShadow ifTrue:[box := self expandFullBoundsForDropShadow: box].
	self hasRolloverBorder ifTrue:[box := self expandFullBoundsForRolloverBorder: box].
	^box! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 00:50'!
outermostMorphThat: conditionBlock
	"Return the outermost containing morph for which aBlock is true, or nil if none"

	| outermost |
	self allOwnersDo: [:m | (conditionBlock value: m) ifTrue: [outermost := m]].
	^ outermost! !

!Morph methodsFor: 'menu' stamp: 'nk 3/10/2004 19:51'!
outermostOwnerWithYellowButtonMenu
	"Answer me or my outermost owner that is willing to contribute menu items to a context menu.
	Don't include the world."

	| outermost |
	outermost := self outermostMorphThat: [ :ea |
		ea isWorldMorph not and: [ ea hasYellowButtonMenu ]].
	^outermost ifNil: [ self hasYellowButtonMenu ifTrue: [ self ] ifFalse: []] ! !

!Morph methodsFor: 'structure' stamp: 'marcus.denker 7/24/2009 14:07'!
outermostWorldMorph

	^World.! !

!Morph methodsFor: 'geometry' stamp: 'IgorStasenko 12/22/2012 03:13'!
overlapsShadowForm: itsShadow bounds: itsBounds
	"Answer true if itsShadow and my shadow overlap at all"
	| andForm overlapExtent |
	overlapExtent := (itsBounds intersect: self fullBounds ifNone: [ ^ false ]) extent.
	overlapExtent > (0 @ 0)
		ifFalse: [^ false].
	andForm := self shadowForm.
	overlapExtent ~= self fullBounds extent
		ifTrue: [andForm := andForm
						contentsOfArea: (0 @ 0 extent: overlapExtent)].
	andForm := andForm
				copyBits: (self fullBounds translateBy: itsShadow offset negated)
				from: itsShadow
				at: 0 @ 0
				clippingBox: (0 @ 0 extent: overlapExtent)
				rule: Form and
				fillColor: nil.
	^ andForm bits
		anySatisfy: [:w | w ~= 0]! !

!Morph methodsFor: 'structure' stamp: ''!
owner
	"Returns the owner of this morph, which may be nil."

	^ owner! !

!Morph methodsFor: 'debug and other' stamp: 'dgd 2/22/2003 19:05'!
ownerChain
	"Answer a list of objects representing the receiver and all of its owners.   The first element is the receiver, and the last one is typically the world in which the receiver resides"

	| c next |
	c := OrderedCollection with: self.
	next := self.
	[(next := next owner) notNil] whileTrue: [c add: next].
	^c asArray! !

!Morph methodsFor: 'change reporting' stamp: 'sw 7/8/1998 13:21'!
ownerChanged
	"The receiver's owner, some kind of a pasteup, has changed its layout."

	self snapToEdgeIfAppropriate! !

!Morph methodsFor: 'structure' stamp: 'di 11/12/2000 16:18'!
ownerThatIsA: aClass
	"Return the first enclosing morph that is a kind of aClass, or nil if none"

	^ self firstOwnerSuchThat: [:m | m isKindOf: aClass]! !

!Morph methodsFor: 'Morphic-Base-Worlds' stamp: 'sw 7/1/1998 18:02'!
pasteUpMorph
	"Answer the closest containing morph that is a PasteUp morph"
	^ self ownerThatIsA: PasteUpMorph! !

!Morph methodsFor: 'Morphic-Base-Worlds' stamp: 'dgd 8/28/2004 18:43'!
pasteUpMorphHandlingTabAmongFields
	"Answer the nearest PasteUpMorph in my owner chain that has the tabAmongFields property, or nil if none"

	| aPasteUp |
	aPasteUp := self owner.
	[aPasteUp notNil] whileTrue:
		[aPasteUp tabAmongFields ifTrue:
			[^ aPasteUp].
		aPasteUp := aPasteUp owner].
	^ nil! !

!Morph methodsFor: 'thumbnail' stamp: 'sw 6/16/1999 11:29'!
permitsThumbnailing
	^ true! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:01'!
point: aPoint from: aReferenceMorph

	owner ifNil: [^ aPoint].
	^ (owner transformFrom: aReferenceMorph) globalPointToLocal: aPoint.
! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:01'!
point: aPoint in: aReferenceMorph

	owner ifNil: [^ aPoint].
	^ (owner transformFrom: aReferenceMorph) localPointToGlobal: aPoint.
! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:02'!
pointFromWorld: aPoint
	^self point: aPoint from: self world! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:03'!
pointInWorld: aPoint
	^self point: aPoint in: self world! !

!Morph methodsFor: 'geometry' stamp: ''!
position

	^ bounds topLeft! !

!Morph methodsFor: 'geometry' stamp: 'wiz 11/25/2004 12:54'!
position: aPoint 
	"Change the position of this morph and and all of its
	submorphs. "
	| delta box |
	delta := aPoint asNonFractionalPoint - bounds topLeft.
	(delta x = 0
			and: [delta y = 0])
		ifTrue: [^ self].
	"Null change"
	box := self fullBounds.
	(delta dotProduct: delta)
			> 100
		ifTrue: ["e.g., more than 10 pixels moved"
			self invalidRect: box.
			self
				invalidRect: (box translateBy: delta)]
		ifFalse: [self
				invalidRect: (box
						merge: (box translateBy: delta))].
	self privateFullMoveBy: delta.
	owner
		ifNotNil: [owner layoutChanged]! !

!Morph methodsFor: 'geometry' stamp: 'di 9/30/1998 12:11'!
positionInWorld

	^ self pointInWorld: self position.
! !

!Morph methodsFor: 'geometry' stamp: 'sw 10/9/1998 08:56'!
positionSubmorphs
	self submorphsDo:
		[:aMorph | aMorph snapToEdgeIfAppropriate]! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 9/13/2013 15:50'!
potentialEmbeddingTargets
	"Return the potential targets for embedding the receiver"

	| oneUp topRend |
	(oneUp := (topRend := self topRendererOrSelf) owner) ifNil: [ ^ #() ].
	^ (oneUp morphsAt: topRend referencePosition behind: topRend unlocked: true) reject: [ :m | m isFlexMorph ]! !

!Morph methodsFor: 'meta-actions' stamp: 'wiz 7/17/2004 22:17'!
potentialTargets
	"Return the potential targets for the receiver.
	This is derived from Morph>>potentialEmbeddingTargets."
	owner ifNil:[^#()].
	^owner morphsAt: self referencePosition behind: self unlocked: true not! !

!Morph methodsFor: 'meta-actions' stamp: 'wiz 11/9/2006 23:04'!
potentialTargetsAt: aPoint 
	"Return the potential targets for the receiver.  
	This is derived from Morph>>potentialEmbeddingTargets."
	| realOwner |
	realOwner := self topRendererOrSelf
	owner
		ifNil: [^ #()].
	^ realOwner
		morphsAt: aPoint
		! !

!Morph methodsFor: 'halos and balloon help' stamp: 'stephane.ducasse 9/20/2008 21:57'!
preferredDuplicationHandleSelector
	"Answer the selector, either #addMakeSiblingHandle: or addDupHandle:, to be offered as the default in a halo open on me"

	^ #addDupHandle:! !

!Morph methodsFor: 'event handling' stamp: 'fbs 1/7/2005 15:43'!
preferredKeyboardBounds

	^ self bounds: self bounds in: World.
! !

!Morph methodsFor: 'event handling' stamp: 'fbs 1/7/2005 15:42'!
preferredKeyboardPosition

	^ (self bounds: self bounds in: World) topLeft.
! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'AlainPlantec 5/7/2010 21:27'!
prepareForScaling
	"If this morph requires a flex shell to scale,
	then wrap it in one and return it.
	Polygons, eg, may override to return themselves."

	^ self addFlexShell! !

!Morph methodsFor: 'copying' stamp: 'AlainPlantec 12/1/2010 09:35'!
prepareToBeSaved
	"Prepare this morph to be saved to disk. Subclasses should nil out any instance variables that holds state that should not be saved, such as cached Forms. Note that this operation may take more drastic measures than releaseCachedState; for example, it might discard the transcript of an interactive chat session."

	self releaseCachedState.
	self formerOwner: nil.
	self formerPosition: nil.
	fullBounds := nil! !

!Morph methodsFor: 'menus' stamp: 'sw 6/12/2001 21:08'!
presentHelp
	"Present a help message if there is one available"

	self inform: 'Sorry, no help has been
provided here yet.'! !

!Morph methodsFor: 'structure' stamp: 'RAA 6/13/2000 15:01'!
primaryHand

        | outer |
        outer := self outermostWorldMorph ifNil: [^ nil].
        ^ outer activeHand ifNil: [outer firstHand]! !

!Morph methodsFor: 'printing' stamp: 'MarcusDenker 10/28/2010 14:02'!
printOn: aStream 
	super printOn: aStream.
	aStream nextPutAll: '('.
	aStream
		print: self identityHash;
		nextPutAll: ')'! !

!Morph methodsFor: 'printing' stamp: 'jm 5/28/1998 18:00'!
printStructureOn: aStream indent: tabCount

	tabCount timesRepeat: [aStream tab].
	self printOn: aStream.
	aStream cr.
	self submorphsDo: [:m | m printStructureOn: aStream indent: tabCount + 1].
! !

!Morph methodsFor: 'private' stamp: 'AlainPlantec 10/3/2011 01:21'!
privateAddAllMorphs: aCollection atIndex: index
	"Private. Add aCollection of morphs to the receiver"
	| myWorld otherSubmorphs |
	myWorld := self world.
	otherSubmorphs := submorphs copyWithoutAll: aCollection.
	(index between: 0 and: otherSubmorphs size)
		ifFalse: [^ self error: 'index out of range'].
	index = 0
		ifTrue:[	submorphs := aCollection asArray, otherSubmorphs]
		ifFalse:[	index = otherSubmorphs size
			ifTrue:[	submorphs := otherSubmorphs, aCollection]
			ifFalse:[	submorphs := otherSubmorphs copyReplaceFrom: index + 1 to: index with: aCollection ]].
	aCollection do: [:m | | itsOwner itsWorld |
		itsOwner := m owner.
		itsOwner ifNotNil: [
			itsWorld := m world.
			(itsWorld == myWorld) ifFalse: [
				itsWorld ifNotNil: [self privateInvalidateMorph: m].
				m outOfWorld: itsWorld].
			(itsOwner ~~ self) ifTrue: [
				m owner privateRemove: m.
				m owner removedMorph: m ]].
		m privateOwner: self.
		myWorld ifNotNil: [self privateInvalidateMorph: m].
		(myWorld == itsWorld) ifFalse: [m intoWorld: myWorld].
		itsOwner == self ifFalse: [
			self addedMorph: m.
			m noteNewOwner: self ].
	].
	self layoutChanged.
! !

!Morph methodsFor: 'private' stamp: 'MarcusDenker 10/21/2013 14:22'!
privateAddMorph: aMorph atIndex: index
	| oldIndex myWorld itsWorld oldOwner |
	(index between: 1 and: submorphs size + 1)
		ifFalse: [ ^ self error: 'index out of range' ].
	myWorld := self world.
	oldOwner := aMorph owner.
	(oldOwner == self and: [ (oldIndex := submorphs indexOf: aMorph) > 0 ])
		ifTrue: [ 
			"aMorph's position changes within in the submorph chain"
			oldIndex < index
				ifTrue: [ 
					"moving aMorph to back"
					submorphs
						replaceFrom: oldIndex
						to: index - 2
						with: submorphs
						startingAt: oldIndex + 1.
					submorphs at: index - 1 put: aMorph ]
				ifFalse: [ 
					"moving aMorph to front"
					oldIndex - 1 to: index by: -1 do: [ :i | submorphs at: i + 1 put: (submorphs at: i) ].
					submorphs at: index put: aMorph ] ]
		ifFalse: [ 
			"adding a new morph"
			oldOwner
				ifNotNil: [ 
					itsWorld := aMorph world.
					itsWorld ifNotNil: [ self privateInvalidateMorph: aMorph ].
					itsWorld == myWorld
						ifFalse: [ aMorph outOfWorld: itsWorld ].
					oldOwner privateRemove: aMorph.
					oldOwner removedMorph: aMorph ].
			aMorph privateOwner: self.
			submorphs := submorphs copyReplaceFrom: index to: index - 1 with: (Array with: aMorph).
			itsWorld == myWorld
				ifFalse: [ aMorph intoWorld: myWorld ] ].
	myWorld ifNotNil: [ self privateInvalidateMorph: aMorph ].
	self layoutChanged.
	oldOwner == self
		ifFalse: [ 
			self addedMorph: aMorph.
			aMorph noteNewOwner: self ]! !

!Morph methodsFor: 'private' stamp: ''!
privateBounds: boundsRect
	"Private!! Use position: and/or extent: instead."

	fullBounds := nil.
	bounds := boundsRect.! !

!Morph methodsFor: 'private' stamp: 'jm 5/29/1998 21:28'!
privateColor: aColor

	color := aColor.
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 8/10/2003 18:31'!
privateDelete
	"Remove the receiver as a submorph of its owner"
	owner ifNotNil:[owner removeMorph: self].! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:46'!
privateExtension: aMorphExtension
	"private - change the receiver's extension"
	extension := aMorphExtension! !

!Morph methodsFor: 'layout' stamp: 'IgorStasenko 12/22/2012 03:11'!
privateFullBounds
	"Private. Compute the actual full bounds of the receiver"

	| box |
	submorphs isEmpty ifTrue: [^self outerBounds].
	box := self outerBounds copy.
	box := box quickMerge: (self clipSubmorphs 
						ifTrue: [self submorphBounds intersect: self clippingBounds ifNone: [ self clippingBounds ]]
						ifFalse: [self submorphBounds]).
	^box origin asIntegerPoint corner: box corner asIntegerPoint! !

!Morph methodsFor: 'private' stamp: 'tk 8/30/1998 09:58'!
privateFullBounds: boundsRect
	"Private!! Computed automatically."

	fullBounds := boundsRect.! !

!Morph methodsFor: 'private' stamp: 'MarcusDenker 10/22/2013 14:56'!
privateFullMoveBy: delta
	"Private!! Relocate me and all of my subMorphs by recursion. Subclasses that implement different coordinate systems may override this method."

	self privateMoveBy: delta.
	submorphs do: [:each | each privateFullMoveBy: delta].
! !

!Morph methodsFor: 'change reporting' stamp: 'ar 8/12/2003 22:26'!
privateInvalidateMorph: aMorph
	"Private. Invalidate the given morph after adding or removing.
	This method is private because a) we're invalidating the morph 'remotely'
	and b) it forces a fullBounds computation which should not be necessary
	for a general morph c) the morph may or may not actually invalidate
	anything (if it's not in the world nothing will happen) and d) the entire
	mechanism should be rewritten."
	aMorph fullBounds.
	aMorph changed! !

!Morph methodsFor: 'private' stamp: ''!
privateOwner: aMorph
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."

	owner := aMorph.! !

!Morph methodsFor: 'private' stamp: 'di 10/18/2004 21:49'!
privateRemove: aMorph
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."

	submorphs := submorphs copyWithout: aMorph.
	self layoutChanged.! !

!Morph methodsFor: 'private' stamp: ''!
privateSubmorphs
	"Private!! Use 'submorphs' instead."

	^ submorphs! !

!Morph methodsFor: 'private' stamp: ''!
privateSubmorphs: aCollection
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."

	submorphs := aCollection.! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/13/2000 17:14'!
processEvent: anEvent
	"Process the given event using the default event dispatcher."
	^self processEvent: anEvent using: self defaultEventDispatcher! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/18/2000 19:14'!
processEvent: anEvent using: defaultDispatcher
	"This is the central entry for dispatching events in morphic. Given some event and a default dispatch strategy, find the right receiver and let him handle it.
	WARNING: This is a powerful hook. If you want to use a different event dispatcher from the default, here is the place to hook it in. Depending on how the dispatcher is written (e.g., whether it calls simply #processEvent: or #processEvent:using:) you can change the dispatch strategy for entire trees of morphs. Similarly, you can disable entire trees of morphs from receiving any events whatsoever. Read the documentation in class MorphicEventDispatcher before playing with it. "
	(self rejectsEvent: anEvent) ifTrue:[^#rejected].
	^defaultDispatcher dispatchEvent: anEvent with: self! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/12/2000 11:16'!
programmedMouseDown: anEvent for: aMorph

	aMorph addMouseActionIndicatorsWidth: 15 color: (Color blue alpha: 0.7).

! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/12/2000 11:16'!
programmedMouseEnter: anEvent for: aMorph

	aMorph addMouseActionIndicatorsWidth: 10 color: (Color blue alpha: 0.3).

! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/12/2000 11:10'!
programmedMouseLeave: anEvent for: aMorph

	self deleteAnyMouseActionIndicators.
! !

!Morph methodsFor: 'debug and other' stamp: 'MarcusDenker 2/25/2010 08:47'!
programmedMouseUp: anEvent for: aMorph 
	| aCodeString |
	self deleteAnyMouseActionIndicators.
	aCodeString := self valueOfProperty: #mouseUpCodeToRun ifAbsent: [^self].
	(self fullBounds containsPoint: anEvent cursorPoint) ifFalse: [^self].
	aCodeString value.
		! !

!Morph methodsFor: 'accessing' stamp: 'dgd 3/7/2003 15:24'!
raisedColor
	"Return the color to be used for shading raised borders. The 
	default is my own color, but it might want to be, eg, my 
	owner's color. Whoever's color ends up prevailing, the color 
	itself gets the last chance to determine, so that when, for 
	example, an InfiniteForm serves as the color, callers won't choke 
	on some non-Color object being returned"
	(color isColor
			and: [color isTransparent
					and: [owner notNil]])
		ifTrue: [^ owner raisedColor].
	^ color asColor raisedColor! !

!Morph methodsFor: 'geometry' stamp: 'ar 9/22/2000 20:12'!
referencePosition
	"Return the current reference position of the receiver"
	| box |
	box := self bounds.
	^box origin + (self rotationCenter * box extent).
! !

!Morph methodsFor: 'geometry' stamp: 'ar 9/27/2000 14:04'!
referencePosition: aPosition
	"Move the receiver to match its reference position with aPosition"
	| newPos intPos |
	newPos := self position + (aPosition - self referencePosition).
	intPos := newPos asIntegerPoint.
	newPos = intPos 
		ifTrue:[self position: intPos]
		ifFalse:[self position: newPos].! !

!Morph methodsFor: 'geometry' stamp: 'sw 10/25/1999 16:49'!
referencePositionInWorld

	^ self pointInWorld: self referencePosition
! !

!Morph methodsFor: 'geometry' stamp: 'sw 10/25/1999 23:33'!
referencePositionInWorld: aPoint
	| localPosition |
	localPosition := owner
		ifNil: [aPoint]
		ifNotNil: [(owner transformFrom: self world) globalPointToLocal: aPoint].

	self referencePosition: localPosition
! !

!Morph methodsFor: 'drawing' stamp: 'sw 10/10/1999 23:25'!
refreshWorld
	| aWorld |
	(aWorld := self world) ifNotNil: [aWorld displayWorldSafely]
! !

!Morph methodsFor: 'events-processing' stamp: 'ar 10/5/2000 19:25'!
rejectDropEvent: anEvent
	"This hook allows the receiver to repel a drop operation currently executed. The method is called prior to checking children so the receiver must validate that the event was really designated for it.
	Note that the ordering of the tests below is designed to avoid a (possibly expensive) #fullContainsPoint: test. If the receiver doesn't want to repel the morph anyways we don't need to check after all."
	(self repelsMorph: anEvent contents event: anEvent) ifFalse:[^self]. "not repelled"
	(self fullContainsPoint: anEvent position) ifFalse:[^self]. "not for me"
	"Throw it away"
	anEvent wasHandled: true.
	anEvent contents rejectDropMorphEvent: anEvent.! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'AlainPlantec 12/1/2010 09:35'!
rejectDropMorphEvent: evt
	"The receiver has been rejected, and must be put back somewhere.  There are three cases:
	(1)  It remembers its former owner and position, and goes right back there
	(2)  It remembers its former position only, in which case it was torn off from a parts bin, and the UI is that it floats back to its donor position and then vanishes."

	self formerOwner notNil 
		ifTrue: [^ self slideBackToFormerSituation: evt].

	self formerPosition  "Position but no owner -- can just make it vanish"
		ifNotNil: [^ self vanishAfterSlidingTo: self formerPosition event: evt].! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 19:47'!
relativeTextAnchorPosition
	^self valueOfProperty: #relativeTextAnchorPosition! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 19:22'!
relativeTextAnchorPosition: aPoint
	^self setProperty: #relativeTextAnchorPosition toValue: aPoint! !

!Morph methodsFor: 'events-removing' stamp: 'rw 4/25/2002 07:18'!
releaseActionMap
	"Release the action map"
	
 	self removeProperty: #actionMap! !

!Morph methodsFor: 'caching' stamp: 'md 4/3/2006 12:02'!
releaseCachedState
	"Release any state that can be recomputed on demand, such as the pixel values for a color gradient or the editor state for a TextMorph. This method may be called to save space when a morph becomes inaccessible. Implementations of this method should do 'super releaseCachedState'."
	self borderStyle releaseCachedState. 
! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/14/2000 12:14'!
removeAlarm: aSelector
	"Remove the given alarm"
	| scheduler |
	scheduler := self alarmScheduler.
	scheduler ifNotNil:[scheduler removeAlarm: aSelector for: self].! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'nk 10/16/2003 14:08'!
removeAllMorphs
	| oldMorphs myWorld |
	myWorld := self world.
	(fullBounds notNil or:[myWorld notNil]) ifTrue:[self invalidRect: self fullBounds].
	submorphs do: [:m | myWorld ifNotNil: [ m outOfWorld: myWorld ]. m privateOwner: nil].
	oldMorphs := submorphs.
	submorphs := EmptyArray.
	oldMorphs do: [ :m | self removedMorph: m ].
	self layoutChanged.
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'nk 10/16/2003 14:02'!
removeAllMorphsIn: aCollection
	"greatly speeds up the removal of *lots* of submorphs"
	| set myWorld |
	set := IdentitySet new: aCollection size * 4 // 3.
	aCollection do: [:each | each owner == self ifTrue: [ set add: each]].
	myWorld := self world.
	(fullBounds notNil or:[myWorld notNil]) ifTrue:[self invalidRect: self fullBounds].
	set do: [:m | myWorld ifNotNil: [ m outOfWorld: myWorld ]. m privateOwner: nil].
	submorphs := submorphs reject: [ :each | set includes: each].
	set do: [ :m | self removedMorph: m ].
	self layoutChanged.
! !

!Morph methodsFor: 'updating' stamp: 'BenjaminVanRyseghem 7/31/2013 14:14'!
removeDependent: anObject

	self flag: #GSoC. "Flag added by Benjamin Van Ryseghem the July 22, 2013 to remember to remove this line in a while"
	[ super removeDependent: anObject ]
		on: Error
		do: [].
	
	self announcer unsubscribe: anObject! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 11/12/2000 18:57'!
removeDropShadow
	self hasDropShadow ifFalse:[^self].
	self changed.
	self hasDropShadow: false.
	fullBounds ifNotNil:[fullBounds := self privateFullBounds].
	self changed.! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'AlainPlantec 5/8/2010 00:33'!
removeFlexShell
	self isFlexed ifTrue: [self owner removeFlexShell]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ClementBera 7/30/2013 11:04'!
removeHalo
	"remove the surrounding halo (if any)"
	self halo ifNotNil: [self primaryHand removeHalo]! !

!Morph methodsFor: 'event handling' stamp: 'tk 8/10/1998 16:02'!
removeLink: actionCode
	self eventHandler ifNotNil:
		[self eventHandler on: actionCode send: nil to: nil]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'di 10/18/2004 21:50'!
removeMorph: aMorph
	"Remove the given morph from my submorphs"
	| aWorld |
	aMorph owner == self ifFalse:[^self].
	aWorld := self world.
	aWorld ifNotNil:[
		aMorph outOfWorld: aWorld.
		self privateInvalidateMorph: aMorph.
	].
	self privateRemove: aMorph.
	aMorph privateOwner: nil.
	self removedMorph: aMorph.
! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/7/2000 16:43'!
removeMouseUpAction

	self primaryHand showTemporaryCursor: nil.
	self removeProperty: #mouseUpCodeToRun.
	#(mouseUp mouseEnter mouseLeave mouseDown) do: [ :sym |
		self
			on: sym 
			send: #yourself 
			to: nil.
	]

! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:48'!
removeProperty: aSymbol 
	"removes the property named aSymbol if it exists"
	extension ifNil:  [^ self].
	extension removeProperty: aSymbol! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 8/12/2003 22:01'!
removedMorph: aMorph
	"Notify the receiver that aMorph was just removed from its children"
! !

!Morph methodsFor: 'testing' stamp: 'stephane.ducasse 11/14/2008 21:48'!
renameTo: aName 
	"Set The morph name."

	self topRendererOrSelf setNameTo: aName.
	^aName! !

!Morph methodsFor: 'structure' stamp: 'wiz 12/7/2006 15:12'!
renderedMorph
	"This now  gets overridden by rendering morphs."

	^self! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'sw 1/11/1999 20:07'!
repelsMorph: aMorph event: ev
	^ false! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 10/25/1999 23:34'!
replaceSubmorph: oldMorph by: newMorph
	| index itsPosition w |
	oldMorph stopStepping.
	itsPosition := oldMorph referencePositionInWorld.
	index := submorphs indexOf: oldMorph.
	oldMorph privateDelete.
	self privateAddMorph: newMorph atIndex: index.
	newMorph referencePositionInWorld: itsPosition.
	(w := newMorph world) ifNotNil:
		[w startSteppingSubmorphsOf: newMorph]! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:47'!
resetExtension
	"reset the extension slot if it is not needed"
	(extension notNil and: [extension isDefault]) ifTrue: [extension := nil] ! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 12:02'!
resetHighlightForDrop
	self highlightForDrop: false! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/15/2001 16:33'!
resistsRemoval
	"Answer whether the receiver is marked as resisting removal"

	^ self hasProperty: #resistsRemoval! !

!Morph methodsFor: 'accessing' stamp: 'StephaneDucasse 4/22/2012 16:35'!
resistsRemoval: aBoolean
	"Set the receiver's resistsRemoval property as indicated"

	aBoolean
		ifTrue: [ self setProperty: #resistsRemoval toValue: true ]
		ifFalse: [ self removeProperty: #resistsRemoval ]! !

!Morph methodsFor: 'menus' stamp: 'StephaneDucasse 4/22/2012 16:53'!
resistsRemovalString
	"Answer the string to be shown in a menu to represent the 
	'resistsRemoval' status"
	^ (self resistsRemoval) -> 'resist being deleted' translated! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 11/27/2001 14:59'!
resizeFromMenu
	"Commence an interaction that will resize the receiver"

	self resizeMorph: ActiveEvent! !

!Morph methodsFor: 'meta-actions' stamp: 'StephaneDucasse 6/28/2013 11:26'!
resizeMorph: evt
	| handle |
	handle := HandleMorph new forEachPointDo: [ :newPoint | self extent: newPoint - self bounds topLeft ].
	evt hand attachMorph: handle.
	handle startStepping! !

!Morph methodsFor: 'event handling' stamp: 'sw 11/16/1998 08:06'!
restoreSuspendedEventHandler
	| savedHandler |
	(savedHandler := self valueOfProperty: #suspendedEventHandler) ifNotNil:
		[self eventHandler: savedHandler].
	submorphs do: [:m | m restoreSuspendedEventHandler]
! !

!Morph methodsFor: 'debug and other' stamp: 'CamilloBruni 5/29/2012 16:51'!
resumeAfterDrawError

	self changed.
	self removeProperty:#errorOnDraw.
	self removeProperty:#drawError.
	self changed.! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 5/24/2000 18:20'!
resumeAfterStepError
	"Resume stepping after an error has occured."

	self startStepping. "Will #step"
	self removeProperty:#errorOnStep. "Will remove prop only if #step was okay"
! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:01'!
reverseTableCells
	"Layout specific. This property describes if the cells should be treated in reverse order of submorphs."
	| props |
	props := self layoutProperties.
	^props ifNil:[false] ifNotNil:[props reverseTableCells].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
reverseTableCells: aBool
	"Layout specific. This property describes if the cells should be treated in reverse order of submorphs."
	self assureTableProperties reverseTableCells: aBool.
	self layoutChanged.! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:16'!
right
	" Return the x-coordinate of my right side "
	^ bounds right! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:15'!
right: aNumber
	" Move me so that my right side is at the x-coordinate aNumber. My extent (width & height) are unchanged "

	self position: ((aNumber - bounds width) @ bounds top)! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
rightCenter

	^ bounds rightCenter! !

!Morph methodsFor: 'structure' stamp: 'dgd 2/22/2003 14:34'!
root
	"Return the root of the composite morph containing the receiver. The owner of the root is either nil, a WorldMorph, or a HandMorph. If the receiver's owner is nil, the root is the receiver itself. This method always returns a morph."

	(owner isNil or: [owner isWorldOrHandMorph]) ifTrue: [^self].
	^owner root! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 10/8/2000 15:44'!
rootMorphsAt: aPoint
	"Return the list of root morphs containing the given point, excluding the receiver.
	ar 11/8/1999: Moved into morph for an incredibly ugly hack in 3D worlds"
self flag: #arNote. "check this at some point"
	^ self submorphs select:
		[:m | (m fullContainsPoint: aPoint) and: [m isLocked not]]! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'AlainPlantec 5/8/2010 00:13'!
rotationCenter
	"Return the rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	^self valueOfProperty: #rotationCenter ifAbsent: [0.5@0.5]
! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'ClementBera 7/30/2013 11:05'!
rotationCenter: aPointOrNil
	"Set the new rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	aPointOrNil 
		ifNil: [self removeProperty: #rotationCenter]
		ifNotNil:[self setProperty: #rotationCenter toValue: aPointOrNil]
! !

!Morph methodsFor: 'rounding' stamp: 'AlainPlantec 12/20/2009 00:20'!
roundedCornersString
	"Answer the string to put in a menu that will invite the user to 
	switch to the opposite corner-rounding mode"
	^ (self wantsRoundedCorners
		ifTrue: ['<yes>' translated]
		ifFalse: ['<no>' translated])
		, ('round corners' translated)! !

!Morph methodsFor: 'Morphic-Base-MorphTreeWidget' stamp: 'AlainPlantec 9/30/2011 15:59'!
rowMorphForNode: aNode inColumn: aColumn
	| rm |
	rm := Morph new color: Color transparent;
			layoutPolicy: RowLayout new;
			hResizing: #shrinkWrap;
			vResizing: #shrinkWrap;
			layoutInset: aColumn container columnInset @ aColumn container rowInset;
			listDirection: #leftToRight;
			cellPositioning: #leftCenter;
			cellInset: 4@0;
			yourself.
			
	rm addMorph: self.
	
	aColumn isFirstColumn 
		ifTrue: [ | icon |
			icon := aColumn container iconBlock value: aNode.
			icon ifNotNil: [ rm addMorph: icon asMorph ].
			"for first column we don't use horizontal inset"
			rm  layoutInset: 0 @ aColumn container rowInset.
		].

	^ rm
! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:01'!
rubberBandCells
	"Layout specific. This property describes if a parent that is #shrinkWrapped around its children should ignore any #spaceFill children. E.g., when #rubberBandCells is true, the compound layout will always stay at the smallest available size, even though some child may be able to grow."
	| props |
	props := self layoutProperties.
	^props ifNil:[false] ifNotNil:[props rubberBandCells].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
rubberBandCells: aBool
	"Layout specific. This property describes if a parent that is #shrinkWrapped around its children should ignore any #spaceFill children. E.g., when #rubberBandCells is true, the compound layout will always stay at the smallest available size, even though some child may be able to grow."
	self assureTableProperties rubberBandCells: aBool.
	self layoutChanged.! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'nk 9/4/2004 11:00'!
scale: newScale
	"Backstop for morphs that don't have to do something special to set their scale"
! !

!Morph methodsFor: 'accessing' stamp: 'nk 9/4/2004 10:49'!
scaleFactor
	^self valueOfProperty: #scaleFactor ifAbsent: [ 1.0 ]
! !

!Morph methodsFor: 'geometry' stamp: 'bf 1/5/2000 19:08'!
screenLocation
	"For compatibility only"

	^ self fullBounds origin! !

!Morph methodsFor: 'geometry' stamp: 'sma 2/5/2000 13:58'!
screenRectangle
	"For compatibility only"

	^ self fullBounds! !

!Morph methodsFor: 'selected object' stamp: 'dgd 8/28/2004 16:30'!
selectedObject
	"answer the selected object for the hand or nil is none"
	^ self primaryHand selectedObject! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:24'!
separateDragAndDrop
	"Conversion only. Separate the old #dragNDropEnabled into #dragEnabled and #dropEnabled and remove the old property."
	| dnd |
	(self hasProperty: #dragNDropEnabled) ifFalse:[^self].
	dnd := (self valueOfProperty: #dragNDropEnabled) == true.
	self dragEnabled: dnd.
	self dropEnabled: dnd.
	self removeProperty: #dragNDropEnabled.
! !

!Morph methodsFor: 'menus' stamp: 'CamilleTeruel 12/6/2013 14:38'!
setArrowheads
	"Let the user edit the size of arrowheads for this object"

	| aParameter result  |
	aParameter := self renderedMorph valueOfProperty:  #arrowSpec ifAbsent: [PolygonMorph defaultArrowSpec].
	result := Morph obtainArrowheadFor: 'Head size for arrowheads: ' translated defaultValue: aParameter asString.
	result ifNotNil:
			[self renderedMorph  setProperty: #arrowSpec toValue: result]
		ifNil:
			[ self inform: 'Invalid input']! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 12/19/2009 23:40'!
setBalloonText: stringOrText
	"Set receiver's balloon help text. Pass nil to remove the help."

	self setBalloonText: stringOrText maxLineLength: self theme settings maxBalloonHelpLineLength! !

!Morph methodsFor: 'halos and balloon help' stamp: 'md 2/27/2006 09:54'!
setBalloonText: stringOrText maxLineLength: aLength 
	"Set receiver's balloon help text. Pass nil to remove the help."
	(extension isNil and: [stringOrText isNil]) ifTrue: [^ self].
	self assureExtension balloonText: 
		(stringOrText ifNotNil: [stringOrText asString withNoLineLongerThan: aLength])! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/26/2001 16:16'!
setBorderStyle: aSymbol
	"Set the border style of my costume"

	| aStyle |
	aStyle := self borderStyleForSymbol: aSymbol.
	aStyle ifNil: [^ self].
	(self canDrawBorder: aStyle)
		ifTrue:
			[self borderStyle: aStyle]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 10/29/1999 17:38'!
setCenteredBalloonText: aString
	self setBalloonText: aString.
	self setProperty: #helpAtCenter toValue: true! !

!Morph methodsFor: 'geometry' stamp: 'tk 7/14/2001 11:11'!
setConstrainedPosition: aPoint hangOut: partiallyOutside
	"Change the position of this morph and and all of its submorphs to aPoint, but don't let me go outside my owner's bounds.  Let me go within two pixels of completely outside if partiallyOutside is true."

	| trialRect delta boundingMorph bRect |
	owner ifNil:[^self].
	trialRect := aPoint extent: self bounds extent.
	boundingMorph := self topRendererOrSelf owner.
	delta := boundingMorph
			ifNil:    [0@0]
			ifNotNil: [
				bRect := partiallyOutside 
					ifTrue: [boundingMorph bounds insetBy: 
								self extent negated + boundingMorph borderWidth + (2@2)]
					ifFalse: [boundingMorph bounds].
				trialRect amountToTranslateWithin: bRect].
	self position: aPoint + delta.
	self layoutChanged  "So that, eg, surrounding text will readjust"
! !

!Morph methodsFor: 'miscellaneous' stamp: 'sw 7/20/2001 00:15'!
setExtentFromHalo: anExtent
	"The user has dragged the grow box such that the receiver's extent would be anExtent.  Do what's needed"

	self extent: anExtent! !

!Morph methodsFor: 'naming' stamp: 'dgd 2/16/2003 21:57'!
setNamePropertyTo: aName 
	"change the receiver's externalName"
	self assureExtension externalName: aName! !

!Morph methodsFor: 'naming' stamp: 'yo 12/3/2004 17:02'!
setNameTo: aName 
	| nameToUse nameString |
	nameToUse := aName ifNotNil: 
					[(nameString := aName asString) notEmpty ifTrue: [nameString] ifFalse: ['*']].
	self setNamePropertyTo: nameToUse	"no Texts here!!"! !

!Morph methodsFor: 'accessing - properties' stamp: 'tk 10/9/2002 08:30'!
setProperties: aList
	"Set many properties at once from a list of prop, value, prop, value"

	1 to: aList size by: 2 do: [:ii |
		self setProperty: (aList at: ii) toValue: (aList at: ii+1)].! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:48'!
setProperty: aSymbol toValue: anObject 
	"change the receiver's property named aSymbol to anObject"
	anObject ifNil: [^ self removeProperty: aSymbol].
	self assureExtension setProperty: aSymbol toValue: anObject! !

!Morph methodsFor: 'menus' stamp: 'ar 9/22/2000 20:14'!
setRotationCenterFrom: aPoint
	self rotationCenter: (aPoint - self bounds origin) / self bounds extent asFloatPoint.! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 18:58'!
setShadowOffset: evt
	| handle |
	handle := HandleMorph new forEachPointDo:
		[:newPoint | self shadowPoint: newPoint].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'menus' stamp: 'di 12/21/2000 17:18'!
setToAdhereToEdge: anEdge
	anEdge ifNil: [^ self].
	anEdge == #none ifTrue: [^ self removeProperty: #edgeToAdhereTo].
	self setProperty: #edgeToAdhereTo toValue: anEdge.
! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 18:59'!
shadowColor
	^self valueOfProperty: #shadowColor ifAbsent:[Color black]! !

!Morph methodsFor: 'drop shadows' stamp: 'AlainPlantec 12/14/2009 11:09'!
shadowColor: aColor
	self shadowColor = aColor ifFalse: [self changed].
	self setProperty: #shadowColor toValue: aColor.! !

!Morph methodsFor: 'drawing' stamp: 'JW 7/12/2005 20:12'!
shadowForm
	"Return a form representing the 'shadow' of the receiver - e.g., all pixels that are occupied by the receiver are one, all others are zero."
	| canvas |
	canvas := (Display defaultCanvasClass extent: self fullBounds extent depth: 1)
				asShadowDrawingCanvas: Color black. "Color black represents one for 1bpp"
	canvas translateBy: bounds topLeft negated
		during:[:tempCanvas| tempCanvas fullDrawMorph: self].
	^ canvas form offset: bounds topLeft
! !

!Morph methodsFor: 'drop shadows' stamp: 'marcus.denker 8/24/2008 22:50'!
shadowOffset
	"Return the current shadow offset"
	
	extension ifNil: [^0@0].
	^self valueOfProperty: #shadowOffset ifAbsent:[0@0]! !

!Morph methodsFor: 'drop shadows' stamp: 'ClementBera 9/30/2013 11:00'!
shadowOffset: aPoint
	"Set the current shadow offset"
	(aPoint isNil or:[(aPoint x isZero) and: [ aPoint y isZero ]])
		ifTrue:[ self removeProperty: #shadowOffset ]
		ifFalse:[ self setProperty: #shadowOffset toValue: aPoint ].! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 11/12/2000 18:58'!
shadowPoint: newPoint
	self changed.
	self shadowOffset: newPoint - self center // 5.
	fullBounds ifNotNil:[fullBounds := self privateFullBounds].
	self changed.! !

!Morph methodsFor: 'geometry' stamp: 'sw 2/16/1999 22:05'!
shiftSubmorphsOtherThan: listNotToShift by: delta
	| rejectList |
	rejectList := listNotToShift ifNil: [OrderedCollection new].
	(submorphs copyWithoutAll: rejectList) do:
		[:m | m position: (m position + delta)]! !

!Morph methodsFor: 'testing' stamp: 'ar 12/3/2001 12:33'!
shouldDropOnMouseUp
	| former |
	former := self formerPosition ifNil:[^false].
	^(former dist: self position) > 10! !

!Morph methodsFor: 'testing' stamp: 'WilliamSix 1/14/2013 19:43'!
shouldFlex
	^ self isFlexMorph.! !

!Morph methodsFor: 'wiw support' stamp: 'ar 3/18/2001 00:14'!
shouldGetStepsFrom: aWorld
	^self world == aWorld! !

!Morph methodsFor: 'drawing' stamp: 'MarcusDenker 10/21/2013 14:22'!
show
	"Make sure this morph is on-stage."

	self visible
		ifTrue: [ ^ self ].
	self visible: true.
	self changed! !

!Morph methodsFor: 'meta-actions' stamp: 'BernardoContreras 1/18/2012 23:06'!
showActions
	"Put up a message list browser of all the code that this morph  
	would run for mouseUp, mouseDown, mouseMove, mouseEnter,  
	mouseLeave, and mouseLinger."
	
	| list cls selector adder |
	list := SortedCollection new.
	adder := [:mrClass :mrSel | list
				add: (RGMethodDefinition realClass: mrClass selector: mrSel)].
	"the eventHandler"
	self eventHandler
		ifNotNil: [list := self eventHandler methodRefList.
			(self eventHandler handlesMouseDown: nil)
				ifFalse: [adder value: HandMorph value: #grabMorph:]].
	"If not those, then non-default raw events"
	#(#keyStroke: #mouseDown: #mouseEnter: #mouseLeave: #mouseMove: #mouseUp: #doButtonAction )
		do: [:sel | 
			cls := self class whichClassIncludesSelector: sel.
			cls
				ifNotNil: ["want more than default behavior"
					cls == Morph
						ifFalse: [adder value: cls value: sel]]].
	"The mechanism on a Button"
	(self respondsTo: #actionSelector)
		ifTrue: ["A button"
			selector := self actionSelector.
			cls := self target class whichClassIncludesSelector: selector.
			cls
				ifNotNil: ["want more than default behavior"
					cls == Morph
						ifFalse: [adder value: cls value: selector]]].
	Smalltalk tools messageList openMessageList: list name: 'Actions of ' , self printString autoSelect: ''! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 10/3/2000 17:06'!
showBalloon: msgString
	"Pop up a balloon containing the given string,
	first removing any existing BalloonMorphs in the world."
	| w |
	self showBalloon: msgString hand: ((w := self world) ifNotNil:[w activeHand]).! !

!Morph methodsFor: 'accessing' stamp: 'wiz 11/6/2005 17:10'!
simplySetVisible: aBoolean
	"Set the receiver's visibility property.  This mild circumlocution is because my TransfomationMorph #visible: method would also set the visibility flag of my flexee, which in this case is pointless because it's the flexee that calls this.
	This appears in morph as a backstop for morphs that don't inherit from TFMorph"

	self visible: aBoolean! !

!Morph methodsFor: 'testing' stamp: 'BenjaminVanRyseghem 11/28/2012 00:58'!
simulateKeyStroke: aCharacter
	| event |

	event := KeyboardEvent new 
				setType: #keystroke 
				buttons: 0 
				position: 0@0 
				keyValue: aCharacter charCode 
				charCode: aCharacter charCode
				hand: ActiveHand 
				stamp: 0.
	self keyStroke: event! !

!Morph methodsFor: 'testing' stamp: 'BenjaminVanRyseghem 11/28/2012 00:59'!
simulateKeyStrokes: aString

	aString do: [:c | self simulateKeyStroke: c ].! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 8/12/2003 23:35'!
slideBackToFormerSituation: evt 
	| slideForm formerOwner formerPosition aWorld startPoint endPoint trans |
	formerOwner := self formerOwner.
	formerPosition := self formerPosition.
	aWorld := evt hand world.
	trans := formerOwner transformFromWorld.
	slideForm := trans isPureTranslation 
				ifTrue: [self imageForm offset: 0 @ 0]
				ifFalse: 
					[((TransformationMorph new asFlexOf: self) transform: trans) imageForm 
						offset: 0 @ 0]. 
	startPoint := evt hand fullBounds origin.
	endPoint := trans localPointToGlobal: formerPosition.
	owner removeMorph: self.
	aWorld displayWorld.
	slideForm 
		slideFrom: startPoint
		to: endPoint
		nSteps: 12
		delay: 15.
	formerOwner addMorph: self.
	self position: formerPosition.
	self justDroppedInto: formerOwner event: evt! !

!Morph methodsFor: 'menus' stamp: 'sw 8/30/1998 09:42'!
snapToEdgeIfAppropriate
	| edgeSymbol oldBounds aWorld |
	(edgeSymbol := self valueOfProperty: #edgeToAdhereTo) ifNotNil:
		[oldBounds := bounds.
		self adhereToEdge: edgeSymbol.
		bounds ~= oldBounds ifTrue: [(aWorld := self world) ifNotNil: [aWorld viewBox ifNotNil:
			[aWorld displayWorld]]]]! !

!Morph methodsFor: 'layout-properties' stamp: 'dgd 2/16/2003 20:02'!
spaceFillWeight
	"Layout specific. This property describes the relative weight that 
	should be given to the receiver when extra space is distributed 
	between different #spaceFill cells."

	^ self
		valueOfProperty: #spaceFillWeight
		ifAbsent: [1]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/15/2000 14:16'!
spaceFillWeight: aNumber
	"Layout specific. This property describes the relative weight that should be given to the receiver when extra space is distributed between different #spaceFill cells."
	aNumber = 1
		ifTrue:[self removeProperty: #spaceFillWeight]
		ifFalse:[self setProperty: #spaceFillWeight toValue: aNumber].
	self layoutChanged.! !

!Morph methodsFor: 'stepping and presenter' stamp: ''!
start
	"Start running my script. For ordinary morphs, this means start stepping."

	self startStepping.
! !

!Morph methodsFor: 'event handling' stamp: 'mir 5/23/2000 17:43'!
startDrag: evt
	"Handle a double-click event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing."

	self eventHandler ifNotNil:
		[self eventHandler startDrag: evt fromMorph: self].! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'mir 1/4/2001 11:02'!
startDrag: anItem with: anObject
	self currentHand attachMorph: anObject! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 1/31/2001 13:07'!
startStepping
	"Start getting sent the 'step' message."
	self startStepping: #stepAt: at: Time millisecondClockValue arguments: nil stepTime: nil.! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 10/22/2000 16:36'!
startStepping: aSelector at: scheduledTime arguments: args stepTime: stepTime
	"Start stepping the receiver"
	| w |
	w := self world.
	w ifNotNil: [
		w startStepping: self at: scheduledTime selector: aSelector arguments: args stepTime: stepTime.
		self changed].! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 7/19/1998 11:51'!
startSteppingIn: aWorld
	"Start getting sent the 'step' message in aWorld"

	self step.  "one to get started!!"
	aWorld ifNotNil: [aWorld startStepping: self].
	self changed! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 10/22/2000 16:42'!
startSteppingSelector: aSelector
	"Start getting sent the 'step' message."
	self startStepping: aSelector at: Time millisecondClockValue arguments: nil stepTime: nil.! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 2/12/2001 17:04'!
step
	"Do some periodic activity. Use startStepping/stopStepping to start and stop getting sent this message. The time between steps is specified by this morph's answer to the stepTime message.  The generic version dispatches control to the player, if any.  The nasty circumlocation about owner's transformation is necessitated by the flexing problem that the player remains in the properties dictionary both of the flex and the real morph.  In the current architecture, only the top renderer's pointer to the player should actually be honored for the purpose of firing."
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'stephane.ducasse 11/27/2008 22:31'!
stepAt: millisecondClockValue
	"Do some periodic activity. Use startStepping/stopStepping to start and stop getting sent this message. The time between steps is specified by this morph's answer to the stepTime message.
	The millisecondClockValue parameter gives the value of the millisecond clock at the moment of dispatch.
	Default is to dispatch to the parameterless step method for the morph, but this protocol makes it possible for some morphs to do differing things depending on the clock value"

	self step
! !

!Morph methodsFor: 'menus' stamp: 'StephaneDucasse 4/22/2012 16:35'!
stickinessString
	"Answer the string to be shown in a menu to represent the stickiness status"
	
	^ (self isSticky) ->  'resist being picked up' translated! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 08:33'!
sticky: aBoolean 
	"change the receiver's sticky property"
	extension sticky: aBoolean! !

!Morph methodsFor: 'stepping and presenter' stamp: ''!
stop
	"Stop running my script. For ordinary morphs, this means stop stepping."

	self stopStepping.
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 12/15/2000 00:00'!
stopStepping
	"Stop getting sent the 'step' message."

	| w |
	w := self world.
	w ifNotNil: [w stopStepping: self].
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 12/15/2000 00:00'!
stopSteppingSelector: aSelector
	"Stop getting sent the given message."
	| w |
	w := self world.
	w ifNotNil: [w stopStepping: self selector: aSelector].
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 10/11/1999 12:59'!
stopSteppingSelfAndSubmorphs
	self allMorphsDo: [:m | m stopStepping]
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'tk 10/20/2000 13:12'!
submorphAfter
	"Return the submorph after (behind) me, or nil"
	| ii |
	owner ifNil: [^ nil].
	^ (ii := owner submorphIndexOf: self) = owner submorphs size 
		ifTrue: [nil]
		ifFalse: [owner submorphs at: ii+1].
	
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'tk 10/20/2000 13:13'!
submorphBefore
	"Return the submorph after (behind) me, or nil"
	| ii |
	owner ifNil: [^ nil].
	^ (ii := owner submorphIndexOf: self) = 1 
		ifTrue: [nil]
		ifFalse: [owner submorphs at: ii-1].
	
! !

!Morph methodsFor: 'layout' stamp: 'nice 1/5/2010 15:59'!
submorphBounds
	"Private. Compute the actual full bounds of the receiver"
	| box |
	submorphs do: [:m | | subBox | 
		(m visible) ifTrue: [
			subBox := m fullBounds.
			box 
				ifNil:[box := subBox copy]
				ifNotNil:[box := box quickMerge: subBox]]].
	box ifNil:[^self bounds]. "e.g., having submorphs but not visible"
	^ box origin asIntegerPoint corner: box corner asIntegerPoint
! !

!Morph methodsFor: 'submorphs-accessing' stamp: ''!
submorphCount

	^ submorphs size! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/1/2000 10:16'!
submorphIndexOf: aMorph
	"Assuming aMorph to be one of my submorphs, answer where it occurs in my submorph list"

	^ submorphs indexOf: aMorph ifAbsent: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 4/9/98 14:26'!
submorphNamed: aName
	^ self submorphNamed: aName ifNone: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'nice 1/5/2010 15:59'!
submorphNamed: aName ifNone: aBlock 
	"Find the first submorph with this name, or a button with an action selector of that name"

	
	self submorphs do: 
			[:button | | sub args | 
			(button respondsTo: #actionSelector) 
				ifTrue: [button actionSelector == aName ifTrue: [^button]].
			((button respondsTo: #arguments) and: [(args := button arguments) notNil]) 
				ifTrue: [(args at: 2 ifAbsent: [nil]) == aName ifTrue: [^button]].
			(button isAlignmentMorph) 
				ifTrue: [(sub := button submorphNamed: aName ifNone: [nil]) ifNotNil: [^sub]]].
	^aBlock value! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ClementBera 10/3/2013 10:29'!
submorphThat: block1 ifNone: block2
	^ submorphs detect: [:m | (block1 value: m) == true] ifNone: block2
	! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 7/3/1998 18:47'!
submorphWithProperty: aSymbol
	^ submorphs detect: [:aMorph | aMorph hasProperty: aSymbol] ifNone: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'efc 8/6/2005 11:35'!
submorphs
	"This method returns my actual submorphs collection. Modifying the collection directly could be dangerous; make a copy if you need to alter it."
	^ submorphs ! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'di 11/4/97 14:30'!
submorphsBehind: aMorph do: aBlock
	| behind |
	behind := false.
	submorphs do:
		[:m | m == aMorph ifTrue: [behind := true]
						ifFalse: [behind ifTrue: [aBlock value: m]]].
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:35'!
submorphsDo: aBlock 
	submorphs do: aBlock! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'di 11/4/97 14:29'!
submorphsInFrontOf: aMorph do: aBlock
	| behind |
	behind := false.
	submorphs do:
		[:m | m == aMorph ifTrue: [behind := true]
						ifFalse: [behind ifFalse: [aBlock value: m]]].
! !

!Morph methodsFor: 'submorphs-accessing' stamp: ''!
submorphsReverseDo: aBlock

	submorphs reverseDo: aBlock.! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 8/15/97 22:03'!
submorphsSatisfying: aBlock
	^ submorphs select: [:m | (aBlock value: m) == true]! !

!Morph methodsFor: 'event handling' stamp: 'sw 11/16/1998 08:07'!
suspendEventHandler
	self eventHandler ifNotNil:
		[self setProperty: #suspendedEventHandler toValue: self eventHandler.
		self eventHandler: nil].
	submorphs do: [:m | m suspendEventHandler].  "All those rectangles"! !

!Morph methodsFor: 'event handling' stamp: 'AlainPlantec 12/19/2009 23:25'!
tabAmongFields
	^ self theme settings tabAmongFields
		or: [self hasProperty: #tabAmongFields] ! !

!Morph methodsFor: 'card in a stack' stamp: 'stephane.ducasse 11/15/2008 14:34'!
tabHitWithEvent: anEvent
	"The tab key was hit.  The keyboard focus has referred this event to me, though this perhaps seems rather backwards.  Anyway, the assumption is that I have the property #tabAmongFields, so now the task is to tab to the next field."

	| currentFocus fieldList anIndex itemToHighlight |
	currentFocus := anEvent hand keyboardFocus.
	fieldList := self allMorphs select:
		[:aMorph | (aMorph wouldAcceptKeyboardFocusUponTab) and: [aMorph isLocked not]].

	fieldList isEmpty ifTrue:[^ self].

	anIndex := fieldList indexOf: currentFocus ifAbsent: [nil].
	itemToHighlight := fieldList atWrap: 
		(anIndex ifNotNil: [anEvent shiftPressed ifTrue: [anIndex - 1] ifFalse: [anIndex + 1]]
				ifNil: [1]).
	anEvent hand newKeyboardFocus: itemToHighlight. self flag: #arNote. "really???"
	itemToHighlight editor selectAll.
	itemToHighlight invalidRect: itemToHighlight bounds ! !

!Morph methodsFor: 'accessing-backstop' stamp: 'wiz 2/14/2006 19:02'!
target: aMorph
"Morphs with targets will override. This backstop does nothing."
"This is here because targeting meta-actions are taken at morph level. 
Do not remove."! !

!Morph methodsFor: 'meta-actions' stamp: 'MarcusDenker 11/28/2009 14:25'!
targetWith: evt
	"Some other morph become target of the receiver"
	|  newTarget |
	newTarget := UIManager default
				chooseFrom: (self potentialTargets
						collect: [:t | t class name asString])
				values: self potentialTargets
				title: (self externalName, ' targets...' translated).
	newTarget ifNil:[^self].
	self target: newTarget.! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 18:36'!
textAnchorType
	^self valueOfProperty: #textAnchorType ifAbsent:[#document]! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 18:37'!
textAnchorType: aSymbol
	aSymbol == #document
		ifTrue:[^self removeProperty: #textAnchorType]
		ifFalse:[^self setProperty: #textAnchorType toValue: aSymbol].! !

!Morph methodsFor: 'rounding' stamp: 'ar 12/25/2001 19:44'!
toggleCornerRounding
	self cornerStyle == #rounded
		ifTrue: [self cornerStyle: #square]
		ifFalse: [self cornerStyle: #rounded].
	self changed! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:46'!
toggleDragNDrop
	"Toggle this morph's ability to add and remove morphs via drag-n-drop."

		self enableDragNDrop: self dragNDropEnabled not.
! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 20:16'!
toggleDropShadow
	self hasDropShadow
		ifTrue:[self removeDropShadow]
		ifFalse:[self addDropShadow].! !

!Morph methodsFor: 'accessing' stamp: 'RAA 2/19/2001 17:38'!
toggleLocked
	
	self lock: self isLocked not! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/15/2001 12:21'!
toggleResistsRemoval
	"Toggle the resistsRemoval property"

	self resistsRemoval
		ifTrue:
			[self removeProperty: #resistsRemoval]
		ifFalse:
			[self setProperty: #resistsRemoval toValue: true]! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 08:53'!
toggleStickiness
	"togle the receiver's Stickiness"
	extension ifNil: [^ self beSticky].
	extension sticky: extension sticky not! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:17'!
top
	" Return the y-coordinate of my top side "

	^ bounds top! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:14'!
top: aNumber
	" Move me so that my top is at the y-coordinate aNumber. My extent (width & height) are unchanged "

	self position: (bounds left @ aNumber)! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
topCenter

	^ bounds topCenter! !

!Morph methodsFor: 'geometry' stamp: 'di 6/12/97 11:07'!
topLeft

	^ bounds topLeft! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:10'!
topLeft: aPoint
	" Move me so that my top left corner is at aPoint. My extent (width & height) are unchanged "

	self position: aPoint
! !

!Morph methodsFor: 'Morphic-Base-Worlds' stamp: 'sw 8/30/1998 09:47'!
topPasteUp
	"If the receiver is in a world, return that; otherwise return the outermost pasteup morph"
	^ self outermostMorphThat: [:m | m isKindOf: PasteUpMorph]! !

!Morph methodsFor: 'structure' stamp: 'dgd 2/22/2003 19:06'!
topRendererOrSelf
	"Answer the topmost renderer for this morph, or this morph itself if it has no renderer. See the comment in Morph>isRenderer."

	| top topsOwner |
	owner ifNil: [^self].
	self isWorldMorph ifTrue: [^self].	"ignore scaling of this world"
	top := self.
	topsOwner := top owner.
	[topsOwner notNil and: [topsOwner isRenderer]] whileTrue: 
			[top := topsOwner.
			topsOwner := top owner].
	^top! !

!Morph methodsFor: 'geometry' stamp: 'sw 8/20/97 23:04'!
topRight

	^ bounds topRight! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:12'!
topRight: aPoint
	" Move me so that my top right corner is at aPoint. My extent (width & height) are unchanged "

	self position: ((aPoint x - bounds width) @ (aPoint y))
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 12/12/2009 07:10'!
transferHalo: event from: formerHaloOwner
	"Progressively transfer the halo to the next likely recipient"
	| localEvt w target |

	self flag: #workAround. "For halo's distinction between 'target' and 'innerTarget' we need to bypass any renderers."
	(formerHaloOwner == self and:[self isRenderer and:[self wantsHaloFromClick not]]) ifTrue:[
		event shiftPressed ifTrue:[
			target := owner.
			localEvt := event transformedBy: (self transformedFrom: owner).
		] ifFalse:[
			target := self renderedMorph.
			localEvt := event transformedBy: (target transformedFrom: self).
		].
		^target transferHalo: localEvt from: target].

"	formerHaloOwner == self ifTrue:[^ self removeHalo]."

	"Never transfer halo to top-most world"
	(self isWorldMorph and:[owner isNil]) ifFalse:[
		(self wantsHaloFromClick and:[formerHaloOwner ~~ self]) 
			ifTrue:[^self addHalo: event from: formerHaloOwner]].

	event shiftPressed ifTrue:[
		"Pass it outwards"
		owner ifNotNil:[^owner transferHalo: event from: formerHaloOwner].
		"We're at the top level; throw the event back in to find recipient"
		formerHaloOwner removeHalo.
		^self processEvent: event copy resetHandlerFields.
	].
	self submorphsDo:[:m|
		localEvt := event transformedBy: (m transformedFrom: self).
		(m fullContainsPoint: localEvt position) 
			ifTrue:[^m transferHalo: event from: formerHaloOwner].
	].
	"We're at the bottom most level; throw the event back up to the root to find recipient"
	formerHaloOwner removeHalo.

	
	(w := self world) ifNil: [ ^self ].
	localEvt := event transformedBy: (self transformedFrom: w) inverseTransformation.
	^ w processEvent: localEvt resetHandlerFields.
! !

!Morph methodsFor: 'menus' stamp: 'MarcusDenker 11/28/2009 14:25'!
transferStateToRenderer: aRenderer
	"Transfer knownName, and visible over to aRenderer, which is being imposed above me as a transformation shell"

	aRenderer simplySetVisible: self visible



 

		! !

!Morph methodsFor: 'event handling' stamp: 'dgd 2/22/2003 14:36'!
transformFrom: uberMorph 
	"Return a transform to be used to map coordinates in a morph above me into my childrens coordinates, or vice-versa. This is used to support scrolling, scaling, and/or rotation. This default implementation just returns my owner's transform or the identity transform if my owner is nil. 
	Note:  This method cannot be used to map into the receiver's coordinate system!!"

	(self == uberMorph or: [owner isNil]) ifTrue: [^IdentityTransform new].
	^owner transformFrom: uberMorph! !

!Morph methodsFor: 'event handling' stamp: 'RAA 6/19/2000 07:13'!
transformFromOutermostWorld
	"Return a transform to map world coordinates into my local coordinates"

	"self isWorldMorph ifTrue: [^ MorphicTransform identity]."
	^ self transformFrom: self outermostWorldMorph! !

!Morph methodsFor: 'event handling' stamp: ''!
transformFromWorld
	"Return a transform to map world coordinates into my local coordinates"

	^ self transformFrom: nil! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/22/2000 18:03'!
transformedBy: aTransform
	aTransform isIdentity ifTrue:[^self].
	aTransform isPureTranslation ifTrue:[
		^self position: (aTransform localPointToGlobal: self position).
	].
	^self addFlexShell transformedBy: aTransform! !

!Morph methodsFor: 'events-processing' stamp: 'marcus.denker 8/24/2008 22:02'!
transformedFrom: uberMorph
	"Return a transform to map coordinates of uberMorph, a morph above me in my owner chain, into the coordinates of MYSELF not any of my children."
	"self flag: #arNote." "rename this method"
	owner ifNil:[^IdentityTransform basicNew].
	^ (owner transformFrom: uberMorph)! !

!Morph methodsFor: 'utilities' stamp: 'sw 10/23/1998 11:50'!
transparentSpacerOfSize: aPoint
	^ (Morph new extent: aPoint) color: Color transparent! !

!Morph methodsFor: 'undo' stamp: 'CamilleTeruel 12/6/2013 14:38'!
undoMove: cmd redo: redo owner: formerOwner bounds: formerBounds predecessor: formerPredecessor 
	"Handle undo and redo of move commands in morphic"

	self owner ifNil: [ ^ self inform: 'No owner' ].

	formerOwner ifNotNil: 
			[formerPredecessor ifNil: [formerOwner addMorphFront: self]
				ifNotNil: [formerOwner addMorph: self after: formerPredecessor]].
	self bounds: formerBounds.
	(self isSystemWindow) ifTrue: [self activate]! !

!Morph methodsFor: 'accessing' stamp: 'di 8/11/1998 12:33'!
unlock
	self lock: false! !

!Morph methodsFor: 'accessing' stamp: 'sw 8/15/97 23:59'!
unlockContents
	self submorphsDo:
		[:m | m unlock]! !

!Morph methodsFor: 'updating' stamp: 'BenjaminVanRyseghem 7/24/2013 13:35'!
update: anAspect

	^ self! !

!Morph methodsFor: 'events-accessing' stamp: 'gvc 9/11/2009 17:43'!
updateableActionMap
	"Answer an updateable action map, saving it in my #actionMap property"
	
	| actionMap |
	self assureExtension.
	actionMap := extension actionMap.
	actionMap ifNil:
		[actionMap := self createActionMap.
		extension actionMap: actionMap].
	^actionMap! !

!Morph methodsFor: 'visual properties' stamp: 'nk 8/28/2003 15:57'!
useBitmapFill
	"Make receiver use a solid fill style (e.g., a simple color)"
	| fill |
	self fillStyle isBitmapFill ifTrue:[^self]. "Already done"
	fill := BitmapFillStyle fromForm: self defaultBitmapFillForm.
	"Note: Must fix the origin due to global coordinates"
	fill origin: self bounds origin.
	self fillStyle: fill.! !

!Morph methodsFor: 'visual properties' stamp: 'ar 6/25/1999 11:11'!
useDefaultFill
	"Make receiver use a solid fill style (e.g., a simple color)"
	self fillStyle: self defaultColor.! !

!Morph methodsFor: 'visual properties' stamp: 'nk 2/27/2003 11:48'!
useGradientFill
	"Make receiver use a solid fill style (e.g., a simple color)"
	| fill color1 color2 |
	self fillStyle isGradientFill ifTrue:[^self]. "Already done"
	color1 := self color asColor.
	color2 := color1 negated.
	fill := GradientFillStyle ramp: {0.0 -> color1. 1.0 -> color2}.
	fill origin: self topLeft.
	fill direction: 0 @ self bounds extent y.
	fill normal: self bounds extent x @ 0.
	fill radial: false.
	self fillStyle: fill! !

!Morph methodsFor: 'rounding' stamp: 'MarcusDenker 10/26/2011 14:58'!
useRoundedCorners
	self cornerStyle: #rounded! !

!Morph methodsFor: 'visual properties' stamp: 'ar 6/18/1999 06:57'!
useSolidFill
	"Make receiver use a solid fill style (e.g., a simple color)"
	self fillStyle isSolidFill ifTrue:[^self]. "Already done"
	self fillStyle: self fillStyle asColor. "Try minimizing changes"! !

!Morph methodsFor: 'accessing' stamp: 'tk 12/16/1998 11:54'!
userString
	"Do I have a text string to be searched on?"

	^ nil! !

!Morph methodsFor: 'layout-properties' stamp: 'tk 10/30/2001 18:39'!
vResizeToFit: aBoolean
	aBoolean ifTrue:[
		self vResizing: #shrinkWrap.
	] ifFalse:[
		self vResizing: #rigid.
	].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:02'!
vResizing
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#rigid] ifNotNil:[props vResizing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
vResizing: aSymbol
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	self assureLayoutProperties vResizing: aSymbol.
	self layoutChanged.
! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/31/2000 20:45'!
vResizingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self vResizing! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 08:53'!
valueOfProperty: aSymbol 
	"answer the value of the receiver's property named aSymbol"
	^ extension ifNotNil: [extension valueOfProperty: aSymbol]! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 08:50'!
valueOfProperty: aSymbol ifAbsent: aBlock 
	"if the receiver possesses a property of the given name, answer  
	its value. If not then evaluate aBlock and answer the result of  
	this block evaluation"
	^ extension 
		ifNotNil: [extension valueOfProperty: aSymbol ifAbsent: aBlock]
		ifNil: [aBlock value]! !

!Morph methodsFor: 'accessing - properties' stamp: 'dgd 2/16/2003 20:55'!
valueOfProperty: aSymbol ifAbsentPut: aBlock 
	"If the receiver possesses a property of the given name, answer  
	its value. If not, then create a property of the given name, give 
	it the value obtained by evaluating aBlock, then answer that  
	value"
	^ self assureExtension valueOfProperty: aSymbol ifAbsentPut: aBlock! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 08:49'!
valueOfProperty: aSymbol ifPresentDo: aBlock 
	"If the receiver has a property of the given name, evaluate  
	aBlock on behalf of the value of that property"
	extension ifNil:  [^ self].
	^ aBlock value: (extension valueOfProperty: aSymbol ifAbsent: [^ self])! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'adrian_lienhard 7/19/2009 17:35'!
vanishAfterSlidingTo: aPosition event: evt

	| aForm aWorld startPoint endPoint |
	aForm := self imageForm offset: 0@0.
	aWorld := self world.
	startPoint := evt hand fullBounds origin.
	self delete.
	aWorld displayWorld.
	endPoint := aPosition.
	aForm slideFrom: startPoint  to: endPoint nSteps: 12 delay: 15.! !

!Morph methodsFor: 'copying' stamp: 'tk 1/6/1999 17:27'!
veryDeepCopyWith: deepCopier
	"Copy me and the entire tree of objects I point to.  An object in the tree twice is copied once, and both references point to him.  deepCopier holds a dictionary of objects we have seen.  See veryDeepInner:, veryDeepFixupWith:"

	self prepareToBeSaved.
	^ super veryDeepCopyWith: deepCopier! !

!Morph methodsFor: 'copying' stamp: 'tk 2/3/2001 14:29'!
veryDeepFixupWith: deepCopier
	"If some fields were weakly copied, fix new copy here."

	"super veryDeepFixupWith: deepCopier.	Object has no fixups, so don't call it"

	"If my owner is being duplicated too, then store his duplicate.
	 If I am owned outside the duplicated tree, then I am no longer owned!!"
	owner := deepCopier references at: owner ifAbsent: [nil].

! !

!Morph methodsFor: 'copying' stamp: 'pmm 3/13/2010 11:32'!
veryDeepInner: deepCopier 
	"The inner loop, so it can be overridden when a field should not  
	be traced."
	"super veryDeepInner: deepCopier.	know Object has no inst vars"
	bounds := bounds shallowCopy.
	"Points are shared with original"
	"owner := owner.	special, see veryDeepFixupWith:"
	submorphs := submorphs veryDeepCopyWith: deepCopier.
	"each submorph's fixup will install me as the owner"
	"fullBounds := fullBounds.	fullBounds is shared with original!!"
	color := color veryDeepCopyWith: deepCopier.
	"color, if simple, will return self. may be complex"
	extension := (extension veryDeepCopyWith: deepCopier)! !

!Morph methodsFor: 'Morphic-Base-Worlds' stamp: 'dgd 9/27/2004 11:45'!
viewBox
	^ self pasteUpMorph viewBox! !

!Morph methodsFor: 'drawing' stamp: 'md 2/27/2006 08:49'!
visible
	"answer whether the receiver is visible"
	extension ifNil: [^ true].
	^ extension visible! !

!Morph methodsFor: 'accessing' stamp: 'IgorStasenko 12/22/2012 03:10'!
visibleClearArea
	"Answer the receiver visible clear area. The intersection 
	between the clear area and the viewbox."
	^ self viewBox intersect: self clearArea ifNone: [ (0@0 corner: 0@0 ) ]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 12/19/2009 23:17'!
wantsBalloon
	"Answer true if receiver wants to show a balloon help text is a few moments."

	^ (self balloonText notNil) and: [self balloonHelpEnabled]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 12/19/2009 23:40'!
wantsDirectionHandles
	^self valueOfProperty: #wantsDirectionHandles ifAbsent:[false]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 12/19/2009 23:41'!
wantsDirectionHandles: aBool
	self setProperty: #wantsDirectionHandles toValue: aBool
! !

!Morph methodsFor: 'event handling' stamp: 'ar 1/10/2001 21:28'!
wantsDropFiles: anEvent
	"Return true if the receiver wants files dropped from the OS."
	^false! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:24'!
wantsDroppedMorph: aMorph event: evt
	"Return true if the receiver wishes to accept the given morph, which is being dropped by a hand in response to the given event. Note that for a successful drop operation both parties need to agree. The symmetric check is done automatically via aMorph wantsToBeDroppedInto: self."

	^self dropEnabled! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 4/8/98 13:26'!
wantsHaloFor: aSubMorph
	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 1/25/2000 17:43'!
wantsHaloFromClick
	^ true! !

!Morph methodsFor: 'halos and balloon help' stamp: 'AlainPlantec 5/7/2010 22:56'!
wantsHaloHandleWithSelector: aSelector inHalo: aHaloMorph
	"Answer whether the receiver would like to offer the halo handle with the given selector (e.g. #addCollapseHandle:)"

	(#(addDismissHandle:) includes: aSelector) ifTrue:
		[^ self resistsRemoval not].

	(#( addDragHandle: ) includes: aSelector) ifTrue:
		[^ self okayToBrownDragEasily].

	(#(addGrowHandle: addScaleHandle:) includes: aSelector) ifTrue:
		[^ self okayToResizeEasily].

	(#( addRotateHandle: ) includes: aSelector) ifTrue:
		[^ self okayToRotateEasily].

	(#(addRecolorHandle:) includes: aSelector) ifTrue:
		[^ self renderedMorph wantsRecolorHandle].

	^ true
	! !

!Morph methodsFor: 'event handling' stamp: 'sw 11/3/97 02:11'!
wantsKeyboardFocusFor: aSubmorph
	"Answer whether a plain mouse click on aSubmorph, a text-edit-capable thing, should result in a text selection there"
	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/27/2001 14:52'!
wantsRecolorHandle
	"Answer whether the receiver would like a recoloring halo handle to be put up.  Since this handle also presently affords access to the property-sheet, it is presently always allowed, even though SketchMorphs don't like regular recoloring"

	^ true
	
! !

!Morph methodsFor: 'rounding' stamp: 'ar 12/22/2001 22:45'!
wantsRoundedCorners
	"Return true if the receiver wants its corners rounded"
	^ self cornerStyle == #rounded! !

!Morph methodsFor: 'testing' stamp: 'StephaneDucasse 7/18/2010 16:33'!
wantsSteps
	"Return true if the receiver overrides the default Morph step method."
	"Details: Find first class in superclass chain that implements #step and return true if it isn't class Morph."

	| c |
	c := self class.
	[c includesSelector: #step] whileFalse: [c := c superclass].
	^ c ~= Morph! !

!Morph methodsFor: 'accessing' stamp: 'StephaneDucasse 10/13/2013 21:50'!
wantsToBeCachedByHand
	"Return true if the receiver wants to be cached by the hand when it is dragged around.
	Note: The default implementation queries all submorphs since subclasses may have shapes that do not fill the receiver's bounds completely."

	self isTranslucentButNotTransparent ifTrue: [ ^ false ].
	self
		submorphsDo: [ :m | 
			m wantsToBeCachedByHand
				ifFalse: [ ^ false ] ].
	^ true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 9/18/2000 18:34'!
wantsToBeDroppedInto: aMorph
	"Return true if it's okay to drop the receiver into aMorph. This check is symmetric to #wantsDroppedMorph:event: to give both parties a chance of figuring out whether they like each other."
	^true! !

!Morph methodsFor: 'accessing' stamp: 'AlainPlantec 10/17/2009 17:11'!
wantsToBeTopmost
	"Answer if the receiver want to be one of the topmost objects in its owner"
	^ false! !

!Morph methodsFor: 'menu' stamp: 'AlainPlantec 12/10/2009 13:05'!
wantsYellowButtonMenu
	"Answer true if the receiver wants a yellow button menu"
	self
		valueOfProperty: #wantsYellowButtonMenu
		ifPresentDo: [:value | ^ value].
	self isInSystemWindow ifTrue: [^ false].
	^ self defaultYellowButtonMenuEnabled! !

!Morph methodsFor: 'menu' stamp: 'dgd 9/18/2004 18:35'!
wantsYellowButtonMenu: aBoolean 
	"Change the receiver to wants or not a yellow button menu"
	self setProperty: #wantsYellowButtonMenu toValue: aBoolean! !

!Morph methodsFor: 'geometry' stamp: ''!
width

	^ bounds width! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:22'!
width: aNumber
	" Set my width; my position (top-left corner) and height will remain the same "

	self extent: aNumber asInteger@self height.
! !

!Morph methodsFor: 'Morphic-Base-Widgets' stamp: 'BenjaminVanRyseghem 2/12/2012 00:22'!
widthToDisplayInList: aList

	^ self minExtent x! !

!Morph methodsFor: 'event handling' stamp: 'tbn 3/12/2010 01:55'!
windowEvent: anEvent
	"Host window event"! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 00:59'!
withAllOwners
	"Return the receiver and all its owners"

	^ Array streamContents: [:strm | self withAllOwnersDo: [:m | strm nextPut: m]]! !

!Morph methodsFor: 'structure' stamp: 'ar 9/14/2000 16:48'!
withAllOwnersDo: aBlock
	"Evaluate aBlock with the receiver and all of its owners"
	aBlock value: self.
	owner ifNotNil:[^owner withAllOwnersDo: aBlock].! !

!Morph methodsFor: 'structure' stamp: 'ClementBera 7/30/2013 11:05'!
world
	^owner ifNotNil: [owner world]! !

!Morph methodsFor: 'geometry' stamp: 'di 2/23/98 11:36'!
worldBounds
	^ self world bounds! !

!Morph methodsFor: 'event handling' stamp: 'sw 5/6/1998 12:54'!
wouldAcceptKeyboardFocus
	"Answer whether a plain mouse click on the receiver should result in a text selection there"
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'sw 8/29/2000 14:57'!
wouldAcceptKeyboardFocusUponTab
	"Answer whether the receiver is in the running as the new keyboard focus if the tab key were hit at a meta level.  This provides the leverage for tabbing among fields of a card, for example."

	^ false! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:02'!
wrapCentering
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of secondary direction
		#bottomRight - center at end of secondary direction
		#center - center in the middle of secondary direction
		#justified - insert extra space inbetween rows/columns
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#topLeft] ifNotNil:[props wrapCentering].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
wrapCentering: aSymbol
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of secondary direction
		#bottomRight - center at end of secondary direction
		#center - center in the middle of secondary direction
		#justified - insert extra space inbetween rows/columns
	"
	self assureTableProperties wrapCentering: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 03:00'!
wrapCenteringString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self wrapCentering! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
wrapDirection
	"Layout specific. This property describes the direction along which a list-like layout should be wrapped. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
		#none
	indicating in which direction wrapping should occur. This direction must be orthogonal to the list direction, that is if listDirection is #leftToRight or #rightToLeft then wrapDirection must be #topToBottom or #bottomToTop and vice versa."
	| props |
	props := self layoutProperties.
	^props ifNil:[#none] ifNotNil:[props wrapDirection].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
wrapDirection: aSymbol
	"Layout specific. This property describes the direction along which a list-like layout should be wrapped. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
		#none
	indicating in which direction wrapping should occur. This direction must be orthogonal to the list direction, that is if listDirection is #leftToRight or #rightToLeft then wrapDirection must be #topToBottom or #bottomToTop and vice versa."
	self assureTableProperties wrapDirection: aSymbol.
	self layoutChanged.
! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 03:00'!
wrapDirectionString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self wrapDirection ! !

!Morph methodsFor: 'event handling' stamp: 'CamilloBruni 2/4/2012 14:43'!
yellowButtonActivity: shiftState 
	"Find me or my outermost owner that has items to add to a  
	yellow button menu.  
	shiftState is true if the shift was pressed.  
	Otherwise, build a menu that contains the contributions from  
	myself and my interested submorphs,  
	and present it to the user."
	| menu |
	self isWorldMorph
		ifFalse: [| outerOwner | 
			outerOwner := self outermostOwnerWithYellowButtonMenu.
			outerOwner
				ifNil: [^ false].
			outerOwner == self
				ifFalse: [^ outerOwner yellowButtonActivity: shiftState]].
	menu := self buildYellowButtonMenu: ActiveHand.
	menu
		addTitle: self externalName
		icon: (self iconOrThumbnailOfSize: 28).
	menu popUpInWorld: self currentWorld.
	^ true! !

Object subclass: #MorphExtension
	instanceVariableNames: 'locked visible sticky balloonText externalName eventHandler otherProperties fillStyle layoutPolicy layoutFrame layoutProperties borderStyle cornerStyle actionMap clipSubmorphs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Kernel'!

!MorphExtension commentStamp: '<historical>' prior: 0!
MorphExtension provides access to extra instance state that is not required in most simple morphs.  This allows simple morphs to remain relatively lightweight while still admitting more complex structures as necessary.  The otherProperties field takes this policy to the extreme of allowing any number of additional named attributes, albeit at a certain cost in speed and space.!

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:41'!
actionMap
	"Answer the value of actionMap"

	^actionMap ifNil: [self valueOfProperty: #actionMap ifAbsent: []]! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:39'!
actionMap: anObject
	"Set the value of actionMap"

	actionMap := anObject! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:41'!
assureOtherProperties
	"creates an otherProperties for the receiver if needed"
	otherProperties ifNil: [self initializeOtherProperties].
	^ otherProperties! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:52'!
balloonText
	^ balloonText! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:55'!
balloonText: newValue
	balloonText := newValue! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 18:10'!
borderStyle
	"Answer the value of borderStyle"

	^borderStyle! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:08'!
borderStyle: anObject
	"Set the value of borderStyle"

	borderStyle := anObject! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 18:10'!
clipSubmorphs
	"Answer the value of clipSubmorphs"

	^clipSubmorphs! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:58'!
clipSubmorphs: anObject
	"Set the value of clipSubmorphs"

	clipSubmorphs := anObject! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'StephaneDucasse 4/3/2011 22:27'!
copyWeakly
	"list of names of properties whose values should be weak-copied when veryDeepCopying a morph.  See DeepCopier."

	^ #(formerOwner)
	"add yours to this list" 

	"formerOwner should really be nil at the time of the copy, but this will work just fine."! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 18:10'!
cornerStyle
	"Answer the value of cornerStyle"

	^cornerStyle! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 17:23'!
cornerStyle: anObject
	"Set the value of cornerStyle"

	cornerStyle := anObject! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:51'!
eventHandler
	"answer the receiver's eventHandler"
	^ eventHandler ! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:56'!
eventHandler: newValue
	eventHandler := newValue! !

!MorphExtension methodsFor: 'viewer' stamp: 'di 8/10/1998 14:47'!
externalName
	^ externalName! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:57'!
externalName: aString 
	"change the receiver's externalName"
	externalName := aString! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 18:12'!
fillStyle
	"Answer the value of fillStyle"

	^ fillStyle! !

!MorphExtension methodsFor: 'accessing' stamp: 'gvc 9/11/2009 16:55'!
fillStyle: anObject
	"Set the value of fillStyle"

	fillStyle := anObject! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'ClementBera 7/30/2013 11:07'!
hasProperty: aSymbol 
	"Answer whether the receiver has the property named aSymbol"
	| property |
	otherProperties ifNil: [^ false].
	property := otherProperties at: aSymbol ifAbsent: [].
	property ifNil: [^ false].
	property == false ifTrue: [^ false].
	^ true! !

!MorphExtension methodsFor: 'initialization' stamp: 'StephaneDucasse 7/18/2010 16:28'!
initialize
	
	locked := false.
	visible := true.
	sticky := false.

! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'marcus.denker 9/17/2008 17:39'!
initializeOtherProperties
	"private - initializes the receiver's otherProperties"
	otherProperties :=  SmallIdentityDictionary new! !

!MorphExtension methodsFor: 'other' stamp: 'alain.plantec 2/6/2009 17:09'!
inspectElement
	"Create and schedule an Inspector on the otherProperties and the 
	named properties."
	| key obj |
	key := UIManager default 
				chooseFrom: self sortedPropertyNames 
				values: self sortedPropertyNames
				title: 'Inspect which property?' translated.
	key
		ifNil: [^ self].
	obj := otherProperties
				at: key
				ifAbsent: ['nOT a vALuE'].
	obj = 'nOT a vALuE'
		ifTrue: [(self perform: key) inspect
			"named properties"]
		ifFalse: [obj inspect]! !

!MorphExtension methodsFor: 'other' stamp: 'ClementBera 7/30/2013 11:07'!
isDefault
	"Return true if the receiver is a default and can be omitted"
	locked == true
		ifTrue: [^ false].
	visible == false
		ifTrue: [^ false].
	sticky == true
		ifTrue: [^ false].
	balloonText ifNotNil: [^ false].
	externalName ifNotNil: [^ false].
	eventHandler ifNotNil: [^ false].
	otherProperties ifNotNil: [otherProperties isEmpty ifFalse: [^ false]].
	^ true! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'gvc 9/11/2009 18:10'!
layoutFrame

	^layoutFrame! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'gvc 9/11/2009 16:45'!
layoutFrame: aLayoutFrame

	layoutFrame := aLayoutFrame! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'gvc 9/11/2009 18:12'!
layoutPolicy

	^layoutPolicy! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'gvc 9/11/2009 16:47'!
layoutPolicy: aLayoutPolicy 
	
	layoutPolicy := aLayoutPolicy! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'gvc 9/11/2009 18:12'!
layoutProperties

	^layoutProperties! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'gvc 9/11/2009 16:47'!
layoutProperties: newProperties 
	"Return the current layout properties associated with the receiver"

	layoutProperties := newProperties! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:38'!
locked
	"answer whether the receiver is Locked"
	^ locked! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:48'!
locked: aBoolean 
	"change the receiver's locked property"
	locked := aBoolean! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'dgd 2/16/2003 21:04'!
otherProperties
	"answer the receiver's otherProperties"
	^ otherProperties! !

!MorphExtension methodsFor: 'printing' stamp: 'MarcusDenker 10/21/2013 14:23'!
printOn: aStream
	"Append to the argument, aStream, a sequence of characters that 
	identifies the receiver."

	super printOn: aStream.
	aStream
		space;
		nextPut: $(;
		print: self identityHash;
		nextPut: $).
	locked == true
		ifTrue: [ aStream nextPutAll: ' [locked] ' ].
	visible == false
		ifTrue: [ aStream nextPutAll: '[not visible] ' ].
	sticky == true
		ifTrue: [ aStream nextPutAll: ' [sticky] ' ].
	balloonText ifNotNil: [ aStream nextPutAll: ' [balloonText] ' ].
	externalName
		ifNotNil: [ 
			aStream
				nextPutAll: ' [externalName = ' , externalName;
				nextPutAll: ' ] ' ].
	eventHandler
		ifNotNil: [ 
			aStream
				nextPutAll: ' [eventHandler = ' , eventHandler printString;
				nextPutAll: '] ' ].
	(otherProperties isNil or: [ otherProperties isEmpty ])
		ifTrue: [ ^ self ].
	aStream nextPutAll: ' [other: '.
	self otherProperties
		keysDo: [ :aKey | 
			aStream
				nextPutAll: ' (' , aKey , ' -> ' , (self otherProperties at: aKey) printString;
				nextPutAll: ')' ].
	aStream nextPut: $]! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'nk 5/1/2004 17:23'!
propertyNamesNotCopied
	"list of names of properties whose values should be deleted when veryDeepCopying a morph.
	See DeepCopier."

	^ #(connectedConstraints connectionHighlights highlightedTargets)
	"add yours to this list" 
! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:37'!
removeOtherProperties
	"Remove the 'other' properties"
	otherProperties := nil! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:43'!
removeProperty: aSymbol 
	"removes the property named aSymbol if it exists"
	otherProperties ifNil: [^ self].
	otherProperties removeKey: aSymbol ifAbsent: [].
	otherProperties isEmpty ifTrue: [self removeOtherProperties]! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'dgd 2/16/2003 21:49'!
setProperty: aSymbol toValue: abObject 
	"change the receiver's property named aSymbol to anObject"
	self assureOtherProperties at: aSymbol put: abObject! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'ClementBera 7/30/2013 11:07'!
sortedPropertyNames
	"answer the receiver's property names in a sorted way"

	| props |
	props := (Array new: 10) writeStream.
	locked == true ifTrue: [props nextPut: #locked].
	visible == false ifTrue: [props nextPut: #visible].
	sticky == true ifTrue: [props nextPut: #sticky].
	balloonText ifNotNil: [props nextPut: #balloonText].
	externalName ifNotNil: [props nextPut: #externalName].
	eventHandler ifNotNil: [props nextPut: #eventHandler].
	 otherProperties ifNotNil: [otherProperties associationsDo: [:a | props nextPut: a key]].
	^props contents sort: [:s1 :s2 | s1 <= s2]! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/14/1998 13:07'!
sticky
	^ sticky! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:47'!
sticky: aBoolean 
	"change the receiver's sticky property"
	sticky := aBoolean! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'dgd 2/16/2003 21:00'!
valueOfProperty: aSymbol 
"answer the value of the receiver's property named aSymbol"
	^ self
		valueOfProperty: aSymbol
		ifAbsent: []! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:43'!
valueOfProperty: aSymbol ifAbsent: aBlock 
	"if the receiver possesses a property of the given name, answer  
	its value. If not then evaluate aBlock and answer the result of  
	this block evaluation"
	otherProperties ifNil: [^ aBlock value].
	^ otherProperties at: aSymbol ifAbsent: [^ aBlock value]! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'dgd 2/16/2003 21:28'!
valueOfProperty: aSymbol ifAbsentPut: aBlock 
	"If the receiver possesses a property of the given name, answer  
	its value. If not, then create a property of the given name, give 
	it the value obtained by evaluating aBlock, then answer that  
	value"
	^self assureOtherProperties at: aSymbol ifAbsentPut: aBlock! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'nk 5/1/2004 17:39'!
veryDeepFixupWith: deepCopier 
	"If target and arguments fields were weakly copied, fix them here.
	If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

	super veryDeepFixupWith: deepCopier.
	otherProperties ifNil: [ ^self ].

	"Properties whose values are only copied weakly replace those values if they were copied via another path"
	self copyWeakly do: [ :propertyName |
		otherProperties at: propertyName ifPresent: [ :property |
			otherProperties at: propertyName
				put: (deepCopier references at: property ifAbsent: [ property ])]].
! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'MarcusDenker 9/7/2010 17:00'!
veryDeepInner: deepCopier 
	"Copy all of my instance variables.
	Some otherProperties need to be not copied at all, but shared. Their names are given by copyWeakly.
	Some otherProperties should not be copied or shared. Their names are given by propertyNamesNotCopied.
	This is special code for the dictionary. See DeepCopier, and veryDeepFixupWith:."

	| namesOfWeaklyCopiedProperties weaklyCopiedValues |
	super veryDeepInner: deepCopier.
	locked := locked veryDeepCopyWith: deepCopier.
	visible := visible veryDeepCopyWith: deepCopier.
	sticky := sticky veryDeepCopyWith: deepCopier.
	balloonText := balloonText veryDeepCopyWith: deepCopier.
	externalName := externalName veryDeepCopyWith: deepCopier.
	eventHandler := eventHandler veryDeepCopyWith: deepCopier. 	"has its own restrictions"
	
	fillStyle := fillStyle veryDeepCopyWith: deepCopier.
	layoutPolicy := layoutPolicy veryDeepCopyWith: deepCopier.
	layoutFrame := layoutFrame veryDeepCopyWith: deepCopier.
	layoutProperties := layoutProperties veryDeepCopyWith: deepCopier.
	borderStyle := borderStyle  veryDeepCopyWith: deepCopier.
	cornerStyle := cornerStyle veryDeepCopyWith: deepCopier.
	actionMap := actionMap veryDeepCopyWith: deepCopier.
	clipSubmorphs := clipSubmorphs veryDeepCopyWith: deepCopier.

	otherProperties ifNil: [ ^self ].

	otherProperties := otherProperties copy.
	self propertyNamesNotCopied do: [ :propName | otherProperties removeKey: propName ifAbsent: [] ].

	namesOfWeaklyCopiedProperties := self copyWeakly.
	weaklyCopiedValues := namesOfWeaklyCopiedProperties collect: [  :propName | otherProperties removeKey: propName ifAbsent: [] ].

	"Now copy all the others."
	otherProperties := otherProperties veryDeepCopyWith: deepCopier.

	"And replace the weak ones."
	namesOfWeaklyCopiedProperties with: weaklyCopiedValues do: [ :name :value | value ifNotNil: [ otherProperties at: name put: value ]].
! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:41'!
visible
	"answer whether the receiver is visible"
	^ visible! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:55'!
visible: newValue
	visible := newValue! !

Object subclass: #MorphicEvent
	instanceVariableNames: 'timeStamp source windowIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MorphicEvent commentStamp: '<historical>' prior: 0!
This class represents the base for all events.

Instance variables:
	stamp	<Integer>	The millisecond clock time stamp (based on Time millisecondClock)
	source	<Hand | nil>	If non-nil the hand that generated the event.!

MorphicEvent subclass: #DropEvent
	instanceVariableNames: 'position contents wasHandled'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!DropEvent commentStamp: 'LaurentLaffont 3/15/2011 20:47' prior: 0!
I model the event produced when a hand drops a (grabbed) morph into another morph.
I know the position of the drop, the dropped morph and the hand that produced me.

To handle this event a morph should override one of these methods:
#acceptDroppingMorph:event:
#justDroppedInto:event:

Additionaly, a morph can specify if it wants to accept a dropped morph by overriding #wantsDroppedMorph:event:. 
Symmetrically, the morph being dropped can specify if it wants to be dropped in another morph by overriding #wantsToBeDroppedInto:. 

Note that for a successful drop operation both parties need to agree.

See HandMorph>>dropMorph:event: for an example of usage.
!

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:33'!
contents
	^contents! !

!DropEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:19'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
	wasHandled := anEvent wasHandled.! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 19:21'!
cursorPoint
	"For compatibility with mouse events"
	^position! !

!DropEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 18:33'!
isDropEvent
	^true! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:33'!
position
	^position! !

!DropEvent methodsFor: 'printing' stamp: 'JMM 9/29/2004 13:24'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self position printString; space.
	aStream nextPutAll: self type; space.
	aStream	 nextPutAll: self windowIndex printString.
	aStream nextPut: $].! !

!DropEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:18'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"
	wasHandled := false.! !

!DropEvent methodsFor: 'dispatching' stamp: 'ar 1/10/2001 21:24'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	self type == #dropEvent ifTrue:[^anObject handleDropMorph: self].! !

!DropEvent methodsFor: 'private' stamp: 'ar 9/13/2000 19:23'!
setPosition: pos contents: aMorph hand: aHand
	position := pos.
	contents := aMorph.
	source := aHand.
	wasHandled := false.! !

!DropEvent methodsFor: 'transforming' stamp: 'ar 10/7/2000 18:28'!
transformBy: aMorphicTransform
	"Transform the receiver into a local coordinate system."
	position :=  aMorphicTransform globalPointToLocal: position.! !

!DropEvent methodsFor: 'transforming' stamp: 'ar 10/7/2000 18:28'!
transformedBy: aMorphicTransform
	"Return the receiver transformed by the given transform into a local coordinate system."
	^self shallowCopy transformBy: aMorphicTransform! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:33'!
type
	^#dropEvent! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:44'!
wasHandled
	^wasHandled! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:44'!
wasHandled: aBool
	wasHandled := aBool.! !

!MorphicEvent methodsFor: 'comparing' stamp: 'ar 9/13/2000 15:36'!
= anEvent
	anEvent isMorphicEvent ifFalse:[^false].
	^self type = anEvent type! !

!MorphicEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:18'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 10/10/2000 21:28'!
cursorPoint
	"Backward compatibility. Use #position instead"
	^ self position! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 16:48'!
hand
	"Return the source that generated the event"
	^source! !

!MorphicEvent methodsFor: 'comparing' stamp: 'ar 9/13/2000 15:36'!
hash
	^self type hash! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/22/2000 10:36'!
isDraggingEvent
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:17'!
isDropEvent
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:19'!
isKeyboard
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 10/10/2000 21:27'!
isKeystroke
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:37'!
isMorphicEvent
	^true! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:19'!
isMouse
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/14/2000 18:21'!
isMouseOver
	^self type == #mouseOver! !

!MorphicEvent methodsFor: 'testing' stamp: 'JMM 10/6/2004 21:23'!
isMove
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'JMM 10/6/2004 21:35'!
isWindowEvent
	^false! !

!MorphicEvent methodsFor: 'accessing' stamp: 'wiz 12/8/2004 23:13'!
position
	"Since cursorPoint is defined and refers to position it should be defined
	here as well"
	^ self subclassResponsibility! !

!MorphicEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:18'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"! !

!MorphicEvent methodsFor: 'dispatching' stamp: 'ar 9/15/2000 21:12'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	^anObject handleUnknownEvent: self! !

!MorphicEvent methodsFor: 'private' stamp: 'ar 10/25/2000 21:26'!
setHand: aHand
	source := aHand! !

!MorphicEvent methodsFor: 'private' stamp: 'ar 10/25/2000 20:53'!
setTimeStamp: stamp
	timeStamp := stamp.! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:29'!
timeStamp
	"Return the millisecond clock value at which the event was generated"
	^timeStamp ifNil:[timeStamp := Time millisecondClockValue]! !

!MorphicEvent methodsFor: 'transforming' stamp: 'ar 9/13/2000 15:47'!
transformedBy: aMorphicTransform
	"Return the receiver transformed by the given transform into a local coordinate system."
! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:34'!
type
	"Return a symbol indicating the type this event."
	^self subclassResponsibility! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 10/10/2000 01:19'!
wasHandled
	"Return true if this event was handled. May be ignored for some types of events."
	^false! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 10/10/2000 01:20'!
wasHandled: aBool
	"Determine if this event was handled. May be ignored for some types of events."! !

!MorphicEvent methodsFor: 'accessing' stamp: 'JMM 7/20/2004 22:10'!
windowIndex
	^windowIndex! !

!MorphicEvent methodsFor: 'accessing' stamp: 'JMM 7/20/2004 22:10'!
windowIndex: aValue
	windowIndex := aValue! !

MorphicEvent subclass: #UserInputEvent
	instanceVariableNames: 'type buttons position handler wasHandled'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants'
	category: 'Morphic-Core-Events'!

!UserInputEvent commentStamp: 'TorstenBergmann 2/20/2014 18:30' prior: 0!
A general user input event like for keyboard or mouse!

UserInputEvent subclass: #KeyboardEvent
	instanceVariableNames: 'keyValue charCode scanCode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!KeyboardEvent commentStamp: '<historical>' prior: 0!
I am a keyboard event. I contain the char code of the event pressed, the modifiers pressed, and the keycode of the key pressed.!

!KeyboardEvent methodsFor: 'comparing' stamp: 'ar 10/24/2000 17:44'!
= aMorphicEvent
	super = aMorphicEvent ifFalse:[^false].
	buttons = aMorphicEvent buttons ifFalse: [^ false].
	keyValue = aMorphicEvent keyValue ifFalse: [^ false].
	^ true
! !

!KeyboardEvent methodsFor: 'testing' stamp: 'DenisKudryashov 7/20/2013 17:29'!
hasSpecialCTRLKeyValue
"
	4 - Character end
	1 - Character home
"

	^ self controlKeyPressed and: [ keyValue <= 26 & (keyValue ~= 4) & (keyValue ~= 1) ]! !

!KeyboardEvent methodsFor: 'comparing' stamp: 'ar 9/13/2000 15:50'!
hash
	^buttons hash + keyValue hash
! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeyDown
	^self type == #keyDown! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeyUp
	^self type == #keyUp! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeyboard
	^true! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeystroke
	^self type == #keystroke! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 10/9/2000 00:43'!
isMouseMove
	^false! !

!KeyboardEvent methodsFor: 'testing' stamp: 'SeanDeNigris 1/29/2013 11:22'!
isUserInterrupt

	^ UserInterruptHandler cmdDotEnabled and: [ self keyCharacter = $. and: [ self commandKeyPressed ] ].
	! !

!KeyboardEvent methodsFor: 'keyboard' stamp: 'michael.rueger 3/11/2009 11:21'!
keyCharacter
	"Answer the character corresponding this keystroke. This is defined only for keystroke events."

	^Unicode value: charCode! !

!KeyboardEvent methodsFor: 'keyboard' stamp: 'nk 10/13/2004 10:43'!
keyString
	"Answer the string value for this keystroke. This is defined only for keystroke events."

	^ String streamContents: [ :s | self printKeyStringOn: s ]! !

!KeyboardEvent methodsFor: 'keyboard' stamp: 'ar 9/13/2000 15:51'!
keyValue
	"Answer the ascii value for this keystroke. This is defined only for keystroke events."

	^ keyValue! !

!KeyboardEvent methodsFor: 'printing' stamp: 'DenisKudryashov 7/20/2013 17:32'!
printKeyStringOn: aStream
	"Print a readable string representing the receiver on a given stream"

	| kc inBrackets firstBracket keyString |
	kc := self keyCharacter.
	inBrackets := false.
	firstBracket := [ inBrackets ifFalse: [ aStream nextPut: $<. inBrackets := true ]].
	self controlKeyPressed ifTrue: [ 	firstBracket value. aStream nextPutAll: 'Ctrl-' ].
	self commandKeyPressed ifTrue: [ firstBracket value. aStream nextPutAll: 'Cmd-' ].
	(buttons anyMask: 32) ifTrue: [ firstBracket value. aStream nextPutAll: 'Opt-' ].
	(self shiftPressed and: [ keyValue between: 1 and: 31 ])
		ifTrue: [ firstBracket value. aStream nextPutAll: 'Shift-' ].

	self hasSpecialCTRLKeyValue
			ifTrue:
				[aStream nextPut: (keyValue + $a asciiValue - 1) asCharacter]
			ifFalse: 
				[keyString := (kc caseOf: {
					[ Character space ] -> [ ' ' ].
					[ Character tab ] -> [ 'tab' ].
					[ Character cr ] -> [ 'cr' ].
					[ Character lf ] -> [ 'lf' ].
					[ Character enter ] -> [ 'enter' ].

					[ Character backspace ] -> [ 'backspace' ].
					[ Character delete ] -> [ 'delete' ].

					[ Character escape ] -> [ 'escape' ].

					[ Character arrowDown ] -> [ 'down' ].
					[ Character arrowUp ] -> [ 'up' ].
					[ Character arrowLeft ] -> [ 'left' ].
					[ Character arrowRight ] -> [ 'right' ].

					[ Character end ] -> [ 'end' ].
					[ Character home ] -> [ 'home' ].
					[ Character pageDown ] -> [ 'pageDown' ].
					[ Character pageUp ] -> [ 'pageUp' ].

					[ Character euro ] -> [ 'euro' ].
					[ Character insert ] -> [ 'insert' ].

				} otherwise: [ String with: kc ]).
				keyString size > 1 ifTrue: [ firstBracket value ].
				aStream nextPutAll: keyString].

	inBrackets ifTrue: [aStream nextPut: $> ]! !

!KeyboardEvent methodsFor: 'printing' stamp: 'nk 10/13/2004 10:42'!
printOn: aStream
	"Print the receiver on a stream"

	aStream nextPut: $[.
	aStream nextPutAll: type; nextPutAll: ' '''.
	self printKeyStringOn: aStream.
	aStream nextPut: $'.
	aStream nextPut: $]! !

!KeyboardEvent methodsFor: 'keyboard' stamp: 'michael.rueger 2/25/2009 22:19'!
scanCode
	^scanCode! !

!KeyboardEvent methodsFor: 'initialize' stamp: 'michael.rueger 2/24/2009 14:08'!
scanCode: anInt
	scanCode := anInt! !

!KeyboardEvent methodsFor: 'dispatching' stamp: 'ar 9/15/2000 21:13'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #keystroke ifTrue:[^anObject handleKeystroke: self].
	type == #keyDown ifTrue:[^anObject handleKeyDown: self].
	type == #keyUp ifTrue:[^anObject handleKeyUp: self].
	^super sentTo: anObject.! !

!KeyboardEvent methodsFor: 'private' stamp: 'michael.rueger 2/23/2009 11:49'!
setType: aSymbol buttons: anInteger position: pos keyValue: aValue charCode: anInt hand: aHand stamp: stamp
	type := aSymbol.
	buttons := anInteger.
	position := pos.
	keyValue := aValue.
	charCode := anInt.
	source := aHand.
	wasHandled := false.
	timeStamp := stamp.! !

!KeyboardEvent methodsFor: 'private' stamp: 'ar 10/5/2000 23:54'!
setType: aSymbol buttons: anInteger position: pos keyValue: aValue hand: aHand stamp: stamp
	type := aSymbol.
	buttons := anInteger.
	position := pos.
	keyValue := aValue.
	source := aHand.
	wasHandled := false.
	timeStamp := stamp.! !

!KeyboardEvent methodsFor: 'printing' stamp: 'SeanDeNigris 1/15/2014 15:24'!
storeOn: aStream

	aStream nextPutAll: type.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	position x asInteger storeOn: aStream.
	aStream space.
	position y asInteger storeOn: aStream.
	aStream space.
	buttons storeOn: aStream.
	aStream space.
	keyValue storeOn: aStream.
	aStream space.
	charCode storeOn: aStream.
	aStream space.
	scanCode storeOn: aStream.! !

UserInputEvent subclass: #MouseEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MouseEvent commentStamp: 'TorstenBergmann 2/20/2014 18:29' prior: 0!
Superclass for mouse events!

MouseEvent subclass: #MouseButtonEvent
	instanceVariableNames: 'whichButton'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MouseButtonEvent commentStamp: 'TorstenBergmann 2/20/2014 18:29' prior: 0!
A button event from a mouse!

!MouseButtonEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 19:58'!
blueButtonChanged
	"Answer true if the blue mouse button has changed. This is the third mouse button or cmd+click on the Mac."

	^ whichButton anyMask: 1! !

!MouseButtonEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 19:58'!
redButtonChanged
	"Answer true if the red mouse button has changed. This is the first mouse button."

	^ whichButton anyMask: 4! !

!MouseButtonEvent methodsFor: 'dispatching' stamp: 'ar 9/16/2000 13:05'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseDown ifTrue:[^anObject handleMouseDown: self].
	type == #mouseUp ifTrue:[^anObject handleMouseUp: self].
	^super sentTo: anObject! !

!MouseButtonEvent methodsFor: 'private' stamp: 'ar 10/5/2000 23:55'!
setType: evtType position: evtPos which: button buttons: evtButtons hand: evtHand stamp: stamp
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.
	whichButton := button.
	timeStamp := stamp.! !

!MouseButtonEvent methodsFor: 'printing' stamp: 'ar 10/24/2000 16:29'!
storeOn: aStream
	super storeOn: aStream.
	aStream space.
	whichButton storeOn: aStream.! !

!MouseButtonEvent methodsFor: 'accessing' stamp: 'nk 3/11/2004 17:44'!
whichButton
	^whichButton! !

!MouseButtonEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 19:59'!
yellowButtonChanged
	"Answer true if the yellow mouse button has changed. This is the second mouse button or option+click on the Mac."

	^ whichButton anyMask: 2! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
anyButton
	^ 7! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
blueButton
	^ 1! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
redButton
	^ 4! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
yellowButton
	^ 2! !

!MouseEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:50'!
= aMorphicEvent
	super = aMorphicEvent ifFalse:[^false].
	position = aMorphicEvent position ifFalse: [^ false].
	buttons = aMorphicEvent buttons ifFalse: [^ false].
	^ true
! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
anyButtonPressed
	"Answer true if any mouse button is being pressed."

	^ buttons anyMask: self class anyButton! !

!MouseEvent methodsFor: 'converting' stamp: 'pmm 3/13/2010 11:33'!
asMouseEnter
	^self shallowCopy setType: #mouseEnter! !

!MouseEvent methodsFor: 'converting' stamp: 'pmm 3/13/2010 11:33'!
asMouseLeave
	^self shallowCopy setType: #mouseLeave! !

!MouseEvent methodsFor: 'converting' stamp: 'marcus.denker 8/24/2008 21:41'!
asMouseMove
	"Convert the receiver into a mouse move"
	^MouseMoveEvent basicNew setType: #mouseMove startPoint: position endPoint: position trail: {position. position} buttons: buttons hand: source stamp: Time millisecondClockValue.! !

!MouseEvent methodsFor: 'converting' stamp: 'marcus.denker 8/24/2008 21:42'!
asMouseOver
	"Convert the receiver into a mouse over event"
	^MouseEvent basicNew setType: #mouseOver position: position buttons: buttons hand: source! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
blueButtonPressed
	"Answer true if the blue mouse button is being pressed. This is the third mouse button or cmd+click on the Mac."

	^ buttons anyMask: self class blueButton! !

!MouseEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 22:51'!
cursorPoint
	"Answer the location of the cursor's hotspot when this event occured."

	^ position! !

!MouseEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:47'!
hash
	^ position hash + buttons hash! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 10/5/2000 19:43'!
isDraggingEvent
	source ifNil:[^false].
	source hasSubmorphs ifTrue:[^true].
	self anyButtonPressed ifTrue:[^true].
	^false! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:30'!
isMouse
	^true! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseDown
	^self type == #mouseDown! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseEnter
	^self type == #mouseEnter! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseLeave
	^self type == #mouseLeave! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseMove
	^self type == #mouseMove! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseUp
	^self type == #mouseUp! !

!MouseEvent methodsFor: 'testing' stamp: 'gvc 1/28/2008 13:14'!
isMouseWheel
	"Answer whether the receiver is a mouse wheel event."
	
	^false! !

!MouseEvent methodsFor: 'printing' stamp: 'JMM 9/29/2004 13:25'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString; space.
	aStream	 nextPutAll: self windowIndex printString.
	aStream nextPut: $].! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
redButtonPressed
	"Answer true if the red mouse button is being pressed. This is the first mouse button."

	^ buttons anyMask: self class redButton! !

!MouseEvent methodsFor: 'dispatching' stamp: 'ar 10/10/2000 21:15'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseOver ifTrue:[^anObject handleMouseOver: self].
	type == #mouseEnter ifTrue:[^anObject handleMouseEnter: self].
	type == #mouseLeave ifTrue:[^anObject handleMouseLeave: self].
	^super sentTo: anObject.! !

!MouseEvent methodsFor: 'private' stamp: 'ar 10/10/2000 21:15'!
setType: aSymbol
	"For quick conversion between event types"
	type := aSymbol.! !

!MouseEvent methodsFor: 'private' stamp: 'ar 9/15/2000 22:53'!
setType: evtType position: evtPos buttons: evtButtons hand: evtHand
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.! !

!MouseEvent methodsFor: 'printing' stamp: 'ar 10/25/2000 22:09'!
storeOn: aStream

	aStream nextPutAll: type.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	position x storeOn: aStream.
	aStream space.
	position y storeOn: aStream.
	aStream space.
	buttons storeOn: aStream.! !

!MouseEvent methodsFor: 'button state' stamp: 'ar 9/15/2000 22:51'!
targetPoint
	"Answer the location of the cursor's hotspot, adjusted by the offset
	of the last mouseDown relative to the recipient morph."

	^ position - source targetOffset! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
yellowButtonPressed
	"Answer true if the yellow mouse button is being pressed. This is the second mouse button or option+click on the Mac."

	^ buttons anyMask: self class yellowButton! !

MouseEvent subclass: #MouseMoveEvent
	instanceVariableNames: 'startPoint trail'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MouseMoveEvent commentStamp: 'TorstenBergmann 2/20/2014 18:30' prior: 0!
A move event from the mouse!

!MouseMoveEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:49'!
= aMorphicEvent
	super = aMorphicEvent ifFalse:[^false].
	position = aMorphicEvent position ifFalse: [^ false].
	startPoint = aMorphicEvent startPoint ifFalse: [^ false].
	buttons = aMorphicEvent buttons ifFalse: [^ false].
	^ true
! !

!MouseMoveEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 22:51'!
endPoint
	"Return the point where the movement ended."
	^position! !

!MouseMoveEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:49'!
hash
	^ position hash + startPoint hash + buttons hash! !

!MouseMoveEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:29'!
isMove
	^true! !

!MouseMoveEvent methodsFor: 'printing' stamp: 'JMM 9/29/2004 13:25'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self startPoint printString; space.
	aStream nextPutAll: self endPoint printString; space.
	aStream nextPutAll: self type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString; space.
	aStream	 nextPutAll: self windowIndex printString.
	aStream nextPut: $].! !

!MouseMoveEvent methodsFor: 'dispatching' stamp: 'ar 10/10/2000 21:15'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseMove ifTrue:[^anObject handleMouseMove: self].
	^super sentTo: anObject.
! !

!MouseMoveEvent methodsFor: 'private' stamp: 'ar 10/5/2000 23:55'!
setType: evtType startPoint: evtStart endPoint: evtEnd trail: evtTrail buttons: evtButtons hand: evtHand stamp: stamp
	type := evtType.
	startPoint := evtStart.
	position := evtEnd.
	trail := evtTrail.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.
	timeStamp := stamp.! !

!MouseMoveEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 16:25'!
startPoint
	"Return the point where the movement started."
	^startPoint! !

!MouseMoveEvent methodsFor: 'printing' stamp: 'ar 10/24/2000 16:30'!
storeOn: aStream
	super storeOn: aStream.
	aStream space.
	self startPoint x storeOn: aStream.
	aStream space.
	self startPoint y storeOn: aStream.
	aStream space.
	"trail storeOn: aStream."! !

!MouseMoveEvent methodsFor: 'accessing' stamp: 'ar 10/24/2000 16:33'!
trail
	"Return any immediate points that have been assembled along the move"
	^trail ifNil:[#()]! !

!MouseMoveEvent methodsFor: 'transforming' stamp: 'ar 9/15/2000 22:52'!
transformBy: aMorphicTransform
	"Transform the receiver into a local coordinate system."
	position :=  aMorphicTransform globalPointToLocal: position.
	startPoint :=  aMorphicTransform globalPointToLocal: startPoint.! !

!MouseMoveEvent methodsFor: 'transforming' stamp: 'ar 9/15/2000 22:52'!
translateBy: delta
	"add delta to cursorPoint, and return the new event"
	position := position + delta.
	startPoint := startPoint + delta.! !

MouseEvent subclass: #MouseWheelEvent
	instanceVariableNames: 'direction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MouseWheelEvent commentStamp: 'gvc 9/23/2008 11:46' prior: 0!
A mouse event generated by intercepting the keyboard events (ctrl+up/down arrow) generated by the VM in response to mouse wheel activity.!

!MouseWheelEvent methodsFor: 'accessing' stamp: 'gvc 1/25/2008 17:40'!
direction
	"Answer the value of direction"

	^ direction! !

!MouseWheelEvent methodsFor: 'accessing' stamp: 'gvc 1/25/2008 17:40'!
direction: anObject
	"Set the value of direction"

	direction := anObject! !

!MouseWheelEvent methodsFor: 'testing' stamp: 'gvc 1/28/2008 13:14'!
isMouseWheel
	"Answer whether the receiver is a mouse wheel event."
	
	^true! !

!MouseWheelEvent methodsFor: 'printing' stamp: 'GaryChambers 9/5/2011 13:44'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; space.
	aStream nextPutAll: self direction; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString; space.
	aStream	 nextPutAll: self windowIndex printString.
	aStream nextPut: $].! !

!MouseWheelEvent methodsFor: 'dispatching' stamp: 'gvc 1/25/2008 17:42'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	
	type == #mouseWheel ifTrue:[^anObject handleMouseWheel: self].
	^super sentTo: anObject.
! !

!MouseWheelEvent methodsFor: 'accessing' stamp: 'gvc 1/28/2008 15:31'!
setType: evtType position: evtPos direction: dirSymbol buttons: evtButtons hand: evtHand stamp: stamp
	"Set the state for the receiver."
	
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.
	direction := dirSymbol.
	timeStamp := stamp.! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'GuillermoPolito 4/21/2013 19:28'!
altKeyPressed
	"Answer true if the alt/option key on the keyboard was being held down when this event occurred."

	"00100000 is the bit that marks that option/alt has been pressed"
	^self optionKeyPressed! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'GuillermoPolito 4/21/2013 19:31'!
anyModifierKeyPressed
	"ignore, however, the shift keys 'cause that's not REALLY a command key "

	^ self buttons anyMask: 2r01110000	"cmd | opt | ctrl"! !

!UserInputEvent methodsFor: 'printing' stamp: 'ar 10/7/2000 21:57'!
buttonString
	"Return a string identifying the currently pressed buttons"
	| string |
	string := ''.
	self redButtonPressed ifTrue:[string := string,'red '].
	self yellowButtonPressed ifTrue:[string := string,'yellow '].
	self blueButtonPressed ifTrue:[string := string,'blue '].
	^string! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:45'!
buttons
	"Return the a word encoding the mouse and modifier buttons for this event."

	^ buttons! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'GuillermoPolito 4/21/2013 19:32'!
commandKeyPressed
	"Answer true if the command key on the keyboard was being held down when this event occurred."

	"The mask for command bit pressed"
	^ buttons anyMask: 2r01000000! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'GuillermoPolito 4/21/2013 19:32'!
controlKeyPressed
	"Answer true if the control key on the keyboard was being held down when this event occurred."

	"The mask for control bit pressed"
	^ buttons anyMask: 2r00010000! !

!UserInputEvent methodsFor: 'initialize' stamp: 'ar 9/13/2000 15:54'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
	handler := anEvent handler.
	wasHandled := anEvent wasHandled.! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
handler
	^handler! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
handler: anObject
	handler := anObject! !

!UserInputEvent methodsFor: 'printing' stamp: 'ar 10/7/2000 21:56'!
modifierString
	"Return a string identifying the currently pressed modifiers"
	| string |
	string := ''.
	self commandKeyPressed ifTrue:[string := string,'CMD '].
	self shiftPressed ifTrue:[string := string,'SHIFT '].
	self controlKeyPressed ifTrue:[string := string,'CTRL '].
	^string! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'GuillermoPolito 4/21/2013 19:28'!
optionKeyPressed
	"Answer true if the alt/option key on the keyboard was being held down when this event occurred."

	"00100000 is the bit that marks that option/alt has been pressed"
	^ buttons anyMask: 2r00100000! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 22:45'!
position
	^position! !

!UserInputEvent methodsFor: 'initialize' stamp: 'ar 9/13/2000 15:54'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"
	handler := nil.
	wasHandled := false.! !

!UserInputEvent methodsFor: 'private' stamp: 'ar 10/24/2000 16:33'!
setPosition: aPoint
	position := aPoint! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'ar 9/13/2000 15:44'!
shiftPressed
	"Answer true if the shift key on the keyboard was being held down when this event occurred."

	^ buttons anyMask: 8
! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:38'!
transformBy: aMorphicTransform
	"Transform the receiver into a local coordinate system."
	position :=  aMorphicTransform globalPointToLocal: position.! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:38'!
transformedBy: aMorphicTransform
	"Return the receiver transformed by the given transform into a local coordinate system."
	^self shallowCopy transformBy: aMorphicTransform! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:37'!
translateBy: delta
	"add delta to cursorPoint, and return the new event"
	position := position + delta.! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:38'!
translatedBy: delta
	"add delta to cursorPoint, and return the new event"
	^self shallowCopy translateBy: delta! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:45'!
type
	"Return a symbol indicating the type this event."

	^ type! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
wasHandled
	^wasHandled! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
wasHandled: aBool
	wasHandled := aBool.! !

Object subclass: #MorphicEventDispatcher
	instanceVariableNames: 'lastType lastDispatch'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MorphicEventDispatcher commentStamp: '<historical>' prior: 0!
The class represents a strategy for dispatching events to some immediate child of a morph. It is used by morphs to delegate the somewhat complex action of dispatching events accurately. !

!MorphicEventDispatcher methodsFor: 'dispatching' stamp: 'ar 10/10/2000 01:20'!
dispatchDefault: anEvent with: aMorph
	"Dispatch the given event. The event will be passed to the front-most visible submorph that contains the position wrt. to the event."
	| localEvt index child morphs inside |
	"See if we're fully outside aMorphs bounds"
	(aMorph fullBounds containsPoint: anEvent position) ifFalse:[^#rejected]. "outside"
	"Traverse children"
	index := 1.
	morphs := aMorph submorphs.
	inside := false.
	[index <= morphs size] whileTrue:[
		child := morphs at: index.
		localEvt := anEvent transformedBy: (child transformedFrom: aMorph).
		(child processEvent: localEvt using: self) == #rejected ifFalse:[
			"Not rejected. The event was in some submorph of the receiver"
			inside := true.
			localEvt wasHandled ifTrue:[anEvent copyHandlerState: localEvt].
			index := morphs size. "break"
		].
		index := index + 1.
	].

	"Check for being inside the receiver"
	inside ifFalse:[inside := aMorph containsPoint: anEvent position event: anEvent].
	inside ifTrue:[^aMorph handleEvent: anEvent].
	^#rejected
! !

!MorphicEventDispatcher methodsFor: 'dispatching' stamp: 'ar 10/10/2000 21:13'!
dispatchDropEvent: anEvent with: aMorph
	"Find the appropriate receiver for the event and let it handle it. The dispatch is similar to the default dispatch with one difference: Morphs are given the chance to reject an entire drop operation. If the operation is rejected, no drop will be executed."
	| inside index morphs child localEvt |
	"Try to get out quickly"
	(aMorph fullBounds containsPoint: anEvent cursorPoint)
		ifFalse:[^#rejected].
	"Give aMorph a chance to repel the dropping morph"
	aMorph rejectDropEvent: anEvent.
	anEvent wasHandled ifTrue:[^self].

	"Go looking if any of our submorphs wants it"
	index := 1.
	inside := false.
	morphs := aMorph submorphs.
	[index <= morphs size] whileTrue:[
		child := morphs at: index.
		localEvt := anEvent transformedBy: (child transformedFrom: aMorph).
		(child processEvent: localEvt using: self) == #rejected ifFalse:[
			localEvt wasHandled ifTrue:[^anEvent wasHandled: true]. "done"
			inside := true.
			index := morphs size]. "break"
		index := index + 1.
	].

	inside ifFalse:[inside := aMorph containsPoint: anEvent cursorPoint event: anEvent].
	inside ifTrue:[^aMorph handleEvent: anEvent].
	^#rejected! !

!MorphicEventDispatcher methodsFor: 'dispatching' stamp: 'tbn 3/12/2010 01:53'!
dispatchEvent: anEvent with: aMorph
	"Dispatch the given event for a morph that has chosen the receiver to dispatch its events. The method implements a shortcut for repeated dispatches of events using the same dispatcher."
	anEvent type == lastType ifTrue:[^self perform: lastDispatch with: anEvent with: aMorph].
	"Otherwise classify"
	lastType := anEvent type.
	anEvent isMouse ifTrue:[
		anEvent isMouseDown ifTrue:[
			lastDispatch := #dispatchMouseDown:with:.
			^self dispatchMouseDown: anEvent with: aMorph]].
	anEvent type == #dropEvent ifTrue:[
		lastDispatch := #dispatchDropEvent:with:.
		^self dispatchDropEvent: anEvent with: aMorph].
	anEvent isWindowEvent ifTrue:[
		lastDispatch := #dispatchWindowEvent:with:.
		^self dispatchWindowEvent: anEvent with: aMorph].	
	lastDispatch := #dispatchDefault:with:.
	^self dispatchDefault: anEvent with: aMorph! !

!MorphicEventDispatcher methodsFor: 'dispatching' stamp: 'ar 10/10/2000 21:14'!
dispatchMouseDown: anEvent with: aMorph
	"Find the appropriate receiver for the event and let it handle it. Default rules:
	* The top-most chain of visible, unlocked morphs containing the event position will get a chance to handle the event.
	* When travelling down the hierarchy a prospective handler for the event is installed. This prospective handler can be used by submorphs wishing to handle the mouse down for negotiating who the receiver is.
	* When travelling up, the prospective handler is always executed. The handler needs to check if the event was handled before as well as checking if somebody else's handler has been installed.
	* If another handler has been installed but the event was not handled it means that somebody up in the hierarchy wants to handle the event.
"
	| globalPt localEvt index child morphs handler inside lastHandler |
	"Try to get out quickly"
	globalPt := anEvent cursorPoint.
	(aMorph fullBounds containsPoint: globalPt) ifFalse:[^#rejected].

	"Install the prospective handler for the receiver"
	lastHandler := anEvent handler. "in case the mouse wasn't even in the receiver"
	handler := aMorph handlerForMouseDown: anEvent.
	handler ifNotNil:[anEvent handler: handler].

	"Now give our submorphs a chance to handle the event"
	index := 1.
	morphs := aMorph submorphs.
	[index <= morphs size] whileTrue:[
		child := morphs at: index.
		localEvt := anEvent transformedBy: (child transformedFrom: aMorph).
		(child processEvent: localEvt using: self) == #rejected ifFalse:[
			"Some child did contain the point so we're part of the top-most chain."
			inside := false.
			localEvt wasHandled ifTrue:[anEvent copyHandlerState: localEvt].
			index := morphs size].
		index := index + 1.
	].

	(inside == false or:[aMorph containsPoint: anEvent cursorPoint event: anEvent]) ifTrue:[
		"Receiver is in the top-most unlocked, visible chain."
		handler ifNotNil:[handler handleEvent: anEvent].
		"Note: Re-installing the handler is not really necessary but good style."
		anEvent handler: lastHandler.
		^self
	].
	"Mouse was not on receiver nor any of its children"
	anEvent handler: lastHandler.
	^#rejected! !

!MorphicEventDispatcher methodsFor: 'dispatching' stamp: 'tbn 3/12/2010 01:53'!
dispatchWindowEvent: anEvent with: aMorph
	"Host window events do not have a position and are only dispatched to the World"
	aMorph isWorldMorph ifFalse: [^#rejected].
	anEvent wasHandled ifTrue:[^self].
	^aMorph handleEvent: anEvent
! !

Object subclass: #MorphicShortcutHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MorphicShortcutHandler commentStamp: '<historical>' prior: 0!
I am a null shortcut handler for morphic. I register myself as a tool to be replaced by somebody else who really wants to handle shortcuts.!

!MorphicShortcutHandler class methodsFor: 'tools' stamp: 'MarcusDenker 7/2/2014 10:09'!
registerToolsOn: aToolRegistry
	"commente out as Smalltalk resetTools might register this handler and break keybindings"
	"aToolRegistry register: self new as: #shortcuts"! !

!MorphicShortcutHandler methodsFor: 'shortcut-handling' stamp: 'GuillermoPolito 11/7/2013 16:47'!
handleKeystroke: aKeystrokeEvent inMorph: aMorph

	"I do nothing. I'm a null object"! !

Object subclass: #MouseOverHandler
	instanceVariableNames: 'mouseOverMorphs enteredMorphs overMorphs leftMorphs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Core-Events'!

!MouseOverHandler commentStamp: 'TorstenBergmann 2/20/2014 18:29' prior: 0!
Mouse over handler in a hand morph!

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 13:33'!
handleAsMouseEnter: anEvent 

	| asMouseEnterEvent |
	
	asMouseEnterEvent := anEvent asMouseEnter.
	enteredMorphs := enteredMorphs contents.
	enteredMorphs reverseDo: [ :anEnteredMorph | 
		self inform: asMouseEnterEvent to: anEnteredMorph originatedFrom: anEvent ifNotFocusedDo: [] ]! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 08:44'!
handleAsMouseLeave: anEvent 

	self keepLeftMorphsOrder.
	self informMouseLeaveToLeftMorphsUsing: anEvent
! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 11:59'!
handleAsMouseOver: anEvent 

	anEvent hand handleEvent: anEvent asMouseOver.
! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 11:56'!
hasLeftMorphsChanged

	^(leftMorphs isEmpty and: [ enteredMorphs position = 0 ]) not! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 13:36'!
inform: evt to: aLeftMorph originatedFrom: anEvent ifNotFocusedDo: aBlock

	^ (self is: anEvent withFocusOver: aLeftMorph) 
		ifTrue: [ self transform: evt from: anEvent andSendTo: aLeftMorph ]
		ifFalse: aBlock! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 11:41'!
informMouseLeaveToLeftMorphsUsing: anEvent 

	| asMouseLeaveEvent |
	
	asMouseLeaveEvent := anEvent asMouseLeave.
	
	leftMorphs do: [ :aLeftMorph | 
		self inform: asMouseLeaveEvent to: aLeftMorph originatedFrom: anEvent ifNotFocusedDo: [ overMorphs nextPut: aLeftMorph ] ]! !

!MouseOverHandler methodsFor: 'initialization' stamp: 'HernanWilkinson 6/4/2009 14:10'!
initialize
	mouseOverMorphs := #().
	
	self initializeTrackedMorphs
	! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 13:29'!
initializeProcessMouseOver

	leftMorphs := mouseOverMorphs asIdentitySet.
	overMorphs := WriteStream on: (Array new: leftMorphs size).
	enteredMorphs := WriteStream on: #()! !

!MouseOverHandler methodsFor: 'initialization' stamp: 'HernanWilkinson 6/12/2009 13:48'!
initializeTrackedMorphs

	leftMorphs := OrderedCollection new.
	overMorphs := WriteStream on: #().
	enteredMorphs := WriteStream on: #().
	! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 13:37'!
is: anEvent withFocusOver: aMorph 

	| focusedMorph |
	
	focusedMorph := anEvent hand mouseFocus.
	^ aMorph = focusedMorph or: [ aMorph hasOwner: focusedMorph ]! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 13:32'!
keepLeftMorphsOrder

	leftMorphs size > 1 ifTrue: [ leftMorphs := mouseOverMorphs intersection: leftMorphs ]
	! !

!MouseOverHandler methodsFor: 'event handling' stamp: 'HernanWilkinson 6/12/2009 13:40'!
noticeMouseOver: aMorph event: anEvent
	"Remember that the mouse is currently over some morph"

	leftMorphs remove: aMorph ifAbsent: [ enteredMorphs nextPut: aMorph ].
	overMorphs nextPut: aMorph.
! !

!MouseOverHandler methodsFor: 'event handling' stamp: 'HernanWilkinson 6/12/2009 13:51'!
processMouseOver: anEvent 

	self initializeProcessMouseOver.
	self handleAsMouseOver: anEvent.
	self hasLeftMorphsChanged ifTrue: [ 
		self handleAsMouseLeave: anEvent.
		self handleAsMouseEnter: anEvent.
		self rememberOverList ].
	
	self initializeTrackedMorphs 
! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 11:58'!
rememberOverList

	mouseOverMorphs := overMorphs contents.
! !

!MouseOverHandler methodsFor: 'event handling - private' stamp: 'HernanWilkinson 6/12/2009 13:38'!
transform: anEvent from: originalEvent andSendTo: aMorph

	| transformedEvent |

	transformedEvent := anEvent transformedBy: (aMorph transformedFrom: originalEvent hand).
	^ aMorph handleEvent: transformedEvent! !

!Object methodsFor: '*Morphic-Core' stamp: 'JuanVuletich 11/1/2010 15:18'!
currentWorld
	"Answer a morphic world that is the current UI focus."
	^UIManager default currentWorld! !

Object subclass: #WorldState
	instanceVariableNames: 'hands viewBox canvas damageRecorder stepList lastStepTime lastStepMessage lastCycleTime alarms lastAlarmTime menuBuilder eventThreads'
	classVariableNames: 'CanSurrenderToOS DebugShowDamage DeferredUIMessages DesktopMenuPragmaKeyword DesktopMenuTitle EasySelectingWorld LastCycleTime MinCycleLapse ServerMode ShowUpdateOptionInWorldMenu'
	poolDictionaries: ''
	category: 'Morphic-Core-Worlds'!

!WorldState commentStamp: 'ls 7/10/2003 19:30' prior: 0!
The state of a Morphic world.  (This needs some serious commenting!!!!)


The MinCycleLapse variable holds the minimum amount of time that a morphic cycle is allowed to take.  If a cycle takes less than this, then interCyclePause: will wait until the full time has been used up.!

!WorldState class methodsFor: 'accessing' stamp: 'AlainPlantec 12/6/2009 21:40'!
MinCycleLapse: milliseconds
	"set the minimum amount of time that may transpire between two calls to doOneCycle"
	MinCycleLapse := milliseconds ifNotNil: [ milliseconds rounded ].! !

!WorldState class methodsFor: 'deferred message' stamp: 'RAA 7/15/2000 12:58'!
addDeferredUIMessage: valuableObject

	self deferredUIMessages nextPut: valuableObject.

! !

!WorldState class methodsFor: 'testing os' stamp: 'RAA 8/14/2000 16:40'!
canSurrenderToOS: aBoolean

	CanSurrenderToOS := aBoolean! !

!WorldState class methodsFor: 'world menu items' stamp: 'SeanDeNigris 8/29/2011 17:33'!
clearHaltOnce
	Halt disableHaltOnce.
	World resetWorldMenu! !

!WorldState class methodsFor: 'accessing' stamp: 'AlainPlantec 12/6/2009 21:42'!
debugShowDamage
	^ DebugShowDamage ifNil: [DebugShowDamage := false]! !

!WorldState class methodsFor: 'accessing' stamp: 'AlainPlantec 12/6/2009 21:42'!
debugShowDamage: aBoolean
	DebugShowDamage := aBoolean! !

!WorldState class methodsFor: 'settings' stamp: 'HilaireFernandes 9/9/2010 15:25'!
defaultWorldMenu
	self desktopMenuTitle: 'World'.
	self desktopMenuPragmaKeyword: 'worldMenu'.! !

!WorldState class methodsFor: 'deferred message' stamp: 'AlainPlantec 7/9/2013 12:48'!
defer: valuableObject

	self addDeferredUIMessage: valuableObject.

! !

!WorldState class methodsFor: 'deferred message' stamp: 'IgorStasenko 6/22/2012 01:42'!
deferredUIMessages

	^DeferredUIMessages ifNil: [DeferredUIMessages := WaitfreeQueue  new].
! !

!WorldState class methodsFor: 'settings' stamp: 'HilaireFernandes 9/9/2010 15:25'!
desktopMenuPragmaKeyword: aString
	DesktopMenuPragmaKeyword := aString.
	World resetWorldMenu.
! !

!WorldState class methodsFor: 'settings' stamp: 'AlainPlantec 12/19/2009 23:56'!
desktopMenuTitle: aString
	DesktopMenuTitle := aString! !

!WorldState class methodsFor: 'settings' stamp: 'SeanDeNigris 6/14/2013 01:52'!
discoveredMenuPragmaKeyword
	^ DesktopMenuPragmaKeyword ifNil: [DesktopMenuPragmaKeyword := 'worldMenu']! !

!WorldState class methodsFor: 'settings' stamp: 'SeanDeNigris 6/14/2013 02:15'!
discoveredMenuTitle
	^ DesktopMenuTitle ifNil: [DesktopMenuTitle := 'World']! !

!WorldState class methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 11:34'!
easySelectingWorld
	^ EasySelectingWorld ifNil: [EasySelectingWorld := false]! !

!WorldState class methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 11:34'!
easySelectingWorld: aBoolean
	EasySelectingWorld := aBoolean! !

!WorldState class methodsFor: 'world menu items' stamp: 'EstebanLorenzano 5/14/2013 09:44'!
helpOn: aBuilder 

	<worldMenu>  
	(aBuilder item: #Help)
		order: 4.0;
		icon: Smalltalk ui icons smallHelpIcon;
		withSeparatorAfter 
		! !

!WorldState class methodsFor: 'initialization' stamp: 'IgorStasenko 6/22/2012 01:45'!
initialize
	"WorldState initialize"

	MinCycleLapse := 20.		"allows 50 frames per second..."
	DeferredUIMessages := nil.! !

!WorldState class methodsFor: 'accessing' stamp: 'RAA 7/23/2000 17:28'!
lastCycleTime

	^LastCycleTime! !

!WorldState class methodsFor: 'world menu items' stamp: 'IgorStasenko 4/26/2011 16:57'!
mostUsedToolsOn: aBuilder   
	<worldMenu>
	(aBuilder group: #MostUsedTools)
		withSeparatorAfter;
		order: 0;
		with: [
			(aBuilder item: #'System Browser') action: [Smalltalk tools openClassBrowser]; icon: Smalltalk tools browser taskbarIcon.
			(aBuilder item: #Workspace) action: [Smalltalk tools openWorkspace]; icon: Smalltalk tools workspace taskbarIcon.
			
			Smalltalk globals at: #TestRunner ifPresent: [ :class |
				(aBuilder item: #'Test Runner') action: [Smalltalk tools openTestRunner]; icon: class taskbarIcon.
			].
			(aBuilder item: #'Monticello Browser') action: [Smalltalk tools openMonticelloBrowser]; icon: Smalltalk tools monticelloBrowser taskbarIcon]! !

!WorldState class methodsFor: 'world menu items' stamp: 'EstebanLorenzano 5/14/2013 09:44'!
quitItemsOn: aBuilder 
	<worldMenu>
	(aBuilder group: #QuitPharo)
		order: 9999; 
		with: [
			(aBuilder item: #'Save')
				target: self;
				selector: #saveSession;
				help: 'save the current version of the image on disk';
				icon: Smalltalk ui icons smallSaveIcon.
			(aBuilder item: #'Save as...')
				target: self;
				selector: #saveAs;
				help: 'save the current version of the image on disk under a new name.';
				icon: Smalltalk ui icons smallSaveAsIcon.
			(aBuilder item: #'Save and quit')
				target: self;
				selector: #saveAndQuit;
				help: 'save the current image on disk, and quit Pharo.';
				icon: Smalltalk ui icons smallQuitIcon.
			(aBuilder item: #'Quit')
				target: self;
				selector: #quitSession;
				help: 'quit Pharo.';
				icon: Smalltalk ui icons smallQuitIcon ]! !

!WorldState class methodsFor: 'world menu items' stamp: 'MarcusDenker 1/10/2014 15:43'!
quitSession
	| save |
	save := Smalltalk tools userManager canSaveImage
		ifFalse: [ false ]
		ifTrue: [ self confirm: 'Save changes before quitting?' translated orCancel: [ ^ self ] ].
	save 
		ifTrue: [Cursor write showWhile: [ Smalltalk snapshot: true andQuit: true ]] 
		ifFalse: [Smalltalk snapshot: false andQuit: true]
 ! !

!WorldState class methodsFor: 'world menu items' stamp: 'GuillermoPolito 3/16/2012 11:54'!
saveAndQuit
	Cursor write showWhile: [
		Smalltalk snapshot: true andQuit: true
	]! !

!WorldState class methodsFor: 'world menu items' stamp: 'StephaneDucasse 6/11/2012 18:12'!
saveAs
	Smalltalk saveAs.! !

!WorldState class methodsFor: 'world menu items' stamp: ''!
saveSession
	Cursor write showWhile: [
		Smalltalk
			snapshot: true
			andQuit: false.
	]! !

!WorldState class methodsFor: 'world menu items' stamp: 'SeanDeNigris 2/16/2014 20:49'!
screenShotCommandOn: aBuilder 
	<worldMenu>
	(aBuilder item: #Screenshot)
		parent: #Tools;
		target: World;
		order: 0.59;
		selector: #makeAScreenshot;
		label: 'Screenshot' translated;
		icon: Smalltalk ui icons smallScreenshotIcon! !

!WorldState class methodsFor: 'settings' stamp: 'AlainPlantec 12/14/2009 14:02'!
serverMode
	^ ServerMode ifNil: [ServerMode := false]! !

!WorldState class methodsFor: 'settings' stamp: 'AlainPlantec 12/14/2009 14:07'!
serverMode: aBoolean
	(aBoolean = true or: [aBoolean = false or: [aBoolean isNil]])
		ifTrue: [ServerMode := aBoolean]! !

!WorldState class methodsFor: 'world menu items' stamp: 'SeanDeNigris 8/29/2011 17:30'!
setHaltOnce
	Halt enableHaltOnce.
	World resetWorldMenu! !

!WorldState class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 15:59'!
showUpdateOptionInWorldMenu
	^ ShowUpdateOptionInWorldMenu ifNil: [ShowUpdateOptionInWorldMenu := true]! !

!WorldState class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 15:59'!
showUpdateOptionInWorldMenu: aBoolean
	ShowUpdateOptionInWorldMenu := aBoolean! !

!WorldState class methodsFor: 'world menu items' stamp: 'IgorStasenko 1/2/2012 19:09'!
startMessageTally
	(self confirm: 'MessageTally will start now,
and stop when the cursor goes
to the top of the screen' translated) 
		ifTrue: [TimeProfiler spyAllOn: [[World activeHand position y > 0] 
					whileTrue: [World doOneCycle]]]! !

!WorldState class methodsFor: 'world menu items' stamp: 'IgorStasenko 1/2/2012 19:10'!
startThenBrowseMessageTally
	"Tally only the UI process"
	
	(self confirm: 'MessageTally the UI process until the
mouse pointer goes to the top of the screen')
		ifTrue: [TimeProfiler
				onBlock: [[World activeHand position y > 10]
						whileTrue: [World doOneCycle]]]! !

!WorldState class methodsFor: 'world menu items' stamp: 'MarcusDenker 10/5/2013 17:41'!
systemOn: aBuilder 
	
	<worldMenu>  
	(aBuilder item: #System)
		order: 3.0;
		icon: Smalltalk ui icons smallConfigurationIcon;
		with: [
			(aBuilder item: #'About...') 
				order: 0;
				icon: Smalltalk ui icons smallLanguageIcon;
				action: [Smalltalk aboutThisSystem].
			(aBuilder item: #'Software update')
				order: 2;
				icon: Smalltalk ui icons smallUpdateIcon;
				precondition: [self showUpdateOptionInWorldMenu];
				action: [ UpdateStreamer new updateFromServer ];
				help: 'Load latest code updates via the internet';
			withSeparatorAfter.
			
			(aBuilder item: #'Start profiling all Processes') 
				action: [self startMessageTally].
			(aBuilder item: #'Start profiling UI ') 
				action: [self startThenBrowseMessageTally].
			(aBuilder item: #'Space left') 
				action: [Smalltalk informSpaceLeftAfterGarbageCollection].
			(aBuilder item: #'Do Image Cleanup') 
				action: [Smalltalk cleanUp: true ];
			withSeparatorAfter.
			
			
			(aBuilder item: #'Start drawing again') 
				action: [World  resumeAfterDrawError].
			(aBuilder item: #'Start stepping again') 
				action: [World  resumeAfterStepError]; 
			withSeparatorAfter.
			
			(aBuilder item: #'Restore display (r)') 
				action: [World restoreMorphicDisplay].
			(aBuilder item: #'Clear all breakpoints') 
				action: [BreakpointManager clear].
			Halt isHaltOnceEnabled
				ifTrue: [(aBuilder item: 'Disable halt/inspect once') action: [self clearHaltOnce]]
				ifFalse: [(aBuilder item: 'Enable halt/inspect once') action: [self setHaltOnce]]].! !

!WorldState class methodsFor: 'accessing' stamp: 'EstebanLorenzano 5/14/2013 09:43'!
theme
	^ Smalltalk ui theme! !

!WorldState class methodsFor: 'world menu items' stamp: 'SvenVanCaekenberghe 1/15/2014 19:17'!
windowsOn: aBuilder 
	<worldMenu>
	(aBuilder item: #Windows)
		order: 4.0;
		withSeparatorAfter ;
		icon: Smalltalk ui icons smallWindowIcon;
		with: [
			(aBuilder item: # 'Collapse all windows') 
				action: [World collapseAll];
				help: 'Reduce all open windows to collapsed forms that only show titles' translated.
			(aBuilder item: # 'Expand all windows') 
				action: [World expandAll];
				help: 'Expand all collapsed windows back to their expanded forms' translated.
			(aBuilder item: # 'Fit all windows') 
				action: [World fitAll];
				help: 'Fit all open windows as visible in World' translated.
			(aBuilder item: #'Close all debuggers') 
				action: [Smalltalk tools debugger closeAllDebuggers].
			(aBuilder item: # 'Send top window to back (\)') 
				action: [SystemWindow sendTopWindowToBack];
				help: 'Make the topmost window become the backmost one, and activate the window just beneath it.' translated.
			(aBuilder item: # 'Move windows onscreen') 
				action: [World bringWindowsFullOnscreen];
				help: 'Make all windows fully visible on the screen' translated.
			(aBuilder item: # 'Delete unchanged windows') 
				action: [World closeUnchangedWindows];
				help: 'Deletes all windows that do not have unsaved text edits.' translated.
			(aBuilder item: # 'Delete all windows discarding edits') 
				action: [World closeAllWindowsDiscardingChanges];
				help: 'Deletes all windows even if they have unsaved text edits.' translated.
			(aBuilder item: # 'Toggle full screen mode') 
				action: [Display toggleFullscreen];
				help: 'Turn full screen mode on when it is off, off when it is on.' translated.
			]! !

!WorldState methodsFor: 'hands' stamp: 'ar 1/22/2001 14:26'!
activeHand

	^ ActiveHand! !

!WorldState methodsFor: 'alarms' stamp: 'ar 9/11/2000 16:43'!
addAlarm: aSelector withArguments: argArray for: aTarget at: scheduledTime
	"Add a new alarm with the given set of parameters"
	self alarms add: 
		(MorphicAlarm 
			scheduledAt: scheduledTime
			receiver: aTarget
			selector: aSelector
			arguments: argArray).! !

!WorldState methodsFor: 'hands' stamp: 'ar 10/26/2000 14:51'!
addHand: aHandMorph
	"Add the given hand to the list of hands for this world."

	hands := (hands copyWithout: aHandMorph) copyWith: aHandMorph.
! !

!WorldState methodsFor: 'alarms' stamp: 'ar 9/11/2000 17:11'!
adjustAlarmTimes: nowTime
	"Adjust the alarm times after some clock weirdness (such as roll-over, image-startup etc)"
	| deltaTime |
	deltaTime := nowTime - lastAlarmTime.
	self alarms do:[:alarm| alarm scheduledTime: alarm scheduledTime + deltaTime].! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:23'!
adjustWakeupTimes: now
	"Fix the wakeup times in my step list. This is necessary when this world has been restarted after a pause, say because some other view had control, after a snapshot, or because the millisecond clock has wrapped around. (The latter is a rare occurence with a 32-bit clock!!)"
	| deltaTime |
	deltaTime := now - lastStepTime.
	stepList do:[:entry| entry scheduledTime: entry scheduledTime + deltaTime].
	lastStepTime := now.
! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 15:22'!
adjustWakeupTimesIfNecessary
	"Fix the wakeup times in my step list if necessary. This is needed after a snapshot, after a long pause (say because some other view had control or because the user was selecting from an MVC-style menu) or when the millisecond clock wraps around (a very rare occurence with a 32-bit clock!!)."

	| now |
	now := Time millisecondClockValue.
	((now < lastStepTime) or: [(now - lastStepTime) > 5000])
		 ifTrue: [self adjustWakeupTimes: now].  "clock slipped"
! !

!WorldState methodsFor: 'alarms' stamp: 'nice 4/16/2009 19:03'!
alarmSortBlock
	^[ :alarm1 :alarm2 | 
		alarm1 scheduledTime < alarm2 scheduledTime.
	]! !

!WorldState methodsFor: 'alarms' stamp: 'RAA 1/5/2001 10:46'!
alarms

	^alarms ifNil: [alarms := Heap sortBlock: self alarmSortBlock]! !

!WorldState methodsFor: 'canvas' stamp: 'stephane.ducasse 9/25/2008 18:10'!
assuredCanvas

	(canvas isNil or: [(canvas extent ~= viewBox extent) or: [canvas form depth ~= Display depth]])
		ifTrue:
			["allocate a new offscreen canvas the size of the window"
			self canvas: (Display defaultCanvasClass extent: viewBox extent)].
	^ self canvas! !

!WorldState methodsFor: 'canvas' stamp: 'di 6/7/1999 17:44'!
canvas

	^ canvas! !

!WorldState methodsFor: 'canvas' stamp: 'ClementBera 7/26/2013 17:06'!
canvas: x 
	canvas := x.
	damageRecorder 
		ifNil: [damageRecorder := DamageRecorder new]
		ifNotNil: [damageRecorder doFullRepaint]! !

!WorldState methodsFor: 'update cycle' stamp: 'RAA 5/24/2000 13:13'!
checkIfUpdateNeeded

	damageRecorder updateIsNeeded ifTrue: [^true].
	hands do: [:h | (h hasChanged and: [h needsToBeDrawn]) ifTrue: [^true]].
	^false  "display is already up-to-date"
! !

!WorldState methodsFor: 'stepping' stamp: 'AlainPlantec 10/17/2009 19:02'!
cleanseStepListForWorld: aWorld
	"Remove morphs from the step list that are not in this World"

	| deletions morphToStep |
	deletions := nil.
	stepList do: [:entry |
		morphToStep := entry receiver.
		morphToStep world == aWorld ifFalse:[
			deletions ifNil: [deletions := OrderedCollection new].
			deletions addLast: entry]].

	deletions ifNotNil:[
		deletions do: [:entry|
			self stopStepping: entry receiver]].

	self alarms copy do:[:entry|
		morphToStep := entry receiver.
		(morphToStep isMorph and:[morphToStep world == aWorld]) 
			ifFalse:[self removeAlarm: entry selector for: entry receiver]].! !

!WorldState methodsFor: 'object filein' stamp: 'RAA 1/5/2001 10:51'!
convertAlarms
	
	alarms ifNotNil: [alarms sortBlock: self alarmSortBlock].	"ensure cleaner block"

! !

!WorldState methodsFor: 'object filein' stamp: 'nice 1/5/2010 15:59'!
convertStepList
	"Convert the old-style step list (an Array of Arrays) into the new-style StepMessage heap"

	| newList |
	(stepList isKindOf: Heap) 
		ifTrue: 
			[^stepList sortBlock: self stepListSortBlock	"ensure that we have a cleaner block"].
	newList := Heap sortBlock: self stepListSortBlock.
	stepList do: 
			[:entry | | wakeupTime morphToStep | 
			wakeupTime := entry second.
			morphToStep := entry first.
			newList add: (StepMessage 
						scheduledAt: wakeupTime
						stepTime: nil
						receiver: morphToStep
						selector: #stepAt:
						arguments: nil)].
	stepList := newList! !

!WorldState methodsFor: 'deferred message' stamp: 'AlainPlantec 7/9/2013 12:49'!
defer: aValuable
	self class defer: aValuable! !

!WorldState methodsFor: 'worldmenu building' stamp: 'SeanDeNigris 6/14/2013 02:53'!
discoveredMenuOn: aBuilder

	^ aBuilder menuEntitled: self discoveredMenuTitle! !

!WorldState methodsFor: 'settings' stamp: 'SeanDeNigris 6/14/2013 01:52'!
discoveredMenuPragmaKeyword
	^ self class discoveredMenuPragmaKeyword! !

!WorldState methodsFor: 'settings' stamp: 'SeanDeNigris 6/14/2013 02:15'!
discoveredMenuTitle
	^ self class discoveredMenuTitle! !

!WorldState methodsFor: 'update cycle' stamp: 'nice 1/5/2010 15:59'!
displayWorld: aWorld submorphs: submorphs
	"Update this world's display."

	| deferredUpdateMode handsToDraw allDamage |

	submorphs do: [:m | m fullBounds].  "force re-layout if needed"
	self checkIfUpdateNeeded ifFalse: [^ self].  "display is already up-to-date"

	deferredUpdateMode := self doDeferredUpdatingFor: aWorld.
	deferredUpdateMode ifFalse: [self assuredCanvas].
	canvas roundCornersOf: aWorld during:[ | worldDamageRects handDamageRects |
		worldDamageRects := self drawWorld: aWorld submorphs: submorphs invalidAreasOn: canvas.  "repair world's damage on canvas"
		"self handsDo:[:h| h noticeDamageRects: worldDamageRects]."
		handsToDraw := self selectHandsToDrawForDamage: worldDamageRects.
		handDamageRects := handsToDraw collect: [:h | h savePatchFrom: canvas].
		allDamage := worldDamageRects, handDamageRects.

		handsToDraw reverseDo: [:h | canvas fullDrawMorph: h].  "draw hands onto world canvas"
	].
	"*make this true to flash damaged areas for testing*"
	self class debugShowDamage ifTrue: [aWorld flashRects: allDamage color: Color black].

	canvas finish.
	"quickly copy altered rects of canvas to Display:"
	deferredUpdateMode
		ifTrue: [self forceDamageToScreen: allDamage]
		ifFalse: [canvas showAt: aWorld viewBox origin invalidRects: allDamage].
	handsToDraw do: [:h | h restoreSavedPatchOn: canvas].  "restore world canvas under hands"
	Display deferUpdates: false; forceDisplayUpdate.
! !

!WorldState methodsFor: 'update cycle' stamp: 'ar 6/28/2003 01:07'!
displayWorldSafely: aWorld
	"Update this world's display and keep track of errors during draw methods."

	[aWorld displayWorld] ifError: [:err :rcvr |
		"Handle a drawing error"
		| errCtx errMorph |
		errCtx := thisContext.
		[
			errCtx := errCtx sender.
			"Search the sender chain to find the morph causing the problem"
			[errCtx notNil and:[(errCtx receiver isMorph) not]] 
				whileTrue:[errCtx := errCtx sender].
			"If we're at the root of the context chain then we have a fatal drawing problem"
			errCtx ifNil:[^self handleFatalDrawingError: err].
			errMorph := errCtx receiver.
			"If the morph causing the problem has already the #drawError flag set,
			then search for the next morph above in the caller chain."
			errMorph hasProperty: #errorOnDraw
		] whileTrue.
		errMorph setProperty: #errorOnDraw toValue: true.
		"Install the old error handler, so we can re-raise the error"
		rcvr error: err.
	].! !

!WorldState methodsFor: 'update cycle' stamp: 'MarcusDenker 7/9/2012 15:28'!
doDeferredUpdatingFor: aWorld
        "If this platform supports deferred updates, then make my canvas be the Display (or a rectangular portion of it), set the Display to deferred update mode, and answer true. Otherwise, do nothing and answer false. One can set the class variable DisableDeferredUpdates to true to completely disable the deferred updating feature."

	(Display deferUpdates: true) ifNil: [^ false].  "deferred updates not supported"
	(canvas notNil and: [canvas form == Display]) ifFalse: [
			aWorld viewBox: Display boundingBox.    "do first since it may clear canvas"
			self canvas: (Display getCanvas copyClipRect: Display boundingBox)].
	^ true
! !

!WorldState methodsFor: 'canvas' stamp: 'RAA 5/25/2000 15:12'!
doFullRepaint

	damageRecorder doFullRepaint
! !

!WorldState methodsFor: 'update cycle' stamp: 'adrian_lienhard 7/18/2009 15:30'!
doOneCycleFor: aWorld
	"Do one cycle of the interaction loop. This method is called repeatedly when the world is running. This is a moderately private method; a better alternative is usually either to wait for events or to check the state of things from #step methods."

	self interCyclePause: MinCycleLapse.
	self doOneCycleNowFor: aWorld.! !

!WorldState methodsFor: 'event-handling' stamp: 'md 4/30/2008 16:33'!
doOneCycleNowFor: aWorld
	"Immediately do one cycle of the interaction loop.
	This should not be called directly, but only via doOneCycleFor:"

	DisplayScreen checkForNewScreenSize.

	"process user input events"
	LastCycleTime := Time millisecondClockValue.
	self handsDo: [:h |
		ActiveHand := h.
		h processEvents.
		ActiveHand := nil
	].

	"the default is the primary hand"
	ActiveHand := self hands first.

	aWorld runStepMethods.		"there are currently some variations here"
	self displayWorldSafely: aWorld.! !

!WorldState methodsFor: 'update cycle' stamp: 'ar 1/22/2001 14:26'!
doOneSubCycleFor: aWorld
	"Like doOneCycle, but preserves activeHand."

	| currentHand |
	currentHand := ActiveHand.
	self doOneCycleFor: aWorld.
	ActiveHand := currentHand! !

!WorldState methodsFor: 'update cycle' stamp: 'nice 1/5/2010 15:59'!
drawWorld: aWorld submorphs: submorphs invalidAreasOn: aCanvas 
	"Redraw the damaged areas of the given canvas and clear the damage list. Return a collection of the areas that
were redrawn."

	| rectList n morphs rects validList |
	rectList := damageRecorder invalidRectsFullBounds: aWorld viewBox.
	"sort by areas to draw largest portions first"
	rectList := rectList asArray sort: [:r1 :r2 | r1 area > r2 area].
	damageRecorder reset.
	n := submorphs size.
	morphs := OrderedCollection new: n * 2.
	rects := OrderedCollection new: n * 2.
	validList := OrderedCollection new: n * 2.
	rectList do: 
			[:dirtyRect | 
			dirtyRect allAreasOutsideList: validList
				do: 
					[:r | | mm rect i c remnantIntersects remnants rectToFill | 
					"Experimental top-down drawing --
			Traverses top to bottom, stopping if the entire area is filled.
			If only a single rectangle remains, then continue with the reduced rectangle."

					rectToFill := r.
					remnants := OrderedCollection with: r.
					i := 1.
					[remnants isEmpty or: [i > n]] whileFalse: 
							[mm := submorphs at: i.
							((remnantIntersects := remnants select: [:each | (mm fullBounds intersects: each)]) notEmpty and: [mm visible]) 
								ifTrue: 
									[morphs addLast: mm.
									
									rects addLast: (Rectangle merging: (remnantIntersects collect: [:each | mm fullBounds intersect: each])).
									remnants removeAll: remnantIntersects.
									remnantIntersects do: [:eachIntersect | remnants addAll: (mm areasRemainingToFill: eachIntersect)].
									remnants size = 1 ifTrue: [rectToFill := remnants first].
									remnants isEmpty ifTrue: [rectToFill := nil]].
							i := i + 1].
					"Now paint from bottom to top, but using the reduced rectangles."
					rectToFill 
						ifNotNil: [aWorld drawOn: (c := aCanvas copyClipRect: rectToFill)].
					[morphs isEmpty] whileFalse: 
							[(rect := rects removeLast) == rectToFill 
								ifFalse: [c := aCanvas copyClipRect: (rectToFill := rect)].
							c fullDrawMorph: morphs removeLast].
					morphs reset.
					rects reset.
					validList add: r]].
	^validList! !

!WorldState methodsFor: 'event-handling' stamp: 'NickPapoulias 12/28/2014 21:12'!
eventThreads

	^ eventThreads ifNil: [ eventThreads := OrderedCollection new ]! !

!WorldState methodsFor: 'worldmenu building' stamp: 'MarcusDenker 9/29/2013 09:00'!
fallbackMenuOn: menu
	"Build the menu that is put up if something is going wrong with the menubuilder"
	
	menu addTitle: 'Fallback menu'.
	menu commandKeyHandler: self.
	menu addLine.
	menu
		defaultTarget: Smalltalk tools;
		addList: Smalltalk tools menuItems.
	menu addLine.
	menu add: 'Save'
		target: self class
		selector: #saveSession.
	menu add: 'Save as...'
		target: self class
		selector: #saveAs.
	menu add: 'Save and quit' 
		target: self class
		selector: #saveAndQuit.
	menu add: 'Quit' 
		target: self class
		selector: #quitSession.! !

!WorldState methodsFor: 'update cycle' stamp: 'stephane.ducasse 9/25/2008 18:11'!
forceDamageToScreen: allDamage

	Display forceDamageToScreen: allDamage.
	! !

!WorldState methodsFor: 'update cycle' stamp: 'alain.plantec 6/2/2008 08:45'!
handleFatalDrawingError: errMsg
	"Handle a fatal drawing error."
	Display deferUpdates: false. "Just in case"
	self primitiveError: errMsg.

	"Hm... we should jump into a 'safe' worldState here, but how do we find it?!!"! !

!WorldState methodsFor: 'hands' stamp: 'di 6/7/1999 17:40'!
hands

	^ hands! !

!WorldState methodsFor: 'hands' stamp: 'RAA 5/24/2000 10:13'!
handsDo: aBlock

	^ hands do: aBlock! !

!WorldState methodsFor: 'hands' stamp: 'RAA 5/24/2000 12:09'!
handsReverseDo: aBlock

	^ hands reverseDo: aBlock! !

!WorldState methodsFor: 'initialization' stamp: 'StephaneDucasse 7/2/2013 16:48'!
initialize

	super initialize.
	hands := Array new.
	damageRecorder:= DamageRecorder new.
	stepList := Heap sortBlock: self stepListSortBlock.
	lastStepTime := 0.
	lastAlarmTime := 0.
	self viewBox: Display boundingBox! !

!WorldState methodsFor: 'update cycle' stamp: 'AlainPlantec 1/7/2010 22:54'!
interCyclePause: milliSecs
	"delay enough that the previous cycle plus the amount of delay will equal milliSecs.  If the cycle is already expensive, then no delay occurs.  However, if the system is idly waiting for interaction from the user, the method will delay for a proportionally long time and cause the overall CPU usage of Squeak to be low.
	If self serverMode returns true then, always do a complete delay of 50ms, independant of my argument. This prevents the freezing problem described in Mantis #6581"

	| currentTime wait |
	self serverMode
		ifFalse: [
			(lastCycleTime notNil and: [CanSurrenderToOS ~~ false]) ifTrue: [ 
				currentTime := Time millisecondClockValue.
				wait := lastCycleTime + milliSecs - currentTime.
				(wait > 0 and: [ wait <= milliSecs ] ) ifTrue: [
					(Delay forMilliseconds: wait) wait ] ] ]
		ifTrue: [ (Delay forMilliseconds: 50) wait ].

	lastCycleTime := Time millisecondClockValue.
	CanSurrenderToOS := true.! !

!WorldState methodsFor: 'hands' stamp: 'GabrielOmarCotelli 11/30/2013 17:13'!
is: hand overlappingDamagedAreaIn: damageList
	| handBounds |
	handBounds := hand fullBounds.
	^ damageList anySatisfy: [ :r | r intersects: handBounds ]! !

!WorldState methodsFor: 'settings' stamp: 'AlainPlantec 12/10/2009 11:37'!
isEasySelecting
	^ self class easySelectingWorld! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:09'!
isStepping: aMorph
	"Return true if the given morph is in the step list."
	lastStepMessage ifNotNil:[(lastStepMessage receiver == aMorph) ifTrue:[^true]].
	stepList do:[:entry| entry receiver == aMorph ifTrue:[^true]].
	^ false! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:09'!
isStepping: aMorph selector: aSelector
	"Return true if the given morph is in the step list."
	lastStepMessage ifNotNil:[
		(lastStepMessage receiver == aMorph and:[lastStepMessage selector == aSelector])
			ifTrue:[^true]].
	stepList do:[:entry| (entry receiver == aMorph and:[entry selector == aSelector]) ifTrue:[^true]].
	^ false! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:05'!
listOfSteppingMorphs
	^stepList collect:[:entry| entry receiver].
! !

!WorldState methodsFor: 'worldmenu building' stamp: 'SeanDeNigris 6/14/2013 03:09'!
menuBuilder
	^ menuBuilder ifNil: [menuBuilder := PragmaMenuBuilder pragmaKeyword: self discoveredMenuPragmaKeyword model: self ]! !

!WorldState methodsFor: 'event-handling' stamp: 'NickPapoulias 12/29/2014 12:42'!
newDoOneCycleNowFor: aWorld
	"Immediately do one cycle of the interaction loop.
	This should not be called directly, but only via doOneCycleFor:"

	self suspendEventThreadScheduling.  

	"process user input events"
	self scheduleEventThread: [
			DisplayScreen checkForNewScreenSize.
			LastCycleTime := Time millisecondClockValue.
			ActiveHand := self hands first. "only one hand -- backward compatible code"
			ActiveHand processEvents.
			aWorld runStepMethods.	"there are currently some variations here"
	].
	
	"note try scheduling instead of directly forking on creation and run on thread round-robing at a time on resume.
	if everything fails fork doIts -- or consider scheduling per system window"
	
	self displayWorldSafely: aWorld.		
	
	self resumeEventThreadScheduling.! !

!WorldState methodsFor: 'canvas' stamp: 'ar 1/30/2001 23:25'!
recordDamagedRect: damageRect

	damageRecorder ifNotNil: [damageRecorder recordInvalidRect: damageRect truncated]
! !

!WorldState methodsFor: 'alarms' stamp: 'GabrielOmarCotelli 11/30/2013 17:09'!
removeAlarm: aSelector for: aTarget
	"Remove the alarm with the given selector"

	self alarms
		detect: [ :any | any receiver == aTarget and: [ any selector == aSelector ] ]
		ifFound: [ :alarm | self alarms remove: alarm ]! !

!WorldState methodsFor: 'hands' stamp: 'ar 1/22/2001 14:26'!
removeHand: aHandMorph
	"Remove the given hand from the list of hands for this world."

	(hands includes: aHandMorph) ifFalse: [^self].
	hands := hands copyWithout: aHandMorph.
	ActiveHand == aHandMorph ifTrue: [ActiveHand := nil].
! !

!WorldState methodsFor: 'event-handling' stamp: 'NickPapoulias 12/29/2014 12:30'!
rescheduleRunningThread

	self eventThreads 
		addLast: (self eventThreads 
						remove: self eventThreads first).! !

!WorldState methodsFor: 'canvas' stamp: 'RAA 5/25/2000 15:10'!
resetDamageRecorder

	damageRecorder reset
! !

!WorldState methodsFor: 'worldmenu building' stamp: 'HilaireFernandes 9/9/2010 14:55'!
resetWorldMenu
	menuBuilder 
		ifNotNil: [menuBuilder reset.
			menuBuilder := nil]! !

!WorldState methodsFor: 'event-handling' stamp: 'NickPapoulias 12/29/2014 12:26'!
resumeEventThreadScheduling

	self eventThreads first resume.! !

!WorldState methodsFor: 'stepping' stamp: 'dgd 2/22/2003 13:31'!
runLocalStepMethodsIn: aWorld 
	"Run morph 'step' methods (LOCAL TO THIS WORLD) whose time has come. Purge any morphs that are no longer in this world.
	ar 3/13/1999: Remove buggy morphs from the step list so that they don't raise repeated errors."

	| now morphToStep stepTime priorWorld |
	now := Time millisecondClockValue.
	priorWorld := ActiveWorld.
	ActiveWorld := aWorld.
	self triggerAlarmsBefore: now.
	stepList isEmpty 
		ifTrue: 
			[ActiveWorld := priorWorld.
			^self].
	(now < lastStepTime or: [now - lastStepTime > 5000]) 
		ifTrue: [self adjustWakeupTimes: now].	"clock slipped"
	[stepList isEmpty not and: [stepList first scheduledTime < now]] 
		whileTrue: 
			[lastStepMessage := stepList removeFirst.
			morphToStep := lastStepMessage receiver.
			(morphToStep shouldGetStepsFrom: aWorld) 
				ifTrue: 
					[lastStepMessage value: now.
					lastStepMessage ifNotNil: 
							[stepTime := lastStepMessage stepTime ifNil: [morphToStep stepTime].
							lastStepMessage scheduledTime: now + (stepTime max: 1).
							stepList add: lastStepMessage]].
			lastStepMessage := nil].
	lastStepTime := now.
	ActiveWorld := priorWorld! !

!WorldState methodsFor: 'stepping' stamp: 'StephaneDucasse 1/31/2011 16:51'!
runStepMethodsIn: aWorld
	"Perform periodic activity inbetween event cycles"
	
	| queue nextInQueue|
	"If available dispatch some deferred UI Message"
	queue := self class deferredUIMessages.
	[(nextInQueue := queue nextOrNil) isNil]
		whileFalse: [ nextInQueue value].
	self runLocalStepMethodsIn: aWorld.! !

!WorldState methodsFor: 'event-handling' stamp: 'NickPapoulias 12/28/2014 21:18'!
scheduleEventThread: aBlock

	self eventThreads 
		addLast: (aBlock newProcess 
					priority: Processor userSchedulingPriority;
					name: 'a UI Event Thread')! !

!WorldState methodsFor: 'hands' stamp: 'GabrielOmarCotelli 11/30/2013 17:13'!
selectHandsToDrawForDamage: damageList
	"Select the set of hands that must be redrawn because either (a) the hand itself has changed or (b) the hand intersects some damage rectangle."

	^ hands
		select: [ :hand | hand needsToBeDrawn and: [ hand hasChanged or: [ self is: hand overlappingDamagedAreaIn: damageList ] ] ]! !

!WorldState methodsFor: 'settings' stamp: 'AlainPlantec 12/14/2009 14:01'!
serverMode
	^ self class serverMode! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:36'!
startStepping: aMorph at: scheduledTime selector: aSelector arguments: args stepTime: stepTime
	"Add the given morph to the step list. Do nothing if it is already being stepped."

	self stopStepping: aMorph selector: aSelector.
	self adjustWakeupTimesIfNecessary.
	stepList add:(
		StepMessage 
			scheduledAt: scheduledTime
			stepTime: stepTime
			receiver: aMorph
			selector: aSelector
			arguments: args)! !

!WorldState methodsFor: 'initialization' stamp: 'sw 9/5/2000 06:39'!
stepListSize
	^ stepList size! !

!WorldState methodsFor: 'stepping' stamp: 'nice 4/16/2009 19:03'!
stepListSortBlock
	^[ :stepMsg1 :stepMsg2 | 
		stepMsg1 scheduledTime <= stepMsg2 scheduledTime
	]! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:10'!
stopStepping: aMorph
	"Remove the given morph from the step list."
	lastStepMessage ifNotNil:[
		(lastStepMessage receiver == aMorph) ifTrue:[lastStepMessage := nil]].
	stepList removeAll: (stepList select:[:stepMsg| stepMsg receiver == aMorph]).
! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:10'!
stopStepping: aMorph selector: aSelector
	"Remove the given morph from the step list."
	lastStepMessage ifNotNil:[
		(lastStepMessage receiver == aMorph and:[lastStepMessage selector == aSelector])
			ifTrue:[lastStepMessage := nil]].
	stepList removeAll: (stepList select:[:stepMsg| stepMsg receiver == aMorph and:[stepMsg selector == aSelector]]).! !

!WorldState methodsFor: 'event-handling' stamp: 'NickPapoulias 12/29/2014 12:28'!
suspendEventThreadScheduling

	self eventThreads do: [ :et | et suspend ].
	self rescheduleRunningThread.
	self unRefCompletedEventThreads.
	! !

!WorldState methodsFor: 'alarms' stamp: 'ar 10/22/2000 16:55'!
triggerAlarmsBefore: nowTime
	"Trigger all pending alarms that are to be executed before nowTime."
	| pending |
	lastAlarmTime ifNil:[lastAlarmTime := nowTime].
	(nowTime < lastAlarmTime or:[nowTime - lastAlarmTime > 10000])
		ifTrue:[self adjustAlarmTimes: nowTime].
	pending := self alarms.
	[pending isEmpty not and:[pending first scheduledTime < nowTime]]
		whileTrue:[pending removeFirst value: nowTime].
	lastAlarmTime := nowTime.! !

!WorldState methodsFor: 'event-handling' stamp: 'NickPapoulias 12/28/2014 21:46'!
unRefCompletedEventThreads

	|completedThreads|
	
	completedThreads := OrderedCollection new.
	self eventThreads do: [ :et | et isTerminated ifTrue: [ completedThreads addLast: et ] ].
	completedThreads do: [:et | self eventThreads remove: et ]! !

!WorldState methodsFor: 'canvas' stamp: 'di 6/7/1999 17:44'!
viewBox

	^ viewBox! !

!WorldState methodsFor: 'canvas' stamp: 'di 6/7/1999 17:58'!
viewBox: x

	viewBox := x! !

!WorldState methodsFor: 'worldmenu building' stamp: 'MarcusDenker 6/28/2013 12:48'!
worldMenu
	^self  menuBuilder menuEntitled: self discoveredMenuTitle.! !
Morph initialize!
WorldState initialize!
HandMorph initialize!
